package brain;

import java.awt.GraphicsEnvironment;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.FileWriter;
import java.io.IOException;
import java.io.InputStream;
import java.io.ObjectInput;
import java.io.ObjectInputStream;
import java.io.ObjectOutput;
import java.io.ObjectOutputStream;
import java.io.OutputStream;
import java.io.PrintWriter;
import java.io.Serializable;
import java.lang.reflect.AccessibleObject;
import java.lang.reflect.Constructor;
import java.lang.reflect.Field;
import java.lang.reflect.InvocationTargetException;
import java.lang.reflect.Method;
import java.text.DateFormat;
import java.text.SimpleDateFormat;
import java.util.ArrayList;
import java.util.Date;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Locale;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Random;
import java.util.Set;
import java.util.Vector;
import java.util.logging.Level;
import java.util.logging.Logger;

import javax.swing.SwingUtilities;
import javax.swing.UIManager;
import javax.swing.event.ChangeEvent;
import javax.swing.event.ChangeListener;
import javax.swing.event.EventListenerList;

import org.apache.commons.cli.CommandLine;
import org.apache.commons.cli.CommandLineParser;
import org.apache.commons.cli.HelpFormatter;
import org.apache.commons.cli.Options;
import org.apache.commons.cli.ParseException;
import org.apache.commons.cli.PosixParser;

import utils.ProgressPrintTimer;
import utils.Timer;
import utils.Utils;
import brain.gui.BrainGUI;
import brain.gui.DisplayManager;

import com.thoughtworks.xstream.XStream;
import com.thoughtworks.xstream.converters.reflection.PureJavaReflectionProvider;
import com.thoughtworks.xstream.io.xml.DomDriver;

/**
 * <h2>modules for the cognitive system</h2>
 * <p>
 * {@link Module}: general interface for all components (modules)
 * {@link DhanModule}: the dHAN module, implements the `Module' interface
 * {@link PatternInputModule}: the input module, implements the `Module'
 * interface
 * </p>
 * 
 * <h2>classical auxiliary classes</h2>
 * <dl>
 * <dt>OneAdaptVar</dt>
 * <dd>a single adaptive signal</dd>
 * <dt>OneStimulus</dt>
 * <dd>class representing an individual external stimuli</dd>
 * <dt>OneStimulusContainer</dt>
 * <dd>class managing sets of stimuli</dd>
 * <dt>OneCenter</dt>
 * <dd>representing a globally defined active center (AC)</dd>
 * <dt>OneCluster</dt>
 * <dd>a general clique; contains a sequence of active centers</dd>
 * <dt>OneCycle</dt>
 * <dd>a general cycle; contains a sequence of active clusters</dd>
 * </dl>
 * 
 * <h2>abstract auxiliary facilities</h2>
 * <dl>
 * <dt>FacilPrint</dt>
 * <dd>general printing to file facility</dd>
 * <dt>FacilCliqueNet</dt>
 * <dd>generates networks with a given distribution of cliques</dd>
 * <dt>RandomGraphs</dt>
 * <dd>facility for the analysis of general random graphs</dd>
 * <dt>FacilStimuli</dt>
 * <dd>facility to generate sets of test stimuli</dd>
 * </dl>
 * 
 * <h2>classes for the statistical analysis</h2>
 * <dl>
 * <dt>StatNet</dt>
 * <dd>general network-theoretical statistics facility</dd>
 * <dt>StatActiClus</dt>
 * <dd>facility for finding/evaluation of cliques/cycles/history</dd>
 * <dt>StatBiasLearn</dt>
 * <dd>monitors from-bias-learning of DhanModule-type modules</dd>
 * <dt>StatTimeManagement</dt>
 * <dd>monitors time to finish a simulation run</dd>
 * </dl>
 * 
 * 
 * 
 * <h1>list of important output files</h1>
 * 
 * <h2>time series of data</h2>
 * <dl>
 * <dt>Brain_dhan.acti</dt>
 * <dd>activities x_i(t)</dd>
 * <dt>Brain_dhan.rate</dt>
 * <dd>growth rate r_i(t)</dd>
 * <dt>Brain_dhan.rateInt</dt>
 * <dd>internal contribution to growth rate r_i(t)</dd>
 * <dt>Brain_dhan.res</dt>
 * <dd>reservoir phi_i(t)</dd>
 * <dt>Brain_dhan.bias</dt>
 * <dd>bias b(t)</dd>
 * <dt>Brain_dhan.backRate</dt>
 * <dd>input grows rates</dd>
 * <dt>Brain_dhan.backStandDev</dt>
 * <dd>standard deviation of incoming link-strength</dd>
 * <dt>Brain_dhan.diffModuleSignal</dt>
 * <dd>the diffusive learning signal</dd>
 * <dt>Brain_dhan.synchro</dt>
 * <dd>an effective measure of clique formation</dd>
 * <dt>Brain_dhan.actiClus</dt>
 * <dd>time series of clusters (cliques) activated</dd>
 * <dt>Brain_dhan.actiClusStat</dt>
 * <dd>how many times a given cluster was activated</dd>
 * <dt>Brain_dhan.actiRate</dt>
 * <dd>auxiliary rates</dd>
 * <dt>Brain_input.stim</dt>
 * <dd>time series of input stimuli</dd>
 * </dl>
 * 
 * <h2>input files</h2>
 * <dl>
 * <dt>Brain_dhan.ass</dt>
 * <dd>cliques (associations) when reading network from file</dd>
 * <dt>Brain_input.patt</dt>
 * <dd>input stimuli</dd>
 * <dt>Brain_input.refPatt</dt>
 * <dd>reference patterns for analyzing purposes</dd>
 * </dl>
 * 
 * <h2>output files (generated at the end)</h2>
 * <dl>
 * <dt>Brain_dhan_input.cliqRecFields</dt>
 * <dd>the receptive fields in (input) of cliques in (input)</dd>
 * <dt>Brain_dhan_input.cliqRefPatt</dt>
 * <dd>the response of cliques in (dhanModule) with respect to reference
 * patterns in (input)</dd>
 * 
 * <dt>Brain_dhan.cliq</dt>
 * <dd>all cliques in dhanModule module</dd>
 * <dt>Brain_dhan.stat</dt>
 * <dd>statistics like degree distribution</dd>
 * <dt>Brain_dhan.statBackLinks</dt>
 * <dd>bin-statistics of the backward links</dd>
 * 
 * <dt>Brain_dhan_input.backLinks</dt>
 * <dd>backlinks connecting (input)-->(dhanModule)</dd>
 * <dt>Brain_dhan.acSequence</dt>
 * <dd>time sequences of active clusters of a module formed out of
 * Brain_dhan.actiClus</dd>
 * <dt>Brain_dhan.statCycle</dt>
 * <dd>statistical properties of cycles of active clusters in
 * Brain_dhan.actiClus</dd>
 * </dl>
 * 
 * <h2>auxiliary files</h2>
 * <dl>
 * <dt>Brain_dhan.fig</dt>
 * <dd>xfig picture of network</dd>
 * <dt>Brain_dhan.learn</dt>
 * <dd>an auxiliary learning analysis</dd>
 * </dl>
 * 
 * 
 * <h1>output-file name convention</h1> Brain_(module1)_(module2)_(...).datatyp
 * where (module..) are the name of the modules involved in data extraction.
 * 
 * 
 * 
 * <h1>
 * instructions for developers :: compatibility with the monitoring interface
 * (GUI)</h1>
 * 
 * module member variable `memberVariable' is observable in GUI if present in
 * member variable 'guiObservables[]'
 * 
 * module member variable `memberVariable' is changable by the GUI if public and
 * not static
 * 
 * if a member variable `memberVariable' has only a finite number of allowed
 * values put these in `memberVariableAllowedValues'; when
 * `memberVariableAllowedValues' is present, then `memberVariable' is changeable
 * through the GUI
 * 
 * 
 * 
 * <h1>
 * actually implemented types of diffusive module signals</h1> DhanModule ::
 * novelty signal, cliques activated by backward links
 * 
 * 
 * 
 * <h1>
 * actually used type of receptors for the inter- and inter-modular diffusive
 * control</h1> "backwardLearning" :: DhanModule - learning of backward links
 * 
 * 
 * Entry point with global variables.
 * 
 * Modules like `DhanModule' extend this Brain and acquire such access to the
 * HashMap `modules' which contains all (other) components.
 * <p>
 * 
 * @author Claudius Gros
 * @author Gregor Kaczor
 * @author Mathias Linkerhand
 * @version 2013-01-26
 * @since 2008
 */

public class Brain implements Runnable, Serializable {
  private static final long                                         serialVersionUID              = 1L;

  public transient final File                                       dataDirectory                 = DEFAULT_DATA_DIRECTORY;
  public transient final File                                       dataInputDirectory            = DEFAULT_DATA_INPUT_DIRECTORY;
  public transient File                                             dataOutputDirectory           = DEFAULT_DATA_OUTPUT_DIRECTORY;

  public static final DateFormat                                    DATE_FORMAT                   = new SimpleDateFormat("yyyy-MM-dd HH:mm:ss z");
  public static final File                                          DEFAULT_DATA_DIRECTORY        = new File("data");
  public static final File                                          DEFAULT_DATA_INPUT_DIRECTORY  = new File(DEFAULT_DATA_DIRECTORY, "input");
  public static File                                                DEFAULT_DATA_OUTPUT_DIRECTORY = new File(DEFAULT_DATA_DIRECTORY, "output");
  public static final File                                          RESOURCES_DIRECTORY           = new File("lib/resources");
  final public static String                                        VERSION                       = "2013-01-26";

  /**
   * Listeners to be notified of module changes.
   * 
   * This field is marked transient, because AccessibleObjects (Fields, …) are
   * not serializable. There (de-)serialization of that field is done by the
   * methods <code>readObject</code> and <code>writeObject</code>.
   */
  protected transient EventListenerList                             changeListeners               = new EventListenerList();

  /** time of the simulation */
  private transient double                                          time;
  /** The time differential. */
  private double                                                    timeDelta                     = 1e-2;
  /** The number of time steps. */
  private long                                                      timeStepNumber                = (long) 1E2;

  /**
   * This field is marked transient, because AccessibleObjects (Fields, …) are
   * not serializable. The (de-)serialization of that field is done by the
   * methods <code>readObject</code> and <code>writeObject</code>.
   */
  private transient Map<Object, Map<AccessibleObject, PrintWriter>> accessibleObjectsToLog        = new HashMap<Object, Map<AccessibleObject, PrintWriter>>();

  private static final Logger                                       logger                        = Logger.getLogger(Brain.class.getName());

  /** Delay specification for observation purposes. */
  private int                                                       simulationSleepDelay          = 0;

  private Set<Module>                                               modules                       = new HashSet<Module>();

  /**
   * HashMap of `StatActiClus' for monitoring (global) clusters of activities
   * centers.
   */
  private Set<StatActiClus>                                         clusterActivityObserver       = new HashSet<StatActiClus>();

  /**
   * Time interval for searching for active clusters. The default value for
   * `clusterSearchTimeInterval' may be changed at run-time.
   */
  public double                                                     clusterSearchTimeInterval     = 4.0;

  /**
   * HashMap of `StatBiasLearn' for monitoring of learning of patterns presented
   * to individual modules via the respective `bias[]'.
   * 
   * <p>
   * The key is identical to the name of the module to be monitored. <br>
   * This observer is activated at the start of `*_IF.oneTimeStep()', when the
   * bias has its nominal value (the `bias[]' is set to zero at the end of
   * `*_IF.oneTimeStep()').
   */
  private Map<Module, StatBiasLearn>                                learnBiasObserver             = new HashMap<Module, StatBiasLearn>();

  // @see
  // http://java.sun.com/javase/6/docs/technotes/guides/concurrency/threadPrimitiveDeprecation.html
  private volatile transient boolean                                isSuspended                   = false;

  private transient double                                          lastLogTime                   = Double.NEGATIVE_INFINITY;
  private double                                                    logInterval                   = 1.0;

  public Brain() {
    final Brain brain = this;

    // cleanup on exit
    final Thread onExitThread = new Thread() {
      @Override
      public void run() {
        // cleanup shutdown
        // close and save all open files
        brain.closeLoggers();
      }
    };
    Runtime.getRuntime().addShutdownHook(onExitThread);
  }

  public static Brain createDefaultGrosBrain() throws FileNotFoundException {
    final Brain brain = new Brain();
    final double duration = 1000;
    brain.setTimeDelta(1e-2);
    brain.setTimeStepNumber((int) Math.round(duration / brain.getTimeDelta()));

    // dhan module
    final Module dhan = new DhanModule(new File(brain.dataInputDirectory, "dhan.cliques"));
    dhan.setName("DEFAULT dhan");
    brain.modules.add(dhan);
    brain.addLogger(dhan, "getActivities");
    brain.addLogger(dhan, "getBiases");
    brain.addLogger(dhan, "afferentRates");
    brain.addLogger(dhan, "afferentInverseRates");
    brain.addLogger(dhan, "backwardMemMean");
    brain.addLogger(dhan, "backwardMemStandDev");
    brain.addLogger(dhan, "diffusiveModuleSignal");
    brain.addLogger(dhan, "rates");
    brain.addLogger(dhan, "ratesActive");
    brain.addLogger(dhan, "ratesInternal");
    brain.addLogger(dhan, "reservoir");
    brain.addLogger(dhan, "synchronizationStrength");

    // input module
    final Module input = new PatternInputModule(new File(brain.dataInputDirectory, "input.patterns"));
    input.setName("DEFAULT input");
    brain.modules.add(input);
    brain.addLogger(input, "getActivities");
    brain.addLogger(input, "getCurrentStimuli");

    // connect modules
    input.addForwardModule(dhan, "default");
    dhan.addAfferentModule(input);

    return brain;
  }

  public static Brain createDefaultLinkerhandBrain() {
    final Brain brain = new Brain();
    final double duration = 1000;
    brain.setTimeDeltaAndDuration(1e-2, duration);

    // adaptive leaky integrator
    final Random random = new Random(0);
    final Module module1 = new LeakyIntegratorWithIntrinsicPlasticityModule(3, 0.6, random);
    module1.setName("DEFAULT LeakyIntegrator 1");
    brain.addModule(module1);
    brain.addLogger(module1, "transferFunctionGain");
    brain.addLogger(module1, "transferFunctionThreshold");
    brain.addLogger(module1, "getActiveClique");

    final Module module2 = new LeakyIntegratorWithIntrinsicPlasticityModule(3, 0.6, random);
    module2.setName("DEFAULT LeakyIntegrator 2");
    module1.addAfferentModule(module2);
    module2.addAfferentModule(module1);
    brain.addModule(module2);

    return brain;
  }

  /**
   * Main routine to start the Brain simulation program.
   * 
   * @param args Arguments for the program. Use --help to get a list.
   */
  public static void main(final String[] args) {
    // Set log level.
    final Logger brainLogger = Logger.getLogger("brain");
    brainLogger.setLevel(Level.ALL);
    Logger.getLogger("").getHandlers()[0].setLevel(Level.WARNING); // Only error output.

    boolean useGui = false;
    Brain brain = null;

    // parse command line arguments
    final Options options = new Options();
    options.addOption("b", "brain", true, "brain file to use for simulation");
    options.getOption("b").setArgName("file");
    options.addOption("d", "debug", false, "debug");
    options.addOption("g", "gui", false, "use GUI" + (useGui ? " (default)" : ""));
    options.addOption("h", "help", false, "print this help and exit");
    options.addOption("o", "output", true, "data output directory");
    options.getOption("o").setArgName("directory");
    options.addOption("s", "stimuli", false, "generate new input stimuli patterns and reference patterns");
    options.addOption("V", "version", false, "print version and exit");
    options.addOption("v", "verbose", false, "increase verbosity");
    final CommandLineParser parser = new PosixParser();
    try {
      final CommandLine cmd = parser.parse(options, args);
      if (cmd.hasOption("h")) {
        // Help, printing usage information.
        final String header = "Brain is a general platform for autonomous cognitive systems written by Claudius Gros, Gregor Kaczor and Mathias Linkerhand.";
        final String footer = Brain.VERSION;
        final HelpFormatter formatter = new HelpFormatter();
        formatter.printHelp(Brain.class.getSimpleName() + " [OPTION] … [FILE] …", header, options, footer);
        System.exit(0);
      }
      if (cmd.hasOption("v")) {
        Logger.getLogger("").getHandlers()[0].setLevel(Level.INFO); // Show more log events.
      }
      if (cmd.hasOption("d")) {
        Logger.getLogger("").getHandlers()[0].setLevel(Level.ALL); // Show all log events.
      }
      if (cmd.hasOption("V")) {
        System.out.println(Brain.VERSION);
        System.exit(0);
      }

      if (cmd.hasOption("s")) {
        // generate and save new stimuli
        final OneStimulus[] barStimuli = OneStimulus.generateBarStimuli(5);
        final OneStimulus[] superimposedBarStimuli = OneStimulus.superimposeStimuli(barStimuli, 0.1, 3000);
        try {
          OneStimulus.saveStimuliToFile(barStimuli, new File(".", "input.referencepatterns"));
          OneStimulus.saveStimuliToFile(superimposedBarStimuli, new File(".", "input.patterns"));
        } catch (FileNotFoundException exception) {
          System.out.println(exception.getMessage());
        }
        System.exit(0);
      }

      useGui = cmd.hasOption("g");

      if (cmd.hasOption("o")) {
        Brain.DEFAULT_DATA_OUTPUT_DIRECTORY = new File(cmd.getOptionValue("o"));
      }

      // Load brain from file. This has to be done after changing the data output directory to make sure the output files are at the correct place specified.
      if (cmd.getArgs().length > 0) {
        final File brainFile = new File(cmd.getArgs()[0]);
        try {
          logger.log(Level.INFO, "Load brain from file " + brainFile + ".");
          brain = Brain.load(brainFile);
          final String outputDirectoryName = brainFile.getName() + ".data";
          final File outputDirectory = new File(Brain.DEFAULT_DATA_OUTPUT_DIRECTORY, outputDirectoryName);
          outputDirectory.mkdirs();
          brain.setDataOutputDirectory(outputDirectory);
        } catch (IllegalArgumentException e) {
          logger.log(Level.SEVERE, "Error loading brain from file " + brainFile + ".");
          e.printStackTrace();
        } catch (ClassNotFoundException e) {
          logger.log(Level.SEVERE, "Error loading brain from file " + brainFile + ".");
          e.printStackTrace();
        } catch (IOException e) {
          logger.log(Level.SEVERE, "Error loading brain from file " + brainFile + ".");
          e.printStackTrace();
        }
      }
    } catch (ParseException e) {
      e.printStackTrace();
      System.exit(1);
    }

    // Process SIGINT signal (<Crtl> + <C>).
    final Thread onFailureThread = new Thread() {
      @Override
      public void run() {
        logger.log(Level.INFO, "Simulation aborted. Please wait for the program to clean up!");

        // cleanup shutdown

        logger.log(Level.INFO, "Exiting. Bye.");
      }
    };
    Runtime.getRuntime().addShutdownHook(onFailureThread);

    // cleanup on exit
    final Thread onExitThread = new Thread() {
      @Override
      public void run() {
        // cleanup shutdown
        logger.log(Level.INFO, "Exiting. Bye.");
      }
    };
    Runtime.getRuntime().addShutdownHook(onExitThread);

    if (brain == null) {
      logger.log(Level.INFO, "No brain file specified. Using standard brain configuration.");
      // load default brain
      final boolean gros = true;
      if (gros) {
        try {
          brain = Brain.createDefaultGrosBrain();
          final String outputDirectoryName = "gros.data";
          final File outputDirectory = new File(Brain.DEFAULT_DATA_OUTPUT_DIRECTORY, outputDirectoryName);
          outputDirectory.mkdirs();
          brain.setDataOutputDirectory(outputDirectory);
        } catch (FileNotFoundException e1) {
          logger.log(Level.SEVERE, e1.getMessage());
          System.exit(1);
        }
      } else {
        brain = Brain.createDefaultLinkerhandBrain();
        final String outputDirectoryName = "linkerhand.data";
        final File outputDirectory = new File(Brain.DEFAULT_DATA_OUTPUT_DIRECTORY, outputDirectoryName);
        outputDirectory.mkdirs();
        brain.setDataOutputDirectory(outputDirectory);
      }
    }

    if (useGui) {
      if (GraphicsEnvironment.isHeadless()) {
        logger.log(Level.SEVERE, "No display, keyboard or mouse found. Exiting …");
        System.exit(1);
      }

      try {
        // set default system’s look and feel
        //UIManager.setLookAndFeel(UIManager.getSystemLookAndFeelClassName());
      } catch (Exception e) {
        logger.log(Level.WARNING, "error while loading " + UIManager.getSystemLookAndFeelClassName() + ": " + e.toString());
      }

      // start the GUI
      final Brain guiBrain = brain;
      SwingUtilities.invokeLater(new Runnable() {
        @Override
        public void run() {
          final BrainGUI brainGui = new BrainGUI(guiBrain);
          brainGui.setLocation(DisplayManager.calculateCoordinates(brainGui));
          brainGui.setVisible(true);
        }
      });
    } else {
      // run simulation
      System.out.println("data output: " + brain.dataOutputDirectory.getPath());
      Thread brainThread = new Thread(brain);
      brainThread.start();
      try {
        // wait for the simulation coming to end
        brainThread.join();
      } catch (InterruptedException exception) {
        exception.printStackTrace();
      }
    }

    // Program ends normal, so do not start the shutdown hook.
    // This should be the last command in the main method.
    Runtime.getRuntime().removeShutdownHook(onFailureThread);
  }

  public void printModuleSetup() {
    String message = "Modules:";
    for (final Module module : this.modules) {
      message += "\n\t" + module.toString();
      for (final Module afferentModule : module.getAfferentModules()) {
        message += "\n\t\t" + "connected afferent (←) to " + afferentModule.getClass().getSimpleName() + " " + afferentModule.getName();
      }
      for (final Module efferentModule : module.getForwardModules()) {
        message += "\n\t\t" + "connected efferent (→) to " + efferentModule.getClass().getSimpleName() + " " + efferentModule.getName();
      }
    }
    logger.log(Level.INFO, message);
  }

  /**
   * Prints default Graphviz files for the current network. All Modules with
   * intra- and inter-links are printed. Prints three versions: only excitatory
   * links, only inhibitory links, all links.
   */
  public void printGraphvizGraphs() {
    try {
      // print graphviz graphs
      for (int trial = 0; trial < 3; trial++) {
        final File brainGraphvizFile;
        final double lowerLinkBound;
        final double upperLinkBound;
        switch (trial) {
          case 0:
            // graph with excitatory links
            brainGraphvizFile = new File(this.dataOutputDirectory, "brain.excitatory-links.gv");
            lowerLinkBound = 0;
            upperLinkBound = Double.POSITIVE_INFINITY;
            break;
          case 1:
            // graph with inhibitory links
            brainGraphvizFile = new File(this.dataOutputDirectory, "brain.inhibitory-links.gv");
            lowerLinkBound = Double.NEGATIVE_INFINITY;
            upperLinkBound = 0;
            break;
          case 2:
            // graph with all edges
            brainGraphvizFile = new File(this.dataOutputDirectory, "brain.all-links.gv");
            lowerLinkBound = Double.NEGATIVE_INFINITY;
            upperLinkBound = Double.POSITIVE_INFINITY;
            break;
          default:
            throw new IllegalStateException("no such trial");
        }
        this.exportGraphvizGraph(brainGraphvizFile, lowerLinkBound, upperLinkBound);
      }
    } catch (FileNotFoundException e) {
      e.printStackTrace();
    }
  }

  /**
   * The global dynamics.
   * 
   * Contains the loop over time-steps for the numerical integration of the
   * differential equations for the time-development.
   */
  @Override
  public void run() {
    // Create all necessary output directories.
    if (!this.dataOutputDirectory.isDirectory()) {
      this.dataOutputDirectory.mkdirs();
    }

    // Print network setup.
    this.printGraphvizGraphs();

    // Print log file header information.
    this.printLogFileHeaders();

    // Save a copy of the current brain (at the beginning of the simulation)
    // in the standard data output directory for later use or re-check.
    final File brainBeginFile = new File(this.dataOutputDirectory, "brain " + Brain.DATE_FORMAT.format(new Date()) + ".brain.xml");
    this.save(brainBeginFile);

    this.clusterActivityObserver.clear();
    this.learnBiasObserver.clear();
    this.setTime(0.0);

    logger.log(Level.INFO, "Simulation started.");

    // output current setup
    this.printModuleSetup();

    // observe clusters in dhanModule
    // TODO: legacy, to be removed, to be moved to the dhan module
    final Module[] modulesToObserveClusters = this.getModules(DhanModule.class);
    this.clusterActivityObserver.add(new StatActiClus(modulesToObserveClusters));
    String moduleNames = null;
    for (final Module module : modulesToObserveClusters) {
      if (moduleNames == null) {
        moduleNames = "";
      } else {
        moduleNames += ", ";
      }
      moduleNames += module.getName();
    }
    logger.log(Level.FINE, "Modules to observe clusters: " + moduleNames);

    // observe learning in dhanModule
    // TODO: legacy, to be removed, to be moved to the dhan module
    final Module[] modulesToObserveLearning = modulesToObserveClusters;
    moduleNames = null;

    for (final Module module : modulesToObserveLearning) {
      if (moduleNames == null) {
        moduleNames = "";
      } else {
        moduleNames += ", ";
      }
      moduleNames += module.getName();
      learnBiasObserver.put(module, new StatBiasLearn((DhanModule) module));
    }
    logger.log(Level.FINE, "Modules to observe learning: " + moduleNames);

    // the actual simulation
    final Timer timer;
    if (GraphicsEnvironment.isHeadless()) {
      // Use standard timer if there is no keyboard, mouse, monitor …
      timer = new Timer();
    } else {
      // Print progress frequently if there is keyboard, mouse, monitor …
      timer = new ProgressPrintTimer(System.out, "Simulation");
    }
    timer.start();

    logger.log(Level.INFO, "Simulation Start: " + Brain.DATE_FORMAT.format(timer.getStart()));

    // loop over discrete time-steps
    for (long timeStep = 0; timeStep < this.getTimeStepNumber(); timeStep++) {
      // simulation paused?
      while (this.isSuspended) {
        // @see
        // http://java.sun.com/javase/6/docs/technotes/guides/concurrency/threadPrimitiveDeprecation.html
        synchronized (this) {
          try {
            // wait for resume
            this.wait();
          } catch (InterruptedException exception) {
            logger.log(Level.INFO, "Simulation interrupted during suspend.");
            this.resume();
            Thread.currentThread().interrupt();
          }
        }
      }

      // simulation stopped?
      if (Thread.currentThread().isInterrupted()) {
        break;
      }

      // perform one time step
      this.performOneTimeStep(this.timeDelta);

      // notify observers about new state
      this.notifyStateChanged(new ChangeEvent(this));

      // simulation delay
      if (this.simulationSleepDelay > 0) {
        try {
          Thread.sleep(this.simulationSleepDelay);
        } catch (InterruptedException exception) {
          // Simulation was interrupted during sleep. So interrupt it
          // again after sleep to finally stop the simulation.
          Thread.currentThread().interrupt();
        }
      }

      timer.setProgress((timeStep + 1.0) / this.timeStepNumber);
    }
    timer.stop();

    // time statistics
    logger.log(Level.INFO, "Simulation duration: " + timer.getElapsed() / 1000 + " s");
    logger.log(Level.INFO, "Simulation End: " + Brain.DATE_FORMAT.format(timer.getStop()));
    logger.log(Level.INFO, "Simulation terminated.");

    // Save a copy of the current brain (at the end of the simulation) in
    // the standard data output directory for later use or re-check.
    final File brainEndFile = new File(this.dataOutputDirectory, "brain " + Brain.DATE_FORMAT.format(new Date()) + ".brain.xml");
    this.save(brainEndFile);

    // end statistics and output for all modules
    this.calculateSimulationStatistics();
  }

  public void performOneTimeStep(final double timeDelta) {
    // $t \mapsto t + \Delta t$
    // $t(n+1) = t(n) + \Delta t$
    this.setTime(this.getTime() + timeDelta);

    // call bias learning observer (at this point `this.bias[]' has its
    // nominal value)
    // TODO: legacy, to be removed, to be moved to the dhan module
    int nDeltaTime = (int) (this.getTime() / this.getTimeDelta());
    double timeToPrint = ((int) (this.getTime() / this.clusterSearchTimeInterval)) * this.clusterSearchTimeInterval;
    if ((nDeltaTime * this.getTimeDelta() <= timeToPrint) && ((nDeltaTime + 1) * this.getTimeDelta() > timeToPrint)) {
      for (final Module module : this.learnBiasObserver.keySet()) {
        this.learnBiasObserver.get(module).findBiasPattern(this.getTime());
      }
    }

    // for all modules
    for (final Module module : this.getModules()) {
      module.performOneTimeStep(timeDelta);
    }

    // for all observers
    // TODO: legacy, to be removed, to be moved to the dhan module
    if ((nDeltaTime * this.getTimeDelta() <= timeToPrint) && ((nDeltaTime + 1) * this.getTimeDelta() > timeToPrint)) {
      for (final StatActiClus statActiClus : this.clusterActivityObserver) {
        statActiClus.findActiveCluster(this.getTime());
      }
    }

    // log system and module parameters and variables
    this.performLogging();
  }

  public double getLogInterval() {
    return this.logInterval;
  }

  public void setLogInterval(final double logInterval) {
    this.logInterval = logInterval;
  }

  public void performLogging() {
    if (this.getTime() >= this.lastLogTime + this.getLogInterval()) {
      for (final Map.Entry<Object, Map<AccessibleObject, PrintWriter>> fieldToLog : this.accessibleObjectsToLog.entrySet()) {
        final Object object = fieldToLog.getKey();
        final Map<AccessibleObject, PrintWriter> fields = fieldToLog.getValue();
        for (final Map.Entry<AccessibleObject, PrintWriter> fieldToPrintWriter : fields.entrySet()) {
          final AccessibleObject accessibleObject = fieldToPrintWriter.getKey();
          final PrintWriter out = fieldToPrintWriter.getValue();

          accessibleObject.setAccessible(true); // Make sure also
          // private
          // fields/methods
          // can be accessed.

          final int decimalPlaces = 5;
          // time
          out.printf(Locale.ENGLISH, "%." + decimalPlaces + "g", this.getTime());

          if (accessibleObject instanceof Field) {
            final Field field = (Field) accessibleObject;
            try {
              final Object fieldValue = field.get(object);

              // object
              logObject(fieldValue, out);
            } catch (IllegalArgumentException e) {
              e.printStackTrace();
            } catch (IllegalAccessException e) {
              e.printStackTrace();
            }
          } else if (accessibleObject instanceof Method) {
            final Method method = (Method) accessibleObject;
            try {
              final Object methodResult = method.invoke(object);

              // object
              logObject(methodResult, out);
            } catch (InvocationTargetException e) {
            } catch (IllegalArgumentException e) {
            } catch (IllegalAccessException e) {
            }
          }
          // new line
          out.println();
        }
      }

      /*
       * this.lastLogTime = this.getTime() does not give the result expected.
       * For example consider an average log interval of 3, but a time delta of
       * 2.
       */
      if (this.lastLogTime == Double.NEGATIVE_INFINITY) {
        // this.lastLogTime = this.getLogInterval();
        this.lastLogTime = this.getTime();
      } else {
        while (this.getTime() >= this.lastLogTime + this.getLogInterval()) {
          this.lastLogTime += this.getLogInterval();
        }
      }
    }
  }

  /**
   * Used to write an object to a textual representation. It uses the object’s
   * toString method. If the object is iterable the values are called
   * recursively.
   * 
   * @param object
   * @param out
   */
  private static void logObject(Object object, final PrintWriter out) {
    // object
    if (object == null) {
      out.print("\t" + null);
      // } else if (object instanceof Float || object instanceof Double) {
      // out.printf(Locale.ENGLISH, "\t%e", object);
    } else if (object instanceof Iterable<?>) {
      final Iterable<?> iterable = (Iterable<?>) object;
      for (final Object value : iterable) {
        logObject(value, out);
      }
    } else if (object instanceof double[]) {
      for (final double value : (double[]) object) {
        logObject(value, out);
      }
    } else if (object instanceof Object[]) {
      for (final Object value : (Object[]) object) {
        logObject(value, out);
      }
    } else if (object instanceof OneStimulus) {
      final OneStimulus stimulus = (OneStimulus) object;
      out.printf("%7.1f %7.1f ", stimulus.getStartTime(), stimulus.getEndTime());
      for (final int node : stimulus.getNodes()) {
        out.printf(" (%2d|%4.2f)", node, stimulus.getStrength(node));
      }
    } else if (object instanceof Double) {
      // final int decimalPlaces = 4;
      // out.print("\t" + String.format(Locale.ENGLISH, "%6." +
      // decimalPlaces + "g", object));
      out.print("\t" + String.format(Locale.ENGLISH, "%g", object));
    } else {
      out.print("\t" + object.toString());
    }
  }

  public void setDataOutputDirectory(final File file) {
    if (!file.isDirectory()) {
      throw new IllegalArgumentException("No such directory " + file + ".");
    }

    // Set local variable.
    this.dataOutputDirectory = file;

    // Relocate log files.
  }

  public void printLogFileHeaders() {
    for (final Entry<Object, Map<AccessibleObject, PrintWriter>> entry : this.accessibleObjectsToLog.entrySet()) {
      final Object object = entry.getKey();
      final Map<AccessibleObject, PrintWriter> fields = entry.getValue();
      for (final Map.Entry<AccessibleObject, PrintWriter> entry2 : fields.entrySet()) {
        final AccessibleObject accessibleObject = entry2.getKey();
        PrintWriter out = entry2.getValue();
        if (out == null) {
          // Create default logger file.
          final String filenameSuffix = ".data";
          final String accessibleObjectName = (accessibleObject instanceof Method) ? ((Method) accessibleObject).getName()
              : (accessibleObject instanceof Field) ? ((Field) accessibleObject).getName() : null;
          final File file = new File(this.dataOutputDirectory, (object + " " + accessibleObjectName + filenameSuffix).replace(" ", "_")); // replace spaces by underscores
          final boolean autoFlush = false; // Do not flush files after
          // every println. (Using
          // buffers is faster.)
          file.getParentFile().mkdirs(); // Create parent directories.
          try {
            out = new PrintWriter(new FileWriter(file), autoFlush);
          } catch (IOException e) {
            e.printStackTrace();
          }
          fields.put(accessibleObject, out);
        }

        // print log file header
        if (object instanceof Module) {
          final Module module = (Module) object;
          module.printLogFileHeader(out);
        }
      }
    }
  }

  public void addLogger(final Object object, final AccessibleObject accessibleObject, final File file) {
    if (!this.accessibleObjectsToLog.containsKey(object)) {
      this.accessibleObjectsToLog.put(object, new HashMap<AccessibleObject, PrintWriter>());
    }

    try {
      final PrintWriter out;
      if (file != null) {
        final boolean autoFlush = false; // Do not flush files after
        // every println. (Using
        // buffers is faster.)
        out = new PrintWriter(new FileWriter(file), autoFlush);
      } else {
        out = null;
      }
      this.accessibleObjectsToLog.get(object).put(accessibleObject, out);
    } catch (IOException e) {
      throw new IllegalArgumentException(e);
    }
  }

  /**
   * User friendly version of method <code>addLogger</code>.
   * 
   * The accessible object (field, method, …) will be search for given a string.
   * First fields, than methods.
   * 
   * The filename will be created automatically according to the observed object
   * and the accessible object, viz. the method or field to observe.
   * 
   * @param object
   * @param fieldOrMethod
   */
  public void addLogger(final Object object, final String fieldOrMethod) {
    try {
      // try field first
      this.addLogger(object, Utils.getField(object.getClass(), fieldOrMethod), null);
    } catch (SecurityException e) {
      throw new IllegalArgumentException(e);
    } catch (NoSuchFieldException e) {
      try {
        // try method next
        this.addLogger(object, Utils.getMethod(object.getClass(), fieldOrMethod), null);
      } catch (NoSuchMethodException e2) {
        throw new IllegalArgumentException(e2);
      }
    }
  }

  public void removeLogger(final Object object) {
    this.accessibleObjectsToLog.remove(object);
  }

  public void removeLogger(final Object object, final Field field) {
    this.accessibleObjectsToLog.get(object).remove(field);
  }

  public void closeLoggers() {
    for (final Map.Entry<Object, Map<AccessibleObject, PrintWriter>> fieldToLog : this.accessibleObjectsToLog.entrySet()) {
      final Map<AccessibleObject, PrintWriter> fields = fieldToLog.getValue();
      for (final Map.Entry<AccessibleObject, PrintWriter> fieldToPrintWriter : fields.entrySet()) {
        final PrintWriter out = fieldToPrintWriter.getValue();
        if (out != null) {
          out.flush();
          out.close();
        }
      }
    }
  }

  public void addChangeListener(ChangeListener listener) {
    this.changeListeners.add(ChangeListener.class, listener);
  }

  public void removeChangeListener(ChangeListener listener) {
    this.changeListeners.remove(ChangeListener.class, listener);
  }

  protected synchronized void notifyStateChanged(ChangeEvent event) {
    for (final ChangeListener listener : this.changeListeners.getListeners(ChangeListener.class)) {
      listener.stateChanged(event);
    }
  }

  public boolean hasModule(final Module module) {
    return this.modules.contains(module);
  }

  /**
   * Adds a module to this brain. Adds its afferent modules as well as they are
   * also necessarily being updated as time evolves.
   * 
   * @param module Module to add.
   */
  public void addModule(final Module module) {
    // Check presence to circumvent infinity recursion.
    if (this.hasModule(module)) {
      //throw new IllegalArgumentException("Module " + module.getClass().getSimpleName() + " " + module.getName() + " already present.");
    } else {
      this.modules.add(module);

      // Also add afferent and efferent modules as they are also necessary
      // to being updated as time evolves.
      for (final Module afferentModule : module.getAfferentModules()) {
        this.addModule(afferentModule);
      }
    }

    // default operations:
    // this.addLogger(module, "activities");
  }

  /**
   * Removes a module from this brain.
   * 
   * @param module Module to remove.
   */
  public void removeModule(final Module module) {
    // this module
    module.removeAfferentModules();
    module.removeForwardModules();

    // other modules
    for (final Module otherModule : this.getModules()) {
      if (otherModule.hasAfferentModule(module)) {
        otherModule.removeAfferentModule(module);
      }
      if (otherModule.hasForwardModule(module)) {
        otherModule.removeForwardModule(module);
      }
    }

    // brain
    this.modules.remove(module);
    this.removeLogger(module);
  }

  public Module[] getModules(final Class<?> clas) {
    return Brain.getModules(clas, this.modules.toArray(new Module[0]));
  }

  /**
   * Get modules of given class (or subclasses) from given modules.
   * 
   * @param clas Class to filter the modules for.
   * @param modules Modules to search for instances.
   * @return Modules of desired type.
   */
  public static Module[] getModules(final Class<?> clas, final Module[] modules) {
    final ArrayList<Module> returnModules = new ArrayList<Module>();
    for (final Module module : modules) {
      if (clas.isInstance(module)) {
        returnModules.add(module);
      }
    }
    return returnModules.toArray(new Module[0]);
  }

  /**
   * Get modules of brain.
   * 
   * @return modules of brain.
   */
  public Set<Module> getModules() {
    return this.modules;
  }

  /**
   * Returns the timeStepNumber for Online Monitoring Purposes
   * 
   * @return timeStepNumber
   */
  public long getTimeStepNumber() {
    return this.timeStepNumber;
  }

  /**
   * Sets the timeStepNumber member variable.
   * 
   * @param timeStepNumber
   */
  public void setTimeStepNumber(final long timeStepNumber) {
    this.timeStepNumber = timeStepNumber;
  }

  /**
   * Returns the timeDelta for Online Monitoring Purposes
   * 
   * @return timeDelta
   */
  public double getTimeDelta() {
    return this.timeDelta;
  }

  /**
   * Sets the timeDelta member variable.
   * 
   * @param timeDelta
   */
  public void setTimeDelta(final double timeDelta) {
    this.timeDelta = timeDelta;

    // Logging more often than changing time should not be possible.
    this.setLogInterval(Math.max(timeDelta, this.getLogInterval()));
  }

  /**
   * Sets the timeDelta member variable.
   * 
   * @param timeDelta
   */
  public void setTimeDeltaAndDuration(final double timeDelta, final double duration) {
    this.setTimeDelta(timeDelta);
    this.setTimeStepNumber((long) (duration / timeDelta));
  }

  /**
   * Returns the time for Online Monitoring Purposes
   * 
   * @return time
   */
  public double getTime() {
    return this.time;
  }

  /**
   * Loads a module to the system given the concrete constructor and all of its
   * arguments.
   * 
   * @param constructor Constructor of the module class to create the module
   * instance from.
   * @param arguments Arguments for the given constructor.
   */
  public static Module loadModule(final Constructor<?> constructor, final Object[] arguments) throws IllegalArgumentException {
    logger.log(Level.INFO, "loading “" + constructor.getDeclaringClass().getName() + "” …");

    try {
      Object object = constructor.newInstance(arguments);
      return (Module) object;
    } catch (SecurityException e) {
      logger.log(Level.SEVERE, e.toString());
      throw new IllegalArgumentException();
    } catch (InstantiationException e) {
      logger.log(Level.SEVERE, e.toString());
      throw new IllegalArgumentException();
    } catch (IllegalAccessException e) {
      logger.log(Level.SEVERE, e.toString());
      throw new IllegalArgumentException();
    } catch (InvocationTargetException e) {
      logger.log(Level.SEVERE, e.toString());
      throw new IllegalArgumentException();
    }
  }

  /**
   * Sets the time member variable.
   * 
   * @param time Time of the simulation.
   */
  public void setTime(final double time) {
    this.time = time;
  }

  /**
   * Tests whether the simulation is suspended.
   * 
   * @return true if the simulation is suspended, false otherwise.
   */
  public boolean isSuspended() {
    return this.isSuspended;
  }

  /**
   * Suspends the current simulation.
   */
  public void suspend() {
    this.isSuspended = true;
  }

  /**
   * Resumes the current simulation.
   */
  public synchronized void resume() {
    this.isSuspended = false;
    this.notify();
  }

  /**
   * Basic toString method.
   * 
   * @return Returns the string "Brain".
   */
  @Override
  public String toString() {
    return "Brain";
  }

  /**
   * Serializes the current brain (and its modules) to the given file.
   * 
   * The filename has to end with either ".ser" or ".xml". If it ends with
   * ".ser" the writeObject method is used, while if it ends with ".xml" the
   * XStream serialization is used.
   */
  public void save(final File file) {
    Brain.save(this, file);
  }

  /**
   * Serializes the given brain object (and its modules) to the given file.
   * 
   * The filename has to end with either ".ser" or ".xml". If it ends with
   * ".ser" the writeObject method is used, while if it ends with ".xml" the
   * XStream serialization is used.
   * 
   * @param brain Object of class Brain to serialize.
   * @param file File to write the serialization to.
   */
  public static void save(final Brain brain, final File file) {
    try {
      if (file.getName().toLowerCase().endsWith(".ser")) {
        final ObjectOutput out = new ObjectOutputStream(new FileOutputStream(file));
        out.writeObject(brain);
        out.close();
      } else if (file.getName().toLowerCase().endsWith(".xml")) {
        final OutputStream out = new FileOutputStream(file);
        final XStream xstream = new XStream(new DomDriver());
        xstream.toXML(brain, out);
        out.close();
      } else {
        throw new IllegalArgumentException("Wrong file name ending.");
      }
    } catch (IOException exception) {
      exception.printStackTrace();
    }
  }

  /**
   * Deserializes brain (and its modules) from the given file.
   * 
   * The filename has to end with either ".ser" or ".xml". If it ends with
   * ".ser" the readObject method is used, while if it ends with ".xml" the
   * XStream deserialization is used.
   * 
   * @throws ClassNotFoundException
   * @throws IOException
   */
  public static Brain load(final File file) throws ClassNotFoundException, IOException {
    if (file.getName().toLowerCase().endsWith(".ser")) {
      // Java Serialized Objects
      final ObjectInput in = new ObjectInputStream(new FileInputStream(file));
      final Object object = in.readObject();
      in.close();
      return (Brain) object;
    } else if (file.getName().toLowerCase().endsWith(".xml")) {
      // XML objects
      final InputStream in = new FileInputStream(file);
      final XStream xstream = new XStream(new PureJavaReflectionProvider(), new DomDriver());
      // final XStream xstream = new XStream(new DomDriver());
      final Object object = xstream.fromXML(in);
      in.close();
      return (Brain) object;
    } else {
      throw new IllegalArgumentException("File not .ser or .xml.");
    }
  }

  /**
   * Calculates statistics for data analysis.
   */
  public void calculateSimulationStatistics() {
    // end statistics for all dhanModule modules
    for (final Module module : this.modules) {
      if (module instanceof DhanModule) {
        final DhanModule dhan = (DhanModule) module;
        dhan.statOutputToFile(); // local statistics

        final StatNet myStat = new StatNet(module); // module statistics
        // object
        myStat.degreeDistribution();
        // evaluate and print degree distribution
        myStat.cliqueStat();
        // evaluates and prints clique statistics
        myStat.cliqueReceptiveFields();
        // receptive fields with respect input
        myStat.cliqueReferencePatterns();
        // response of cliques with respect
        // to certain given reference patterns

        try {
          module.exportXfigCircle(new File(Brain.DEFAULT_DATA_OUTPUT_DIRECTORY, "Brain_" + module.getName() + ".circle.fig"));
          module.exportXfigLattice(new File(Brain.DEFAULT_DATA_OUTPUT_DIRECTORY, "Brain_" + module.getName() + ".lattice.fig"));
          module.exportGraphvizGraph(new File(Brain.DEFAULT_DATA_OUTPUT_DIRECTORY, "Brain_" + module.getName() + ".gv"));
        } catch (FileNotFoundException e) {
          e.printStackTrace();
        }
      }
    }

    // statistics and output for all biasPattObservers
    for (final StatBiasLearn observer : this.learnBiasObserver.values()) {
      observer.dataOutputToFile();
    }

    // statistics and output for all acitClusObservers
    for (final StatActiClus ss : clusterActivityObserver) {
      ss.evaluateStatClusters();
      ss.evaluateSequencesOfActiveClusters();
      ss.evaluateStatCycles();
    }
  }

  /**
   * Returns getGlobalTimeDelayMillis.
   * 
   * @return Number of milliseconds the program has to sleep in every loop in
   * oder to delay execution of small and fast systems.
   * @see Brain#simulationSleepDelay
   */
  public int getSimulationSleepDelay() {
    return this.simulationSleepDelay;
  }

  /**
   * Sets simulationSleepDelay.
   * 
   * @param simulationSleepDelay - Number of milliseconds the program has to
   * sleep in every loop in oder to delay execution of small and fast systems. *
   * @see Brain#simulationSleepDelay
   */
  public void setSimulationSleepDelay(final int simulationSleepDelay) {
    this.simulationSleepDelay = simulationSleepDelay;
  }

  /**
   * Necessary for (de-)serialization in order to initialize transient fields.
   */
  @SuppressWarnings("unchecked")
  private void readObject(final ObjectInputStream in) throws IOException, ClassNotFoundException {
    // First use the default readObject method. Everything but transient
    // fields will be deserialized.
    in.defaultReadObject();

    // Initialize all the transient fields.
    this.changeListeners = new EventListenerList();
    this.accessibleObjectsToLog = new HashMap<Object, Map<AccessibleObject, PrintWriter>>();

    // Deserialize the fields to log.
    final Map<String, List<String>> loggers = (HashMap<String, List<String>>) in.readObject();
    for (final Map.Entry<String, List<String>> value : loggers.entrySet()) {
      final String moduleName = value.getKey();
      final List<String> fieldNames = value.getValue();
      for (final String fieldName : fieldNames) {
        Module module = null;
        for (final Module candidate : this.getModules()) {
          if (moduleName.equals(candidate.getName())) {
            module = candidate;
            break;
          }
        }
        this.addLogger(module, fieldName);
      }
    }
  }

  /**
   * Necessary for (de-)serialization in order to initialize transient fields.
   */
  private void writeObject(final ObjectOutputStream out) throws IOException {
    // First use the default writeObject method. Everything but transient
    // fields will be serialized.
    out.defaultWriteObject();

    // Serialize all the transient fields.
    this.changeListeners = new EventListenerList();

    // Serialize which fields to log.
    final Map<String, List<String>> loggers = new HashMap<String, List<String>>();
    for (final Map.Entry<Object, Map<AccessibleObject, PrintWriter>> accessibleObjectToLog : this.accessibleObjectsToLog.entrySet()) {
      final Object object = accessibleObjectToLog.getKey();

      // Consider only modules.
      if (!(object instanceof Module)) {
        continue;
      }

      final Module module = (Module) object;

      final Map<AccessibleObject, PrintWriter> accessibleObjectsToPrintWriter = accessibleObjectToLog.getValue();
      final List<String> fieldList = new Vector<String>();
      for (final Map.Entry<AccessibleObject, PrintWriter> accessibleObjectToPrintWriter : accessibleObjectsToPrintWriter.entrySet()) {
        final AccessibleObject accessibleObject = accessibleObjectToPrintWriter.getKey();

        final String accessibleObjectName;
        if (accessibleObject instanceof Field) {
          accessibleObjectName = ((Field) accessibleObject).getName();
        } else if (accessibleObject instanceof Method) {
          accessibleObjectName = ((Method) accessibleObject).getName();
        } else {
          accessibleObjectName = null;
        }

        fieldList.add(accessibleObjectName);
      }

      if (!fieldList.isEmpty()) {
        loggers.put(module.getName(), fieldList);
      }
    }
    out.writeObject(loggers);
  }

  public void exportGraphvizGraph(final File gvFile) throws FileNotFoundException {
    this.exportGraphvizGraph(gvFile, Double.NEGATIVE_INFINITY, Double.POSITIVE_INFINITY);
  }

  public void exportGraphvizGraph(final File gvFile, final double lowerLinkBound, final double upperLinkBound) throws FileNotFoundException {
    final PrintWriter dotOut = new PrintWriter(gvFile);
    dotOut.println("digraph G {");
    dotOut.println("graph [bgcolor=transparent]");

    // module vertices
    for (final Module module : this.getModules()) {
      module.writeGraphvizSubgraph(dotOut, lowerLinkBound, upperLinkBound);
    }

    // inter module edges
    for (final Module module : this.getModules()) {
      for (final Module afferentModule : module.getAfferentModules()) {
        for (int i = 0; i < module.getSize(); i++) {
          for (int j = 0; j < afferentModule.getSize(); j++) {
            final double w_ij = module.getAfferentLink(afferentModule, i, j);
            final String color = w_ij >= 0 ? "#008000" : "#800000";
            if (w_ij != 0.0 && w_ij >= lowerLinkBound && w_ij <= upperLinkBound) {
              // do print only links with strength not zero, viz.
              // really existing links
              final String iNodeName = module.getName() + "_" + i;
              final String jNodeName = afferentModule.getName() + "_" + j;
              dotOut.println("\"" + jNodeName + "\" -> \"" + iNodeName + "\" [label=\"" + w_ij + "\" color=\"" + color + "\" fontcolor=\"" + color + "\"];");
            }
          }
        }
      }
    }

    dotOut.println("}");
    dotOut.close();
  }
}
