package brain;

import java.io.ByteArrayOutputStream;
import java.io.File;
import java.io.FileNotFoundException;
import java.io.OutputStreamWriter;
import java.io.PrintWriter;
import java.io.Serializable;
import java.util.Arrays;
import java.util.Collections;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Locale;
import java.util.Map;
import java.util.Random;
import java.util.Scanner;
import java.util.Set;
import java.util.Vector;

import utils.Utils;

/**
 * The DhanModule module of the cognitive system (CS).
 * 
 * All parameters for time development of activities and reservoirs are
 * generally public. Exceptions are parameters which need additional computation
 * or checks when changed.
 *
 * @author Claudius Gros
 * @author Mathias Linkerhand
 * @version 2011-11-08
 */
public class DhanModule extends Module implements Serializable {
  /** Random serial number, used for serialization. */
  static final long       serialVersionUID                 = -8719357346526504937L;

  // for the reservoir function f_w^{post}(phi)
  public final static int RES_FW_POST                      = 0;
  // for the reservoir function f_w^{pre}(phi)
  public final static int RES_FW_PRE                       = 1;
  // for the reservoir function f_z(phi)
  public final static int RES_FZ_PRE                       = 2;
  // for the reservoir function f_z(phi)
  public final static int RES_GZ_PRE                       = 3;

  // excitatory links - long-term memory
  public final static int LINK_EXCITATORY_LONGTERM         = 0;
  // inhibitory links
  public final static int LINK_INHIBITORY_LONGTERM         = 1;
  // excitatory links - short-term memory
  public final static int LINK_EXITATORY_SHORTTERM         = 2;

  /** Default value for nearly active reservoir/activities level. */
  public static double    nearlyActive                     = 1 - 1E-3;
  /** Default value for nearly empty reservoir/activities level. */
  public static double    nearlyInactive                   = 1E-7;

  /** Array with the total growth-rates. */
  private double          rates[];
  /** Array with the integral contribution to the growth-rates. */
  private double          ratesInternal[];
  /** Array with the reservoirs. */
  private double          reservoir[];
  /**
   * Array with the input through the backward links, viz. the input received from
   * other cognitiveSystems by active presynaptic neurons.
   */
  private double          afferentRates[];
  /**
   * Array with the inverse input through the backward links, viz. the input
   * received from other cognitiveSystems by inactive presynaptic neurons.
   */
  private double          afferentInverseRates[];

  /**
   * Matrix with the internal synaptic weights of the ACs.
   * 
   * [0][.][.] : excitatory; long-term
   * [1][.][.] : inhibitory
   * [2][.][.] : excitatory; short-term
   */
  private double          internalLinks[][][];
  /**
   * The number of different link types in internalLinks[INTERNAL_LINK_TYPES][][].
   */
  public final static int INTERNAL_LINK_TYPES_NUMBER       = 3;

  /**
   * Has a large value when the current clique is activated due to the influence
   * of the incoming signal due to the backward connections.
   */
  public double           diffusiveModuleSignal            = 0.0;
  /**
   * An effective measure (in [0,1]) of clique formation strength, interpreted as
   * poor-man's synchronization.
   */
  public double           synchronizationStrength          = 0.0;

  /** Standard strength of inhibitory links */
  public double           defaultInhibitoryLinkStrength;
  /** Standard strength of excitatory links */
  public double           defaultExcitatoryLinkStrength;

  /**
   * The total inhibitory contribution to the growth rates are normalized, with a
   * width `rateInhiWidth' for the sigmoidal.
   */
  public double           rateInhiWidth;

  /** Number of different parameter sets for the reservoir functions. */
  final public int        nResFunctions                    = 4;
  /** The reservoir depletes when the activities is bigger than `phiXActive'. */
  public double           phiXActive;
  /** The reservoir growths when the activities is smaller than `phiXInactive'. */
  public double           phiXInactive;
  /** Depletion-rate for the inhibitory reservoir. */
  public double           phiGammaDown;
  /** Growth-rate for the inhibitory reservoir. */
  public double           phiGammaUp;
  /** Averaging-rate for the inhibitory reservoirs of the winning clique. */
  public double           phiGammaCoh;
  /**
   * Critical phi for the reservoir functionsf_w() and f_z() - public because used
   * indirectly.
   */
  private double[]        phiCritical                      = new double[nResFunctions];
  /**
   * Width of the reservoir functions f_w() and f_z() - public because used
   * indirectly.
   */
  private double[]        phiWidth                         = new double[nResFunctions];
  /**
   * Minimal value for the reservoir functions f_w() and f_z() - public because
   * used indirectly.
   */
  private double[]        phiMinValue                      = new double[nResFunctions];
  //
  // --------- parameters for the short-term memory ----------- //
  //
  /** Depletion-rate for short-term synaptic plasticity. */
  public double           shortTermMemGammaDown;
  /** Activation-rate for short-term synaptic plasticity. */
  public double           shortTermMemGammaUp;
  /** Maximal value for the short-term synaptic strength. */
  public double           shortTermMemMaxStrength;
  /** Critical activities strength for short-term memory to be activated. */
  public double           shortTermMemCriticalX;
  /** Minimal external stimulus-strength for short-term memory to be activated. */
  public double           shortTermMemMinExtBias;
  //
  /* ========= parameters for the long-term memory ========= */
  //
  /** Optimization-rate (long-term) for excitatory links */
  public double           longTermMemGammaOpt;
  /** Mininal values for excitatory links */
  public double           longTermMemMinStrength;
  /**
   * Optimal growth rate -- central parameter for the working-point optimization.
   */
  public double           longTermMemOptRate;
  /** Critical presynaptic activities for long-term memory dynamics. */
  public double           longTermMemCriticalX;
  //
  /* ========== diffusive module control signals ========== */
  //
  /** Activation rate for the diffusive module (learning) signal. */
  public double           diffusiveModuleSignalRateUp;
  /** Deactivation rate for the diffusive module (learning) signal. */
  public double           diffusiveModuleSignalRateDown;
  /**
   * When above this critical value, the diffusive (learning) signal activates
   * generalized Hebbian learning for the backward links.
   */
  public double           diffusiveModuleSignalCriticalValue;
  /**
   * When the backrate is above this threshold, the diffusive (learning) signal is
   * activated.
   */
  public double           diffusiveModuleSignalBackRateThreshold;

  /**
   * Critical activities level for a module-neuron to be considered active.
   */
  public double           backwardMemXActive;
  /**
   * Critical activities level for a module-neuron to be considered inactive.
   */
  public double           backwardMemXInactive;
  /** Rate for active-to-active backward links to optimize. */
  public double           backwardMemGammaActAct;
  /** Rate for inactive-to-active backward links to optimize. */
  public double           backwardMemGammaInaAct;
  /** Rate for active-to-inactive backward links to optimize. */
  public double           backwardMemGammaActIna;
  /** Rate for inactive-to-inactive backward links to optimize. */
  public double           backwardMemGammaInaIna;
  /** Target growth-rate-input for active-to-active backward input. */
  public double           backwardMemOptRateActAct;
  /** Target growth-rate-input for inactive-to-active backward input. */
  public double           backwardMemOptRateInaAct;
  /** Target growth-rate-input for active-to-inactive backward input. */
  public double           backwardMemOptRateActIna;
  /** Target growth-rate-input for inactive-to-inactive backward input. */
  public double           backwardMemOptRateInaIna;
  /** The mean value of all backward links reaching a given site. */
  public double           backwardMemMean[];
  /** The standard deviation of all backward links reaching a given site. */
  public double           backwardMemStandDev[];
  //
  /* ========== local variables for long-term memory ========== */
  //
  /**
   * Growth rates without f_w(phi) - internal variable.<br>
   * 
   * Without the influence of the postsynaptic reservoir. For use for the
   * working-point optimization.
   */
  private double          ratesActive[];
  /**
   * Positive contribution to the growth-rate - internal variable.
   * 
   * Without the influence of the postsynaptic reservoir. For use for the
   * working-point optimization.
   */
  private double          ratesPositive[];
  /**
   * Negative contribution to the growth-rate - internal variable.
   * 
   * For use for the working-point optimization.
   */
  private double          ratesNegative[];

  /* ========== local variables for reservoir function ========== */
  /**
   * Inverse of the reservoir-function width - internal variable.
   * 
   * For use for the reservoir functions.
   */
  public double[]         phiInverseWidth                  = new double[nResFunctions];
  /**
   * Auxiliary variable for a fast evaluation of the reservoir function.
   */
  public double[]         phiAuxShift                      = new double[nResFunctions];
  /**
   * Auxiliary variable for a fast evaluation of the reservoir function.
   */
  public double[]         phiAuxFactor                     = new double[nResFunctions];

  /* ========== auxiliary variables, not directly needed for the CS ========== */
  /** The number of initial random clusters (all identical). */
  public int              auxInitialCliqueNumber;
  /** The size of initial random clusters (all identical). */
  public int              auxInitialCliqueSize;
  /**
   * Determines the file names containing the associations.
   * <p>
   */
  public String[]         auxAssInputFileNameAllowedValues = null;

  public DhanModule() throws FileNotFoundException {
    this(new File(Brain.DEFAULT_DATA_INPUT_DIRECTORY, "dhan.cliques"));
    this.setName("dhan");
  }

  public DhanModule(final File cliquesFile) throws FileNotFoundException {
    this.reset();

    this.setName(cliquesFile.getName());

    // set excitatory internal links
    this.readCliques(cliquesFile);
  }

  @Override
  public void reset() {
    this.setName("dhan");
    final int size = 16;

    // initialize activities
    for (int i = 0; i < size; i++) {
      this.addNeuron(DhanModule.nearlyInactive);
    }

    // initialize reservoirs
    this.reservoir = new double[size];
    final Random random = new Random(0);
    Utils.randomFill(this.reservoir, random, 0.8, 0.804);

    this.rates = new double[size];
    this.ratesInternal = new double[size];
    Collections.fill(this.biases, 0.0);
    this.afferentRates = new double[size];
    this.afferentInverseRates = new double[size];
    this.internalLinks = new double[INTERNAL_LINK_TYPES_NUMBER][size][size];
    this.ratesActive = new double[size];
    this.ratesPositive = new double[size];
    this.ratesNegative = new double[size];
    this.backwardMemMean = new double[size];
    this.backwardMemStandDev = new double[size];

    // needed for default, overwritten in setDefaultParameter() ---
    Arrays.fill(this.phiWidth, 0.1);
    Arrays.fill(this.phiCritical, 0.5);
    Arrays.fill(this.phiMinValue, 0.5);

    // the DhanModule diffusive learning signal is a novelty signal and modulates the learning of the backward links
    this.addForwardModule(this, "backwardLearning");

    // receptor fields test consistency with allowed values
    this.addReceptor("backwardLearning", 0);
    this.addReceptor("default", 0);

    // how to select the internal links at the start
    this.defaultInhibitoryLinkStrength = -1; // strength of inhibitory links
    this.defaultExcitatoryLinkStrength = 0.18; // strength of excitatory links (0.12)
    this.auxInitialCliqueNumber = 40; // number of initial random cliques
    this.auxInitialCliqueSize = 2; // initial random cliques-size

    // parameters for the rates
    this.rateInhiWidth = 0.5;

    // width for the normalization of inhibitory contribution [0.5]
    // parameters for the reservoir
    this.phiXActive = 0.900; // activities level for reservoir to deplete
    this.phiXInactive = 0.100; // activities level for reservoir to growth
    this.phiGammaDown = 0.0300; // depletion-rate (0.030)
    this.phiGammaUp = 0.0100; // growth-rate (0.010)
    this.phiGammaCoh = 0.0500; // coherence: averaging (0.050)

    this.setPhiCritical(0.700, DhanModule.RES_FW_POST); // critical phi for f_w^{post}()
    this.setPhiCritical(0.500, DhanModule.RES_FW_PRE); // critical phi for f_w^{pre}()
    this.setPhiCritical(0.300, DhanModule.RES_FZ_PRE); // critical value f_z()
    this.setPhiCritical(0.700, DhanModule.RES_GZ_PRE); // critical value g_z()

    this.setPhiWidth(0.100, DhanModule.RES_FW_POST); // width of f_w^{post}()
    this.setPhiWidth(0.100, DhanModule.RES_FW_PRE); // width of f_w^{pre}()
    this.setPhiWidth(0.100, DhanModule.RES_FZ_PRE); // width of f_z()
    this.setPhiWidth(0.100, DhanModule.RES_GZ_PRE); // width of g_z()

    this.setPhiMinValue(0.100, DhanModule.RES_FW_POST); // minimal values for f_w^{post}() [0.1]
    this.setPhiMinValue(0.100, DhanModule.RES_FW_PRE); // minimal values for f_w^{pre}()  [0.15]
    this.setPhiMinValue(0.000, DhanModule.RES_FZ_PRE); // minimal values for f_z() [0.0]
    this.setPhiMinValue(0.000, DhanModule.RES_GZ_PRE); // minimal values for f_z() [0.0]

    // parameters for the short-term memory
    this.shortTermMemGammaUp = 0.40; // growth-rate (0.1)
    this.shortTermMemGammaDown = 0.005;// depletion-rate
    this.shortTermMemMaxStrength = 0.06; // maximal value
    this.shortTermMemCriticalX = 0.35; // critical activities
    this.shortTermMemMinExtBias = 0.10; // minimal stimulus strength

    // parameters for the long-term memory
    this.longTermMemGammaOpt = 0.01; // rate for changing (0.01)
    this.longTermMemMinStrength = -0.030; // minimal value
    this.longTermMemCriticalX = 0.8; // critical presynaptic activities
    this.longTermMemOptRate = 0.300; // optimal growth rate [0.3]

    // parameters for the diffusive control signals
    this.diffusiveModuleSignalRateUp = 4.0; // activities rate
    this.diffusiveModuleSignalRateDown = 0.15; // deactivation rate
    this.diffusiveModuleSignalCriticalValue = 0.25; // threshold for learning
    this.diffusiveModuleSignalBackRateThreshold = 0.15; // backRates threshold

    // parameters for the backward links
    this.backwardMemGammaActAct = 0.00200; // act-to-act optimization
    this.backwardMemGammaInaAct = 0.00100; // inact-to-act optimization
    this.backwardMemGammaActIna = 0.00100; // act-to-inact optim. (0.0020)
    this.backwardMemGammaInaIna = 0.0000; // inact-to-inact optim.

    this.backwardMemOptRateActAct = 0.80; // act-to-act input rate
    this.backwardMemOptRateInaAct = -0.20; // inact-to-act inverse-input rate
    this.backwardMemOptRateActIna = 0.20; // act-to-inact input rate
    this.backwardMemOptRateInaIna = 0.00; // inact-to-inact inverse input rate

    this.backwardMemXActive = 0.4; // critical-x for active
    this.backwardMemXInactive = 0.2; // critical-x for inactive

    // set the sizes of the arrays
    this.internalLinks = new double[DhanModule.INTERNAL_LINK_TYPES_NUMBER][size][size];

    // set non excitatory parts of link matrices
    for (int i = 0; i < size; i++) {
      for (int j = 0; j < size; j++) {
        if (this.getLink(i, j, DhanModule.LINK_EXCITATORY_LONGTERM) > 0.0) {
          this.setLink(i, j, DhanModule.LINK_INHIBITORY_LONGTERM, 0.0);
          this.setLink(i, j, DhanModule.LINK_EXITATORY_SHORTTERM, 0.0);
        } else {
          this.setLink(i, j, DhanModule.LINK_EXCITATORY_LONGTERM, this.longTermMemMinStrength);
          this.setLink(i, j, DhanModule.LINK_INHIBITORY_LONGTERM, this.defaultInhibitoryLinkStrength);
          this.setLink(i, j, DhanModule.LINK_EXITATORY_SHORTTERM, 0.0);
        }
      }
    }
    for (int i = 0; i < size; i++) {
      this.setLink(i, i, DhanModule.LINK_EXCITATORY_LONGTERM, 0.0);
      this.setLink(i, i, DhanModule.LINK_INHIBITORY_LONGTERM, 0.0);
      this.setLink(i, i, DhanModule.LINK_EXITATORY_SHORTTERM, 0.0);
    }

    // start with a finite number of
    final int activeNodesNumber = 1; // number of active initial sites
    for (int node = 0; node < activeNodesNumber; node++) {
      this.setActivity(node, 0.8 + 0.002 * random.nextDouble());
    }

    // initialize standard deviation of backward links
    Arrays.fill(this.backwardMemMean, 0); // just to have a value
    Arrays.fill(this.backwardMemStandDev, 0.001); // just to avoid floating-point errors
  }

  /**
   * Sets the number of nodes.
   */
  @Override
  public void setSize(final int size) {
    if (size <= 0) {
      throw new IllegalArgumentException("size has to be positive");
    }

    final int oldSize = this.getSize();

    super.setSize(size);

    // update array sizes when size has changed. retain as much prior information as possible
    while (this.getSize() < size) {
      this.addNeuron(DhanModule.nearlyInactive);
    }
    while (this.getSize() > size) {
      this.removeNeuron(this.getSize() - 1);
    }

    this.afferentInverseRates = Arrays.copyOf(this.afferentInverseRates, size);
    this.afferentRates = Arrays.copyOf(this.afferentRates, size);
    this.backwardMemMean = Arrays.copyOf(this.backwardMemMean, size);
    this.backwardMemStandDev = Arrays.copyOf(this.backwardMemStandDev, size);
    this.rates = Arrays.copyOf(this.rates, size);
    this.ratesActive = Arrays.copyOf(this.ratesActive, size);
    this.ratesInternal = Arrays.copyOf(this.ratesInternal, size);
    this.ratesNegative = Arrays.copyOf(this.ratesNegative, size);
    this.ratesPositive = Arrays.copyOf(this.ratesPositive, size);
    this.reservoir = Arrays.copyOf(this.reservoir, size);

    for (int t = 0; t < this.internalLinks.length; t++) {
      this.internalLinks[t] = Utils.copyOf(this.internalLinks[t], size, size);
      if (size > oldSize) {
        final double value;
        switch (t) {
          case 0:
            value = this.longTermMemMinStrength;
            break;
          case 1:
            value = this.defaultInhibitoryLinkStrength;
            break;
          default:
            value = 0;
            break;
        }
        for (int i = 0; i < size; i++) {
          for (int j = oldSize; j < size; j++) {
            this.setLink(i, j, t, value);
            this.setLink(j, i, t, value);
          }
        }
      }
    }
  }

  /** Returns the reservoir level of a given center. */
  public double getReservoir(int i) {
    return this.reservoir[i];
  }

  /** Returns the value of a specific internal link. */
  @Override
  public double getLink(int i, int j) {
    return this.getLink(i, j, 0);
  }

  /** Returns the value of a specific internal link. */
  public double getLink(int i, int j, int t) {
    return this.internalLinks[t][i][j];
  }

  /** Sets the value of a specific internal link. */
  @Override
  public void setLink(int i, int j, double value) {
    this.setLink(i, j, 0, value);
  }

  /** Sets the value of a specific internal link. */
  public void setLink(int i, int j, int t, double value) {
    if (i < 0 || i > this.getSize() - 1) {
      throw new IllegalArgumentException("input `i' incorrect");
    }
    if (j < 0 || j > this.getSize() - 1) {
      throw new IllegalArgumentException("input `j' incorrect");
    }
    if (t < 0 || t > DhanModule.INTERNAL_LINK_TYPES_NUMBER - 1) {
      throw new IllegalArgumentException("input `t' incorrect");
    }

    this.internalLinks[t][i][j] = value;

    super.setLink(i, j, value);
  }

  public void addInternalLink(int i, int j, int t, double value) {
    this.internalLinks[t][i][j] += value;
  }

  /** Returns critical reservoir levels for the reservoir functions. */
  public double[] getPhiCritical() {
    return this.phiCritical;
  }

  /**
   * Sets critical reservoir level for the reservoir functions.
   * 
   * The auxiliary variables <br>
   * `inversePhiWidth' and <br>
   * `inDiffPhi' and `constPoint' <br>
   * are newly evaluated at the same time.
   */
  public void setPhiCritical(double phiCritical_in, int whichOne) {
    assert ((phiCritical_in > 0) && (phiCritical_in < 1)) : "DhanModule.setPhiCritical - wrong phiCritical ";
    assert ((whichOne >= 0) && (whichOne < nResFunctions)) : "DhanModule.setCriticalPhi - wrong whichOne ";
    this.phiCritical[whichOne] = phiCritical_in;
    //
    double zeroPoint = Math.tanh(phiInverseWidth[whichOne] * (0.0 - this.phiCritical[whichOne]));
    double onePoint = Math.tanh(phiInverseWidth[whichOne] * (1.0 - this.phiCritical[whichOne]));
    this.phiAuxFactor[whichOne] = (1.0 - phiMinValue[whichOne]) / (onePoint - zeroPoint);
    this.phiAuxShift[whichOne] = zeroPoint * phiAuxFactor[whichOne] - phiMinValue[whichOne];
  }

  public void setPhiMinValue(double[] _PhiMinValue) {
    this.phiMinValue = _PhiMinValue;
  }

  public void setPhiWidth(double[] _PhiWidth) {
    this.phiWidth = _PhiWidth;
  }

  public void setPhiCritical(double[] _PhiCritical) {
    this.phiCritical = _PhiCritical;
  }

  /** Returns the reservoir function width. */
  public double[] getPhiWidth() {
    return this.phiWidth;
  }

  /**
   * Sets the reservoir function width.
   * 
   * The auxiliary variables <br>
   * `inversePhiWidth' and <br>
   * `inDiffPhi' and `constPoint' <br>
   * are newly evaluated at the same time.
   */
  public void setPhiWidth(final double argument, final int reservoirFunction) {
    if (argument <= 0) {
      throw new IllegalArgumentException("DhanModule.setPhiWidth - invalid phiWidth");
    }
    if (reservoirFunction < 0 || reservoirFunction >= this.nResFunctions) {
      throw new IllegalArgumentException("DhanModule.setPhiWidth - wrong whichOne");
    }
    this.phiWidth[reservoirFunction] = argument;
    this.phiInverseWidth[reservoirFunction] = 1.0 / argument;

    final double zeroPoint = Math.tanh(this.phiInverseWidth[reservoirFunction] * (0.0 - this.phiCritical[reservoirFunction]));
    final double onePoint = Math.tanh(this.phiInverseWidth[reservoirFunction] * (1.0 - this.phiCritical[reservoirFunction]));

    this.phiAuxFactor[reservoirFunction] = (1.0 - this.phiMinValue[reservoirFunction]) / (onePoint - zeroPoint);
    this.phiAuxShift[reservoirFunction] = zeroPoint * this.phiAuxFactor[reservoirFunction] - this.phiMinValue[reservoirFunction];
  }

  /**
   * Returns the minimal value for the reservoir functions.
   * 
   * @return result of the reservoir function
   */
  public double[] getPhiMinValue() {
    return this.phiMinValue;
  }

  /**
   * Sets the minimal value for the reservoir functions.
   * 
   * The auxiliary variables <br>
   * `inversePhiWidth' and <br>
   * `inDiffPhi' and `constPoint' <br>
   * are newly evaluated at the same time.
   * 
   * @param argument -
   *          the minimal value for phi
   * @param reservoirFunction -
   *          the reservoir function
   */
  public void setPhiMinValue(final double argument, final int reservoirFunction) {
    if (argument < 0 || argument >= 1) {
      throw new IllegalArgumentException("DhanModule.setPhiWidth - invalid phiWidth");
    }
    if (reservoirFunction < 0 || reservoirFunction >= this.nResFunctions) {
      throw new IllegalArgumentException("DhanModule.setPhiWidth - wrong whichOne");
    }

    this.phiMinValue[reservoirFunction] = argument;

    final double zeroPoint = Math.tanh(phiInverseWidth[reservoirFunction] * (0.0 - this.phiCritical[reservoirFunction]));
    final double onePoint = Math.tanh(phiInverseWidth[reservoirFunction] * (1.0 - this.phiCritical[reservoirFunction]));

    this.phiAuxFactor[reservoirFunction] = (1.0 - phiMinValue[reservoirFunction]) / (onePoint - zeroPoint);
    this.phiAuxShift[reservoirFunction] = zeroPoint * phiAuxFactor[reservoirFunction] - phiMinValue[reservoirFunction];
  }

  /**
   * The reservoir functions.
   * 
   * For a fast evaluation it uses the auxiliary variables <br>
   * `phiInverseWidth[]', `phiCritical[]' <br>
   * `phiAuxFactor[]' and `phiAuxShift[]'.
   * 
   * @param argument
   * @param reservoirFunction -
   *          whichOne: 0 =^= f_w^{post}(phi) <br>
   *          whichOne: 1 =^= f_w^{pre}(phi) <br>
   *          whichOne: 2 =^= f_z(phi) <br>
   * @return result of the reservoir function
   */
  public double reservoirFunctions(final double argument, final int reservoirFunction) {
    assert ((reservoirFunction >= 0) && (reservoirFunction < nResFunctions)) : "DhanModule.reservoirFunctions - wrong whichOne ";
    assert ((argument > 0.0) && (argument < 1.0)) : "DhanModule.reservoirFunctions - reservoir out-off range ";
    return this.phiAuxFactor[reservoirFunction] * Math.tanh(this.phiInverseWidth[reservoirFunction] * (argument - this.phiCritical[reservoirFunction]))
        - this.phiAuxShift[reservoirFunction];
  }

  public static List<List<String>> loadCliques(final File file) throws FileNotFoundException {
    final List<List<String>> cliques = new Vector<List<String>>();
    final Scanner scanner = new Scanner(file);

    while (scanner.hasNextLine()) {
      final String line = scanner.nextLine().trim();
      if (line.isEmpty()) {
        // empty line found
        continue;
      }
      if (line.charAt(0) == '#') {
        // comment found
        continue;
      }

      // read clique
      String[] tokens = line.split("\\s+"); // split with whitespaces
      if (Integer.parseInt(tokens[0], 10) != tokens.length - 1) {
        // The first number states the numbers of following numbers in the line.
        throw new IllegalArgumentException("format error in file " + file.getPath());
      }
      // The first index is the number of vertices of the clique. So it is not a vertex. So remove it.
      tokens = Arrays.copyOfRange(tokens, 1, tokens.length);

      final List<String> clique = new Vector<String>(Arrays.asList(tokens));

      cliques.add(clique);
    }
    return cliques;
  }

  public static List<List<Integer>> cliquesToInteger(final List<List<String>> cliques) {
    final List<List<Integer>> integerCliques = new Vector<List<Integer>>(cliques.size());
    for (final List<String> clique : cliques) {
      final Vector<Integer> integerClique = new Vector<Integer>(clique.size());
      for (final String node : clique) {
        integerClique.add(Integer.parseInt(node, 10));
      }
      integerCliques.add(integerClique);
    }
    return integerCliques;
  }

  public static <Type> Map<Type, Set<Type>> cliquesToNetwork(final List<List<Type>> cliques) {
    final Map<Type, Set<Type>> edges = new HashMap<Type, Set<Type>>();

    // create associations for all vertices in clique
    for (final List<Type> clique : cliques) {
      for (final Type node1 : clique) {
        for (final Type node2 : clique) {
          if (node1 == node2) {
            // do not create self-loop
            continue;
          }
          if (!edges.containsKey(node1)) {
            edges.put(node1, new HashSet<Type>());
          }
          edges.get(node1).add(node2);
        }
      }
    }

    return edges;
  }

  /**
   * Reads associations in the form of cliques from file.
   * 
   * <p>
   * --class DhanModule-- <br>
   * The link strength might be either set by hand, or from input. In both cases
   * the link strength might be weighted be the number contributing cliques. <br>
   * Normally the number of neurons is determined automatically on input and set.
   * <br>
   */
  public void readCliques(final File file) throws FileNotFoundException {
    double meanCliqueSize = 0.0; // mean number of sites per clique
    int cliqueNumber = 0; // number of input cliques
    final double linkStrength = this.defaultExcitatoryLinkStrength;
    double meanLinkStrength = 0.0; // average input link strength

    final List<List<Integer>> cliques = DhanModule.cliquesToInteger(DhanModule.loadCliques(file));
    final Map<Integer, Set<Integer>> links = DhanModule.cliquesToNetwork(cliques);
    this.setSize(links.size());

    final Random random = new Random(0);

    for (final List<Integer> clique : cliques) {
      // statistics
      cliqueNumber += 1;
      meanCliqueSize += clique.size();
      meanLinkStrength += linkStrength;

      for (final int node1 : clique) {
        for (final int node2 : clique) {
          if (node1 == node2) {
            // do not create self-loop
            continue;
          }

          // add minimal random component to avoid exact degeneracies
          // average when same links is added more than once
          double offset = linkStrength * (0.9 + 0.1 * random.nextDouble());
          if (this.getLink(node1, node2, DhanModule.LINK_EXCITATORY_LONGTERM) > 0) {
            offset = 0.5 * (offset + this.getLink(node1, node2, DhanModule.LINK_EXCITATORY_LONGTERM));
          }
          //TODO: check randomization in link strengths
          this.setLink(node1, node2, DhanModule.LINK_EXCITATORY_LONGTERM, linkStrength);
        }
      }
    }

    // normalize mean clique size by number of cliques
    meanCliqueSize = meanCliqueSize / cliqueNumber;
    // normalize mean link strength by number of cliques
    meanLinkStrength = meanLinkStrength / cliqueNumber;

    // set inhibitory links
    for (int i = 0; i < this.getSize(); i++) {
      for (int j = 0; j < this.getSize(); j++) {
        // link is excitatory?
        if (this.getLink(i, j, DhanModule.LINK_EXCITATORY_LONGTERM) > 0.0) {
          // delete link inhibition
          this.setLink(i, j, DhanModule.LINK_INHIBITORY_LONGTERM, 0.0);
        }
      }
    }
  }

  /**
   * A single time step of the dynamics.
   * 
   * <p>
   * --class DhanModule-- <br>
   * Uses the global variables time, deltaTime defined in Brain.
   */
  @Override
  public synchronized void calculateOneTimeStep(final double timeDelta) {
    // evaluate the input received via afferent links, looping over all afferent cognitiveSystems
    this.afferentRates = new double[this.getSize()];
    this.afferentInverseRates = new double[this.getSize()];
    for (final Module module : this.getAfferentModules()) {
      for (int i = 0; i < this.getSize(); i++) {
        for (int j = 0; j < module.getSize(); j++) {
          this.afferentRates[i] += this.getAfferentLink(module, i, j) * module.getActivity(j);
          this.afferentInverseRates[i] += this.getAfferentLink(module, i, j) * (1.0 - module.getActivity(j));
        }
      }
    }

    // activate memories before the external input (the bias) is manipulated
    //this.shortTermMemory(timeDelta);
    this.longTermMemory(timeDelta);
    this.diffusiveControlSignals(timeDelta);
    this.afferentMemory(timeDelta);

    // the bias corresponds to stimuli multiply with coupling function
    for (int i = 0; i < this.getSize(); i++) {
      this.setBias(i, this.getBias(i) * this.reservoirFunctions(this.reservoir[i], DhanModule.RES_FW_PRE));
    }

    // update reservoir
    for (int i = 0; i < this.getSize(); i++) {
      if (this.getActivity(i) > this.phiXActive) {
        this.reservoir[i] -= this.reservoir[i] * this.phiGammaDown * timeDelta;
      }
      if (this.getActivity(i) < this.phiXInactive) {
        this.reservoir[i] += (1 - this.reservoir[i]) * this.phiGammaUp * timeDelta;
      }
    }

    // synchronization of reservoir
    double phiActive = 0.0;
    int nActive = 0;
    for (int i = 0; i < this.getSize(); i++) {
      if (this.getActivity(i) > this.phiXActive) {
        nActive++;
        phiActive += this.reservoir[i];
      }
    }
    for (int i = 0; i < this.getSize(); i++) {
      if (this.getActivity(i) > this.phiXActive) {
        this.reservoir[i] += this.phiGammaCoh * (phiActive - this.reservoir[i] * nActive) * timeDelta;
      }
    }

    // effective activities
    final double[] actiTimesGzFz = new double[this.getSize()];
    // activities * (incoming inhibition)
    final double[] actiTimesFw = new double[this.getSize()];
    // activities * (incoming excitation)
    for (int j = 0; j < this.getSize(); j++) {
      actiTimesFw[j] = this.getActivity(j) * this.reservoirFunctions(this.reservoir[j], DhanModule.RES_FW_PRE);
      actiTimesGzFz[j] = this.getActivity(j) * this.reservoirFunctions(this.getActivity(j), DhanModule.RES_GZ_PRE)
          * this.reservoirFunctions(this.reservoir[j], DhanModule.RES_FZ_PRE);
    }

    // calculate internal contribution to growth rates
    for (int i = 0; i < this.getSize(); i++) {
      double positiveContribution = 0.0; // positive contribution to growth-rate
      double negativeContribution = 0.0; // negative contribution to growth-rate
      double totalPositive = 0.0;

      for (int j = 0; j < this.getSize(); j++) {
        totalPositive = this.getLink(i, j, DhanModule.LINK_EXCITATORY_LONGTERM) + this.getLink(i, j, DhanModule.LINK_EXITATORY_SHORTTERM);
        if (totalPositive > 0) {
          // excitation
          positiveContribution += totalPositive * actiTimesFw[j];
        } else {
          // inhibition
          negativeContribution += this.getLink(i, j, DhanModule.LINK_INHIBITORY_LONGTERM) * actiTimesGzFz[j];
        }
      }

      // normalize inhibitory contribution
      negativeContribution = this.defaultInhibitoryLinkStrength * (-1.0) * Math.tanh(negativeContribution / this.rateInhiWidth);

      // add contributions to the growth rate, do not deactivate current active clique
      this.ratesInternal[i] = this.reservoirFunctions(this.reservoir[i], DhanModule.RES_FW_POST) * positiveContribution + negativeContribution
          + this.getBias(i);
      this.rates[i] = this.ratesInternal[i];
      if ((this.ratesInternal[i] < 0) || (this.afferentRates[i] > 0))
        this.rates[i] += this.afferentRates[i];
    }

    // new activities
    for (int i = 0; i < this.getSize(); i++) {
      if (this.rates[i] > 0) {
        this.setActivity(i, this.getActivity(i) + (1.0 - this.getActivity(i)) * this.rates[i] * timeDelta);
      } else {
        this.setActivity(i, this.getActivity(i) + this.getActivity(i) * this.rates[i] * timeDelta);
      }
    }

    // check for out-off range values
    for (int i = 0; i < this.getSize(); i++) {
      this.setActivity(i, Math.min(this.getActivity(i), DhanModule.nearlyActive));
      this.reservoir[i] = Math.min(this.reservoir[i], DhanModule.nearlyActive);
      this.setActivity(i, Math.max(this.getActivity(i), DhanModule.nearlyInactive));
      this.reservoir[i] = Math.max(this.reservoir[i], DhanModule.nearlyInactive);
    }

    this.setBias(0.0);

    // zero bias diffusive control receptors,
    // they are set by the other/same module
    // via the forward connections
    for (String receptorType : this.receptors.keySet()) {
      this.receptors.put(receptorType, 0.0);
    }

    // evaluate effective degree of synchronization
    this.synchronizationStrength = 0.0;
    double tempDenominator = 0.0;
    for (int i = 0; i < this.getSize(); i++) {
      if (this.getActivity(i) > this.phiXActive) {
        tempDenominator += 1.0;
        this.synchronizationStrength += this.reservoir[i];
      }
    }
    if (tempDenominator > 0.0) {
      this.synchronizationStrength = this.synchronizationStrength / tempDenominator;
    }

    // forward inter-modular diffusive signals
    for (final Module module : this.forwardModules.keySet()) {
      final String forwardPort = this.forwardModules.get(module);
      module.addReceptor(forwardPort, this.diffusiveModuleSignal);
    }
  }

  /**
   * Updates the diffusive module control signals.
   * <p>
   * --class DhanModule-- <br>
   */
  public void diffusiveControlSignals(final double timeDelta) {
    // update of diffusive learn signal and test printing
    this.diffusiveModuleSignal -= this.diffusiveModuleSignalRateDown * this.diffusiveModuleSignal * timeDelta;

    for (int i = 0; i < this.getSize(); i++) {
      if ((this.rates[i] > 0) && (this.ratesInternal[i] < 0.0) && (this.afferentRates[i] > this.diffusiveModuleSignalBackRateThreshold)
          && (this.getActivity(i) > this.backwardMemXActive)) {
        this.diffusiveModuleSignal += this.diffusiveModuleSignalRateUp * (1.0 - this.diffusiveModuleSignal) * timeDelta;
      }
    }
  }

  /**
   * Evaluates the synaptic plasticities for the backward links.
   * 
   * <p>
   * --class DhanModule-- <br>
   */
  public void afferentMemory(final double timeDelta) {
    // --- plasticity when backward learning signal is large
    // --- `this.diffusiveModuleSignal' is set via the intra-modular
    // --- `this.myForwardModules' to `this.myReceptors'
    if (this.getReceptor("backwardLearning") < this.diffusiveModuleSignalCriticalValue) {
      return;
    }

    // --- evaluate backward link statistics ---
    for (int i = 0; i < this.getSize(); i++) {
      this.backwardMemMean[i] = 0.0;
      this.backwardMemStandDev[i] = 0.0;
    }
    int backwardTotalNumberOfLinks = 0; // total number of incoming links

    for (final Module afferentModule : this.getAfferentModules()) {
      int backwardModuleN = afferentModule.getSize();
      backwardTotalNumberOfLinks += backwardModuleN;

      for (int i = 0; i < this.getSize(); i++) {
        for (int j = 0; j < backwardModuleN; j++) {
          this.backwardMemMean[i] += this.getAfferentLink(afferentModule, i, j);
          this.backwardMemStandDev[i] += this.getAfferentLink(afferentModule, i, j) * this.getAfferentLink(afferentModule, i, j);
        }
      }
    }

    // --- update arrays
    for (int i = 0; i < this.getSize(); i++) {
      this.backwardMemMean[i] = this.backwardMemMean[i] / backwardTotalNumberOfLinks;

      this.backwardMemStandDev[i] = this.backwardMemStandDev[i] / backwardTotalNumberOfLinks;
      this.backwardMemStandDev[i] = Math.sqrt(backwardMemStandDev[i] - this.backwardMemMean[i] * this.backwardMemMean[i]);
    }

    // link-plasticities: optimization of overall contributions
    for (final Module afferentModule : this.getAfferentModules()) {
      int backwardModuleN = afferentModule.getSize();

      for (int i = 0; i < this.getSize(); i++) {
        double contrActAct = 0.0;
        double contrInaAct = 0.0;
        double contrActIna = 0.0;
        double contrInaIna = 0.0;

        // active postsynaptic sites
        if ((this.getActivity(i) > this.backwardMemXActive) && (this.rates[i] > 0)) {
          contrActAct = this.backwardMemGammaActAct * timeDelta * ((this.backwardMemOptRateActAct < this.afferentRates[i]) ? -1.0 : 1.0);
          contrInaAct = this.backwardMemGammaInaAct * timeDelta * ((this.backwardMemOptRateInaAct < this.afferentInverseRates[i]) ? -1.0 : 1.0);
        }

        // inactive postsynaptic sites
        if ((this.getActivity(i) < this.backwardMemXInactive) && (this.rates[i] < 0)) {
          contrActIna = this.backwardMemGammaActIna * timeDelta * ((this.backwardMemOptRateActIna < this.afferentRates[i]) ? -1.0 : 1.0);
          contrInaIna = this.backwardMemGammaInaIna * timeDelta * ((this.backwardMemOptRateInaIna < this.afferentInverseRates[i]) ? -1.0 : 1.0);
        }

        // link plasticities
        for (int j = 0; j < backwardModuleN; j++) {
          this.setAfferentLink(afferentModule, i, j, this.getAfferentLink(afferentModule, i, j) + contrActAct * afferentModule.getActivity(j) + contrInaAct
              * (1.0 - afferentModule.getActivity(j)));
          this.setAfferentLink(afferentModule, i, j, this.getAfferentLink(afferentModule, i, j) + contrActIna * afferentModule.getActivity(j) + contrInaIna
              * (1.0 - afferentModule.getActivity(j)));
        }
      }
    }
  }

  /**
   * Evaluates the long-term synaptic plasticities.
   * 
   * <p>
   * --class DhanModule-- <br>
   */
  public void longTermMemory(final double timeDelta) {
    // --- no adaption of intra-module links if
    // --- dynamics driven by external input -
    // --- `this.diffusiveModuleSignal' is set via the intra-modular
    // --- `this.myForwardModules' to `this.myReceptors'
    if (this.getReceptor("backwardLearning") > this.diffusiveModuleSignalCriticalValue) {
      return;
    }

    // determine the sum of active incoming links
    for (int i = 0; i < this.getSize(); i++) {
      double totalPositive;
      this.ratesActive[i] = 0.0;
      //
      if (this.getActivity(i) > this.longTermMemCriticalX)
        for (int j = 0; j < this.getSize(); j++)
          if (this.getActivity(j) > this.longTermMemCriticalX) {
            totalPositive = this.getLink(i, j, DhanModule.LINK_EXCITATORY_LONGTERM) + this.getLink(i, j, DhanModule.LINK_EXITATORY_SHORTTERM);
            if (totalPositive > 0.0)
              this.ratesActive[i] += totalPositive * this.getActivity(j);
          }
    }

    // loop over all links: asymmetric!
    for (int node = 0; node < this.getSize(); node++) {
      if (this.getActivity(node) > this.longTermMemCriticalX) {
        double deltaRate = this.longTermMemOptRate - this.ratesActive[node];
        for (int j = 0; j < this.getSize(); j++) {
          if ((node != j) && (this.getActivity(j) > this.longTermMemCriticalX)) {
            if (deltaRate > 0) {
              this.setLink(node, j, 0, this.getLink(node, j, 0) + this.longTermMemGammaOpt * deltaRate * timeDelta);
            } else {
              this.setLink(node, j, 0, this.getLink(node, j, 0) + this.longTermMemGammaOpt * (this.getLink(node, j, 0) - longTermMemMinStrength) * deltaRate
                  * timeDelta);
            }
          }
        }
      }
    }
  }

  /**
   * Evaluates the short-term synaptic plasticities.
   * 
   * <p>
   * --class DhanModule-- <br>
   */
  public void shortTermMemory(final double timeDelta) {
    // loop over all internal links
    for (int ownNode = 0; ownNode < this.getSize(); ownNode++) {
      for (int otherNode = 0; otherNode < this.getSize(); otherNode++) {
        if (ownNode != otherNode) {
          this.addInternalLink(ownNode, otherNode, 2, -this.shortTermMemGammaDown * this.getLink(ownNode, otherNode, 2) * timeDelta);

          if ((this.getBias(ownNode) > this.shortTermMemMinExtBias) && (this.getBias(otherNode) > this.shortTermMemMinExtBias)
              && (this.getActivity(otherNode) > this.shortTermMemCriticalX)) {
            this.addInternalLink(ownNode, otherNode, 2, this.shortTermMemGammaUp * (this.shortTermMemMaxStrength - this.getLink(ownNode, otherNode, 2))
                * this.getBias(ownNode) * reservoirFunctions(this.reservoir[otherNode], DhanModule.RES_FW_POST) * timeDelta);
          }
        }
      }
    }
  }

  /**
   * Evaluates some statistical properties, like link-strength distribution, for
   * output on file upon external call and sends the data to the printing
   * facility.
   * 
   * <p>
   * --class DhanModule-- <br>
   */
  public void statOutputToFile() {
    ByteArrayOutputStream baos = new ByteArrayOutputStream();
    OutputStreamWriter osw = new OutputStreamWriter(baos);
    PrintWriter pw = new PrintWriter(osw);
    String outputFileName;

    // --- determine range of back-link values ---
    double lowestValue = 0.0;
    double largestValue = 0.0;
    int nBins = 25;
    double[] statBins = new double[nBins];

    for (int i = 0; i < this.getSize(); i++) {
      // average standard deviation
      largestValue += this.backwardMemStandDev[i];
    }

    largestValue = 0.1 * (int) (10.0 * 4.0 * largestValue / this.getSize());
    if (largestValue < 0.1)
      largestValue = 0.1;
    lowestValue = -largestValue;

    // --- evaluate backward rate statistics ---
    for (int iBin = 0; iBin < nBins; iBin++)
      statBins[iBin] = 0.0;

    for (final Module afferentModule : this.getAfferentModules()) {
      for (int i = 0; i < this.getSize(); i++)
        for (int j = 0; j < afferentModule.getSize(); j++) {
          double relValue = this.getAfferentLink(afferentModule, i, j) - lowestValue;
          int iBin = (int) (relValue * nBins / (largestValue - lowestValue));
          if ((iBin >= 0) && (iBin < nBins))
            statBins[iBin] = statBins[iBin] + 1.0;
        }
    }

    double totalValues = 0.0; // for normalization
    for (int iBin = 0; iBin < nBins; iBin++)
      totalValues += statBins[iBin];
    for (int iBin = 0; iBin < nBins; iBin++)
      statBins[iBin] = statBins[iBin] / totalValues;

    // --- output strings for backward rate statistics ---
    for (int iBin = 0; iBin < nBins; iBin++) {
      double valueBin = lowestValue + (iBin + 0.5) * (largestValue - lowestValue) / nBins;
      pw.printf("%8.3f %8.3f \n", valueBin, statBins[iBin]);
    }

    pw.flush();
    outputFileName = "Brain_" + this.getName() + ".statBackLinks";
    try {
      PrintWriter printWriter = new PrintWriter(new File(Brain.DEFAULT_DATA_OUTPUT_DIRECTORY, outputFileName));
      printWriter.printf(Locale.ENGLISH, baos.toString());
      printWriter.close();
    } catch (FileNotFoundException e) {
      e.printStackTrace();
    }
    baos.reset();

    // --- simple output of backwardLinks module by module ---
    for (final Module afferentModule : this.getAfferentModules()) {
      for (int j = 0; j < afferentModule.getSize(); j++) {
        pw.printf("%5d", j);
        for (int i = 0; i < this.getSize(); i++)
          pw.printf("%8.3f", this.getAfferentLink(afferentModule, i, j));
        pw.printf("\n");
      }
      pw.flush();
      outputFileName = "Brain_" + this.getName() + "_" + afferentModule.getName() + ".backLinks";
      try {
        PrintWriter printWriter = new PrintWriter(new File(Brain.DEFAULT_DATA_OUTPUT_DIRECTORY, outputFileName));
        printWriter.printf(Locale.ENGLISH, baos.toString());
        printWriter.close();
      } catch (FileNotFoundException e) {
        e.printStackTrace();
      }
      baos.reset();
    }

    
    
    pw.printf("# --- reservoir functions --- \n");

    int NN = 100; // number of data points
    for (int i = 0; i < NN; i++) {
      double phi = i * 1.0 / NN;
      pw.printf("%7.2f", phi);
      for (int ff = 0; ff < this.nResFunctions; ff++)
        pw.printf("%8.3f", this.reservoirFunctions(phi, ff));
      pw.printf("\n");
    }

    pw.flush();
    outputFileName = "Brain_" + this.getName() + ".resFunc";
    try {
      PrintWriter printWriter = new PrintWriter(new File(Brain.DEFAULT_DATA_OUTPUT_DIRECTORY, outputFileName));
      printWriter.printf(Locale.ENGLISH, baos.toString());
      printWriter.close();
    } catch (FileNotFoundException e) {
      e.printStackTrace();
    }
    baos.reset();
  }

  public double getPhiXActive_xml() {
    return this.phiXActive;
  }

  public double getRateInhiWidth() {
    return this.rateInhiWidth;
  }

  public void setRateInhiWidth(final double rateInhiWidth_new) {
    this.rateInhiWidth = rateInhiWidth_new;
  }

  public double[] getRates() {
    return this.rates;
  }

  public double[] getRatesInternal() {
    return this.ratesInternal;
  }

  public double[] getReservoir() {
    return this.reservoir;
  }

  public double[] getAfferentRates() {
    return this.afferentRates;
  }

  public double[] getRatesActive() {
    return this.ratesActive;
  }

  public double[] getRatesPositive() {
    return this.ratesPositive;
  }

  public double[] getRatesNegative() {
    return this.ratesNegative;
  }

  public double[] getPhiInverseWidth() {
    return this.phiInverseWidth;
  }

  public double[] getPhiAuxShift() {
    return this.phiAuxShift;
  }

  public double[] getPhiAuxFactor() {
    return this.phiAuxFactor;
  }

  @Override
  public void printLogFileHeader(final PrintWriter out) {
    super.printLogFileHeader(out);

    out.printf(Locale.ENGLISH, "# --------------------------- \n");
    out.printf(Locale.ENGLISH, "# ---   internal links    --- \n");
    out.printf(Locale.ENGLISH, "# --------------------------- \n");
    out.printf(Locale.ENGLISH, "# initialClique-number/size   :%10d%10d\n", this.auxInitialCliqueNumber, this.auxInitialCliqueSize);
    out.printf(Locale.ENGLISH, "#   internal link-type number :%10d\n", DhanModule.INTERNAL_LINK_TYPES_NUMBER);
    out.printf(Locale.ENGLISH, "#   excit/inhbi link-strength :%10.6f%10.6f\n", this.defaultExcitatoryLinkStrength, this.defaultInhibitoryLinkStrength);
    out.printf(Locale.ENGLISH, "# --------------------------- \n");
    out.printf(Locale.ENGLISH, "# ---        rate         --- \n");
    out.printf(Locale.ENGLISH, "# --------------------------- \n");
    out.printf(Locale.ENGLISH, "# inhibi. normalization-width :%10.6f\n", this.rateInhiWidth);
    out.printf(Locale.ENGLISH, "# --------------------------- \n");
    out.printf(Locale.ENGLISH, "# ---      reservoir      --- \n");
    out.printf(Locale.ENGLISH, "# --------------------------- \n");
    out.printf(Locale.ENGLISH, "#  depletion/growth/coh. rate :%10.6f%10.6f%10.6f\n", this.phiGammaDown, this.phiGammaUp, this.phiGammaCoh);
    out.printf(Locale.ENGLISH, "# critical-x for depl./growth :%10.6f%10.6f\n", this.phiXActive, this.phiXInactive);
    //
    out.printf(Locale.ENGLISH, "#   critical-phi: fw post/pre :%10.6f%10.6f\n", this.getPhiCritical()[DhanModule.RES_FW_POST],
        this.getPhiCritical()[DhanModule.RES_FW_PRE]);
    out.printf(Locale.ENGLISH, "#   critical-phi: fz / gz     :%10.6f%10.6f\n", this.getPhiCritical()[DhanModule.RES_FZ_PRE],
        this.getPhiCritical()[DhanModule.RES_GZ_PRE]);

    out.printf(Locale.ENGLISH, "#       width for fw post/pre :%10.6f%10.6f\n", this.getPhiWidth()[DhanModule.RES_FW_POST],
        this.getPhiWidth()[DhanModule.RES_FW_PRE]);
    out.printf(Locale.ENGLISH, "#       width for fz / gz     :%10.6f%10.6f\n", this.getPhiWidth()[DhanModule.RES_FZ_PRE],
        this.getPhiWidth()[DhanModule.RES_GZ_PRE]);

    out.printf(Locale.ENGLISH, "#   min-value for fw post/pre :%10.6f%10.6f\n", this.getPhiMinValue()[DhanModule.RES_FW_POST],
        this.getPhiMinValue()[DhanModule.RES_FW_PRE]);
    out.printf(Locale.ENGLISH, "#   min-value for fz / gz     :%10.6f%10.6f\n", this.getPhiMinValue()[DhanModule.RES_FZ_PRE],
        this.getPhiMinValue()[DhanModule.RES_GZ_PRE]);
    out.printf(Locale.ENGLISH, "# --------------------------- \n");
    out.printf(Locale.ENGLISH, "# ---  short-term memory  --- \n");
    out.printf(Locale.ENGLISH, "# --------------------------- \n");
    out.printf(Locale.ENGLISH, "#       growth/depletion rate :%10.6f%10.6f\n", this.shortTermMemGammaUp, this.shortTermMemGammaDown);
    out.printf(Locale.ENGLISH, "#               maximal value :%10.6f\n", this.shortTermMemMaxStrength);
    out.printf(Locale.ENGLISH, "#     critical activities level :%10.6f\n", this.shortTermMemCriticalX);
    out.printf(Locale.ENGLISH, "#   minimal stimulus-strength :%10.6f\n", this.shortTermMemMinExtBias);
    out.printf(Locale.ENGLISH, "# --------------------------- \n");
    out.printf(Locale.ENGLISH, "# ---   long-term memory  --- \n");
    out.printf(Locale.ENGLISH, "# --------------------------- \n");
    out.printf(Locale.ENGLISH, "#           optimization rate :%10.6f\n", this.longTermMemGammaOpt);
    out.printf(Locale.ENGLISH, "#               minimal value :%10.6f\n", this.longTermMemMinStrength);
    out.printf(Locale.ENGLISH, "#     critical activities level :%10.6f\n", this.longTermMemCriticalX);
    out.printf(Locale.ENGLISH, "#         optimal growth rate :%10.6f\n", this.longTermMemOptRate);
    out.printf(Locale.ENGLISH, "# --------------------------- \n");
    out.printf(Locale.ENGLISH, "# ---  diffusive control   -- \n");
    out.printf(Locale.ENGLISH, "# --------------------------- \n");
    out.printf(Locale.ENGLISH, "#       growth/depletion rate :%10.6f%10.6f\n", this.diffusiveModuleSignalRateUp, this.diffusiveModuleSignalRateDown);
    out.printf(Locale.ENGLISH, "# threshold: diff./back rates :%10.6f%10.6f\n", this.diffusiveModuleSignalCriticalValue,
        this.diffusiveModuleSignalBackRateThreshold);
    out.printf(Locale.ENGLISH, "# --------------------------- \n");
    out.printf(Locale.ENGLISH, "# --- forward connections --- \n");
    out.printf(Locale.ENGLISH, "# --------------------------- \n");
    for (final Module forwardModule : this.forwardModules.keySet()) {
      out.printf(Locale.ENGLISH, "#       target module, port :%10s\n", forwardModule.getName());
    }
    out.printf(Locale.ENGLISH, "# --------------------------- \n");
    out.printf(Locale.ENGLISH, "# --- backward connections -- \n");
    out.printf(Locale.ENGLISH, "# --------------------------- \n");
    out.printf(Locale.ENGLISH, "#           x-active/inactive :%10.6f%10.6f\n", this.backwardMemXActive, this.backwardMemXInactive);
    out.printf(Locale.ENGLISH, "#       Gamma:A-A,I-A/A-I,I-I :%10.6f%10.6f%10.6f%10.6f\n", this.backwardMemGammaActAct, this.backwardMemGammaInaAct,
        this.backwardMemGammaInaAct, this.backwardMemGammaInaIna);
    out.printf(Locale.ENGLISH, "#  input-rate:A-A,I-A/A-I,I-I :%10.6f%10.6f%10.6f%10.6f\n", this.backwardMemOptRateActAct, this.backwardMemOptRateInaAct,
        this.backwardMemOptRateActIna, this.backwardMemOptRateInaIna);
    out.printf(Locale.ENGLISH, "# backward module identifiers :");
    for (final Module afferentModule : this.getAfferentModules()) {
      out.printf(Locale.ENGLISH, "%10s", afferentModule.getName());
    }
    out.printf(Locale.ENGLISH, "\n");
    out.printf(Locale.ENGLISH, "# *************************** \n");
  }
}