package brain;

import java.awt.Color;
import java.awt.Graphics2D;
import java.awt.image.BufferedImage;
import java.io.BufferedReader;
import java.io.BufferedWriter;
import java.io.ByteArrayInputStream;
import java.io.ByteArrayOutputStream;
import java.io.File;
import java.io.FileNotFoundException;
import java.io.FileReader;
import java.io.FileWriter;
import java.io.IOException;
import java.io.ObjectInputStream;
import java.io.ObjectOutputStream;
import java.io.OutputStreamWriter;
import java.io.PrintWriter;
import java.io.Serializable;
import java.text.DecimalFormat;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collection;
import java.util.Collections;
import java.util.Comparator;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;
import java.util.Locale;
import java.util.Map;
import java.util.Random;
import java.util.Set;
import java.util.TreeMap;
import java.util.TreeSet;
import java.util.Vector;
import java.util.logging.Level;
import java.util.logging.Logger;

import javax.imageio.ImageIO;

import org.apache.commons.lang3.ArrayUtils;

import utils.Interval;

public abstract class FacilityLegacy {

}

/**
 * A single external/internal stimulus, characterized by an string-identifier
 * and the stimuli data, such starting-/end-time, sites and strengths.
 * <p>
 * The standard stimulus strength is defined as a static variable. It is less
 * than the maximal value in order to allow for strong stimuli.
 * 
 * @author Claudius Gros
 * @author Mathias Linkerhand
 * @version 2010-03-08
 */
class OneStimulus implements Serializable, Comparable<OneStimulus>, Cloneable {
  private static final long    serialVersionUID         = 1L;

  private static final Logger  logger                   = Logger.getLogger(OneStimulus.class.getName());

  /** Standard stimulus strength. */
  public static double         standardStimulusStrength = 0.95;
  public Interval              interval;

  /** Random number generator. */
  private static final Random  random                   = new Random(0);

  /** Strength of stimulus at affected sites. */
  private Map<Integer, Double> strengths;

  public OneStimulus() {
    this.strengths = new HashMap<Integer, Double>();
  }

  public int getSize() {
    return this.strengths.size();
  }

  public Set<Integer> getNodes() {
    return this.strengths.keySet();
  }

  public double getStrength(final int node) {
    return this.strengths.get(node);
  }

  public void setStrength(final int node, final double strength) {
    this.strengths.put(node, strength);
  }

  public void superimposeWith(final OneStimulus stimulus) {
    for (final int node : stimulus.getNodes()) {
      this.setStrength(node, stimulus.getStrength(node));
    }
  }

  public void superimposeWith(final OneStimulus[] stimuli, final double probability) {
    for (final OneStimulus stimulus : stimuli) {
      if (OneStimulus.random.nextDouble() <= probability) {
        this.superimposeWith(stimulus);
      }
    }
  }

  /**
   * Compares two stimuli chronologically
   */
  @Override
  public int compareTo(final OneStimulus oneStimulus) {
    return this.getTimeInterval().compareTo(oneStimulus.getTimeInterval());
  }

  @Override
  public boolean equals(Object anObject) {
    if (this == anObject) {
      return true;
    }
    if (anObject instanceof OneStimulus) {
      final OneStimulus stimulus = (OneStimulus) anObject;
      return this.strengths.equals(stimulus.strengths);
    }
    return false;
  }

  @Override
  public int hashCode() {
    return this.strengths.hashCode() + this.interval.hashCode();
  }

  @Override
  public OneStimulus clone() {
    try {
      // create a deep copy
      final ByteArrayOutputStream baos = new ByteArrayOutputStream();
      new ObjectOutputStream(baos).writeObject(this);
      final ByteArrayInputStream bais = new ByteArrayInputStream(baos.toByteArray());
      return (OneStimulus) new ObjectInputStream(bais).readObject();
    } catch (IOException e) {
      e.printStackTrace();
    } catch (ClassNotFoundException e) {
      e.printStackTrace();
    }
    return null;
  }

  /**
   * Retruns the start Time of the Stimuli
   * 
   * @return the startTimeList
   */
  public double getStartTime() {
    return this.interval.getStart();
  }

  public double getEndTime() {
    return this.interval.getEnd();
  }

  public Interval getTimeInterval() {
    return this.interval;
  }

  /**
   * Saves a gray scaled picture of a lattice pattern.
   * 
   * As there is no information about the resolution of a two dimensional
   * pattern the user has explicitely give that as a parameter.
   * 
   * @param width
   *            - width of the two dimensional lattice
   * @param height
   *            - height of the two dimensional lattice
   * @param patternCounter
   *            - just an indicator to identify patterns
   * @param scale
   *            - one element of the stimulusLattice would be represented as
   *            one pixel in the image which is a 1x1 filled rectangle.
   *            <code>scale</code> is the width and height of the filled
   *            rectange in the image
   * @param stimulusLattice
   *            - array of elements between 0 and 1 representing the strength
   *            of a signal at given index of the lattice
   */
  public static void savePicture(int width, int height, int patternCounter, int scale, double[] stimulusLattice) {
    // make sure scale is above 0
    scale = scale < 1 ? 1 : scale;

    // create a new image to draw on
    BufferedImage bimage = new BufferedImage(width * scale, height * scale, BufferedImage.TYPE_INT_ARGB);
    Graphics2D g2d = bimage.createGraphics();

    // for the fun of optimization create a color cache
    ArrayList<Color> colorCache = new ArrayList<Color>(256);
    for (int c = 0; c < 256; c++)
      colorCache.add(new Color(c, c, c));

    // draw the stimulus on image
    int stimulusIndex = 0;
    int colorIndex = 0;
    for (int x = 0; x < width * scale; x += scale)
      for (int y = 0; y < height * scale; y += scale) {
        colorIndex = (int) ((1 - stimulusLattice[stimulusIndex]) * 255);
        // use the same color value for RGB as this is the gray scale of
        // the
        // color space
        g2d.setColor(colorCache.get(colorIndex));
        g2d.fillRect(y, x, scale, scale);
        stimulusIndex++;
      }

    // create a file object
    File file = new File("pattern_" + patternCounter + ".png");

    try {
      // write image to file
      String imagefiletype = "png";
      ImageIO.write(bimage, imagefiletype, file);
      logger.log(Level.INFO, "PrintablePattern " + patternCounter + " written as " + imagefiletype + " to " + file.getAbsolutePath());
    } catch (IOException e) {
      logger.log(Level.SEVERE, "Could not write image to " + file.getAbsolutePath());
      e.printStackTrace();
    }
  }

  public static void saveStimuliToFile(final OneStimulus[] stimuli, final File file) throws FileNotFoundException {
    final String nodeSeparator = "\t";

    final PrintWriter out = new PrintWriter(file);
    for (final OneStimulus stimulus : stimuli) {
      final List<Integer> nodes = new Vector<Integer>(stimulus.getNodes());
      Collections.sort(nodes); // sorting for human-friendly output (not
      // really necessary)
      out.println(nodes.size() + nodeSeparator + nodes);
    }
    out.close();
  }

  public static OneStimulus[] loadStimuliFromFile(final File file) throws FileNotFoundException {
    return null;
  }

  /**
   * Generates non-overlapping square stimuli.
   */
  public static OneStimulus[] generateSquareStimuli(final int size) {
    final Set<OneStimulus> elementaryStimuli = new HashSet<OneStimulus>();

    // --- select parameters --- //
    int L = size * size; // consider a L x L overall square

    // --- generate elementary stimuli --- //
    for (int iX = 0; iX < size; iX++) {
      for (int iY = 0; iY < size; iY++) {
        final OneStimulus stimulus = new OneStimulus();
        for (int i = 0; i < size; i++)
          for (int j = 0; j < size; j++) {
            // TODO: check
            // int iStim = j + i * size;
            stimulus.setStrength(iX * L * size + iY * size + i * L + j, 1.0);
          }
        elementaryStimuli.add(stimulus);
      }
    }

    return elementaryStimuli.toArray(new OneStimulus[0]);
  }

  /**
   * Generates stimuli for the bar problem.
   */
  public static OneStimulus[] generateBarStimuli(final int barLength) {
    final Vector<OneStimulus> elementaryStimuli = new Vector<OneStimulus>();

    final double strength = 1.0;

    // horizontal
    for (int y = 0; y < barLength; y++) {
      final OneStimulus stimulus = new OneStimulus();
      for (int x = 0; x < barLength; x++) {
        stimulus.setStrength(x + barLength * y, strength);
      }
      elementaryStimuli.add(stimulus);
    }

    // vertical
    for (int x = 0; x < barLength; x++) {
      final OneStimulus stimulus = new OneStimulus();
      for (int y = 0; y < barLength; y++) {
        stimulus.setStrength(x + barLength * y, strength);
      }
      elementaryStimuli.add(stimulus);
    }

    return elementaryStimuli.toArray(new OneStimulus[0]);
  }

  /**
   * Superimposes stimuli.
   * 
   * @param probability
   *            probability of a given bar to occur
   * @param stimuliNumber
   *            number of statistical independent stimuli to generate
   */
  public static OneStimulus[] superimposeStimuli(final OneStimulus[] stimuli, final double probability, final int stimuliNumber) {
    final Vector<OneStimulus> superimposedStimuli = new Vector<OneStimulus>();

    final boolean zeroStimulusAllowed = false;

    for (int stimulusIndex = 0; stimulusIndex < stimuliNumber; stimulusIndex++) {
      OneStimulus superimposedStimulus;
      if (zeroStimulusAllowed) {
        // Zero-stimulus are allowed. So just superimpose the others to
        // a zero-stimulus.
        superimposedStimulus = new OneStimulus();
        superimposedStimulus.superimposeWith(stimuli, probability);
      } else {
        // Zero-stimulus is not allowed. At least one stimulus should be
        // present. So choose the first randomly.
        final int firstStimulusIndex = OneStimulus.random.nextInt(stimuli.length);
        superimposedStimulus = stimuli[firstStimulusIndex].clone();
        superimposedStimulus.superimposeWith((OneStimulus[]) ArrayUtils.remove(stimuli, firstStimulusIndex), probability);
      }
      superimposedStimuli.add(superimposedStimulus);
    }

    return superimposedStimuli.toArray(new OneStimulus[0]);
  }
}

/**
 * A single center is characterized by the module and an integer-identifier for
 * the index of the center within this module - used for global activities
 * analysis.
 * 
 * @author Claudius Gros
 * @author Mathias Linkerhand
 * @version 2009-11-30
 */
@Deprecated
class OneCenter {
  public Module module;
  /** number of the site */
  public int    index;

  public OneCenter(final Module module, final int centerSite) {
    this.module = module;
    this.index = centerSite;
  }

  /**
   * Compare two centers for equality.
   * 
   * @return true if equal, false otherwise
   */
  @Override
  public boolean equals(final Object other) {
    return other instanceof OneCenter && ((OneCenter) other).index == this.index && ((OneCenter) other).module == this.module;
  }

  /**
   * Sort an array of centers by module-names and indices.
   */
  static public void sortCenterArray(OneCenter[] centerArray) {
    // order first by module names
    Arrays.sort(centerArray, new Comparator<OneCenter>() {
      @Override
      public int compare(OneCenter o1, OneCenter o2) {
        return o1.module.getName().compareTo(o2.module.getName());
      }
    });

    // order then by center indices
    Arrays.sort(centerArray, new Comparator<OneCenter>() {
      @Override
      public int compare(OneCenter o1, OneCenter o2) {
        return o1.index == o2.index ? 0 : (o1.index > o2.index ? 1 : -1);
      }
    });
  }
}

/**
 * Auxiliary class for a single cluster of active sites as an sorted array of
 * `OneCenter', together with some characteristics like the starting time - it
 * contains a member function for comparison of clusters.
 * 
 * @author Claudius Gros
 * @author Mathias Linkerhand
 * @version 2009-12-09
 */
class OneCluster {
  /** Array with activities centers globally indexed via `OneCenter'. */
  public OneCenter[] clusterSites;
  /** This this cluster was first (re-)created dynamically. */
  public double      startTime;
  /** Time-interval during which this cluster was dynamically stable. */
  public double      durationTime;

  /**
   * Standard constructor for OneCluster, cluster sites are sorted on input.
   * 
   */
  public OneCluster(ArrayList<OneCenter> activeACs, double startTime) {
    this.clusterSites = new OneCenter[activeACs.size()]; // set array size
    for (int i = 0; i < activeACs.size(); i++) {
      this.clusterSites[i] = new OneCenter(activeACs.get(i).module, activeACs.get(i).index);
    }
    this.startTime = startTime;
    this.durationTime = 0.0;
    OneCenter.sortCenterArray(clusterSites);
  }

  public OneCluster(ArrayList<OneCenter> activeACs, double startTime1, double durationTime1) {
    this.clusterSites = new OneCenter[activeACs.size()]; // set array size
    for (int i = 0; i < activeACs.size(); i++)
      this.clusterSites[i] = new OneCenter(activeACs.get(i).module, activeACs.get(i).index);
    //
    this.startTime = startTime1;
    this.durationTime = durationTime1;
    OneCenter.sortCenterArray(clusterSites);
  }

  /**
   * Returns true if `OneCluster' a is a true subcluster of `OneCenter' of b -
   * and non-identical.
   * 
   * @param a
   *            - a <code>OneCenter</code> to check if a is a true subcluster
   *            of b
   * @param b
   *            - a <code>OneCenter</code> which gets checked if it does
   *            contain a
   * @return true, if a is a true/complete subcluster of b, false otherwise
   * 
   */
  static public boolean isSubCluster(OneCluster a, OneCluster b) {
    int nSitesA = a.clusterSites.length;
    int nSitesB = b.clusterSites.length;
    if (nSitesA >= nSitesB) {
      return false;
    }
    int iA = 0;
    int iB = 0;
    boolean returnValue = true;

    while ((iA < nSitesA) && (returnValue)) {
      boolean iA_in_B = false;
      while ((iB < nSitesB) && (!iA_in_B)) {
        if (a.clusterSites[iA].equals(b.clusterSites[iB])) {
          iA_in_B = true;
        }
        iB++;
      }
      iA++;
      returnValue = iA_in_B;
    }
    return returnValue;
  }

  /**
   * Compares two clusters for equality - assumes ordered arrays
   * `clusterSites[]' in input `OneCluster'.
   * 
   * @param a
   *            - a <code>OneCluster</code> which is checked whether its from
   *            the same module and contains the same sites like b
   * @param b
   *            - a <code>OneCluster</code> which is checked whether its from
   *            the same module and contains the same sites like a
   * @return true, if all sites of a and b are from equal module and contain
   *         same sites
   * 
   */
  static public boolean equals(OneCluster a, OneCluster b) {
    int nSitesA = a.clusterSites.length;
    int nSitesB = b.clusterSites.length;
    if (nSitesA != nSitesB) {
      return false;
    }
    for (int iA = 0; iA < nSitesA; iA++) {
      if (!a.clusterSites[iA].equals(b.clusterSites[iA])) {
        return false;
      }
    }
    return true;
  }

  /**
   * Returns a HashMap containing the ranges of the module, assumes cluster
   * sites * to be sorted by module names.
   * 
   * @param a
   *            OneCluster of active sites
   * @return A HashMap containing the ranges of the module
   * 
   */
  static public HashMap<String, int[]> moduleRanges(OneCluster a) {
    HashMap<String, int[]> returnMap = new HashMap<String, int[]>();
    int nSites = a.clusterSites.length;
    if (nSites == 0)
      return returnMap;

    int iModuleStart = 0;
    int iA = 0;
    String oldModuleName = a.clusterSites[iA].module.getName();
    String currentModuleName;

    while (iA <= nSites) {
      if (iA < nSites)
        currentModuleName = a.clusterSites[iA].module.getName();
      else
        currentModuleName = "end of array";

      if (currentModuleName.compareTo(oldModuleName) == 0) {
        int[] moduleData = new int[2];
        moduleData[0] = iModuleStart;
        moduleData[1] = iA;
        returnMap.put(oldModuleName, moduleData);
        //
        oldModuleName = currentModuleName;
        iModuleStart = iA;
      }
      iA++;
    }
    return returnMap;
  }

  /**
   * Returns the number of shared sites for clusters, uses the module ranges
   * mapping.
   */
  static public int numberOfSharedSites(OneCluster a, OneCluster b) {
    HashMap<String, int[]> moduleRangesA = OneCluster.moduleRanges(a);
    HashMap<String, int[]> moduleRangesB = OneCluster.moduleRanges(b);
    int returnNumber = 0;

    for (String moduleName : moduleRangesA.keySet()) {
      if (moduleRangesB.containsKey(moduleName)) {
        for (int iA = moduleRangesA.get(moduleName)[0]; iA < moduleRangesA.get(moduleName)[1]; iA++) {
          for (int iB = moduleRangesB.get(moduleName)[0]; iB < moduleRangesB.get(moduleName)[1]; iB++) {
            if (a.clusterSites[iA].index == b.clusterSites[iB].index) {
              returnNumber++;
            }
          }
        }
      }
    }
    return returnNumber;
  }

  /**
   * For a module this method returns whether all sites of a module of this
   * OneCluster are active.
   * 
   * @param module
   *            a module name whose sites are checked
   * @return true, if all sites of a module are true, false otherwise
   */
  public boolean isOveractive(Module module) {
    int activeCount = 0;
    for (int i = 0; i < this.clusterSites.length; i++) {
      activeCount += module.getName().equals(clusterSites[i].module.getName()) ? 1 : 0;
    }
    return activeCount == module.getSize();
  }
}

/**
 * Contains a cycle of active clusters.
 * <p>
 * A cycle is defined as a recurrent sequence of active clusters.
 * 
 * @author Gregor Kaczor
 * @author Mathias Linkerhand
 * @version 2009-12-09
 */

class OneCycle {
  /** Holds a list of active clusters wich are recurrent in time */
  private LinkedList<OneCluster> cycle;
  /** Contains all occurances of the same cycle in time. */
  private LinkedList<OneCycle>   occurenceTime;
  /**
   * Start time of the cycle which is the start time of the first active
   * cluster in the cycle
   */
  private double                 startTime    = 0;
  /**
   * Duration time of the cycle which is the sum of durations of all the
   * active clusters forming the cycle, viz. exclusive transition time between
   * clusters. May be duration of a cycle should be definied as (start time) +
   * (duration of the last active cluster) - (start timee of the fist active
   * cluster)
   */
  private double                 durationTime = 0;
  /** number of active clusters forming the cycle */
  private int                    cycleSize    = 0;

  /**
   * Standard constructor.
   */
  public OneCycle() {
    this.cycle = new LinkedList<OneCluster>();
    this.occurenceTime = new LinkedList<OneCycle>();
  }

  @Override
  public String toString() {
    StringBuffer sb = new StringBuffer(1000);
    for (Iterator<OneCluster> iter = cycle.iterator(); iter.hasNext();) {
      OneCluster element = iter.next();
      for (int i = 0; i < element.clusterSites.length; i++) {
        sb = sb.append("(" + element.clusterSites[i].module.getName() + "|" + element.clusterSites[i].index + ")");
      }
      if (iter.hasNext()) {
        sb = sb.append("-");
      }
    }
    return sb.toString();
  }

  /**
   * Checks if this OneCycle contains a OneCluster.
   * 
   * @param activeCluster
   *            a cluster to check
   * @return true if this OneCycle contains the parameter, false otherwise
   */
  public boolean contains(OneCluster activeCluster) {
    for (Object element2 : cycle) {
      OneCluster element = (OneCluster) element2;
      if (OneCluster.equals(activeCluster, element)) {
        return true;
      }
    }
    return false;
  }

  /**
   * Adds an OneCluster to this.
   * 
   * @param activeCluster
   */
  public void add(OneCluster activeCluster) {
    cycle.add(activeCluster);
    if (cycle.size() == 1) {
      startTime = activeCluster.startTime;
    }
    durationTime += activeCluster.durationTime;
    cycleSize = cycle.size();
  }

  /**
   * Returns a cycle
   * 
   * @return a cycle
   */
  public LinkedList<OneCluster> getCycle() {
    return cycle;
  }

  /**
   * There are same cycles spread at different times and durations. Once we
   * identified a cycle, we just add its occurence.
   * 
   * @param cycle1
   */
  public void addOccurance(OneCycle cycle1) {
    occurenceTime.add(cycle1);
  }

  /**
   * Returns a list of time steps where the cycle occured.
   * 
   * @return Returns a LinkedList of OneCycles
   */
  public LinkedList<OneCycle> getOccurenceTime() {
    return occurenceTime;
  }

  /**
   * Returns the first occurence of cycle
   * 
   * @return Returns start time of the first active cluster in the cycle
   */
  public double getStartTime() {
    return startTime;
  }

  /**
   * Returns the duration of a cycle
   * 
   * @return Duration time of the cycle which is the sum of durations of all
   *         the active clusters forming the cycle, viz. exclusive transition
   *         time between clusters.
   */

  public double getDurationTime() {
    return durationTime;
  }

  /**
   * Returns number of active clusters forming the cycle
   * 
   * @return Returns number of active clusters forming a cycle
   */
  public int getCycleSize() {
    return cycleSize;
  }
}

/**
 * CycleContainer is a class which manages cycles.
 * 
 * @author Gregor Kaczor
 * @author Mathias Linkerhand
 * @version 2009-12-09
 */
class CycleContainer {
  /**
   * Contains all cycles. Every Cycle should appear at least twice to be
   * defined as a cycle Every further appearance is contained within OneCycle.
   */
  private HashMap<String, OneCycle> cycles;
  /** Contains a not yet closed candidate for a cycle.* */
  private OneCycle                  candidate;

  /**
   * A standard constructor for the cycle container.
   */
  public CycleContainer() {
    this.cycles = new HashMap<String, OneCycle>();
    this.candidate = new OneCycle();
  }

  /**
   * Checks if CycleContainer contains a cycle.
   * 
   * @param cycle
   *            a cycle to check
   * @return true if CycleContainer contains the cycle, false otherwise
   */
  public boolean containsCycle(OneCycle cycle) {
    return cycles.containsKey(cycle.toString());
  }

  /**
   * Decides if an active cluster closes a current cycle.
   * 
   * @param activeCluster
   *            the current active cluster
   */
  public void addCandidate(OneCluster activeCluster) {
    if (candidate.contains(activeCluster)) {
      // //must be more then one active cluster to form a cycle
      if (candidate.getCycle().size() > 1) {
        OneCycle oneCycle = cycles.get(candidate.toString());
        if (oneCycle == null) {
          cycles.put(candidate.toString(), candidate);
          candidate.addOccurance(candidate);
        } else {
          oneCycle.addOccurance(candidate);
        }
      }
      candidate = new OneCycle();
      candidate.add(activeCluster);
    } else {
      candidate.add(activeCluster);
    }
  }

  /**
   * Returns the number of cycles.
   * 
   * @return Returns the number of cycles.
   */
  public int numberOfCycles() {
    return cycles.size();
  }

  /**
   * Returns a HashMap which contains cycles
   * 
   * @return a HashMap which contains cycles
   */
  public HashMap<String, OneCycle> getCycles() {
    return cycles;
  }

  /**
   * Returns a candidate for a cycle
   */
  public OneCycle getCandidate() {
    return candidate;
  }

  /**
   * Returns a TreeMap in which cycles are ordered in a chronological order in
   * which they occure
   * 
   * @return a TreeMap with ordered cycles in which the key is the occurence
   *         time
   */
  public TreeMap<Double, OneCycle> getChronoSortedCycles() {
    TreeMap<Double, OneCycle> toSort = new TreeMap<Double, OneCycle>();
    for (Object element : cycles.keySet()) {
      String type = (String) element;
      OneCycle oneCycle = cycles.get(type);
      for (Object cycleOccurence : oneCycle.getOccurenceTime()) {
        OneCycle cycle = (OneCycle) cycleOccurence;
        toSort.put(Double.valueOf(cycle.getStartTime()), cycle);
      }
    }
    return toSort;
  }
}

/**
 * General abstract facility - it is never instantiated - for the generation of
 * networks with a given clique-distribution.
 * 
 * @author Claudius Gros
 * @author Mathias Linkerhand
 * @version 2010-03-11
 */
final class FacilCliqueNetwork {
  /**
   * The network (graph) to be generated, made-up of knot-index and an array
   * containing the indices of its neighbors.
   */
  static Map<Integer, Set<Integer>> network                = new HashMap<Integer, Set<Integer>>();
  /** The initial network size, to be determined. */
  static int                        initialN               = -1;
  /** The actual network size. */
  static int                        nSites                 = 800;
  /** The sizes of all cliques generated. */
  static int[]                      allCliqueSizes         = null;
  /** The distribution of clique sizes. */
  static double[]                   cliqueSizeDistribution = null;
  /** The mean of the clique-size distribution. */
  static double                     meanCliqueSize         = 2.5;

  /**
   * Standard constructor - should never be called.
   */
  private FacilCliqueNetwork() {
  }

  /**
   * Creates the initial distribution of cliques and stores them unconnected
   * in `network'.
   * <p>
   * The clique sizes are in between ]0,2*meanCliqueSize[
   */
  public static void createUniformDistribution(final Random random) {
    // --- set counters --- //
    int nCliques = 0;
    int maxN = nSites;
    int[] tempCliques = new int[maxN];

    nSites = 0;

    // --- generate clique size --- //
    while (nSites < maxN) {
      int size = 2 + random.nextInt((int) (2.0 * meanCliqueSize - 1));
      tempCliques[nCliques] = size;
      nSites += size;
      nCliques++;
    }

    // --- store allcliques --- //
    allCliqueSizes = new int[nCliques];
    for (int i = 0; i < nCliques; i++) {
      allCliqueSizes[i] = tempCliques[i];
    }

    // --- block-diagonal cliques --- //
    int startSite = 0;
    for (int iC = 0; iC < allCliqueSizes.length; iC++) {
      int cSize = allCliqueSizes[iC];
      for (int iCS = 0; iCS < cSize; iCS++) {
        network.put(startSite + iCS, new HashSet<Integer>(cSize - 1));
        for (int jCS = 0; jCS < cSize; jCS++) {
          if (iCS != jCS) {
            network.get(startSite + iCS).add(startSite + jCS);
          }
        }
      }
      startSite += cSize;
    }

    // --- set initial number of sites --- //
    // --- (max key-value of `network´) --- //
    initialN = nSites;

    // --- evaluate clique-size distribution --- //
    int maxCliqueSize = 0;
    for (int i = 0; i < allCliqueSizes.length; i++) {
      if (allCliqueSizes[i] > maxCliqueSize) {
        maxCliqueSize = allCliqueSizes[i];
      }
    }
    cliqueSizeDistribution = new double[maxCliqueSize + 1];
    for (int i = 0; i < (maxCliqueSize + 1); i++) {
      cliqueSizeDistribution[i] = 0.0;
    }
    for (int iC = 0; iC < allCliqueSizes.length; iC++) {
      cliqueSizeDistribution[allCliqueSizes[iC]] += 1.0;
    }
  }
}

/**
 * General abstract facility - it is never instantiated - for the theoretical
 * analysis of general random graphs.
 * 
 * @author Claudius Gros
 * @author Mathias Linkerhand
 * @version 2011-04-05
 */
final class RandomGraphs {
  /** The network size. */
  static int          nSites                 = 0;
  /** The network of cliques contains clique of minimal clique size. */
  static int          minCliqueSize          = 0;
  /** The average link density. */
  static double       linkDensity            = 0d;
  /** The distribution of clique sizes. */
  static double[]     cliqueSizeDistribution = null;

  static final Logger logger                 = Logger.getLogger(RandomGraphs.class.getName());

  /**
   * Standard constructor - should never be called.
   */
  private RandomGraphs() {
  }

  /**
   * Constructs the network of cliques for given network size and link
   * density. <br>
   * 
   * <p>
   * --class RandomGraphs-- <br>
   */
  static void networkOfCliques() {
    assert ((nSites > 1) && (linkDensity > 0) && (linkDensity < 1));
    assert ((minCliqueSize > 1) && (minCliqueSize < nSites));
    // --- ---------------------------- --- //
    // --- the clique size distribution --- //
    // --- numbers, not probabilities --- //
    // --- ---------------------------- --- //
    cliqueSizeDistribution = new double[nSites + 1];
    for (int z = 1; z <= nSites; z++)
      cliqueSizeDistribution[z] = RandomGraphs.numberOfCliques(nSites, z, linkDensity);
    //
    int zMax = 1; // maximum of the distribution
    for (int z = 1; z <= nSites; z++)
      if (cliqueSizeDistribution[z] > cliqueSizeDistribution[zMax])
        zMax = z;
    // --- ------------------------------------- --- //
    // --- evaluate the number of new links --- //
    // --- viz the number of overlapping cliques --- //
    // --- ------------------------------------- --- //
    double nOverlappingCliques = 0.0; // number of cliques with common sites
    double smallNumber = 0.00001;
    //
    for (int z2 = minCliqueSize; z2 <= nSites; z2++)
      // size of second clique
      if ((z2 < zMax) || (cliqueSizeDistribution[z2] > smallNumber * cliqueSizeDistribution[zMax]))
        for (int z1 = minCliqueSize; z1 <= z2; z1++)
          // size of first clique
          if (z1 + z2 <= nSites) {
            double pNotOverlapping; // z1 and z2 have no common
            // sites
            pNotOverlapping = RandomGraphs.choose(nSites - z2, z1) / (RandomGraphs.choose(nSites, z1) - RandomGraphs.choose(z2, z1));
            // System.out.printf(Locale.ENGLISH,
            // "# RandomGraphs:%4d%4d%4d%10.6f\n",
            // nSites,z2,z1,pNotOverlapping);
            double factor = 1.0;
            if (z1 == z2)
              factor = 0.5;
            nOverlappingCliques += (1.0 - pNotOverlapping) * factor * cliqueSizeDistribution[z1] * cliqueSizeDistribution[z2];
          } // end of loop over all clique sizes
    // --- ------------------------------------------- --- //
    // --- update number of sites == number of cliques --- //
    // --- ------------------------------------------- --- //
    nSites = 0;
    for (int z = minCliqueSize; z < cliqueSizeDistribution.length; z++)
      nSites += (int) (cliqueSizeDistribution[z]);
    //
    linkDensity = nOverlappingCliques * 2.0 / (nSites * (nSites - 1.0));
    double coordinationNumber = linkDensity * (nSites - 1);
    logger.log(Level.FINE, String.format(Locale.ENGLISH, "# RandomGraphs.networkOfCliques%9d%4d%8.1f%14.10f\n", nSites, zMax, coordinationNumber, linkDensity));
  }

  /**
   * Number of z-cliques for a n-site Erdős–Rényi random graph and a
   * link-density p.
   * <p>
   * Returns (n over k) p^{k(k-1)/2} (1-p^k)^{n-k}
   */
  //
  static double numberOfCliques(int n, int z, double p) {
    if (n < z)
      return -1.0;
    if (z < 1)
      return -1.0;
    if ((p < 0.0) || (p > 1.0))
      return -1.0;

    double returnValue = RandomGraphs.choose(1.0 * n, 1.0 * z);
    double pToZ = Math.pow(p, 1.0 * z);
    returnValue = returnValue * Math.pow(pToZ, 0.5 * (z - 1));
    returnValue = returnValue * Math.pow(1.0 - pToZ, 1.0 * (n - z));
    return returnValue;
  }

  /**
   * Evaluates n!/( (n-k)!k! ) via log-double arithmetics.
   * <p>
   * Copyright (C) 2002-2006 Alexei Drummond and Andrew Rambaut GPL
   */
  static double choose(double n, double k) {
    n = Math.floor(n + 0.5);
    k = Math.floor(k + 0.5);
    double lchoose = RandomGraphs.lnGamma(n + 1.0) - RandomGraphs.lnGamma(k + 1.0) - RandomGraphs.lnGamma(n - k + 1.0);
    return Math.floor(Math.exp(lchoose) + 0.5);
  }

  /**
   * The log-Gamma function.
   * <p>
   * Copyright (C) 2002-2006 Alexei Drummond and Andrew Rambaut, GPL <br>
   * Pike MC and Hill ID (1966) Algorithm 291: Logarithm of the gamma
   * function. Communications of the Association for Computing Machinery,
   * 9:684
   */
  static double lnGamma(double alpha) {
    double x = alpha, f = 0.0, z;
    if (x < 7) {
      f = 1;
      z = x - 1;
      while (++z < 7) {
        f *= z;
      }
      x = z;
      f = -Math.log(f);
    }
    z = 1 / (x * x);
    return f + (x - 0.5) * Math.log(x) - x + 0.918938533204673
        + (((-0.000595238095238 * z + 0.000793650793651) * z - 0.002777777777778) * z + 0.083333333333333) / x;
  }
}

/**
 * Evaluation of general network statistics.
 * <p>
 * May be instantiated by every component. <br>
 * Contains subroutine for network statistics and some static auxiliary routines
 * for cluster and clique generation and supervision.
 * 
 * @author Claudius Gros
 * @author Gregor Kaczor
 * @author Mathias Linkerhand
 * @version 2011-06-21
 */
class StatNet {
  private static Random       random = new Random(0);
  /** module to be examined. */
  private Module              module;
  /** The network in terms of sites and their neighbors. */
  private List<List<Integer>> nodesAndTheirNeighbors;
  //
  /**
   * The list of pair of vertices in the network for which the first vertex is
   * dominated by the second vertex.
   * <p>
   * A vertex (I) is dominated by a vertex (II) if <br>
   * (a) (I) and (II) are connected by a vertex and if <br>
   * (b) the set of neighbors of (I) is a subset of the set of neighbors of
   * (II).
   */
  private List<List<Integer>> myDominatedVertices;
  // --- ------------------------------- ---
  // --- variables for degree statistics ---
  // --- ------------------------------- ---
  /** The total number of edges of the network. */
  public int                  degreeStatEdgeNumber;
  /** The density of edges relative to a fully connected network. */
  public double               degreeStatEdgeDensity;
  /** The clustering coefficient. */
  public double               degreeStatClusteringCoefficient;
  /** The number of sites with `k' links is given by `degreeStatLinkSum[k]'. */
  public int[]                degreeStatLinkSum;
  /** The maximal occurring degree. */
  public int                  degreeStatMaximalDegree;
  /** The average degree. */
  public double               degreeStatAverageDegree;
  // --- ------------------------------- ---
  // --- variables for clique statistics ---
  // --- ------------------------------- ---
  /** Container for all cliques found in the module. */
  Set<Set<Integer>>           DELETEcliques;
  /** The distribution of clique sizes. */
  int[]                       cliqueStatSizeSum;

  /** The maximal actual clique size. */
  public int                  cliqueStatMaximalSize;
  /** The average clique size. */
  public double               cliqueStatAverageSize;
  // --- ---------------------------------------------- ---
  // --- variables the statistics of dominated vertices ---
  // --- ---------------------------------------------- ---
  /** The number of vertices dominated by at least one other vertex. */
  public int                  nDomintedVertices;

  public StatNet(final Module module) {
    this.module = module;
    // this.cliques = new TreeSet<Set<Integer>>();
    this.myDominatedVertices = new Vector<List<Integer>>();

    // --- store network as a mapping of sites to neighbors ---
    this.nodesAndTheirNeighbors = new Vector<List<Integer>>();
    int size = module.getSize();
    for (int i = 0; i < size; i++) {
      final List<Integer> neighbors = new Vector<Integer>();
      for (int j = 0; j < size; j++) {
        if (module.getLink(i, j) > 0.0) {
          neighbors.add(j);
        }
      }
      this.nodesAndTheirNeighbors.add(neighbors);
    }
  }

  /**
   * Finds all dominated sites for the locally stored network. <br>
   * 
   * A vertex (I) is dominated by a vertex (II) if <br>
   * (a) (I) and (II) are connected by a vertex and if <br>
   * (b) the set of neighbors of (I) is a subset of the set of neighbors of
   * (II).
   */
  public void findAllDominatedVertices() {
    // --- ----------------------------------------- --- //
    // --- clear storage array and check the network --- //
    // --- ----------------------------------------- --- //
    this.myDominatedVertices.clear();
    int N = this.nodesAndTheirNeighbors.size();
    if (N < 3)
      return;
    // --- -------------------------------- --- //
    // --- go through all pairs of vertices --- //
    // --- -------------------------------- --- //
    for (int i = 0; i < N; i++)
      for (int j = 0; j < N; j++)
        if (i != j) {
          List<Integer> temp = this.nodesAndTheirNeighbors.get(i);
          Collections.sort(temp);
          if (Collections.binarySearch(temp, j) >= 0) // i,j connected
            if (StatNet.numberOfSharedSites(this.nodesAndTheirNeighbors.get(i), this.nodesAndTheirNeighbors.get(j)) == this.nodesAndTheirNeighbors.get(i)
                .size() - 1) // neighbors are
            // subset
            {
              List<Integer> newPair = new Vector<Integer>(2);
              newPair.add(i);
              newPair.add(j);
              this.myDominatedVertices.add(newPair);
            }
        }
    // --- -------------------------------------- --- //
    // --- evaluate the number of dominated sites --- //
    // --- -------------------------------------- --- //
    this.nDomintedVertices = 0;
    int lastVertex = -1;
    for (int i = 0; i < myDominatedVertices.size(); i++)
      if (myDominatedVertices.get(i).get(0) > lastVertex) {
        this.nDomintedVertices++;
        lastVertex = myDominatedVertices.get(i).get(0);
      }
  }

  /**
   * Calculates the clique statistics and the backward receptive fields of the
   * cliques. <br>
   * 
   * Examines only the internal links of the input component for the
   * determination cliques. <br>
   * Cliques are clusters of activities centers full connected by exitatory
   * links. <br>
   * Prints also all cliques found on seprated file fur possible further use.
   * In addition it evaluates the receptive fields of all cliques found by
   * looping over all backward modules of the input component seperately.
   */
  public void cliqueStat() {
    // --- local module infos --- //
    int N = this.nodesAndTheirNeighbors.size();
    if (N < 2)
      return;

    // --- `cliqueStatSizeSum[cS]' is the number of ---
    // -- of cliques with size cS\in[1,size] ---
    this.cliqueStatSizeSum = new int[N + 1];
    for (int i = 0; i < this.cliqueStatSizeSum.length; i++)
      this.cliqueStatSizeSum[i] = 0;

    // --- determine clique statistics ---
    for (final Set<Integer> clique : this.module.getCliques()) {
      this.cliqueStatSizeSum[clique.size()]++;
    }

    // --- determine maximal size of cliques ---
    this.cliqueStatMaximalSize = 0;
    for (int i = 0; i < this.cliqueStatSizeSum.length; i++)
      if (this.cliqueStatSizeSum[i] > 0)
        this.cliqueStatMaximalSize = i;

    // --- determine average clique size ---
    this.cliqueStatAverageSize = 0.0;
    for (int i = 0; i <= this.cliqueStatMaximalSize; i++)
      this.cliqueStatAverageSize += i * this.cliqueStatSizeSum[i];
    this.cliqueStatAverageSize = this.cliqueStatAverageSize / this.module.getCliques().size();

    // --- write to ByteArrayOutputStream via PrintWriter --- //
    ByteArrayOutputStream baos = new ByteArrayOutputStream();
    OutputStreamWriter osw = new OutputStreamWriter(baos);
    PrintWriter pw = new PrintWriter(osw);

    // --- evaluate return string for ".stat" output-file --- //
    pw.printf(Locale.ENGLISH, "# ********************* \n");
    pw.printf(Locale.ENGLISH, "# Clique statistics for `%s' \n", this.module.getName());
    pw.printf(Locale.ENGLISH, "# ********************* \n");
    pw.printf(Locale.ENGLISH, "# total number of cliques : %10d\n", this.module.getCliques().size());
    pw.printf(Locale.ENGLISH, "# average size of cliques : %10.2f\n", this.cliqueStatAverageSize);

    pw.printf(Locale.ENGLISH, "#   clique size |");
    for (int i = 1; i <= this.cliqueStatMaximalSize; i++)
      pw.printf(Locale.ENGLISH, " %5d", i);

    pw.printf(Locale.ENGLISH, "# clique number |");
    for (int i = 1; i <= this.cliqueStatMaximalSize; i++)
      pw.printf(Locale.ENGLISH, " %5d", this.cliqueStatSizeSum[i]);
    pw.printf(Locale.ENGLISH, "\n");

    pw.printf(Locale.ENGLISH, "# -------------------------------------------- \n");
    pw.printf(Locale.ENGLISH, "# All cliques found in internalLinks[][][0] of `%s' \n", this.module.getName());
    pw.printf(Locale.ENGLISH, "# -------------------------------------------- \n");
    for (final Set<Integer> clique : this.module.getCliques()) {
      pw.printf(Locale.ENGLISH, "# cliques | %2d |", clique.size());
      for (final Integer node : clique) {
        pw.printf(Locale.ENGLISH, " %4d", node);
      }
      pw.printf(Locale.ENGLISH, "\n");
    }
    //
    pw.flush();

    String outputFileName = "Brain_" + this.module.getName() + ".stat";
    try {
      PrintWriter printWriter = new PrintWriter(new File(Brain.DEFAULT_DATA_OUTPUT_DIRECTORY, outputFileName));
      printWriter.print(baos.toString());
    } catch (FileNotFoundException e) {
      e.printStackTrace();
    }
    baos.reset();

    // --- evaluate return string for ".cliq" output-file --- //
    pw.printf(Locale.ENGLISH, "# size = %d\n", N);
    pw.printf(Locale.ENGLISH, "#   clique size |");
    for (int i = 1; i <= this.cliqueStatMaximalSize; i++)
      pw.printf(Locale.ENGLISH, " %5d", i);
    //
    pw.printf(Locale.ENGLISH, "# clique number |");
    for (int i = 1; i <= this.cliqueStatMaximalSize; i++)
      pw.printf(Locale.ENGLISH, " %5d", this.cliqueStatSizeSum[i]);
    pw.printf(Locale.ENGLISH, "\n");
    //
    pw.printf(Locale.ENGLISH, "#nData %d \n", this.module.getCliques().size());
    //
    for (final Set<Integer> clique : this.module.getCliques()) {
      pw.printf(Locale.ENGLISH, "%d ", clique.size());
      for (final Integer node : clique) {
        pw.printf(Locale.ENGLISH, " %d", node);
      }
      pw.printf(Locale.ENGLISH, "\n");
    }

    pw.flush();
    outputFileName = "Brain_" + this.module.getName() + ".cliq";
    try {
      PrintWriter printWriter = new PrintWriter(new File(Brain.DEFAULT_DATA_OUTPUT_DIRECTORY, outputFileName));
      printWriter.print(baos.toString());
      printWriter.close();
    } catch (FileNotFoundException e) {
      e.printStackTrace();
    }
    baos.reset();
  }

  /**
   * Evaluates the receptive fields of all cliques found by looping over all
   * backward modules of the input component separately.
   */
  public void cliqueReceptiveFields() {
    ByteArrayOutputStream baos = new ByteArrayOutputStream();
    PrintWriter pw = new PrintWriter(new OutputStreamWriter(baos));

    // --- evaluate and print receptive fields of all cliques ---
    DhanModule localDhan = (DhanModule) this.module; // casting
    int inputSiteNumber = 0; // running, over all input modules
    for (final Module afferentModule : localDhan.getAfferentModules()) {
      for (int j = 0; j < afferentModule.getSize(); j++) {
        pw.printf(Locale.ENGLISH, "%5d", inputSiteNumber);
        inputSiteNumber++;
        for (final Set<Integer> clique : this.module.getCliques()) {
          double avField = 0.0;
          for (final Integer node : clique) {
            avField += localDhan.getAfferentLink(afferentModule, node, j);
          }
          avField = avField / clique.size();
          pw.printf(Locale.ENGLISH, "%8.3f", avField);
        }
        pw.printf(Locale.ENGLISH, "\n");
      }
      pw.flush();

      final String outputFileName = this.module.getName() + "_" + afferentModule.getName() + ".cliqueReceptiveFields";
      try {
        PrintWriter printWriter = new PrintWriter(new File(Brain.DEFAULT_DATA_OUTPUT_DIRECTORY, outputFileName));
        printWriter.print(baos.toString());
        printWriter.close();
      } catch (FileNotFoundException e) {
        e.printStackTrace();
      }
    }
  }

  /**
   * Evaluates the receptive fields of all cliques with respect to reference
   * input patters by looping over all backward modules of the input component
   * separately.
   */
  public void cliqueReferencePatterns() {
    final Module module = this.module;
    ByteArrayOutputStream baos = new ByteArrayOutputStream();
    OutputStreamWriter osw = new OutputStreamWriter(baos);
    PrintWriter pw = new PrintWriter(osw);

    // --- evaluate and print receptive fields of all cliques ---
    // --- with respect to the reference patterns of ---
    // --- one single backward input module ---
    for (final Module afferentModule : this.module.getAfferentModules()) {
      DhanModule localDhan = (DhanModule) module; // casting
      final File outputFile = new File(Brain.DEFAULT_DATA_OUTPUT_DIRECTORY, module.getName() + "_" + afferentModule.getName() + ".cliqueReferencePatterns");
      outputFile.delete();
      if (afferentModule instanceof PatternInputModule && localDhan.hasAfferentModule(afferentModule)) {
        final PatternInputModule patternInputModule = (PatternInputModule) afferentModule;
        try {
          patternInputModule.readReferencePattern(new File(Brain.DEFAULT_DATA_INPUT_DIRECTORY, "input.referencepatterns"));
        } catch (IOException e) {
          // TODO Auto-generated catch block
          e.printStackTrace();
        }
        for (int iRefPatt = 0; iRefPatt < patternInputModule.referencePatterns.size(); iRefPatt++) {
          OneStimulus refPatt = patternInputModule.referencePatterns.get(iRefPatt);
          pw.printf(Locale.ENGLISH, "%5d", iRefPatt);
          for (final Set<Integer> clique : this.module.getCliques()) {
            double avField = 0.0;
            for (final int node : refPatt.getNodes()) {
              final double strength = refPatt.getStrength(node);
              for (final Integer vertex : clique) {
                if (node < patternInputModule.getSize()) {
                  avField += localDhan.getAfferentLink(afferentModule, vertex, node) * strength;
                }
              }
            }
            avField = avField / clique.size();
            pw.printf(Locale.ENGLISH, "%8.3f", avField);
          }
          pw.printf(Locale.ENGLISH, "\n");
        }
        pw.flush();

        try {
          PrintWriter printWriter = new PrintWriter(new File(Brain.DEFAULT_DATA_OUTPUT_DIRECTORY, outputFile.getName()));
          printWriter.print(baos.toString());
          printWriter.close();
        } catch (FileNotFoundException e) {
          e.printStackTrace();
        }
        this.patternToClique(outputFile, new Module[] { patternInputModule });
      }
    }
  }

  /**
   * Evaluates the degree distribution and send it to the printing facility. <br>
   * 
   * Examines only the internal links of the input component. via the network
   * stored in myNetwork. <br>
   * Calculates also the clustering coefficient.
   * 
   */
  public void degreeDistribution() {
    // --- ----------------- --- //
    // --- check the network --- //
    // --- ----------------- --- //
    int N = this.nodesAndTheirNeighbors.size();
    if (N < 2)
      return;
    // --- -------------- ---
    // --- zero variables ---
    // --- -------------- ---
    this.degreeStatEdgeNumber = 0;
    this.degreeStatClusteringCoefficient = 0.0;
    this.degreeStatLinkSum = new int[N];
    for (int i = 0; i < N; i++)
      this.degreeStatLinkSum[i] = 0;
    // --- --------------------------------- ---
    // --- calculate the degree distribution ---
    // --- --------------------------------- ---
    for (int i = 0; i < N; i++) {
      this.degreeStatEdgeNumber += this.nodesAndTheirNeighbors.get(i).size();
      this.degreeStatLinkSum[this.nodesAndTheirNeighbors.get(i).size()]++;
    }
    this.degreeStatEdgeNumber = this.degreeStatEdgeNumber / 2;
    // --- ------------------------------------ ---
    // --- calculate the clustering coefficient ---
    // --- ------------------------------------ ---
    int nNetworkClusters = 0; // number of pairs of neighbors
    for (int iS = 0; iS < N; iS++) // loop over all network sites
    {
      int nNeighbors = this.nodesAndTheirNeighbors.get(iS).size();
      nNetworkClusters += nNeighbors * (nNeighbors - 1) / 2;
      //
      for (int i1 = 0; i1 < nNeighbors; i1++) {
        int neighbor1 = this.nodesAndTheirNeighbors.get(iS).get(i1);
        for (int i2 = 0; i2 < i1; i2++) {
          int neighbor2 = this.nodesAndTheirNeighbors.get(iS).get(i2);
          for (int ss = 0; ss < this.nodesAndTheirNeighbors.get(neighbor2).size(); ss++)
            if (this.nodesAndTheirNeighbors.get(neighbor2).get(ss) == neighbor1)
              this.degreeStatClusteringCoefficient++;
        }
      }
    } // end of loop over all vertices
      //
    this.degreeStatClusteringCoefficient = this.degreeStatClusteringCoefficient / nNetworkClusters;
    // --- ------------------------------- --- //
    // --- evaluate averages and densities --- //
    // --- ------------------------------- --- //
    this.degreeStatEdgeDensity = this.degreeStatEdgeNumber * 2 / ((N - 1.0) * N);
    //
    this.degreeStatMaximalDegree = 0;
    for (int k = 0; k < N; k++)
      if (this.degreeStatLinkSum[k] > 0)
        this.degreeStatMaximalDegree = k;
    //
    this.degreeStatAverageDegree = 0.0;
    for (int k = 0; k < N; k++)
      this.degreeStatAverageDegree += k * this.degreeStatLinkSum[k];
    this.degreeStatAverageDegree = this.degreeStatAverageDegree / N;
    //
    int degreeSum = 0;
    for (int k = 0; k <= this.degreeStatMaximalDegree; k++)
      degreeSum += this.degreeStatLinkSum[k];
    // --- ---------------------------------------------- --- //
    // --- write to ByteArrayOutputStream via PrintWriter --- //
    // --- ---------------------------------------------- --- //
    ByteArrayOutputStream baos = new ByteArrayOutputStream();
    OutputStreamWriter osw = new OutputStreamWriter(baos);
    PrintWriter pw = new PrintWriter(osw);
    // --- --------------------- --- //
    // --- evaluate print string --- //
    // --- --------------------- --- //
    pw.printf(Locale.ENGLISH, "# ********************* \n");
    pw.printf(Locale.ENGLISH, "# Network statistics of `%s' \n", this.module.getName());
    pw.printf(Locale.ENGLISH, "# ********************* \n");
    pw.printf(Locale.ENGLISH, "#       total number of edges : %9d %9.4f  : edge density\n", this.degreeStatEdgeNumber, this.degreeStatEdgeDensity);
    pw.printf(Locale.ENGLISH, "#              average degree : %9.4f %9.4f  : clustering coefficient\n", this.degreeStatAverageDegree,
        this.degreeStatClusteringCoefficient);
    pw.printf(Locale.ENGLISH, "#              maximal degree : %9d\n", this.degreeStatMaximalDegree);
    pw.printf(Locale.ENGLISH, "# ----------------------- \n");
    pw.printf(Locale.ENGLISH, "# The degree distribution \n");
    pw.printf(Locale.ENGLISH, "# ----------------------- \n");
    pw.printf(Locale.ENGLISH, "#  k      size(k)      P(k)\n");
    for (int k = 0; k <= this.degreeStatMaximalDegree; k++)
      pw.printf(Locale.ENGLISH, " %3d %9d %9.4f\n", k, this.degreeStatLinkSum[k], this.degreeStatLinkSum[k] / (1.0 * degreeSum));
    //
    pw.flush();

    // --- send string to printing facility --- //
    String outputFileName = "Brain_" + this.module.getName() + ".stat";
    try {
      PrintWriter printWriter = new PrintWriter(new File(Brain.DEFAULT_DATA_OUTPUT_DIRECTORY, outputFileName));
      printWriter.print(baos.toString());
      printWriter.close();
    } catch (FileNotFoundException e) {
      e.printStackTrace();
    }
    baos.reset();
  }

  /**
   * Generate an ordered cluster with `Zin' sites out of `Nin' sites.
   * 
   * Uses the random number generator instantiated as `random' in `Brain'.
   */
  static int[] generateCluster(int Nin, int Zin) {
    assert ((Zin > 0) && (Zin <= Nin)) : "# (StatNet-genClus) call error";
    int[] result = new int[Zin]; // the return object
    // ------------------------------
    // ----- random cluster -----
    // ------------------------------
    result[0] = random.nextInt(Nin); // random integer in [0,Nin[
    for (int i_z = 1; i_z < Zin; i_z++) {
      int nextSiteIsDifferent = 0;
      int nextSite = -1;
      while (nextSiteIsDifferent == 0) {
        nextSiteIsDifferent = 1;
        nextSite = random.nextInt(Nin); // random integer in [0,Nin[
        for (int j_z = 0; j_z < i_z; j_z++)
          if (nextSite == result[j_z])
            nextSiteIsDifferent = 0;
      }
      result[i_z] = nextSite;
    }
    // -----------------------------------
    // ----- order cluster sites -----
    // -----------------------------------
    Arrays.sort(result);
    return result;
  }

  /**
   * Evaluates the number of shared sites in two strictly montonically ordered
   * input integer arrays.
   * 
   * @param cliqueA
   *            a set A of integers as an int[]
   * @param cliqueB
   *            a set B of integers as an int[]
   * @return the size of the intersection of A with B
   */
  //
  static int numberOfSharedSites(int[] cliqueA, int[] cliqueB) {
    if ((cliqueA.length == 0) || (cliqueB.length == 0))
      return 0;
    // --- -------------- --- //
    // --- check ordering --- //
    // --- -------------- --- //
    for (int iA = 0; iA < cliqueA.length - 1; iA++)
      assert (cliqueA[iA] < cliqueA[iA + 1]);
    for (int iB = 0; iB < cliqueB.length - 1; iB++)
      assert (cliqueB[iB] < cliqueB[iB + 1]);
    // --- ---------------- --- //
    // --- evaluate overlap --- //
    // --- ---------------- --- //
    int iA = 0;
    int iB = 0;
    int returnValue = 0;
    while ((iA < cliqueA.length) && (iB < cliqueB.length)) {
      if (cliqueA[iA] == cliqueB[iB]) {
        returnValue++;
        iA++;
        iB++;
      } else if (cliqueA[iA] < cliqueB[iB])
        iA++;
      else
        iB++;
    }
    return returnValue;
  }

  static int numberOfSharedSites(List<Integer> cliqueA, List<Integer> cliqueB) {
    return numberOfSharedSites(ArrayUtils.toPrimitive(cliqueA.toArray(new Integer[0])), ArrayUtils.toPrimitive(cliqueB.toArray(new Integer[0])));
  }

  /**
   * Processes Brain_dhan_input.cliqRefPatt and statistically evalutes the
   * mapping between a pattern and the winning clique(s), the absolute and
   * relative difference to the second clique and the span between the winner
   * clique and the most non matching clique.
   * 
   * The method finally sumarizes the results. First by giving the rate by
   * dividing the number of distinct winner cliques by the number of patterns.
   * 100% does mean there is a one to one mapping between every pattern and
   * clique. Then by giving the average of the absolute and relative
   * difference as well as the span.
   */
  public void patternToClique(final File inputFile, Module[] modules) {
    final ByteArrayOutputStream baos = new ByteArrayOutputStream();
    final PrintWriter pw = new PrintWriter(new OutputStreamWriter(baos));

    final File outputFile = new File(Brain.DEFAULT_DATA_OUTPUT_DIRECTORY, inputFile.getName().concat(".cliqToPatt"));
    // FacilPrint.removeFile(outputFile);
    outputFile.delete();
    BufferedReader bfr = null;
    try {
      bfr = new BufferedReader(new FileReader(inputFile));
    } catch (FileNotFoundException e) {
      e.printStackTrace();
      return;
    }
    double averageRelDist = 0d;
    double averageAbsDist = 0d;
    double averageMinMaxSpan = 0d;
    String line = "";
    int patternCount = 0;
    TreeSet<Integer> distinctWinningcliques = new TreeSet<Integer>();
    // DecimalFormat df1 = new DecimalFormat("00%");
    DecimalFormat df2 = new DecimalFormat("#0.00000");

    pw.printf(Locale.ENGLISH, "pattern->clique\t\tabsolute difference\trelative difference\t\tmax-min=span\n");
    pw.printf(Locale.ENGLISH, "\t\t\t\t\tto second\t\t\tto second\n");
    pw.printf(Locale.ENGLISH, "\t\t\t\t\tclique\t\t\t\tclique\n");
    pw.printf(Locale.ENGLISH, "------------------------------------------------------------------------------------\n");

    try {
      while ((line = bfr.readLine()) != null) {
        line = line.trim();
        if (line.startsWith("#") || line.length() == 0) {
          continue;
        }
        do {
          line = line.replaceAll("  ", " ");
        } while (line.indexOf("  ") > -1);
        patternCount++;
        String tmp[] = line.split(" ");

        pw.printf(Locale.ENGLISH, tmp[0] + "->");
        double max = Double.NEGATIVE_INFINITY;
        double min = Double.MAX_VALUE;
        double value = 0;
        int index = -1;
        for (int i = 1; i < tmp.length; i++) {
          value = Double.parseDouble((tmp[i]));
          if (max < value) {
            max = value;
            index = i;
          }
          if (min > value) {
            min = value;
          }
        }

        pw.printf(Locale.ENGLISH, Integer.toString(index));
        distinctWinningcliques.add(index);

        double dtmp[] = new double[tmp.length];
        HashMap<Double, Integer> row = new HashMap<Double, Integer>();
        for (int i = 1; i < tmp.length; i++) {
          dtmp[i] = Double.valueOf(tmp[i]);
          row.put(dtmp[i], i);
        }
        Arrays.sort(dtmp);
        pw.printf(Locale.ENGLISH, "\t\t\t\t\t" + df2.format((dtmp[dtmp.length - 1] - dtmp[dtmp.length - 2])));
        averageAbsDist += ((dtmp[dtmp.length - 1] - dtmp[dtmp.length - 2]));

        dtmp = new double[tmp.length];
        row = new HashMap<Double, Integer>();
        for (int i = 1; i < tmp.length; i++) {
          dtmp[i] = Double.valueOf(tmp[i]);
          row.put(dtmp[i], i);
        }
        Arrays.sort(dtmp);
        pw.printf(Locale.ENGLISH, "\t\t\t" + df2.format(((dtmp[dtmp.length - 1] - dtmp[dtmp.length - 2]) / (max - min))));
        averageRelDist += ((dtmp[dtmp.length - 1] - dtmp[dtmp.length - 2]) / (max - min));

        pw.printf(Locale.ENGLISH, "\t\t\t\t\t" + df2.format(max) + "-" + df2.format(min) + "=" + df2.format(max - min));
        averageMinMaxSpan += (max - min);
        pw.printf(Locale.ENGLISH, "\n");
      }
      bfr.close();
    } catch (IOException exception) {
      exception.printStackTrace();
    }
    pw.printf(Locale.ENGLISH, "------------------------------------------------------------------------------------\n");
    pw.printf(Locale.ENGLISH, "%3.2f", distinctWinningcliques.size() / (double) patternCount * 100);
    pw.printf(
        Locale.ENGLISH,
        "\t\t\t\t\t" + df2.format(averageAbsDist / patternCount) + "\t\t\t" + df2.format(averageRelDist / patternCount) + "\t\t\t"
            + df2.format(averageMinMaxSpan / patternCount));
    pw.flush();

    try {
      PrintWriter printWriter = new PrintWriter(new File(Brain.DEFAULT_DATA_OUTPUT_DIRECTORY, outputFile.getName()));
      printWriter.print(baos.toString());
      printWriter.close();
    } catch (FileNotFoundException e) {
      e.printStackTrace();
    }
  }
}

/**
 * Auxiliary class for finding, storing and printing of intra- and inter-module
 * cluster of active centers and for the examination of the such generated
 * time-series of clusters. <br>
 * It may be instantiated by every component.
 * 
 * @author Claudius Gros
 * @author Gregor Kaczor
 * @author Mathias Linkerhand
 * @version 2011-11-23
 */
class StatActiClus {
  /** ArrayList with all clusters found so far. */
  public List<OneCluster> allClusters        = new Vector<OneCluster>();
  /** Here we calculate Cycles on the fly* */
  public CycleContainer   cycleContainer     = new CycleContainer();
  /** ActiveCluster searches these modules for combined active cliques. */
  public Set<Module>      modules;
  /**
   * Time-interval for searching for active clusters. The default value for
   * `clusterSearchTimeInterval' may be changed at run-time.
   */
  public double           searchTimeInterval = 4.0;
  /**
   * Activity center is active if `activities[i]' is bigger than
   * `criticalActive'.
   */
  public double           criticalActive     = 0.70;
  /**
   * Activity center is active if `reservoir[i]' is smaller than
   * `criticalReservoir'.
   */
  public double           criticalReservoir  = 0.70;

  public StatActiClus(final Set<Module> modules) {
    this.modules = modules;
  }

  public StatActiClus(final Module[] modules) {
    this(new HashSet<Module>(Arrays.asList(modules)));
  }

  /**
   * Finds the currently active cluster. If an active cluster is found
   * <p>
   * May be called any time, as it determines itself if enough time
   * `clusterSearchTimeInterval' has passed to actually perform the search for
   * an active cluster.
   * <p>
   * -- class StatActiClus -- <br>
   */
  public void findActiveCluster(final double currentTime) {
    // new cluster stored in temporary ArrayList ---
    ArrayList<OneCenter> newActiveCluster = new ArrayList<OneCenter>();

    // search in all modules
    for (final Module module : this.modules) {
      if (module instanceof DhanModule) {
        DhanModule dhan = (DhanModule) module;
        for (int site = 0; site < module.getSize(); site++) {
          if ((dhan.getActivity(site) > this.criticalActive) && (dhan.getReservoir(site) < this.criticalReservoir)) {
            newActiveCluster.add(new OneCenter(dhan, site));
          }
        }
      }
      if (module instanceof PatternInputModule) {
        throw new RuntimeException("not implemented");
      }
    }

    // create new cluster
    OneCluster activeCluster;
    if (newActiveCluster.size() > 1) {
      activeCluster = new OneCluster(newActiveCluster, currentTime);
    } else
      return;
    int indexPrevCluster = this.allClusters.size() - 1; // of previous
    // cluster

    // set new duration-time if active cluster is not new
    if (indexPrevCluster >= 0) {
      if (OneCluster.equals(activeCluster, this.allClusters.get(indexPrevCluster))) {
        this.allClusters.get(indexPrevCluster).durationTime = currentTime - this.allClusters.get(indexPrevCluster).startTime;
        return; // new cluster == old cluster
      }
    }

    // check wether active cluster is subset of old one
    if (indexPrevCluster >= 0)
      if (OneCluster.isSubCluster(activeCluster, this.allClusters.get(indexPrevCluster))
      /* && !this.allClusters.get(indexPrevCluster).isOveractive("dhan") */) {
        return; // new cluster not a clique
      }

    // check wether old cluster is subset of new cluster
    if (indexPrevCluster >= 0)
      if (OneCluster.isSubCluster(this.allClusters.get(indexPrevCluster), activeCluster)
      /* && !this.allClusters.get(indexPrevCluster).isOveractive("dhan") */) {
        this.allClusters.set(indexPrevCluster, activeCluster);
        // printClusterToFile(indexPrevCluster);
        return; // old cluster replaced by new one
      }

    // active cluster is novel
    this.allClusters.add(activeCluster); // add new active cluster
    if (indexPrevCluster >= 0 && this.allClusters.get(indexPrevCluster).durationTime > 0) {
      this.printClusterToFile(indexPrevCluster);
    }
  }

  /**
   * Send the active cluster to the general printing facility.
   * 
   * <p>
   * --class StatActiClus-- <br>
   * 
   * @param whichCluster
   *            index of a cluster
   */
  public void printClusterToFile(int whichCluster) {
    OneCluster printCluster = allClusters.get(whichCluster);
    String outputFileName = "Brain_";
    for (final Module module : this.modules) {
      outputFileName += module.getName();
    }
    outputFileName += ".actiClus";

    final ByteArrayOutputStream baos = new ByteArrayOutputStream();
    final OutputStreamWriter osw = new OutputStreamWriter(baos);
    final PrintWriter pw = new PrintWriter(osw);

    char[] positionIdicator = new char[printCluster.clusterSites[0].module.getSize()];
    Arrays.fill(positionIdicator, '_');

    // evaluate return string
    pw.printf(Locale.ENGLISH, "%5.0f %5.0f", printCluster.startTime, printCluster.durationTime);
    for (int iC = 0; iC < printCluster.clusterSites.length; iC++) {
      positionIdicator[printCluster.clusterSites[iC].index] = Integer.toString(printCluster.clusterSites[iC].index, Character.MAX_RADIX).charAt(0);
      pw.printf(Locale.ENGLISH, "  (%s|%3d)", printCluster.clusterSites[iC].module.getName(), printCluster.clusterSites[iC].index);
    }
    pw.printf(Locale.ENGLISH, "\n");
    pw.flush();

    try {
      PrintWriter printWriter = new PrintWriter(new File(Brain.DEFAULT_DATA_OUTPUT_DIRECTORY, outputFileName));
      printWriter.print(baos.toString());
      printWriter.close();
    } catch (FileNotFoundException e) {
      e.printStackTrace();
    }
  }

  /**
   * Evaluate the statistics of all found active clusters and prints them to
   * file.
   * <p>
   * Should be called at the end.
   * <p>
   * -- class StatActiClus -- <br>
   */
  public void evaluateStatClusters() {
    final Collection<OneCluster> clusters = this.allClusters;

    // diffClusters contains all non-identical clusters found
    // as defined by arrays of activities centers
    final HashMap<OneCluster, Integer> allDiffClusters = new HashMap<OneCluster, Integer>();
    double meanActivationTime = 0; // of all clusters
    double minActiTimeFraction = 0.15;// minimal duration of activities
    // for a cluster to enter stat-evaluation
    for (final OneCluster currClus : clusters) {
      meanActivationTime += currClus.durationTime;
    }
    meanActivationTime = meanActivationTime / clusters.size();

    for (final OneCluster currClus : clusters) {
      if (currClus.durationTime > minActiTimeFraction * meanActivationTime)
        if (allDiffClusters.containsKey(currClus)) {
          allDiffClusters.put(currClus, allDiffClusters.get(currClus) + 1);
        } else {
          allDiffClusters.put(currClus, 1);
        }
    }

    // --- ---------------------------------------------- --- //
    // --- write to ByteArrayOutputStream via PrintWriter --- //
    // --- ---------------------------------------------- --- //
    ByteArrayOutputStream baos = new ByteArrayOutputStream();
    OutputStreamWriter osw = new OutputStreamWriter(baos);
    PrintWriter pw = new PrintWriter(osw);
    String outputFileName = "Brain_";
    for (final Module module : this.modules) {
      outputFileName += module.getName();
    }
    outputFileName += ".actiClusStat";

    // output-string for cluster statistics
    pw.printf(Locale.ENGLISH, "# number of activations of the activated clusters \n");
    pw.printf(Locale.ENGLISH, "# number of activated clusters : %6d \n", allDiffClusters.size());
    pw.printf(Locale.ENGLISH, "#         mean activities time : %6.1f \n", meanActivationTime);
    pw.printf(Locale.ENGLISH, "#       minmal activities time : %6.1f \n", minActiTimeFraction * meanActivationTime);
    pw.printf(Locale.ENGLISH, " \n");

    for (final OneCluster cluster : allDiffClusters.keySet()) {
      pw.printf(Locale.ENGLISH, "%4d ", allDiffClusters.get(cluster));
      for (final OneCenter cite : cluster.clusterSites) {
        int iSite = cite.index;
        pw.printf(Locale.ENGLISH, "  (%s|%3d)", cite.module.getName(), iSite);
      }
      pw.printf(Locale.ENGLISH, "\n");
    }

    // send output string to print facility
    pw.flush();
    try {
      PrintWriter printWriter = new PrintWriter(new File(Brain.DEFAULT_DATA_OUTPUT_DIRECTORY, outputFileName));
      printWriter.print(baos.toString());
      printWriter.close();
    } catch (FileNotFoundException e) {
      e.printStackTrace();
    }
    baos.reset();
  }

  /**
   * Evaluates sequences of active clusters. Saves all sequences to
   * Brain_modulename.acSequence file with the start time, duration and the
   * sequence as String representation. Duration is the sum of durations of
   * every active cluster.
   */
  public void evaluateSequencesOfActiveClusters() {
    // for every module
    for (final Module module : this.modules) {
      String path = preprocessFile(module.getName(), 20);
      CycleContainer cyclesFound = StatActiClus.cycleDetector(path, module);

      // --- ---------------------------------------------- --- //
      // --- write to ByteArrayOutputStream via PrintWriter --- //
      // --- ---------------------------------------------- --- //
      ByteArrayOutputStream baos = new ByteArrayOutputStream();
      OutputStreamWriter osw = new OutputStreamWriter(baos);
      PrintWriter pw = new PrintWriter(osw);
      String outputFileName = "Brain_" + module.getName() + ".acSequence";

      // --- ---------------------------------- ---
      // --- output-string for cycle statistics ---
      // --- ---------------------------------- ---
      // pw.printf(Locale.ENGLISH,
      // "# number of activations of cycles \n");
      pw.printf(Locale.ENGLISH, "# number of activated sequences : %6d \n", cyclesFound.getCycles().size());
      pw.printf(Locale.ENGLISH, " \n");

      for (Double key : cyclesFound.getChronoSortedCycles().keySet()) {
        OneCycle oneCycle = cyclesFound.getChronoSortedCycles().get(key);
        pw.printf(Locale.ENGLISH, "%4d ", (int) (oneCycle.getStartTime()));
        pw.printf(Locale.ENGLISH, "%4d ", (int) oneCycle.getDurationTime());
        pw.printf(Locale.ENGLISH, oneCycle.toString());
        pw.printf(Locale.ENGLISH, " \n");
      }

      pw.flush();
      try {
        PrintWriter printWriter = new PrintWriter(new File(Brain.DEFAULT_DATA_OUTPUT_DIRECTORY, outputFileName));
        printWriter.print(baos.toString());
        printWriter.close();
      } catch (FileNotFoundException e) {
        e.printStackTrace();
      }
      baos.reset();
    }
  } // end of StatActiClus.evaluateSequencesOfActiveClusters()

  /**
   * Evaluates cycles and saves them in a Brain_modulename.statCycle file. A
   * cycle is defined as a sequence of active clusters where an active
   * clusters repeats itself in the data stream. This cycle must be recurrent
   * otherwise it is not a cycle.
   * 
   */
  public void evaluateStatCycles() {
    // for every module
    for (final Module module : this.modules) {
      String path = preprocessFile(module.getName(), 20);
      CycleContainer cyclesFound = cycleDetector(path, module);

      // --- ---------------------------------------------- --- //
      // --- write to ByteArrayOutputStream via PrintWriter --- //
      // --- ---------------------------------------------- --- //
      ByteArrayOutputStream baos = new ByteArrayOutputStream();
      OutputStreamWriter osw = new OutputStreamWriter(baos);
      PrintWriter pw = new PrintWriter(osw);
      String outputFileName = "Brain_" + module.getName() + ".statCycle";

      // --- ---------------------------------- ---
      // --- output-string for cycle statistics ---
      // --- ---------------------------------- ---
      // pw.printf(Locale.ENGLISH,
      // "# number of activations of cycles \n");
      pw.printf(Locale.ENGLISH, "# number of activated cycles   : %6d \n", cyclesFound.getCycles().size());
      pw.printf(Locale.ENGLISH, "\n");

      for (Object element : cyclesFound.getCycles().keySet()) {
        String type = (String) element;
        OneCycle oneCycle = cyclesFound.getCycles().get(type);
        // for(Object cycleOccurence: oneCycle.getOccurenceTime()){
        // OneCycle cycle=(OneCycle)cycleOccurence;
        // pw.printf(Locale.ENGLISH,
        // cycle.getStartTime()+" "+cycle.getDurationTime()+"
        // "+cycle.toString());
        // pw.printf(Locale.ENGLISH, " \n");
        // }
        if (oneCycle.getOccurenceTime().size() > 1) {
          // System.out.println(oneCycle.getOccurenceTime().size()+"x
          // "+oneCycle);
          pw.printf(Locale.ENGLISH, "%4d ", oneCycle.getOccurenceTime().size());
          pw.printf(Locale.ENGLISH, oneCycle.toString());
          // pw.printf(Locale.ENGLISH,
          // oneCycle.getOccurenceTime().size()+" "+oneCycle);
          pw.printf(Locale.ENGLISH, "\n");
        }
      }
      //
      // }
      pw.flush();
      try {
        PrintWriter printWriter = new PrintWriter(new File(Brain.DEFAULT_DATA_OUTPUT_DIRECTORY, outputFileName));
        printWriter.print(baos.toString());
        printWriter.close();
      } catch (FileNotFoundException e) {
        e.printStackTrace();
      }
      baos.reset();
    }
  } // end of StatActiClus.evaluateStatCycles()

  /**
   * Detect cycles in the file .actiClus of a module
   * 
   * @param path
   *            a path to the .actiClus file
   * @param module
   *            the modulename
   * @return a CycleContainer which holds all cycles detected in the file
   */
  private static CycleContainer cycleDetector(String path, Module module) {
    CycleContainer cycleContainer = new CycleContainer();
    // open file for reading
    File inputFile = new File(path);
    FileReader fr;
    BufferedReader bfr = null;
    try {
      fr = new FileReader(inputFile);
      bfr = new BufferedReader(fr);
    } catch (FileNotFoundException e) {
      System.err.println("failed to open " + inputFile.getAbsolutePath());
      e.printStackTrace();
      return cycleContainer;
    }
    String line;
    try {
      while ((line = bfr.readLine()) != null) {
        line = line.trim();
        String data[] = line.split(" ");
        if (data.length < 3)
          continue;
        // System.out.println(line);
        ArrayList<OneCenter> activeACs = new ArrayList<OneCenter>();
        for (int i = 2; i < data.length; i++) {
          OneCenter oneCenter = new OneCenter(module, Integer.parseInt(data[i]));
          activeACs.add(oneCenter);
        }
        OneCluster oneCluster = new OneCluster(activeACs, Double.parseDouble(data[0]), Double.parseDouble(data[1]));
        cycleContainer.addCandidate(oneCluster);
      }
      bfr.close();
    } catch (IOException e) {
      e.printStackTrace();
    }
    return cycleContainer;
  }

  /**
   * Preprocesses the .actiClus file of a module in order to sort out clusters
   * with a duration below the threshold. The method produces a file
   * processing.tmp and returns the files absolute path for further processing
   * 
   * @param modulename
   * @param treshold
   */
  private static String preprocessFile(String modulename, int treshold) {
    final File file = new File(Brain.DEFAULT_DATA_OUTPUT_DIRECTORY, "Brain_" + modulename + ".actiClus");

    // open input file for reading
    FileReader fr;
    BufferedReader bfr = null;
    try {
      fr = new FileReader(file);
      bfr = new BufferedReader(fr);
    } catch (FileNotFoundException e) {
      System.err.println("failed to open " + file.getAbsolutePath());
      return null;
    }

    // open temporary output file for writing
    File outputFile = null;
    BufferedWriter bfw = null;
    try {
      outputFile = File.createTempFile("Brain", ".tmp");
      final FileWriter fw = new FileWriter(outputFile);
      bfw = new BufferedWriter(fw);
    } catch (FileNotFoundException e) {
      // e.printStackTrace();
      System.err.println("failed to open tempory file");
      return null;
    } catch (IOException exception) {
      exception.printStackTrace();
      return null;
    }

    // read Brain_module.actiClus
    String line = "";
    boolean write = false;
    try {
      while ((line = bfr.readLine()) != null) {
        while (line.indexOf("  ") > -1)
          // replace many white-spaces
          line = line.replaceAll("  ", " "); // by a single
        // white-space
        // line = removeDuplicateWhitespace(line).toString();
        if (line.startsWith("#")) {
          write = true;
          continue;
        }

        if (write) {
          line = line.trim();
          String data[] = line.split(" ");
          if (Integer.parseInt(data[1]) <= treshold)
            continue;
          bfw.write(data[0] + " " + data[1] + " ");
          List<Integer> sitesArray = getSites(modulename, line);
          for (int i = 0; i < sitesArray.size(); i++) {
            bfw.write(sitesArray.get(i) + " ");
          }
          bfw.write("\n");
        }
      }
      bfw.close();
      bfr.close();
    } catch (IOException e) {

      e.printStackTrace();
    }
    return outputFile.getAbsolutePath();
  }

  /**
   * Returns an ArrayList of active cluster sites from a String line read from
   * .actiClus file
   * 
   * @param line
   *            line in this format start time durationTime (modulename|site)
   *            (modulename|site)… example: 12 40 (dhanModule| 0) (dhanModule|
   *            10) (dhanModule| 18)
   * @param modulename
   *            name of the module
   * @return sites of the active cluster in an ArrayList
   */
  private static List<Integer> getSites(String modulename, String line) {
    List<Integer> result = new ArrayList<Integer>();
    int pos = 0;
    String site = "";
    while ((pos = line.indexOf("(" + modulename + "|", pos + 1)) > -1) {
      site = line.substring(pos + 2 + modulename.length(), line.indexOf(")", pos + 6));
      result.add(Integer.valueOf(site.trim()));
    }
    return result;
  }

  /**
   * Member method to get allClusters
   * 
   * @return the allClusters
   */
  public List<OneCluster> getAllClusters() {
    return allClusters;
  }
}

/**
 * Evaluates to which extend a given module learns patterns present in its own
 * `bias[ ]'.
 * 
 * @author Claudius Gros
 * @author Mathias Linkerhand
 * @version 2010-01-18
 */
class StatBiasLearn {
  /** Name of the module to monitor. */
  public DhanModule             dhanModule;
  /** ArrayList with all patterns found in the `bias[]' so far. */
  public ArrayList<OneStimulus> allBiasPattern = new ArrayList<OneStimulus>();

  /** A `bias[]' is considered active if above the `criticalBias'. */
  public double                 criticalBias   = 0.001;

  public StatBiasLearn(final DhanModule module) {
    this.dhanModule = module;
  }

  /**
   * Check for patterns in the bias-input.
   * 
   */
  public void findBiasPattern(final double currentTime) {
    // --- new pattern stored in temporary ArrayList ---
    ArrayList<OneCenter> newBiasPattern = new ArrayList<OneCenter>();

    // --- search in given module ---
    Module localModule = this.dhanModule;
    for (int i = 0; i < localModule.getSize(); i++) {
      if (localModule.getBias(i) > criticalBias)
        newBiasPattern.add(new OneCenter(this.dhanModule, i));
    }

    // --- create current pattern as `OneStimulus', if big enough ---
    // --- otherwise check for `durationTimeList' of previous pattern ---
    OneStimulus activePattern;
    if (newBiasPattern.size() > 1)
      activePattern = new OneStimulus();
    else {
      int currentSize = this.allBiasPattern.size();
      if (currentSize > 0)
        if (this.allBiasPattern.get(currentSize - 1).getEndTime() == 0)
          this.allBiasPattern.get(currentSize - 1).interval = new Interval(this.allBiasPattern.get(currentSize - 1).getStartTime(), currentTime);
      return;
    }

    // --- store infos about current pattern ---
    activePattern.interval = new Interval(currentTime, activePattern.getEndTime()); // starting time

    for (int i = 0; i < newBiasPattern.size(); i++) {
      final int node = newBiasPattern.get(i).index;
      final double strength = localModule.getBias(node);
      activePattern.setStrength(node, strength);
    }

    // --- check if active pattern is the first --- //
    int currentSize = this.allBiasPattern.size();
    if (currentSize == 0) {
      this.allBiasPattern.add(activePattern);
      return;
    }

    // check if previous pattern is equal to current one; if yes, add bias
    // strengths, if no, store current pattern
    OneStimulus prevPattern = this.allBiasPattern.get(currentSize - 1);
    if (prevPattern.getNodes().equals(activePattern.getNodes())) {
      for (final int node : activePattern.getNodes()) {
        final double strength = activePattern.getStrength(node);
        prevPattern.setStrength(node, prevPattern.getStrength(node) + strength);
      }
    } else {
      this.allBiasPattern.add(activePattern);
      if (prevPattern.getEndTime() == 0)
        prevPattern.interval = new Interval(prevPattern.getStartTime(), currentTime);
    }
  }

  /**
   * Selects data for output on file and sends the print strings to the
   * printing facility.
   * 
   */
  public void dataOutputToFile() {
    // --- write to ByteArrayOutputStream via PrintWriter --- //
    ByteArrayOutputStream baos = new ByteArrayOutputStream();
    OutputStreamWriter osw = new OutputStreamWriter(baos);
    PrintWriter pw = new PrintWriter(osw);
    String outputFileName;

    // --- output-string for patterns found ---
    pw.printf(Locale.ENGLISH, "# bias patterns \n");
    double rescaleBias; // strength have been accumulated
    boolean patternIsClique; // has the pattern been learned?
    String learnResult; // output-string of learn result
    //
    for (OneStimulus biasPattern : allBiasPattern) {
      // determine learn result
      patternIsClique = this.dhanModule.isClique(biasPattern.getNodes());
      if (patternIsClique) {
        learnResult = "    learned";
      } else {
        learnResult = "not learned";
      }

      rescaleBias = 1.0; // determine scale

      pw.printf(Locale.ENGLISH, "#\n"); // output of sites
      pw.printf(Locale.ENGLISH, "%7.1f %7.1f  ", biasPattern.getStartTime(), biasPattern.getEndTime());
      for (final int node : biasPattern.getNodes()) {
        pw.printf(Locale.ENGLISH, "%6d", node);
      }
      pw.printf(Locale.ENGLISH, "\n"); // output of strengths
      pw.printf(Locale.ENGLISH, "%15s  ", learnResult);
      for (final int node : biasPattern.getNodes()) {
        pw.printf(Locale.ENGLISH, "%6.2f", biasPattern.getStrength(node) * rescaleBias);
      }
      pw.printf(Locale.ENGLISH, "\n");
    }
    pw.flush();
    outputFileName = "Brain_" + this.dhanModule.getName() + ".learn";
    try {
      PrintWriter printWriter = new PrintWriter(new File(Brain.DEFAULT_DATA_OUTPUT_DIRECTORY, outputFileName));
      printWriter.print(baos.toString());
      printWriter.close();
    } catch (FileNotFoundException e) {
      e.printStackTrace();
    }
    baos.reset();
  }
}