package brain;

import java.io.File;
import java.io.IOException;
import java.io.PrintWriter;
import java.io.Serializable;
import java.util.ArrayList;
import java.util.Collections;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Random;
import java.util.Set;
import java.util.TreeMap;
import java.util.Vector;
import java.util.logging.Level;
import java.util.logging.Logger;

import org.jdom.Document;
import org.jdom.Element;
import org.jdom.JDOMException;
import org.jdom.input.SAXBuilder;

import utils.Graphs;
import utils.Utils;
import brain.gui.table.ObjectTableModel;

/**
 * Rate encoding leaky integrator with silent inhibition (RELISI)
 * 
 * @author Mathias Linkerhand
 * @version 2012-07-25
 */
public class LeakyIntegratorModule extends Module implements Serializable {
  private static final long                     serialVersionUID      = 0L;

  protected List<Double>                        leaks;
  protected List<Double>                        membranePotentials;

  private static final Logger                   logger                = Logger.getLogger(ObjectTableModel.class.getName());

  protected transient Set<Integer>              lastActiveClique      = null;
  protected transient Map<Double, Set<Integer>> activeCliqueEvolution = new TreeMap<Double, Set<Integer>>();

  public static enum SilentInhibition {
    INHIBITION_NOT_SILENT, SILENT_INHIBITION_WITH_RESPECT_TO_MEMBRANE_POTENTIAL, SILENT_INHIBITION_WITH_RESPECT_TO_FIRING_RATE
  };

  private SilentInhibition silentInhibition;

  /**
   * Creates a network corresponding to the given adjacency matrix with given
   * activities and leaks.
   * 
   * @param adjacencyMatrix
   * @param activities
   * @param leaks
   */
  public LeakyIntegratorModule(final double[][] adjacencyMatrix, final double[] activities, final double[] leaks) {
    this();

    if (adjacencyMatrix.length != 0 && adjacencyMatrix.length != adjacencyMatrix[0].length) {
      throw new IllegalArgumentException("Adjacency matrix has to be quadratic, but is " + adjacencyMatrix.length + "×" + adjacencyMatrix[0].length + ".");
    }
    if (adjacencyMatrix.length != activities.length) {
      throw new IllegalArgumentException("Wrong length of activities.");
    }
    if (adjacencyMatrix.length != leaks.length) {
      throw new IllegalArgumentException("Wrong length of leaks.");
    }

    // creating vertices
    for (int i = 0; i < adjacencyMatrix.length; i++) {
      final double activity = activities[i];
      final double leak = leaks[i];
      this.addNeuron(activity);
      this.setLeak(0, leak);
    }

    // creating edges
    for (int i = 0; i < adjacencyMatrix.length; i++) {
      for (int j = 0; j < adjacencyMatrix[i].length; j++) {
        this.setLink(i, j, adjacencyMatrix[i][j]);
      }
    }
  }

  /**
   * Creates a network corresponding to the given adjacency matrix with random
   * activities and leaks.
   * 
   * @param adjacencyMatrix
   */
  public LeakyIntegratorModule(final double[][] adjacencyMatrix, final Random random) {
    //		this(adjacencyMatrix, Utils.fill(adjacencyMatrix.length, 0.0), Utils.fill(adjacencyMatrix.length, 0.1));
    this(adjacencyMatrix, Utils.randomArray(adjacencyMatrix.length, random, 0, 1), Utils.fill(adjacencyMatrix.length, 0.1));
  }

  /**
   * Creates an Erdos-Renyi network of size $n$ and edge probability $p$.
   * 
   * @param size The size $n$.
   * @param linkProbability The edge probability $p$.
   */
  public LeakyIntegratorModule(final int size, final double linkProbability) {
    this(Graphs.addInhibitoryLinks(Graphs.createErdosRenyiGraphAdjacencyMatrix(size, linkProbability)), new Random());
  }

  /**
   * Creates an Erdos-Renyi network of size $n$ and edge probability $p$.
   * 
   * @param size The size $n$.
   * @param linkProbability The edge probability $p$.
   */
  public LeakyIntegratorModule(final int size, final double linkProbability, final Random random) {
    this(Graphs.addInhibitoryLinks(Graphs.createErdősRényiGraphAdjacencyMatrix(size, linkProbability, random)), random);
  }

  /**
   * Creates an empty network.
   */
  public LeakyIntegratorModule() {
    this.leaks = new Vector<Double>();
    this.membranePotentials = new Vector<Double>();
    this.setName("LeakyIntegrator");
    this.setSilentInhibition(SilentInhibition.INHIBITION_NOT_SILENT);
  }

  /**
   * Creates a network according to the specified file.
   * 
   * @param file
   * @throws IOException
   */
  public LeakyIntegratorModule(final File file) throws IOException {
    this();

    try {
      Document document = new SAXBuilder().build(file);
      for (Object node : document.getRootElement().getChildren("node")) {
        final Element element = ((Element) node);
        if (element.getAttributeValue("type").toLowerCase().equals("leakyintegrator")) {
          final double activity = element.getAttributeValue("activity").equals("random") ? Math.random() : Double.parseDouble(element
              .getAttributeValue("activity"));
          final double leak = 0.1;
          this.addNeuron(activity);
          this.setLeaks(leak);
        } else {
        }
      }

      final Map<String, Integer> ids = new HashMap<String, Integer>();

      for (Object node : document.getRootElement().getChildren("edge")) {
        final String afferentNeuron = ((Element) node).getAttributeValue("from");
        final String efferentNeuron = ((Element) node).getAttributeValue("to");
        final double weight = Double.valueOf(((Element) node).getAttributeValue("weight"));

        if (!ids.containsKey(afferentNeuron)) {
          ids.put(afferentNeuron, ids.size());
        }
        if (!ids.containsKey(efferentNeuron)) {
          ids.put(efferentNeuron, ids.size());
        }

        this.setLink(ids.get(afferentNeuron), ids.get(efferentNeuron), weight);
      }
    } catch (JDOMException exception) {
      throw new IllegalArgumentException("file “" + file + "” not ok: " + exception);
    }
  }

  public static LeakyIntegratorModule fromXML(final File file) throws IOException {
    return new LeakyIntegratorModule(file);
  }

  /**
   * @param i
   * @param timeDelta time delta $\Delta t$
   * @return activity
   */
  // $x_i(t + \Delta t) = x_i(t) + \left[ \left( \sum_{j = 1}^{N} z_{i j} y_j \Gamma_i \right) x_i + \sum_{j = 1}^{N} w_{i j} y_j \right] \Delta t$
  protected double calculateMembranePotential(final int i, final double timeDelta) {
    // collect inputs and inhibitions
    double input = 0;
    double inhibition = 0;

    // current module
    for (int j = 0; j < this.getSize(); j++) {
      final double w_ij = this.getLink(i, j);
      if (w_ij > 0) {
        // excitation link
        input += w_ij * this.getActivity(j);
      } else if (w_ij < 0) {
        // inhibition link
        inhibition += w_ij * this.getActivity(j);
      }
    }

    // afferent modules (the same of the current module just for the other modules)
    for (final Module module : this.getAfferentModules()) {
      for (int j = 0; j < module.getSize(); j++) {
        final double w_ij = this.getAfferentLink(module, i, j);
        if (w_ij > 0) {
          // excitation link
          input += w_ij * module.getActivity(j);
        } else if (w_ij < 0) {
          // inhibition link
          inhibition += w_ij * module.getActivity(j);
        }
      }
    }

    final double x_i = this.getMembranePotential(i); // membrane potential $x$
    final double y_i = this.getActivity(i); // firing rate $y$
    final double gamma_i = this.getLeak(i); // leak $\Gamma$

    /*
     * discrete iteration of differential equation
     */
    final double delta_x_i;
    switch (this.getSilentInhibition()) {
      case INHIBITION_NOT_SILENT: // no silent inhibition
        delta_x_i = (-gamma_i * x_i + inhibition + input) * timeDelta;
        break;
      case SILENT_INHIBITION_WITH_RESPECT_TO_MEMBRANE_POTENTIAL: // silent inhibition with respect to the membrane potential
        delta_x_i = (-gamma_i * x_i + inhibition * x_i + input) * timeDelta;
        break;
      case SILENT_INHIBITION_WITH_RESPECT_TO_FIRING_RATE: // silent inhibition with respect to the firing rate
        delta_x_i = (-gamma_i * x_i + inhibition * y_i + input) * timeDelta;
        break;
      default:
        throw new IllegalStateException("no such silent inhibition type");
    }
    final double x_new = x_i + delta_x_i;

    return x_new;
  }

  @Override
  protected synchronized void calculateOneTimeStep(final double deltaTime) {
    final List<Double> newMembranePotentials = new Vector<Double>(this.getSize());

    // first: calculating all new states
    for (int i = 0; i < this.getSize(); i++) {
      newMembranePotentials.add(this.calculateMembranePotential(i, deltaTime));
    }
    // second: updating states
    this.setMembranePotentials(newMembranePotentials);

    final boolean findActiveCliques = false;
    if (findActiveCliques) {
      final Set<Integer> activeClique = this.getActiveClique();
      if (this.lastActiveClique != activeClique) {
        logger.log(Level.FINER, "active clique at time " + this.elapsedTime + ": " + activeClique);
        this.activeCliqueEvolution.put(this.elapsedTime, activeClique);
        this.lastActiveClique = activeClique;
      }
    }
  }

  /**
   * Maps the membrane potential x ∈ [0, ∞) to the rate coding activity y ∈ [0,
   * 1].
   * 
   * @param potential
   * @return activity
   */
  public double transferFunction(final double potential, final int neuron) {
    final double exponent = 4.0;
    final double tmp = Math.pow(potential, exponent);
    return tmp / (tmp + 1.0);
  }

  /**
   * vectorial form of @see transferFunction
   */
  public List<Double> transferFunction(final List<Double> potentials) {
    final List<Double> result = new Vector<Double>(potentials.size());
    for (int i = 0; i < potentials.size(); i++) {
      final Double potential = potentials.get(i);
      result.add(this.transferFunction(potential, i));
    }
    return result;
  }

  /**
   * inverse of @see transferFunction
   */
  public double transferFunctionInverse(final double activity, final int neuron) {
    final double exponent = 1.0 / 4.0;
    return Math.pow(activity / (1.0 - activity), exponent);
  }

  /**
   * vectorial form of @see transferFunctionInverse
   */
  public List<Double> transferFunctionInverse(final List<Double> activities) {
    final List<Double> result = new Vector<Double>(activities.size());
    for (int i = 0; i < activities.size(); i++) {
      final Double activity = activities.get(i);
      result.add(this.transferFunctionInverse(activity, i));
    }
    return result;
  }

  @Override
  public void addNeuron(final double activity) {
    super.addNeuron(activity);

    final double defaultLeak = 1.0;
    this.leaks.add(defaultLeak);
    this.membranePotentials.add(this.transferFunctionInverse(activity, this.getSize() - 1)); // size should have been incremented by super function
  }

  @Override
  public void removeNeuron(final int index) {
    super.removeNeuron(index);

    this.leaks.remove(index);
    this.membranePotentials.remove(index);
  }

  /**
   * Sets the activity of a given node. Also sets the corresponding membrane
   * potential of that node.
   */
  @Override
  public void setActivity(final int index, final double activity) {
    super.setActivity(index, activity);
    this.membranePotentials.set(index, this.transferFunctionInverse(activity, index));
  }

  /**
   * Sets the activities of all node. Also sets the corresponding membrane
   * potentials of that nodes.
   */
  @Override
  public void setActivities(final List<Double> activities) {
    super.setActivities(activities);
    this.membranePotentials = this.transferFunctionInverse(activities);
  }

  /**
   * Updates the activities directly. Calling this function is maybe necessary
   * after changing parameters of the transfer function or changing membrane
   * potentials.
   */
  public void updateActivity(final int neuron) {
    this.activities.set(neuron, this.transferFunction(this.getMembranePotential(neuron), neuron));
  }

  /**
   * Updates the activities directly. Calling this function is maybe necessary
   * after changing parameters of the transfer function or changing membrane
   * potentials.
   */
  public void updateActivities() {
    this.activities = this.transferFunction(this.getMembranePotentials());
  }

  /**
   * Gets the membrane potential of a given node.
   */
  public double getMembranePotential(final int index) {
    return this.membranePotentials.get(index);
  }

  /**
   * Gets the membrane potentials of all nodes.
   */
  public List<Double> getMembranePotentials() {
    return this.membranePotentials;
  }

  /**
   * Sets the membrane potential of a given node. Also sets the corresponding
   * activity of that node.
   */
  public void setMembranePotential(final int index, final double membranePotential) {
    super.setActivity(index, this.transferFunction(membranePotential, index));
    this.membranePotentials.set(index, membranePotential);
  }

  /**
   * Sets the membrane potentials of all nodes. Also updates the corresponding
   * activities of that nodes.
   */
  public void setMembranePotentials(final double membranePotential) {
    final List<Double> membranePotentials = new ArrayList<Double>(this.getSize());
    for(int neuron = 0; neuron < this.getSize(); neuron++){
      membranePotentials.add(membranePotential);
    }
    this.setMembranePotentials(membranePotentials);
  }

    /**
   * Sets the membrane potentials of all nodes. Also updates the corresponding
   * activities of that nodes.
   */
  public void setMembranePotentials(final List<Double> membranePotentials) {
    if (membranePotentials.size() != this.getSize()) {
      throw new IllegalArgumentException("Wrong size " + membranePotentials.size() + " of membranePotentials. Should be the graph size " + this.getSize());
    }
    //    super.setActivities(this.transferFunction(membranePotentials));
    for (int i = 0; i < this.getSize(); i++) {
      super.setActivity(i, this.transferFunction(membranePotentials.get(i), i));
    }
    this.membranePotentials = membranePotentials;
  }

  public void setLeak(final int neuron, final double leak) {
    this.leaks.set(neuron, leak);
  }

  /**
   * Sets the leak for the whole network.
   * 
   * @param leak
   */
  public void setLeaks(final double leak) {
    Collections.fill(this.leaks, leak);
  }

  public double getLeak(int i) {
    return this.leaks.get(i);
  }

  public List<Double> getLeaks() {
    return this.leaks;
  }

  public SilentInhibition getSilentInhibition() {
    return this.silentInhibition;
  }

  public void setSilentInhibition(final SilentInhibition silentInhibition) {
    this.silentInhibition = silentInhibition;
  }

  @Override
  public void reset() {
    // TODO Auto-generated method stub
  }

  /**
   * Weight edges according the the node’s in-degree: w_{i j} = 1 / deg^-(j)
   */
  public void normalizeLinks() {
    final Module module = this;
    for (int i = 0; i < module.getSize(); i++) {
      for (int j = 0; j < module.getSize(); j++) {
        final double oldW_ij = module.getLink(i, j); // i ← j
        final double newW_ij;
        if (oldW_ij < 0) {
          // inhibitory link
          //final int inDegree = module.getNodeInDegree(i, -1); // i
          //newW_ij = -1.0 / inDegree; // weighted links according to the node’s degree
          newW_ij = -1.0;
        } else if (oldW_ij > 0) {
          // exhibitory link
          final int inDegree = module.getNodeInDegree(i, 1); // i
          newW_ij = +1.0 / inDegree; // weighted links according to the node’s degree
          //newW_ij = +1.0;
        } else {
          // no link
          newW_ij = 0;
        }
        module.setLink(i, j, newW_ij);
      }
    }
  }

  @Override
  public void printLogFileHeader(final PrintWriter out) {
    super.printLogFileHeader(out);

    final String commentOperator = "# ";
    out.println(commentOperator + "leaks:");
    out.println(commentOperator + this.getLeaks());
    out.println(commentOperator + "membranePotentials:");
    out.println(commentOperator + this.getMembranePotentials());
  }
}
