package brain;

import java.io.File;
import java.io.IOException;
import java.io.PrintWriter;
import java.io.Serializable;
import java.util.Arrays;
import java.util.Collection;
import java.util.List;
import java.util.Random;
import java.util.Vector;

import transientDynamics.TransientDynamicsOfLeakyIntegratorsWithIntrinsicPlasticity;

/**
 * This class extends the <code>LeakyIntegratorModule</code> by adaptive
 * transfer functions in order to implement entropy maximization using intrinsic
 * plasticity.
 * 
 * @author Mathias Linkerhand
 * @version 2013-01-27
 * @since 2012
 */
public class LeakyIntegratorWithIntrinsicPlasticityModule extends LeakyIntegratorModule implements Serializable {
  private static final long    serialVersionUID     = 1L;
  protected List<List<Double>> transferFunctionParameters;
  protected List<List<Double>> targetDistributionParameters;
  protected boolean            intrinsicPlasticity  = true;
  protected List<List<Double>> intrinsicAdaptationLearningRates;
  protected List<List<Double>> transferFunctionParameterChanges;

  private int                  transferFunctionType = 1;        // 0 = old (g(0) = 0), 1 = classical fermi

  /**
   * Default LeakyIntegratorWithIntrinsicPlasticityModule constructor for demonstration purpose. A set of neuron with varius firing rate target distributions is used to show the polyhomeostatic adaption process. Make sure to add an afferent noise input module with white noise in the range [0, 10].
   */
  public LeakyIntegratorWithIntrinsicPlasticityModule() {
    super();
    this.setSize(5); // $N$

    this.setLeaks(1.0); // $\Gamma$
    this.setIntrinsicPlasticity(true); // Activate polyhomeostatic adaption.
    this.setIntrinsicAdaptationLearningRates(0.01); // $\epsilon_a = \epsilon_b$
    this.setTargetDistributionParameters(0, 0.0, 0.0); // neuron, $\lambda_1$, $\lambda_2$
    this.setTargetDistributionParameters(1, +10.0, 0.0); // neuron, $\lambda_1$, $\lambda_2$
    this.setTargetDistributionParameters(2, -10.0, 0.0); // neuron, $\lambda_1$, $\lambda_2$
    this.setTargetDistributionParameters(3, -10.0, +10.0); // neuron, $\lambda_1$, $\lambda_2$
    this.setTargetDistributionParameters(4, +10.0, -10.0); // neuron, $\lambda_1$, $\lambda_2$
    this.setTransferFunctionParameters(1.0, 1.0); // gain $a$, threshold $b$
    this.setMembranePotentials(0.0); // $x$
  }
  
  /**
   * Create a leaky integrator with latching dynamics for demonstration purpose.
   * 
   * @param networkSize Size of the network.
   * @param numberOfPatterns Number of random binary patterns.
   * @param sparseness Ensemble average firing activity for the patterns.
   * @param lambda1 $\lambda_1$ for the time average firing activity $\mu$ for the target distribution.
   */
  public LeakyIntegratorWithIntrinsicPlasticityModule(final int networkSize, final int numberOfPatterns, final double sparseness, final double lambda1){
    super();
    
    /*
     * Create network.
     */
    this.setSize(networkSize);    
    this.setLeaks(1.0); // $\Gamma$
    this.setIntrinsicPlasticity(true);
    this.setIntrinsicAdaptationLearningRates(0.1, 0.01); // $\epsilon_a$, $\epsilon_b$
    this.setTargetDistributionParameters(lambda1, 0.0); // $\lambda_1$, $\lambda_2$
    this.setTransferFunctionParameters(1.0, 1.0); // gain $a$, threshold $b$
    this.setMembranePotentials(0.0); // $x$
    
    /*
     * Store random binary patterns using Hopfield learning.
     */
    final Collection<List<Double>> patterns = TransientDynamicsOfLeakyIntegratorsWithIntrinsicPlasticity.createRandomBinarySparsePatterns(sparseness, networkSize, numberOfPatterns, new Random(0));
    final double[][] w = TransientDynamicsOfLeakyIntegratorsWithIntrinsicPlasticity.hopfieldWeightedAdjacencyMatrix(networkSize, patterns, sparseness);
    this.setLinks(w);
  }

  public LeakyIntegratorWithIntrinsicPlasticityModule(double[][] adjacencyMatrix, final Random random) {
    super(adjacencyMatrix, random);
  }

  public LeakyIntegratorWithIntrinsicPlasticityModule(int size, double linkProbability) {
    super(size, linkProbability);
  }

  public LeakyIntegratorWithIntrinsicPlasticityModule(int size, double linkProbability, Random random) {
    super(size, linkProbability, random);
  }

  public LeakyIntegratorWithIntrinsicPlasticityModule(File file) throws IOException {
    super(file);
  }

  @Override
  public void addNeuron(final double activity) {
    // transfer function parameters
    final Double[] defaultTransferFunctionParameters = new Double[] { 1.0, 1.0 };
    if (this.transferFunctionParameters == null) {
      this.transferFunctionParameters = new Vector<List<Double>>();
    }
    this.transferFunctionParameters.add(Arrays.asList(defaultTransferFunctionParameters));

    // transfer function parameter changes
    final Double[] defaultTransferFunctionParameterChanges = new Double[] { 0.0, 0.0 };
    if (this.transferFunctionParameterChanges == null) {
      this.transferFunctionParameterChanges = new Vector<List<Double>>();
    }
    this.transferFunctionParameterChanges.add(Arrays.asList(defaultTransferFunctionParameterChanges));

    // target distribution
    final Double[] defaultTargetDistributionParameters = new Double[] { -10.0, 0.0 };
    if (this.targetDistributionParameters == null) {
      this.targetDistributionParameters = new Vector<List<Double>>();
    }
    this.targetDistributionParameters.add(Arrays.asList(defaultTargetDistributionParameters));

    // intrinsic adaptation learning rates
    final Double[] defaultIntrinsicAdaptationLearningRates = new Double[] { 1e-2, 1e-2 };
    if (this.intrinsicAdaptationLearningRates == null) {
      this.intrinsicAdaptationLearningRates = new Vector<List<Double>>();
    }
    this.intrinsicAdaptationLearningRates.add(Arrays.asList(defaultIntrinsicAdaptationLearningRates));

    super.addNeuron(activity);
  }

  @Override
  public void removeNeuron(final int neuron) {
    super.removeNeuron(neuron);

    this.transferFunctionParameters.remove(neuron);
    this.transferFunctionParameterChanges.remove(neuron);
    this.targetDistributionParameters.remove(neuron);
    this.intrinsicAdaptationLearningRates.remove(neuron);
  }

  /**
   * Maps the membrane potential x ∈ [0, ∞) to the rate coding activity y ∈ [0,
   * 1).
   * 
   * @param potential
   * @return activity
   */
  @Override
  public double transferFunction(final double potential, final int neuron) {
    final double a = this.getTransferFunctionParameters(neuron).get(0);
    final double b = this.getTransferFunctionParameters(neuron).get(1);
    final double x = potential;

    final double y;
    switch (this.transferFunctionType) {
      case 0:
        if (potential < 0) {
          throw new IllegalArgumentException("Potential has to be non-negative! x ∈ [0, ∞)");
        }
        // g = z/(z+1), z = (x/b)^(a·b)
        final double z = Math.pow(x / b, a * b);
        y = z / (z + 1.0);
        break;
      case 1:
        // g(x) = 1/(1 + e^[-a·(x-b)])
        // $g(x) = \frac{1}{1 + e^{-a \cdot (x - b)}}$
        y = 1.0 / (1.0 + Math.exp(-a * (x - b)));
        break;
      case 2:
        y = (Math.tanh(a * (x - b)) + 1.0) / 2.0;
        break;
      case -1:
        y = a * (x - b) + 0.5;
        break;
      default:
        throw new IllegalStateException("No such transfer function.");
    }

    return y;
  }

  /**
   * inverse of @see transferFunction
   */
  @Override
  public double transferFunctionInverse(final double activity, final int neuron) {
    if (activity < 0 || activity >= 1) {
      throw new IllegalArgumentException("Activity has to be between 0 and 1! y ∈ [0, 1)");
    }

    final double a = this.getTransferFunctionParameters(neuron).get(0);
    final double b = this.getTransferFunctionParameters(neuron).get(1);
    final double y = activity;

    final double x;
    switch (this.transferFunctionType) {
      case 0:
        x = b * Math.pow(y / (1.0 - y), 1.0 / (a * b));
        break;
      case 1:
        if (activity <= 0 || activity >= 1) {
          throw new IllegalArgumentException("Activity has to be between 0 and 1! y ∈ (0, 1)");
        }
        x = Math.log(y / (1.0 - y)) / a + b;
        break;
      default:
        throw new IllegalStateException("No such transfer function.");
    }

    return x;
  }

  @Override
  public void calculateOneTimeStep(final double deltaTime) {
    super.calculateOneTimeStep(deltaTime);

    if (this.isIntrinsicPlasticity()) {
      this.adaptTransferFunctionParameters(deltaTime);
    }
  }

  public void adaptTransferFunctionParameters(final double deltaTime) {
    for (int neuron = 0; neuron < this.getSize(); neuron++) {
      // neuron properties
      final double x = this.getMembranePotential(neuron); // membrane potential
      final double y = this.getActivity(neuron); // firing rate

      // transfer function properties
      final double a = this.getTransferFunctionParameters(neuron).get(0); // gain
      final double b = this.getTransferFunctionParameters(neuron).get(1); // threshold

      // output distribution properties $f(x) = e^{\lambda_1 x + \lambda_2^2 x^2}$
      final double lambda_1 = this.getTargetDistributionParameters(neuron).get(0); // $\lambda_1$
      final double lambda_2 = this.getTargetDistributionParameters(neuron).get(1); // $\lambda_2$

      // adaptation iteration properties
      final double epsilon_a = this.getIntrinsicAdaptationLearningRates(neuron).get(0);
      final double epsilon_b = this.getIntrinsicAdaptationLearningRates(neuron).get(1);

      final double qPrime_over_q = lambda_1 + 2.0 * lambda_2 * y;

      final double partial_d_over_partial_a;
      final double partial_d_over_partial_b;
      switch (this.transferFunctionType) {
        case 0:
          partial_d_over_partial_a = -1.0 / a - b * Math.log(x / b) * (1 - 2 * y + qPrime_over_q * y * (1.0 - y));
          partial_d_over_partial_b = -1.0 / b - a * (Math.log(x / b) - 1) * (1 - 2 * y + qPrime_over_q * y * (1.0 - y));
          break;
        case 1:
          final double termInParenthesis = 1.0 - 2.0 * y + qPrime_over_q * y * (1.0 - y); // For explanation look at the stochastic adaption rules in the paper.
          partial_d_over_partial_a = -1.0 / a - (x - b) * termInParenthesis;
          partial_d_over_partial_b = a * termInParenthesis;
          break;
        default:
          throw new IllegalStateException("No such transfer function.");
      }

      final double delta_a = -epsilon_a * partial_d_over_partial_a * deltaTime;
      final double delta_b = -epsilon_b * partial_d_over_partial_b * deltaTime;

      // set new transfer function parameter changes
      this.setTransferFunctionParameterChanges(neuron, delta_a / deltaTime, delta_b / deltaTime);

      // set new transfer function parameters
      final double newA = a + delta_a;
      final double newB = b + delta_b;
      if (newA < 0) {
        throw new RuntimeException("a(t) < 0, for t=" + this.getElapsedTime());
      }
      this.setTransferFunctionParameters(neuron, newA, newB);
    }
  }

  public List<List<Double>> getTransferFunctionParameters() {
    return this.transferFunctionParameters;
  }

  public List<Double> getTransferFunctionParameters(final int neuron) {
    return this.transferFunctionParameters.get(neuron);
  }

  public void setTransferFunctionParameters(final double gain, final double threshold) {
    for (int neuron = 0; neuron < this.getSize(); neuron++) {
      this.setTransferFunctionParameters(neuron, gain, threshold);
    }
  }

  public void setTransferFunctionParameters(final int neuron, final double gain, final double threshold) {
    this.transferFunctionParameters.get(neuron).set(0, gain);
    this.transferFunctionParameters.get(neuron).set(1, threshold);

    // activities have to be recalculated
    this.updateActivity(neuron);
  }

  public List<List<Double>> getTransferFunctionParameterChanges() {
    return this.transferFunctionParameterChanges;
  }

  private void setTransferFunctionParameterChanges(final int neuron, final double gain, final double threshold) {
    this.transferFunctionParameterChanges.get(neuron).set(0, gain);
    this.transferFunctionParameterChanges.get(neuron).set(1, threshold);
  }

  public boolean isIntrinsicPlasticity() {
    return intrinsicPlasticity;
  }

  public void setIntrinsicPlasticity(boolean intrinsicPlasticity) {
    this.intrinsicPlasticity = intrinsicPlasticity;
  }

  public List<List<Double>> getIntrinsicAdaptationLearningRates() {
    return this.intrinsicAdaptationLearningRates;
  }

  public List<Double> getIntrinsicAdaptationLearningRates(final int neuron) {
    return this.getIntrinsicAdaptationLearningRates().get(neuron);
  }

  public void setIntrinsicAdaptationLearningRates(final double intrinsicAdaptationParameter) {
    this.setIntrinsicAdaptationLearningRates(intrinsicAdaptationParameter, intrinsicAdaptationParameter);
  }

  public void setIntrinsicAdaptationLearningRates(final double intrinsicAdaptationParameter1, final double intrinsicAdaptationParameter2) {
    for (int neuron = 0; neuron < this.getSize(); neuron++) {
      this.setIntrinsicAdaptationLearningRates(neuron, intrinsicAdaptationParameter1, intrinsicAdaptationParameter2);
    }
  }

  public void setIntrinsicAdaptationLearningRates(final int neuron, final double intrinsicAdaptationParameter) {
    this.setIntrinsicAdaptationLearningRates(neuron, intrinsicAdaptationParameter, intrinsicAdaptationParameter);
  }

  public void setIntrinsicAdaptationLearningRates(final int neuron, final double intrinsicAdaptationParameter1, final double intrinsicAdaptationParameter2) {
    this.intrinsicAdaptationLearningRates.get(neuron).set(0, intrinsicAdaptationParameter1);
    this.intrinsicAdaptationLearningRates.get(neuron).set(1, intrinsicAdaptationParameter2);
  }

  public List<List<Double>> getTargetDistributionParameters() {
    return this.targetDistributionParameters;
  }

  public List<Double> getTargetDistributionParameters(final int neuron) {
    return this.getTargetDistributionParameters().get(neuron);
  }

  public void setTargetDistributionParameters(final double targetDistributionParameter1, final double targetDistributionParameter2) {
    for (int neuron = 0; neuron < this.getSize(); neuron++) {
      this.setTargetDistributionParameters(neuron, targetDistributionParameter1, targetDistributionParameter2);
    }
  }

  public void setTargetDistributionParameters(final int neuron, final double targetDistributionParameter1, final double targetDistributionParameter2) {
    this.targetDistributionParameters.get(neuron).set(0, targetDistributionParameter1);
    this.targetDistributionParameters.get(neuron).set(1, targetDistributionParameter2);
  }

  @Override
  public String toString() {
    //return super.toString() + "epsilon_a=" + intrinsicAdaptationLearningRate1 + " epsilon_b=" + intrinsicAdaptationLearningRate2;
    return super.toString();
  }

  @Override
  public void printLogFileHeader(final PrintWriter out) {
    super.printLogFileHeader(out);

    final String commentOperator = "# ";
    out.println(commentOperator + "transferFunctionParameters:");
    out.println(commentOperator + this.getTransferFunctionParameters());
    out.println(commentOperator + "targetDistributionParameters:");
    out.println(commentOperator + this.getTargetDistributionParameters());
    out.println(commentOperator + "isIntrinsicPlasticity:");
    out.println(commentOperator + this.isIntrinsicPlasticity());
    out.println(commentOperator + "intrinsicAdaptationParameters:");
    out.println(commentOperator + this.getIntrinsicAdaptationLearningRates());
  }
}
