package brain;

import java.awt.Point;
import java.io.File;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.ObjectInputStream;
import java.io.OutputStream;
import java.io.OutputStreamWriter;
import java.io.PrintWriter;
import java.io.Serializable;
import java.io.Writer;
import java.util.Arrays;
import java.util.Collections;
import java.util.Comparator;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Locale;
import java.util.Map;
import java.util.Set;
import java.util.TreeMap;
import java.util.TreeSet;
import java.util.Vector;
import java.util.logging.Level;
import java.util.logging.Logger;

import javax.swing.event.ChangeEvent;
import javax.swing.event.ChangeListener;
import javax.swing.event.EventListenerList;

/**
 * This class provides default implementations for the <code>Module</code>
 * interface. Standard behaviors like the get, set, add, remove methods are
 * defined here. This class also provides basic graph calculation. The developer
 * need only subclass this class and implement calculateOneTimeStep and define
 * its own important methods.
 * 
 * @author Mathias Linkerhand
 * @version 2013-01-31
 */
public abstract class Module implements Serializable {
  private static final long               serialVersionUID             = 1L;

  private static final Logger             logger                       = Logger.getLogger(Module.class.getName());

  /** Listeners to be notified of module changes. */
  protected transient EventListenerList   changeListeners              = new EventListenerList();

  /**
   * HashMap with string-identifiers and identifier of the receiving port of the
   * target module.
   * <p>
   * The geometry of the forward connection, specific to the `DhanModule'
   * module. The receiving port (receptors for the neuromodulator) is given by
   * the second argument of `forwardModules'.<br>
   */
  protected Map<Module, String>           forwardModules               = new HashMap<Module, String>();

  private Map<Module, List<List<Double>>> afferentModules              = new HashMap<Module, List<List<Double>>>();
  private Set<Module>                     efferentModules              = new HashSet<Module>();

  /**
   * Array with the activities of the individual neurons.
   */
  protected List<Double>                  activities                   = new Vector<Double>();

  /**
   * Array with the activities of the individual neurons.
   */
  protected List<Double>                  biases                       = new Vector<Double>();

  /**
   * links to connect vertices
   */
  private List<List<Double>>              links                        = new Vector<List<Double>>();

  public static final int                 HEBBIAN_LEARNING             = 0;
  public static final int                 OJAS_RULE_LEARNING           = 1;
  public static final int                 GENERALIZED_HEBBIAN_LEARNING = 2;
  private List<Double>                    learningRates                = new Vector<Double>();
  private int                             learningType                 = Module.HEBBIAN_LEARNING;

  /**
   * HashMap with string-identifiers encoding the open ports for the individual
   * receptors.
   * <p>
   * The receptors are targeted by the diffusive module signals of other (or
   * this) modules, via the respective forward connection, stored in the
   * respective `forwardModules'.<br>
   * The allowed types of receptors, the cues of the HashMap, are stored in
   * `auxMyReceptorsAllowedValues'.
   */
  protected HashMap<String, Double>       receptors                    = new HashMap<String, Double>();

  /**
   * The number of time steps elapsed for this module.
   */
  protected transient long                elapsedTimeSteps;

  /**
   * The time elapsed for this module.
   */
  protected transient double              elapsedTime;

  /**
   * The calculated cliques for this graph.
   */
  private Set<Set<Integer>>               cliques;

  /**
   * Name of the module.
   */
  private String                          name;

  /**
   * Used for @see #getActiveClique(). Stores the last ordered indices. Useful
   * if and only if sorting is stable (@see <a
   * href="http://en.wikipedia.org/wiki/Sorting_algorithm#Stability">sorting
   * algorithm stability</a>).
   */
  private Integer[]                       sortedIndices                = null;

  private boolean                         isLearning                   = false;

  /**
   * returns the number of time steps elapsed
   */
  public long getElapsedTimeSteps() {
    return this.elapsedTimeSteps;
  }

  /**
   * returns the time elapsed
   */
  public double getElapsedTime() {
    return this.elapsedTime;
  }

  /**
   * performs one time step of the simulation calculation
   */
  final public void performOneTimeStep(final double timeDelta) {
    // count the time
    this.elapsedTimeSteps++;
    this.elapsedTime += timeDelta;

    // do the time step
    this.calculateOneTimeStep(timeDelta);

    // learning
    if (this.isLearning()) {
      this.learning(timeDelta);
    }

    // notify observers about new state
    this.notifyStateChanged(new ChangeEvent(this));
  }

  /**
   * calculates one time step of the simulation calculation
   * 
   * This method has to be implemented by subclasses.
   * 
   * @param timeDelta
   */
  protected abstract void calculateOneTimeStep(final double timeDelta);

  public void addNeuron(final double activity) {
    this.activities.add(activity);

    // bias
    final double defaultBias = 0;
    this.biases.add(defaultBias);

    // learning rate
    final double defaultLearningRate = 0;
    this.learningRates.add(defaultLearningRate);

    // cliques have to be recalculated
    this.cliques = null;

    // add edges of current graph
    this.links.add(new Vector<Double>());
    for (final List<Double> adjacencyMatrixRow : this.links) {
      for (int i = adjacencyMatrixRow.size(); i < this.links.size(); i++) {
        // new links are initialized to zero
        adjacencyMatrixRow.add(0.0);
      }
    }

    // add edges to afferent graphs
    for (final Module afferentModule : this.getAfferentModules()) {
      this.afferentModules.get(afferentModule).add(new Vector<Double>());
      for (final List<Double> afferentLinkMatrixRow : this.afferentModules.get(afferentModule)) {
        for (int i = afferentLinkMatrixRow.size(); i < this.afferentModules.get(afferentModule).size(); i++) {
          // new links are initialized to zero
          afferentLinkMatrixRow.add(0.0);
        }
      }
    }

    // add edges to efferent graphs
    for (final Module efferentModule : this.efferentModules) {
      efferentModule.afferentModules.get(this).add(new Vector<Double>());
      for (final List<Double> afferentLinkMatrixRow : efferentModule.afferentModules.get(this)) {
        for (int i = afferentLinkMatrixRow.size(); i < efferentModule.afferentModules.get(this).size(); i++) {
          // new links are initialized to zero
          afferentLinkMatrixRow.add(0.0);
        }
      }
    }
  }

  public void learning(final double timeDelta) {
    // intra-module synapses
    for (int i = 0; i < this.getSize(); i++) {
      final double eta = this.getLearningRate(i); // learning rate
      final double y_post = this.getActivity(i); // postsynaptic activity i

      for (int j = 0; j < this.getSize(); j++) {
        final double y_pre = this.getActivity(j); // presynaptic activity j
        final double w_ij = this.getLink(i, j);

        if (w_ij != 0.0) {
          final double delta_w_ij;
          switch (this.learningType) {
            case Module.HEBBIAN_LEARNING:
              // Hebbian learning rule
              delta_w_ij = timeDelta * (eta * y_post * y_pre);
              break;
            case Module.OJAS_RULE_LEARNING:
              // Oja’s learning rule
              delta_w_ij = timeDelta * (eta * y_post * (y_pre - y_post * w_ij));
              break;
            case Module.GENERALIZED_HEBBIAN_LEARNING:
              // Generalized Hebbian Algorithm (GHA)
              //final double delta_w_ij = timeDelta * (eta * y_post * (y_pre - y_post * w_ij)); // not yet ready, therefore incorrect
              throw new IllegalStateException("not implemented");
            default:
              throw new IllegalStateException("Learning type unknown.");
          }

          final double w_ij_new = w_ij + delta_w_ij;
          this.setLink(i, j, w_ij_new);
        }
      }
    }

    // inter-module synapses
    for (final Module afferentModule : this.getAfferentModules()) {
      for (int post = 0; post < this.getSize(); post++) {
        final double eta = this.getLearningRate(post); // learning rate
        final double y_post = this.getActivity(post); // postsynaptic activity i

        for (int pre = 0; pre < afferentModule.getSize(); pre++) {
          final double y_pre = this.getActivity(pre); // presynaptic activity j
          final double w_ij = this.getAfferentLink(afferentModule, post, pre); // synaptic strength j → i (pre → post)

          if (w_ij != 0.0) {
            final double deltaw_ij;
            switch (this.learningType) {
              case Module.HEBBIAN_LEARNING:
                // Hebbian learning rule
                deltaw_ij = timeDelta * (eta * y_post * y_pre);
                break;
              case Module.OJAS_RULE_LEARNING:
                // Oja’s learning rule
                deltaw_ij = timeDelta * (eta * y_post * (y_pre - y_post * w_ij));
                break;
              case Module.GENERALIZED_HEBBIAN_LEARNING:
                // Generalized Hebbian Algorithm (GHA)
                //final double delta_w_ij = timeDelta * (eta * y_post * (y_pre - y_post * w_ij)); // not yet ready, therefore incorrect
                throw new IllegalStateException("not implemented");
              default:
                throw new IllegalStateException("Learning type unknown.");
            }

            final double w_ij_new = w_ij + deltaw_ij;
            this.setAfferentLink(afferentModule, post, pre, w_ij_new);
          }
        }
      }
    }
  }

  public boolean isLearning() {
    return this.isLearning;
  }

  public void setLearning(final boolean isLearning) {
    this.isLearning = isLearning;
  }

  public double getLearningRate(final int neuron) {
    return this.learningRates.get(neuron);
  }

  public List<Double> getLearningRates() {
    return this.learningRates;
  }

  public int getLearningType() {
    return this.learningType;
  }

  public void setLearningRates(final double learningRate) {
    for (int neuron = 0; neuron < this.getSize(); neuron++) {
      this.learningRates.set(neuron, learningRate);
    }
  }

  public void setLearningRates(final int neuron, final double learningRate) {
    this.learningRates.set(neuron, learningRate);
  }

  public void setLearningType(final int learningType) {
    this.learningType = learningType;
  }

  public void removeNeuron(final int index) {
    this.activities.remove(index);

    // bias
    this.biases.remove(index);

    // learning rates
    this.learningRates.remove(index);

    // cliques have to be recalculated
    this.cliques = null;

    // remove edges from current graph
    this.links.remove(index);
    for (final List<Double> adjacencyMatrixRow : this.links) {
      adjacencyMatrixRow.remove(index);
    }

    // remove edges from afferent graphs
    // ***X*    ****
    // ***X*    ****
    // ***X* -> ****
    // XXXXX    ****
    // ***X*
    for (final Module afferentModule : this.getAfferentModules()) {
      this.afferentModules.get(afferentModule).remove(index);
      for (final List<Double> afferentLinkMatrixRow : this.afferentModules.get(afferentModule)) {
        afferentLinkMatrixRow.remove(index);
      }
    }

    // remove edges from efferent graphs
    for (final Module efferentModule : this.efferentModules) {
      efferentModule.afferentModules.get(this).remove(index);
      for (final List<Double> afferentLinkMatrixRow : efferentModule.afferentModules.get(this)) {
        afferentLinkMatrixRow.remove(index);
      }
    }
  }

  /**
   * Returns the activities levels.
   */
  public List<Double> getActivities() {
    return this.activities;
  }

  /**
   * Sets the activities levels.
   */
  public void setActivities(final List<Double> activities) {
    if (activities.size() != this.getSize()) {
      throw new IllegalArgumentException("wrong size of activities");
    }

    //Collections.copy(this.activities, activities);
    this.activities = activities;
  }

  /**
   * Returns the activities level of a given center.
   */
  public double getActivity(int i) {
    return this.activities.get(i);
  }

  /**
   * Sets the activity level of a given center.
   */
  public void setActivity(final int index, final double activity) {
    this.activities.set(index, activity);
  }

  /**
   * Sets the activity of all the given neurons to the given value and the
   * activities of all the other neurons to the other given value.
   * 
   * @param neurons the set of neurons to set the activities to using the first
   * value
   * @param activeActivity activity for the neurons specified
   * @param inactiveActivity activity for the complement of the neurons
   * specified
   */
  public void setActivities(final Set<Integer> neurons, final double activeActivity, final double inactiveActivity) {
    final TreeSet<Integer> sortedNeurons = new TreeSet<Integer>(neurons);
    for (int neuron = 0; neuron < this.getSize(); neuron++) {
      this.setActivity(neuron, sortedNeurons.contains(neuron) ? activeActivity : inactiveActivity);
    }
  }

  /**
   * Gets the number of nodes.
   */
  public int getSize() {
    return this.getActivities().size();
  }

  /**
   * Sets the number of nodes.
   */
  public void setSize(final int size) {
    if (size < 0) {
      throw new IllegalArgumentException("Size has to be non-negative.");
    }

    final double defaultActivity = 1e-8;

    // Add nodes if size is to small.
    while (size > this.getSize()) {
      this.addNeuron(defaultActivity);
    }

    // Remove nodes if size is to big.
    while (size < this.getSize()) {
      this.removeNeuron(this.getSize() - 1); // Remove the last vertex.
    }
  }

  /**
   * Gets the link w_ij (i ← j).
   * 
   * @param i
   * @param j
   * @return link
   */
  public double getLink(final int i, final int j) {
    return this.links.get(i).get(j);
  }

  /**
   * Gets the links w_ij (i ← j).
   * 
   * @return links
   */
  public List<List<Double>> getLinks() {
    return this.links;
  }

  /**
   * Gets the average positive links.
   */
  public double getPositiveLinkAverage() {
    int positiveLinkNumber = 0;
    double totalLinkStrength = 0;
    /*
     * for (final double link : Utils.flatten(new Double(0), this.getLinks())) {
     * if (link > 0) { positiveLinkNumber++; totalLinkStrength += link; } }
     */
    for (int i = 0; i < this.getSize(); i++) {
      for (int j = 0; j < this.getSize(); j++) {
        final double link = this.getLink(i, j);
        if (link > 0) {
          positiveLinkNumber++;
          totalLinkStrength += link;
        }
      }
    }
    final double average = totalLinkStrength / positiveLinkNumber;

    return average;
  }

  /**
   * Sets the negative links. The current negative links are overridden by the
   * given number.
   * 
   * @param value new link for each of the currently negative links
   */
  public void setNegativeLinks(final double value) {
    for (int i = 0; i < this.getSize(); i++) {
      for (int j = 0; j < this.getSize(); j++) {
        final double link = this.getLink(i, j);
        if (link < 0) {
          this.setLink(i, j, value);
        }
      }
    }
  }

  /**
   * Sets the positive links. The current positive links are overridden by the
   * given number.
   * 
   * @param value new link for each of the currently positive links
   */
  public void setPositiveLinks(final double value) {
    for (int i = 0; i < this.getSize(); i++) {
      for (int j = 0; j < this.getSize(); j++) {
        final double link = this.getLink(i, j);
        if (link > 0) {
          this.setLink(i, j, value);
        }
      }
    }
  }

  /**
   * Sets the positive link average by rescaling all the positive links.
   * 
   * @param value
   */
  public void setPositiveLinkAverage(final double value) {
    final double average = this.getPositiveLinkAverage();

    for (int i = 0; i < this.getSize(); i++) {
      for (int j = 0; j < this.getSize(); j++) {
        final double link = this.getLink(i, j);
        if (link > 0) {
          this.setLink(i, j, link / average * value);
        }
      }
    }
  }

  /**
   * Sets the link w_ij (i ← j).
   * 
   * @param i
   * @param j
   * @param value
   */
  public void setLink(final int i, final int j, final double value) {
    this.setLink(i, j, value, false);
  }

  /**
   * Sets the link w_ij (i ← j) and w_ji (i → j) if symmetric is true.
   * 
   * @param i
   * @param j
   * @param value
   * @param symmetric
   */
  public void setLink(final int i, final int j, final double value, final boolean symmetric) {
    this.links.get(i).set(j, value);
    if (symmetric) {
      this.links.get(j).set(i, value);
    }
  }

  /**
   * Sets the links $w$.
   * 
   * @param w weighted adjacency matrix
   */
  public void setLinks(final double[][] w) {
    for(int i = 0; i < w.length; i++){
      for(int j = 0; j < w[i].length; j++){
        this.setLink(i, j, w[i][j]);
      }
    }
  }

  public void multiplyExhibitoryAndInhibitoryLinksSeparately(final double exhibitoryFactor, final double inhibitoryFactor) {
    for (int i = 0; i < this.getSize(); i++) {
      for (int j = 0; j < this.getSize(); j++) {
        final double oldWeight = this.getLink(i, j);
        final double newWeight;
        if (oldWeight < 0.0) {
          // inhibitory link
          newWeight = inhibitoryFactor * oldWeight;
        } else if (oldWeight > 0.0) {
          // exhibitory link
          newWeight = exhibitoryFactor * oldWeight;
        } else {
          // no link
          newWeight = 0.0;
        }
        this.setLink(i, j, newWeight);
      }
    }
  }

  public double getLinkProbability() {
    if (this.getSize() < 2) {
      // There are not links in N = 0 and N = 1 networks.
      return 0;
    }

    // number of edges in the current graph
    double linksNumber = 0;
    for (int i = 0; i < this.getSize(); i++) {
      for (int j = 0; j < this.getSize(); j++) {
        if (this.getLink(i, j) > 0) {
          linksNumber++;
        }
      }
    }
    // Consider the graph to be undirected.
    linksNumber /= 2.0;

    // Total number of possible edges in an undirected graph of a given size, viz.: N·(N-1)/2.
    int totalLinksNumber = this.getSize() * (this.getSize() - 1) / 2;

    // Ratio of both.
    return linksNumber / (double) totalLinksNumber;
  }

  /**
   * Adds a module to the HashMap `afferentModules'.
   * <p>
   * 
   * @param afferentModule - the module to which this module connects
   */
  public void addAfferentModule(final Module afferentModule, final double defaultLink) {
    if (this.hasAfferentModule(afferentModule)) {
      throw new IllegalArgumentException("module already present");
    }

    final List<List<Double>> afferentLinks = new Vector<List<Double>>(this.getSize());
    for (int i = 0; i < this.getSize(); i++) {
      final List<Double> row = new Vector<Double>(afferentModule.getSize());
      for (int j = 0; j < afferentModule.getSize(); j++) {
        row.add(defaultLink);
      }
      afferentLinks.add(row);
    }
    this.afferentModules.put(afferentModule, afferentLinks);
    afferentModule.efferentModules.add(this);
  }

  /**
   * Adds a module to the HashMap `afferentModules'.
   * <p>
   * 
   * @param afferentModule - the module to which this module connects
   */
  public void addAfferentModule(final Module afferentModule) {
    // new links are initialized to one
    final double defaultLink = 1.0;
    this.addAfferentModule(afferentModule, defaultLink);
  }

  /**
  * Returns all afferent modules.
  */
  public Set<Module> getAfferentModules() {
    return this.afferentModules.keySet();
  }

  public void removeAfferentModule(final Module afferentModule) {
    if (this.afferentModules.remove(afferentModule) == null) {
      throw new IllegalArgumentException("no such module " + afferentModule);
    } else {
      afferentModule.efferentModules.remove(this);
    }
  }

  public void removeAfferentModules() {
    for (final Module module : this.afferentModules.keySet()) {
      module.efferentModules.remove(this);
    }
    this.afferentModules.clear();
  }

  public boolean hasAfferentModule(final Module module) {
    return this.afferentModules.containsKey(module);
  }

  public List<List<Double>> getAfferentLinks(final Module afferentModule) {
    if (!this.afferentModules.containsKey(afferentModule)) {
      throw new IllegalArgumentException("no such module " + afferentModule);
    }
    return this.afferentModules.get(afferentModule);
  }

  public void setAfferentLinks(final Module afferentModule, final List<List<Double>> afferentLinks) {
    if (!this.afferentModules.containsKey(afferentModule)) {
      throw new IllegalArgumentException("no such module " + afferentModule);
    }
    this.afferentModules.put(afferentModule, afferentLinks);
  }

  public double getAfferentLink(final Module afferentModule, final int i, final int j) {
    if (!this.afferentModules.containsKey(afferentModule)) {
      throw new IllegalArgumentException("no such module " + afferentModule);
    }
    return this.afferentModules.get(afferentModule).get(i).get(j);
  }

  public void setAfferentLink(final Module afferentModule, final int i, final int j, final double afferentLink) {
    if (!this.afferentModules.containsKey(afferentModule)) {
      throw new IllegalArgumentException("no such module " + afferentModule);
    }
    this.afferentModules.get(afferentModule).get(i).set(j, afferentLink);
  }

  public void setAfferentLinks(final Module afferentModule, final double afferentLink) {
    for (int i = 0; i < this.getSize(); i++) {
      for (int j = 0; j < afferentModule.getSize(); j++) {
        this.setAfferentLink(afferentModule, i, j, afferentLink);
      }
    }
  }

  public void addForwardModule(final Module module, final String whichPort) {
    this.forwardModules.put(module, whichPort);
  }

  /**
   * Returns all forward module names as an ArrayList of Strings.
   * <p>
   * --class DhanModule-- <br>
   */
  public Module[] getForwardModules() {
    return this.forwardModules.keySet().toArray(new Module[0]);
  }

  /**
   * Removes a particular forward module from the HashMap of forward modules.
   * <p>
   * --class DhanModule-- <br>
   * 
   * @param module name of module to remove
   */
  public void removeForwardModule(final Module module) {
    if (!this.forwardModules.containsKey(module)) {
      throw new IllegalArgumentException("no such module");
    }
    this.forwardModules.remove(module);
  }

  /**
   * Remove all forward modules from the HashMap of forward modules.
   * <p>
   * --class DhanModule-- <br>
   */
  public void removeForwardModules() {
    this.forwardModules.clear();
  }

  public boolean hasForwardModule(final Module module) {
    return this.forwardModules.containsKey(module);
  }

  /**
   * Adds contribution `addValue' to the specified receptor field, returning
   * `false' if specified receptor does not exist.
   */
  public void addReceptor(String receptorType, double addValue) {
    if (this.receptors.containsKey(receptorType)) {
      this.receptors.put(receptorType, this.receptors.get(receptorType) + addValue);
    } else {
      this.receptors.put(receptorType, addValue);
    }
  }

  /** Returns the value of the specified receptor field. */
  public double getReceptor(String receptorType) {
    if (!this.receptors.containsKey(receptorType)) {
      throw new IllegalArgumentException("receptor " + receptorType + " unknown");
    } else {
      return this.receptors.get(receptorType);
    }
  }

  public Map<String, Double> getReceptors() {
    return this.receptors;
  }

  @Override
  public String toString() {
    return this.getClass().getSimpleName() + " " + this.getName();
    //return this.getClass().getSimpleName() + " " + this.getName() + " N=" + this.getSize() + " p=" + this.getLinkProbability();
  }

  public void addChangeListener(final ChangeListener listener) {
    this.changeListeners.add(ChangeListener.class, listener);
  }

  public void removeChangeListener(ChangeListener listener) {
    this.changeListeners.remove(ChangeListener.class, listener);
  }

  protected synchronized void notifyStateChanged(ChangeEvent event) {
    for (final ChangeListener listener : this.changeListeners.getListeners(ChangeListener.class)) {
      listener.stateChanged(event);
    }
  }

  public double getAverageCoordinationNumber() {
    return (this.getSize() - 1.0) / this.getLinkProbability();
  }

  /**
   * Returns the degrees of the vertices.
   */
  public int getNodeDegree(final int node) {
    int nodeDegree = 0;
    for (int i = 0; i < this.getSize(); i++) {
      if (this.getLink(i, node) != 0.0 || this.getLink(node, i) != 0.0) {
        nodeDegree++;
      }
    }
    return nodeDegree;
  }

  public int getNodeInDegree(final int node, final int sign) {
    int degree = 0;
    for (int i = 0; i < this.getSize(); i++) {
      final double link = this.getLink(node, i);
      if ((sign == 0 && link != 0) || (sign < 0 && link < 0) || (sign > 0 && link > 0)) {
        degree++;
      }
    }
    return degree;
  }

  public int getNodeOutDegree(final int node, final int sign) {
    int degree = 0;
    for (int i = 0; i < this.getSize(); i++) {
      final double link = this.getLink(i, node);
      if ((sign == 0 && link != 0.0) || (sign < 0 && link < 0.0) || (sign > 0 && link > 0.0)) {
        degree++;
      }
    }
    return degree;
  }

  /**
   * Returns the degrees of the vertices.
   */
  public List<Integer> getNodeDegrees() {
    final List<Integer> nodeDegrees = new Vector<Integer>();
    for (int i = 0; i < this.getSize(); i++) {
      nodeDegrees.add(this.getNodeDegree(i));
    }
    return nodeDegrees;
  }

  /**
   * Returns the degrees of the vertices.
   */
  public List<Integer> getNodeInDegrees(final int sign) {
    final List<Integer> nodeDegrees = new Vector<Integer>();
    for (int i = 0; i < this.getSize(); i++) {
      nodeDegrees.add(this.getNodeInDegree(i, sign));
    }
    return nodeDegrees;
  }

  /**
   * Returns the degrees of the vertices.
   */
  public List<Integer> getNodeOutDegrees(final int sign) {
    final List<Integer> nodeDegrees = new Vector<Integer>();
    for (int i = 0; i < this.getSize(); i++) {
      nodeDegrees.add(this.getNodeOutDegree(i, sign));
    }
    return nodeDegrees;
  }

  /**
   * Returns the degree distribution of the graph.
   */
  public Map<Integer, Integer> getDegreeDistribution() {
    final Map<Integer, Integer> degreeDistribution = new TreeMap<Integer, Integer>();
    for (final int nodeDegree : this.getNodeDegrees()) {
      if (!degreeDistribution.containsKey(nodeDegree)) {
        degreeDistribution.put(nodeDegree, 1);
        continue;
      }
      degreeDistribution.put(nodeDegree, 1 + degreeDistribution.get(nodeDegree));
    }
    return degreeDistribution;
  }

  /**
   * Returns the cliques of the graph.
   */
  public synchronized Set<Set<Integer>> getCliques() {
    if (this.cliques == null) {
      logger.log(Level.FINER, "start searching for all cliques");
      // The recursion is initiated by setting R and X to be the empty set and P to be the vertex set of the graph.
      final Set<Integer> r = Collections.emptySet();
      final Set<Integer> p = new TreeSet<Integer>();
      for (int i = 0; i < this.getSize(); i++) {
        p.add(i);
      }
      final Set<Integer> x = new TreeSet<Integer>();
      this.cliques = this.cliquesBronKerbosch(r, p, x);
      logger.log(Level.FINER, "searching for all cliques complete");
    }
    return this.cliques;
  }

  /**
   * Determines whether a given cluster is a maximal clique.
   */
  public boolean isClique(final Set<Integer> cluster) {
    if (this.cliques != null) {
      // If all the cliques are already known it could be or is faster to use that information.
      return this.cliques.contains(cluster);
    } else {
      // Is cluster fully connected, so a common known clique?
      for (final int i : cluster) {
        for (final int j : cluster) {
          if (i != j && this.getLink(i, j) <= 0.0 && this.getLink(j, i) <= 0.0) {
            return false;
          }
        }
      }

      // Is that clique maximal?
      for (int i = 0; i < this.getSize(); i++) {
        // Outside of the cluster?
        if (!cluster.contains(i)) {
          // Connected to all?
          for (final int j : cluster) {
            if (this.getLink(i, j) <= 0.0 || this.getLink(j, i) <= 0.0) {
              return true;
            }
          }
        }
      }

      // Clique is not maximal.
      return false;
    }
  }

  /**
   * returns all the neighbors of all the vertices
   * 
   * @return neighbors
   */
  public List<Set<Integer>> getNeighbors() {
    final List<Set<Integer>> neighborList = new Vector<Set<Integer>>();
    for (int i = 0; i < this.getSize(); i++) {
      final Set<Integer> neighbors = this.getNeighbors(i);
      neighborList.add(neighbors);
    }
    return neighborList;
  }

  /**
   * returns all the neighbors of a vertex It considers positive links only and
   * no autapses.
   * 
   * @return neighbors
   */
  public Set<Integer> getNeighbors(final int neuron) {
    final Set<Integer> neighbors = new TreeSet<Integer>();
    for (int j = 0; j < this.getSize(); j++) {
      if (j != neuron && this.getLink(j, neuron) > 0.0) {
        neighbors.add(j);
      }
    }
    return neighbors;
  }

  /**
   * Returns the currently active clique of the graph.
   * 
   * @return cliques
   */
  public Set<Integer> getActiveClique() {
    Set<Integer> result = null;

    // sort activities keeping indices
    final Integer[] sortedIndices;
    if (this.sortedIndices == null) {
      sortedIndices = new Integer[this.getActivities().size()];
      this.sortedIndices = sortedIndices;
    } else {
      sortedIndices = this.sortedIndices;
    }
    for (int i = 0; i < this.getActivities().size(); i++) {
      sortedIndices[i] = i;
    }
    final List<Double> data = this.getActivities();
    Arrays.sort(sortedIndices, new Comparator<Integer>() {
      @Override
      public int compare(final Integer o1, final Integer o2) {
        return data.get(o2).compareTo(data.get(o1));
      }
    });

    final Set<Set<Integer>> remainingCliques = new HashSet<Set<Integer>>(this.getCliques());
    for (int i = 0; i < sortedIndices.length; i++) {
      // remove cliques not having the vertex
      final Set<Set<Integer>> cliquesToRemove = new HashSet<Set<Integer>>();
      for (final Set<Integer> clique : remainingCliques) {
        if (!clique.contains(sortedIndices[i])) {
          cliquesToRemove.add(clique);
        }
      }
      remainingCliques.removeAll(cliquesToRemove);

      // No clique left?
      if (remainingCliques.size() == 0) {
        break;
      }

      // Only 1 clique remaining …?
      if (remainingCliques.size() == 1) {
        final Set<Integer> candidate = remainingCliques.iterator().next();
        // … and its size is i+1 …
        if (candidate.size() == i + 1) {
          final double minimumCliqueActivity = 1e-8; // default: 1e-8;
          // … and there is a special minimum activity …
          if (data.get(sortedIndices[i]) > minimumCliqueActivity) {
            final double cliqueGap = 1.5;
            // … and it’s the last node or the next node’s activity is cliqueGap times smaller
            if (i == sortedIndices.length - 1 || cliqueGap * data.get(sortedIndices[i + 1]) < data.get(sortedIndices[i])) {
              result = candidate;
              break;
            }
          }
        }
      }
    }

    return result;
  }

  /**
   * finds cliques using the Bron-Kerbosch algorithm
   * 
   * http://en.wikipedia.org/wiki/Bron%E2%80%93Kerbosch_algorithm
   * 
   * Bron, Coen; Kerbosch, Joep (1973),
   * "Algorithm 457: finding all cliques of an undirected graph", Commun. ACM
   * (ACM) 16 (9): 575–577, doi:10.1145/362342.362367
   * (http://dx.doi.org/10.1145%2F362342.362367)
   * 
   * @param r
   * @param p
   * @param x
   * @return cliques
   */
  private Set<Set<Integer>> cliquesBronKerbosch(final Set<Integer> r, final Set<Integer> p, final Set<Integer> x) {
    final Set<Set<Integer>> cliques = new HashSet<Set<Integer>>();
    final List<Set<Integer>> neighborList = this.getNeighbors();
    if (p.isEmpty() && x.isEmpty()) {
      cliques.add(r);
    }

    // The copy of the set p is necessary, because p will be changed in the loop.
    for (final Integer v : new Vector<Integer>(p)) {
      final Set<Integer> rNew = new TreeSet<Integer>(r);
      rNew.add(v);

      final Set<Integer> pNew = new TreeSet<Integer>(p);
      pNew.retainAll(neighborList.get(v));

      final Set<Integer> xNew = new TreeSet<Integer>(x);
      xNew.retainAll(neighborList.get(v));

      cliques.addAll(this.cliquesBronKerbosch(rNew, pNew, xNew));

      p.remove(v);
      x.add(v);
    }
    return cliques;
  }

  /**
   * returns the clique graph of the graph as an adjacency matrix
   */
  public List<List<Double>> getCliqueGraph() {
    final Set<Set<Integer>> cliques = this.getCliques();
    final List<List<Double>> adjacencyMatrix = new Vector<List<Double>>(cliques.size());
    for (final Set<Integer> cliqueA : cliques) {
      final List<Double> adjacencyMatrixRow = new Vector<Double>(cliques.size());
      for (final Set<Integer> cliqueB : cliques) {
        if (cliqueA != cliqueB && !Collections.disjoint(cliqueA, cliqueB)) {
          adjacencyMatrixRow.add(1.0);
        } else {
          adjacencyMatrixRow.add(0.0);
        }
      }
      adjacencyMatrix.add(adjacencyMatrixRow);
    }
    return adjacencyMatrix;
  }

  /**
   * http://en.wikipedia.org/wiki/Trivial_Graph_Format
   * 
   * @param tgfFile
   * @throws FileNotFoundException
   */
  public void exportTGFGraph(final File tgfFile) throws FileNotFoundException {
    final PrintWriter tgfOut = new PrintWriter(tgfFile);
    for (int i = 0; i < this.getSize(); i++) {
      tgfOut.println(i + " " + i);
    }
    tgfOut.println("#");
    for (int i = 0; i < this.getSize(); i++) {
      for (int j = 0; j < this.getSize(); j++) {
        final double w_ij = this.getLink(i, j);
        if (w_ij > 0.0) {
          // do print only links with strength not zero, viz. really existing links
          tgfOut.println(j + " " + i + " " + w_ij);
        }
      }
    }
    tgfOut.close();
  }

  public void exportGraphMLGraph(final File graphmlFile) throws FileNotFoundException {
    final PrintWriter graphmlOut = new PrintWriter(graphmlFile);
    graphmlOut.println("<?xml version=\"1.0\" encoding=\"UTF-8\"?>");
    graphmlOut
        .println("<graphml xmlns=\"http://graphml.graphdrawing.org/xmlns\" xmlns:xsi=\"http://www.w3.org/2001/XMLSchema-instance\" xsi:schemaLocation=\"http://graphml.graphdrawing.org/xmlns http://graphml.graphdrawing.org/xmlns/1.0/graphml.xsd\">");
    graphmlOut.println("<graph id=\"G\" edgedefault=\"directed\">");
    for (int i = 0; i < this.getSize(); i++) {
      graphmlOut.println("<node id=\"node" + i + "\"/>");
    }
    for (int i = 0; i < this.getSize(); i++) {
      for (int j = 0; j < this.getSize(); j++) {
        final double w_ij = this.getLink(i, j);
        if (w_ij > 0.0) {
          // do print only links with strength not zero, viz. really existing links
          graphmlOut.println("<edge id=\"edge" + j + "." + i + "\" source=\"node" + j + "\" target=\"node" + i + "\"/>");
        }
      }
    }
    graphmlOut.println("</graph>");
    graphmlOut.println("</graphml>");
    graphmlOut.close();
  }

  public void exportGraphvizGraph(final File gvFile) throws FileNotFoundException {
    this.exportGraphvizGraph(gvFile, Double.NEGATIVE_INFINITY, Double.POSITIVE_INFINITY);
  }

  public void exportGraphvizGraph(final File gvFile, final double lowerLinkBound, final double upperLinkBound) throws FileNotFoundException {
    final PrintWriter dotOut = new PrintWriter(gvFile);
    dotOut.println("digraph G{");
    dotOut.println("graph [bgcolor=transparent]");

    this.writeGraphvizSubgraph(dotOut, lowerLinkBound, upperLinkBound);

    dotOut.println("}");
    dotOut.close();
  }

  public void writeGraphvizSubgraph(final PrintWriter out) {
    this.writeGraphvizSubgraph(out, Double.NEGATIVE_INFINITY, Double.POSITIVE_INFINITY);
  }

  public void writeGraphvizSubgraph(final PrintWriter out, final double lowerLinkBound, final double upperLinkBound) {
    final String name = this.getName() != null ? this.getName() : "unnamed";
    out.println("subgraph cluster_" + name.replaceAll("\\s", "") + " {"); // whitespace removed, maybe \W "non word character" should be used?
    out.println("\tlabel= \"" + name + "\";");
    out.println("\tcolor= black;");
    //out.println("\tstyle= \"filled\";");
    //out.println("\tbgcolor= red;");

    out.println();

    // vertices
    for (int i = 0; i < this.getSize(); i++) {
      out.println("\t\"" + this.getName() + "_" + i + "\" [label=\"" + i + "\"];");
    }

    out.println();

    // intra module edges
    for (int i = 0; i < this.getSize(); i++) {
      for (int j = 0; j < this.getSize(); j++) {
        final double w_ij = this.getLink(i, j);
        final String color = w_ij >= 0 ? "#008000" : "#800000";
        final String iNodeName = name + "_" + i;
        final String jNodeName = name + "_" + j;
        if (w_ij != 0.0 && w_ij >= lowerLinkBound && w_ij <= upperLinkBound) {
          // do print only links with specific strength
          out.println("\t\"" + jNodeName + "\" -> \"" + iNodeName + "\" [label=\"" + w_ij + "\" color=\"" + color + "\" fontcolor=\"" + color + "\"];");
        }
      }
    }
    out.println("}");
  }

  /**
   * Prints the xfig representation of one module with the vertices arranged on
   * a circle.
   */
  public void exportXfigCircle(final File file) throws FileNotFoundException {
    final PrintWriter output = new PrintWriter(file);

    // --- printing xfig header ---
    output.println("#FIG 3.2 ");
    output.println("Landscape ");
    output.println("OneCenter ");
    output.println("Metric ");
    output.println("A4 ");
    output.println("100.00 ");
    output.println("Single ");
    output.println("-2 ");
    output.println("1200 2 ");

    // --- print vertices on a circle ---
    final Point center = new Point(4500, 4500);
    final double circleRadius = 3000.0;
    final int pointRadius = 120;
    final Vector<Point> allPoints = new Vector<Point>(this.getSize()); // position of vertices

    for (int i = 0; i < this.getSize(); i++) {
      final Point point = new Point((int) (center.x + circleRadius * Math.cos(2.0 * i * Math.PI / this.getSize() + 0.5 * Math.PI)),
          (int) (center.y + circleRadius * Math.sin(2.0 * i * Math.PI / this.getSize() + 0.5 * Math.PI)));
      allPoints.add(point);
    }

    // --- print associations as lines ---
    for (int i = 0; i < this.getSize(); i++) {
      for (int j = 0; j < i; j++) {
        if (this.getLink(i, j) > 0) {
          if ((this.getActivity(i) > 0.8) && (this.getActivity(j) > 0.8)) {
            output.println("2 1 1 4 15 7 50 -1 -1 0.000 0 0 -1 0 0 2");
          } else {
            output.println("2 1 0 3  0 7 50 -1 -1 0.000 0 0 -1 0 0 2");
          }
          output.printf(Locale.ENGLISH, "    %d %d    %d %d \n", allPoints.get(i).x, allPoints.get(i).y, allPoints.get(j).x, allPoints.get(j).y);
        }
      }
    }

    // --- draw vertices ---
    for (int i = 0; i < this.getSize(); i++) {
      if (this.getActivity(i) > 0.8) {
        output.printf(Locale.ENGLISH, "1 3 0 4 15  7 49 -1 20 0.000 1 0.0000");
      } else {
        output.printf(Locale.ENGLISH, "1 3 0 1 10 10 49 -1 20 0.000 1 0.0000");
      }
      final int x = allPoints.get(i).x;
      final int y = allPoints.get(i).y;
      output.printf(Locale.ENGLISH, " %d %d %d %d %d %d %d %d\n", x, y, pointRadius, pointRadius, x, y, x + pointRadius, y);
    }

    // --- print site numbers ---
    for (int i = 0; i < this.getSize(); i++) {
      final int x = (int) (1.15 * (allPoints.get(i).x - center.x) + center.x - 200);
      final int y = (int) (1.15 * (allPoints.get(i).y - center.y) + center.y + 100);
      output.printf(Locale.ENGLISH, "4 0 0 50 -1 0 20 0.0000 4 180 660");
      output.printf(Locale.ENGLISH, " %d %d (%d)\\001\n", x, y, i);
    }

    output.close();
  }

  /**
   * Prints the xfig representation of one module with the vertices arranged on
   * a two dimensional square lattice.
   */
  public void exportXfigLattice(final File file) throws FileNotFoundException {
    OutputStream ostream = new FileOutputStream(file);
    Writer writer = new OutputStreamWriter(ostream);
    PrintWriter output = new PrintWriter(writer);

    // --- printing xfig header ---
    output.println("#FIG 3.2 ");
    output.println("Landscape ");
    output.println("OneCenter ");
    output.println("Metric ");
    output.println("A4 ");
    output.println("100.00 ");
    output.println("Single ");
    output.println("-2 ");
    output.println("1200 2 ");
    int dimX = 0;
    int dimY = this.getSize();
    while ((dimX < dimY) || (dimX * dimY != this.getSize())) {
      dimX++;
      dimY = (this.getSize() / dimX);
    }

    // --- print vertices on a two dimensional lattice ---
    int startX = 2000;
    int startY = 2000;
    int stepX = 1600;
    int stepY = 1600;
    int PointRadius = 120;

    int[][] allPoints = new int[this.getSize()][2]; // position of vertices

    int iSite = 0;
    for (int iY = 0; iY < dimY; iY++) {
      for (int iX = 0; iX < dimX; iX++) {
        allPoints[iSite][0] = startX + iX * stepX;
        allPoints[iSite][1] = startY + iY * stepY;
        iSite++;
      }
    }

    // --- print associations as splines ---
    double arcValue = 0.10; // a measure for the curviness
    assert ((arcValue > 0.01) && (arcValue < 0.6));

    for (int i = 0; i < this.getSize(); i++) {
      for (int j = 0; j < i; j++) {
        if (this.getLink(i, j) > 0) {
          int x1 = allPoints[i][0];
          int y1 = allPoints[i][1];
          int x2 = allPoints[j][0];
          int y2 = allPoints[j][1];
          double middleX = 0.5 * (x2 + x1);
          double middleY = 0.5 * (y2 + y1);
          double deltaX = 1.0 * (x2 - x1);
          double deltaY = 1.0 * (y2 - y1);
          double xBetween = middleX + arcValue * deltaY;
          double yBetween = middleY - arcValue * deltaX;

          if ((this.getActivity(i) > 0.8) && (this.getActivity(j) > 0.8)) {
            output.printf(Locale.ENGLISH, "3 2 1 4 15 7 50 -1 -1 0.000 0 0 0 3\n");
          } else {
            output.printf(Locale.ENGLISH, "3 2 0 3  0 7 50 -1 -1 0.000 0 0 0 3\n");
          }
          output.printf(Locale.ENGLISH, "          %d %d %d %d %d %d\n", x1, y1, (int) xBetween, (int) yBetween, x2, y2);
          output.printf(Locale.ENGLISH, "          0.000 -1.000 0.000 \n");
        }
      }
    }

    // --- draw vertices ---
    for (int i = 0; i < this.getSize(); i++) {
      int xx = allPoints[i][0];
      int yy = allPoints[i][1];
      if (this.getActivity(i) > 0.8)
        output.printf(Locale.ENGLISH, "1 3 0 4 15  7 49 -1 20 0.000 1 0.0000");
      else
        output.printf(Locale.ENGLISH, "1 3 0 1 10 10 49 -1 20 0.000 1 0.0000");
      output.printf(Locale.ENGLISH, " %d %d %d %d %d %d %d %d\n", xx, yy, PointRadius, PointRadius, xx, yy, xx + PointRadius, yy);
    }

    output.close();
  }

  public String getName() {
    return this.name;
  }

  public void setName(final String name) {
    this.name = name;
  }

  public void addBias(final double bias) {
    for (int i = 0; i < this.getSize(); i++) {
      this.addBias(i, bias);
    }
  }

  public void addBias(final int index, final double bias) {
    this.biases.set(index, this.biases.get(index) + bias);
  }

  public double getBias(final int index) {
    return this.biases.get(index);
  }

  public List<Double> getBiases() {
    return this.biases;
  }

  public void setBias(final double bias) {
    for (int neuron = 0; neuron < this.getSize(); neuron++) {
      this.setBias(neuron, bias);
    }
  }

  public void setBias(final int index, final double bias) {
    this.biases.set(index, bias);
  }

  public void reset() {
  }

  private Set<Integer> getConnectedComponent(int vertex, final Set<Integer> connectedComponent) {
    connectedComponent.add(vertex);
    for (final int v : this.getNeighbors(vertex)) {
      if (!connectedComponent.contains(v)) {
        getConnectedComponent(v, connectedComponent);
      }
    }

    return connectedComponent;
  }

  private Set<Integer> getConnectedComponent(int vertex) {
    final Set<Integer> connectedComponent = new HashSet<Integer>();
    return getConnectedComponent(vertex, connectedComponent);
  }

  public Set<Set<Integer>> getConnectedComponents() {
    final Set<Integer> remainingNeurons = new HashSet<Integer>(this.getSize());
    for (int i = 0; i < this.getSize(); i++) {
      remainingNeurons.add(i);
    }

    final Set<Set<Integer>> connectedComponents = new HashSet<Set<Integer>>();
    while (!remainingNeurons.isEmpty()) {
      int neuron = remainingNeurons.iterator().next();

      final Set<Integer> connectedComponent = this.getConnectedComponent(neuron);
      connectedComponents.add(connectedComponent);
      remainingNeurons.removeAll(connectedComponent);
    }

    return connectedComponents;
  }

  /**
   * Prints a file header for the log files.
   */
  public void printLogFileHeader(final PrintWriter out) {
    final String commentOperator = "# ";
    out.println(commentOperator + "type:");
    out.println(commentOperator + this.getClass().getCanonicalName());
    out.println(commentOperator + "size:");
    out.println(commentOperator + this.getSize());
    out.println(commentOperator + "name:");
    out.println(commentOperator + this.getName());
    out.println(commentOperator + "links:");
    out.println(commentOperator + this.getLinks());
  }

  /**
   * Necessary for (de-)serialization in order to initialize transient fields.
   */
  private void readObject(final ObjectInputStream in) throws IOException, ClassNotFoundException {
    in.defaultReadObject();

    // Initialize all the transient fields.
    this.changeListeners = new EventListenerList();
  }
}