package brain;

import java.io.BufferedReader;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.io.PrintWriter;
import java.io.Reader;
import java.io.Serializable;
import java.util.ArrayList;
import java.util.HashSet;
import java.util.List;
import java.util.Locale;
import java.util.Random;
import java.util.Set;
import java.util.SortedMap;
import java.util.TreeMap;
import java.util.Vector;
import java.util.Map.Entry;
import java.util.logging.Level;
import java.util.logging.Logger;

import utils.Interval;

/**
 * The PatternInputModule module of the cognitive system (CS).
 * 
 * All parameters for time development of activities and reservoirs are
 * generally public. Exceptions are parameters which need additional computation
 * or checks when changed.
 * 
 * @author Claudius Gros
 * @author Mathias Linkerhand
 * @version 2011-09-20
 */
public class PatternInputModule extends Module implements Serializable {
  static final long                        serialVersionUID      = 1L;

  private static final Logger              logger                = Logger.getLogger(PatternInputModule.class.getName());

  /** String-identifier for the module. */
  public String                            name;

  /**
   * The diffusive signal specific to the input module, it's task yet to be
   * determined.
   */
  public double                            diffusiveModuleSignal = 0.0;

  /**
   * Dynamic LinkedList of `OneStimulus' with the external remainingStimuli - may be
   * changed during run-time.
   */
  private SortedMap<Interval, OneStimulus> remainingStimuli      = null;

  /**
   * Set of `OneStimulus' which have been applied to PatternInputModule.
   */
  public Set<OneStimulus>                  appliedStimuli;

  /**
   * For the first stimulus.
   */
  public double                            stimulusStartingTime  = 100.0;
  /**
   * Length of a stimulus.
   */
  public double                            stimulusDuration      = 40.0;
  public double                            timeBetweenStimuli    = 100.0;

  /** Dynamic ArrayList of `OneStimulus' with the reference patterns. */
  public ArrayList<OneStimulus>            referencePatterns;

  public PatternInputModule() {
    this(new File(Brain.DEFAULT_DATA_INPUT_DIRECTORY, "input.patterns"));
  }

  /**
   * The module-name is set for good. All other parameters are changeable.
   */
  public PatternInputModule(final File patternsFile) {
    this.name = patternsFile.getName();
    this.name = "input";

    this.appliedStimuli = new HashSet<OneStimulus>();
    this.referencePatterns = new ArrayList<OneStimulus>();

    // receptor fields
    this.receptors.put("default", 0.0);

    // remainingStimuli
    this.stimulusStartingTime = 100.0; // for first stimulus
    this.stimulusDuration = 40.0; // length of a stimulus

    try {
      // read remainingStimuli from file
      final OneStimulus[] stimuli = PatternInputModule.loadStimuli(patternsFile, this.stimulusStartingTime, this.stimulusDuration, this.timeBetweenStimuli);
      this.remainingStimuli = new TreeMap<Interval, OneStimulus>();

      int maxStimulusVertexNumber = 0;
      for (final OneStimulus stimulus : stimuli) {
        this.remainingStimuli.put(stimulus.getTimeInterval(), stimulus);

        for (final int node : stimulus.getNodes()) {
          // set this.N as largest stimulus site
          maxStimulusVertexNumber = Math.max(maxStimulusVertexNumber, node);
        }
      }
      this.setSize(maxStimulusVertexNumber + 1);
    } catch (FileNotFoundException e) {
      e.printStackTrace();
    }
  }

  /** Returns the name specific to this module. */
  @Override
  public String getName() {
    return this.name;
  }

  /** Returns the value of a specific internal link. */
  @Override
  public double getLink(int i, int j) {
    //throw new UnsupportedOperationException("There are no links in input module.");
    return 0.0;
  }

  /** Sets the value of a specific internal link. */
  @Override
  public void setLink(int i, int j, double value) {
    throw new UnsupportedOperationException("Cannot change links of input module.");
  }

  @Override
  public void addAfferentModule(final Module module) {
    throw new UnsupportedOperationException("Cannot add afferent module to input module.");
  }

  public OneStimulus[] getStimuli(final double time) {
    final Vector<OneStimulus> stimuli = new Vector<OneStimulus>();
    final SortedMap<Interval, OneStimulus> stimuliStartedBeforeTime = (SortedMap<Interval, OneStimulus>) this.remainingStimuli.headMap(new Interval(time,
        Double.POSITIVE_INFINITY));
    for (final Entry<Interval, OneStimulus> entry : stimuliStartedBeforeTime.entrySet()) {
      final Interval interval = entry.getKey();
      final OneStimulus stimulus = entry.getValue();

      if (stimulus.getTimeInterval().getEnd() >= time) {
        stimuli.add(stimulus);
      } else {
        this.remainingStimuli.remove(interval);
      }
    }
    return stimuli.toArray(new OneStimulus[0]);
  }
  
  public OneStimulus[] getCurrentStimuli(){
    final OneStimulus[] currentStimuli = this.getStimuli(this.elapsedTime);
    return currentStimuli;
  }

  /**
   * A single time step of the dynamics.
   * 
   * <p>
   * --class PatternInputModule-- <br>
   * Sets the input activities determined by the remainingStimuli.
   */
  @Override
  public void calculateOneTimeStep(final double timeDelta) {
    // --- obtain currently active remainingStimuli ---
    final OneStimulus[] currentStimuli = this.getCurrentStimuli();

    final Random random = new Random(0);

    // random nearly zero activities
    for (int i = 0; i < this.getSize(); i++) {
      this.setActivity(i, 0.01 * random.nextDouble()); // 0.01 * [0.0, 1.0) = [0.00, 0.01)
    }

    // add remainingStimuli to activities
    for (final OneStimulus stimulus : currentStimuli) {
      for (final int node : stimulus.getNodes()) {
        if (node < this.getSize()) { // bound check
          this.setActivity(node, this.getActivity(node) + stimulus.getStrength(node));
        }
      }
    }

    // --- check activities-bounds ---
    for (int i = 0; i < this.getSize(); i++) {
      this.setActivity(i, Math.min(this.getActivity(i), DhanModule.nearlyActive));
    }

    // zero bias diffusive control receptors, they are set by the other/same module via the forward connections
    for (String receptorType : this.receptors.keySet()) {
      this.receptors.put(receptorType, 0.0);
    }

    // forward inter-modular diffusive signals
    for (final Module module : this.forwardModules.keySet()) {
      final String forwardPort = this.forwardModules.get(module);
      module.addReceptor(forwardPort, this.diffusiveModuleSignal);
    }
  }

  /**
   * Reads pattern in the form of cliques from file.
   * 
   * <p>
   * --class PatternInputModule-- <br>
   * @throws FileNotFoundException
   */
  public static OneStimulus[] loadStimuli(final File file, final double stimulusStartingTime, final double stimulusDuration, final double timeBetweenStimuli)
      throws FileNotFoundException {
    final Vector<OneStimulus> stimuli = new Vector<OneStimulus>();

    // noise-level of remainingStimuli strengths in [0, 1]
    final double stiRandomNoisePercentage = 0.0;

    // Number of repetitions of the input remainingStimuli.
    final int stiNumberOfStimuliSetRepetitions = 0;

    final Random random = new Random(0);

    double tmp_startingTime = stimulusStartingTime;
    final List<List<Integer>> cliques = DhanModule.cliquesToInteger(DhanModule.loadCliques(file));
    for (final List<Integer> clique : cliques) {
      final OneStimulus stimulus = new OneStimulus();
      for (final int node : clique) {
        final double strength = OneStimulus.standardStimulusStrength * (1.0 + stiRandomNoisePercentage * (random.nextDouble() - 1.0));
        stimulus.setStrength(node, strength);
      }
      tmp_startingTime += timeBetweenStimuli; // update starting time
      stimulus.interval = new Interval(tmp_startingTime, tmp_startingTime + stimulusDuration);
      stimuli.add(stimulus); // add new stimulus

      // add the same stimulus again and again
      for (int counter = 0; counter < stiNumberOfStimuliSetRepetitions; counter++) {
        stimuli.add(stimulus); // add new stimulus
      }
    }

    logger.log(Level.INFO, "patterns loaded: " + cliques.size());

    logger.log(Level.INFO, "duplicated and randomized patterns: " + stimuli.size());
    return stimuli.toArray(new OneStimulus[0]);
  }

  /**
   * Reads the reference patterns in the form of cliques from file.
   * 
   * The reference patterns may be used for controlling learning, e.g. an
   * independent component analysis.
   * <p>
   * --class PatternInputModule-- <br>
   * @throws IOException 
   */
  public void readReferencePattern(final File referenceInputPatternsFile) throws IOException {
    // read via buffered reader
    this.referencePatterns.clear();
    InputStream istream = new FileInputStream(referenceInputPatternsFile);
    Reader reader = new InputStreamReader(istream);
    BufferedReader myInput = new BufferedReader(reader);
    // --- ------------------------- ---
    // --- read line by line of form ---
    // --- Z p1 p2 .... pZ ---
    // --- ------------------------- ---
    for (String line = myInput.readLine(); line != null; line = myInput.readLine()) {
      if (line.indexOf("#") == -1) // skip comments
      {
        line = line.trim(); // removes leading and trailing white spaces
        if (line.length() == 0) {
          continue;
        }

        String[] myTokens = line.split("\\s+"); // split input line
        int nSites = Integer.parseInt(myTokens[0]);
        assert ((nSites == myTokens.length - 1) || // without grey-scale
        (2 * nSites == myTokens.length - 1)); // with grey-scale

        this.referencePatterns.add(new OneStimulus()); // add new stimulus

        for (int iS = 1; iS < (1 + nSites); iS++) {
          this.referencePatterns.get(this.referencePatterns.size() - 1).setStrength(Integer.parseInt(myTokens[iS]), OneStimulus.standardStimulusStrength);
        }
      }
    }
    myInput.close();
  }

  @Override
  public void reset() {
    // TODO Auto-generated method stub
  }

  @Override
  public void printLogFileHeader(final PrintWriter out) {
    super.printLogFileHeader(out);

    out.printf(Locale.ENGLISH, "# --------------------------- \n");
    out.printf(Locale.ENGLISH, "# ---   header for module --- : %s\n", this.getName());
    out.printf(Locale.ENGLISH, "# --------------------------- \n");
    out.printf(Locale.ENGLISH, "#             number of sites :%10d\n", this.getSize());
    out.printf(Locale.ENGLISH, "# --------------------------- \n");
    out.printf(Locale.ENGLISH, "# --- stimuli information --- \n");
    out.printf(Locale.ENGLISH, "# --------------------------- \n");
    out.printf(Locale.ENGLISH, "# --- forward connections --- \n");
    out.printf(Locale.ENGLISH, "# --------------------------- \n");
    for (final Module forwardModule : this.forwardModules.keySet()) {
      out.printf(Locale.ENGLISH, "#       target module, port :%10s\n", forwardModule.getName());
    }
    out.printf(Locale.ENGLISH, "# --------------------------- \n");
    out.printf(Locale.ENGLISH, "# ---  reference patters  --- \n");
    out.printf(Locale.ENGLISH, "# --------------------------- \n");
    out.printf(Locale.ENGLISH, "#    number reference pattern :%10d\n", this.referencePatterns.size());
    out.printf(Locale.ENGLISH, "# *************************** \n");
  }
}