package brain.cliqueCalculation;

import java.io.BufferedReader;
import java.io.File;
import java.io.FileNotFoundException;
import java.io.FileReader;
import java.io.FileWriter;
import java.io.IOException;
import java.util.HashMap;
import java.util.Iterator;
import java.util.TreeSet;
import java.util.Vector;

import brain.Module;

/**
 * This class shows the application of a network generating algorithms which
 * generate a graph out of a given distribution of cliques.
 * 
 * @author Gregor Kaczor
 * @author Mathias Linkerhand
 * @version 2011-01-10
 */

public class BidirectionalGraph extends Graph {
  /**
   * The network consists of a adjacency list realized by a HashMap.
   */
  private HashMap<Integer, TreeSet<Integer>> networkTreeSet    = null;

  private boolean                            debug             = false;

  private HashMap<Integer, Integer>          cliqueDistribution;

  private TreeSet<String>                    connections       = new TreeSet<String>();

  /**
   * Diameter gives the distance of the most distant vertices connected by the
   * shortest path possible. If this was not already computes diameter is -1.
   */
  public int                                 diameter          = -1;

  /**
   * Gives the number of edges in the graph.
   */
  public int                                 edgesCount        = -1;

  private int[]                              distribution      = null;

  // private double clusteringCoefficient = -1;

  // private double linkDensity = -1;

  //private long                               totalCliqueCount;

  private Vector<TreeSet<Integer>>           cliques;

  private TreeSet<Integer>                   visited;

  private boolean                            hasIdependentSets = false;

  /**
   * Standard constructor
   */
  public BidirectionalGraph() {
    super();
    networkTreeSet = new HashMap<Integer, TreeSet<Integer>>();
  }

  /**
   * This constructor allows to instantiate an already constructed graph in a
   * HashMap
   * 
   * @param localpart
   */
  public BidirectionalGraph(HashMap<Integer, TreeSet<Integer>> localpart) {
    super();
    networkTreeSet = localpart;
  }

  /**
   * This constructor allows to instantiate an already constructed graph from
   * internal links of an Module.
   * 
   * @param _internalLinks - internal links of an Module
   */
  public BidirectionalGraph(double[][] _internalLinks) {
    super();
    networkTreeSet = new HashMap<Integer, TreeSet<Integer>>();
    for (int i = 0; i < _internalLinks.length; i++) {
      for (int j = i + 1; j < _internalLinks.length; j++) {
        Integer i_obj = Integer.valueOf(i);
        Integer j_obj = Integer.valueOf(j);
        if (!getNodeExists(Integer.valueOf(i_obj))) {
          addNode(i_obj);
        }
        if (!getNodeExists(Integer.valueOf(j_obj))) {
          addNode(j_obj);
        }
        if (_internalLinks[i][j] > 0) {
          this.addDirectedLink(i_obj, j_obj);
          this.addDirectedLink(j_obj, i_obj);
        }
      }
    }
  }

  /**
   * This constructor allows to instantiate an already constructed graph from
   * internal links of an Module.
   * 
   * @param module
   */
  public BidirectionalGraph(final Module module) {
    super();
    networkTreeSet = new HashMap<Integer, TreeSet<Integer>>();
    for (int i = 0; i < module.getSize(); i++) {
      for (int j = i + 1; j < module.getSize(); j++) {
        Integer i_obj = Integer.valueOf(i);
        Integer j_obj = Integer.valueOf(j);
        if (!getNodeExists(Integer.valueOf(i_obj))) {
          addNode(i_obj);
        }
        if (!getNodeExists(Integer.valueOf(j_obj))) {
          addNode(j_obj);
        }
        if (module.getLink(i, j) > 0) {
          this.addDirectedLink(i_obj, j_obj);
          this.addDirectedLink(j_obj, i_obj);
        }
      }
    }
  }

  /**
   * Adds a node to the graph
   * 
   * @param node
   */
  @Override
  public void addNode(Integer node) {
    networkTreeSet.put(new Integer(node), new TreeSet<Integer>());
  }

  /**
   * Removes a node from the graph and all connections attached to it.
   * 
   * @param node
   */
  public void removeNode(Integer node) {
    TreeSet<Integer> connections = networkTreeSet.get(new Integer(node));
    Iterator<Integer> connectionsIterator = connections.iterator();
    Integer nodeB = null;
    while (connectionsIterator.hasNext()) {
      nodeB = connectionsIterator.next();
      removeConnection(node, nodeB);
      removeConnection(nodeB, node);
    }

    networkTreeSet.remove(node);
  }

  /**
   * Adds a aet of nodes to the graph. Attention! It does not check wheather
   * nodes are overwritten. It that case the connections are lost!
   * 
   * @param nodes a set of nodes
   */
  public void addNodes(TreeSet<Integer> nodes) {
    Iterator<Integer> nodesItrator = nodes.iterator();
    Integer lauf = null;
    while (nodesItrator.hasNext()) {
      lauf = nodesItrator.next();
      networkTreeSet.put(new Integer(lauf), new TreeSet<Integer>());
    }
  }

  /**
   * Adds a directed Link from nodeA to nodeB. For unidirectional links use
   * connect()
   * 
   * @param nodeA node to connect from
   * @param nodeB node to connect to
   */
  public void addDirectedLink(Integer nodeA, Integer nodeB) {
    networkTreeSet.get(nodeA).add(new Integer(nodeB));
    addConnection(nodeA, nodeB);
    // Integer[] con = new Integer[2];
    // con[0] = nodeA;
    // con[1] = nodeB;
    // connections.add(con);
  }

  /**
   * Add a connection between nodeA and nodeB to the connection list.
   * 
   * @param node - a node which links to nodeB
   * @param nodeB a node which links to nodeA
   */
  private void addConnection(Integer nodeA, Integer nodeB) {
    // if(!areConnected(nodeA,nodeB)) {
    // totalCoordinationnumber+=2;;
    // }
    connections.add(generateConnectionArray(nodeA, nodeB));

  }

  /**
   * Remove a connection between nodeA and nodeB from the connection list.
   * 
   * @param node - a node which does not any more link to nodeB
   * @param nodeB a node which does not any more link to nodeA
   */
  private void removeConnection(Integer nodeA, Integer nodeB) {
    // if(areConnected(nodeA,nodeB)) {
    // totalCoordinationnumber-=2;
    // }
    connections.remove(generateConnectionArray(nodeA, nodeB));

  }

  /**
   * generates a connection array to save it in a TreeSet of connections. to
   * avoid doublettes in a bidirectional graph it does check the two
   * combinations of nodeA and nodeB
   * 
   * @param nodeA
   * @param nodeB
   */
  private String generateConnectionArray(Integer nodeA, Integer nodeB) {
    String connection1 = nodeA + "," + nodeB;

    // connection1.add(0, nodeA);
    // connection1.add(1, nodeB);
    // connection1[0] = nodeA;
    // connection1[1] = nodeB;

    if (connections.contains(connection1)) {
      return connection1;
    } else {
      String connection2 = nodeB + "," + nodeA;
      // connection2.add(0, connection1.elementAt(1));
      // connection2.add(1, connection1.elementAt(0));
      // connection2[0] = connection1[1];
      // connection2[1] = connection1[0];
      return connection2;
    }

  }

  /**
   * Gets common neighbours between a pool of vertices and an additional
   * vertice.
   * 
   * @param Vertices - a set of vertices
   * @param can - a vertice
   * @return TreeSet of vertices which are common neighbours to vertices in
   * Vertices and can
   */
  public TreeSet<Integer> getCommonNeighbours(TreeSet<Integer> Vertices, Integer can) {

    TreeSet<Integer> neighbours_of_candidate = networkTreeSet.get(can);

    TreeSet<Integer> neighbours_of_vertices = new TreeSet<Integer>();
    Iterator<Integer> iterateVertices = Vertices.iterator();
    TreeSet<Integer> neighbours_in_common = new TreeSet<Integer>();

    while (iterateVertices.hasNext()) {
      neighbours_of_vertices.addAll(networkTreeSet.get(iterateVertices.next()));
    }

    Iterator<Integer> iterateNeighbours = neighbours_of_vertices.iterator();
    while (iterateNeighbours.hasNext()) {
      Integer tmp = iterateNeighbours.next();
      if (neighbours_of_candidate.contains(tmp)) {
        neighbours_in_common.add(tmp);

      }
    }
    return neighbours_in_common;
  }

  /**
   * Prints a clique summary to the standard output.
   * 
   */
  public void printCliquesSummary() {
    TreeSet<Integer> sortedCliqueDistributionKeySet = new TreeSet<Integer>(getCliqueDistribution().keySet());
    Iterator<Integer> sortedCliqueDistributionKeySetIterator = sortedCliqueDistributionKeySet.iterator();
    Integer cliqueSize = 0;
    Integer counter = 0;
    Integer totalCliques = 0;
    while (sortedCliqueDistributionKeySetIterator.hasNext()) {
      cliqueSize = sortedCliqueDistributionKeySetIterator.next();
      totalCliques += getCliqueDistribution().get(cliqueSize) == null ? 0 : getCliqueDistribution().get(cliqueSize);
      System.out.println("Cliques count of size " + cliqueSize + " = " + getCliqueDistribution().get(cliqueSize));
      counter++;
    }
    System.out.println("Cliques Sizes Count = " + getCliqueDistribution().size());
    System.out.println("Total Cliques Count = " + totalCliques);
  }

  /**
   * Counts all coordination numbers of the network. A coordination number of a
   * vertice is the count of its links to other vertices. The total coordination
   * number is the sum of the coordination numbers of every vertice
   * 
   * @return total coordination number
   */
  public Integer totalCoordinationNumer(TreeSet<Integer> partOfNetwork) {
    /***********************************************************************
     * Integer result = 0;
     * 
     * Iterator keySetIterator = partOfNetwork.iterator(); Integer i = null;
     * while (keySetIterator.hasNext()) { i = (Integer) keySetIterator.next();
     * result += networkTreeSet.get(i).size(); } return result;
     **********************************************************************/
    // Thats 50 % faster version
    int result = 0;

    Integer keys[] = new Integer[partOfNetwork.size()];
    partOfNetwork.toArray(keys);
    int size = keys.length;
    for (int i = 0; i < size; i++) {
      result += networkTreeSet.get(keys[i]).size();
    }
    return result;
  }

  /**
   * Connects two vertices in a graph in both directions from nodeA to nodeB and
   * from nodeB to nodeA (bidirectional)
   * 
   * @param nodeA - vertice to be connected to j
   * @param nodeB - vertice to be connected to i
   */

  @Override
  public void connect(Integer nodeA, Integer nodeB) {
    assert (nodeA.compareTo(nodeB) != 0) : "Same Vertices to connect each other";
    networkTreeSet.get(nodeA).add(new Integer(nodeB));
    networkTreeSet.get(nodeB).add(new Integer(nodeA));
    addConnection(nodeA, nodeB);
  }

  /**
   * Disconnects two vertices in a bidirectional graph in both directions from i
   * to j and from j to i (bidirectional)
   * 
   * @param nodeA - vertice to be disconnected from j
   * @param nodeB - vertice to be disconnected from i
   */
  @Override
  public void disconnect(Integer nodeA, Integer nodeB) {
    networkTreeSet.get(nodeA).remove(nodeB);
    networkTreeSet.get(nodeB).remove(nodeA);
    removeConnection(nodeA, nodeB);
  }

  /**
   * Disconnects a vertice from a bidirectional graph in both directions
   * 
   * @param node - vertice to be disconnected from graph
   */
  public void disconnect(Integer node) {
    TreeSet<Integer> ts = networkTreeSet.get(node);
    Iterator<Integer> i = ts.iterator();
    while (i.hasNext()) {
      disconnect(node, i.next());
    }
  }

  /**
   * Connects a node to a node set
   * 
   * @param clique a node set
   * @param candidate a vertice to be connected to the noe set
   */
  public void connectedToClique(Integer candidate, TreeSet<Integer> clique) {
    Iterator<Integer> iter = clique.iterator();
    Integer key = null;

    while (iter.hasNext()) {
      key = iter.next();
      connect(key, candidate);
    }
    return;

  }

  /**
   * Prints the network to the standard output
   */
  public void printNetwork() {
    TreeSet<Integer> sortedKeySet = new TreeSet<Integer>(networkTreeSet.keySet());
    Iterator<Integer> iter = sortedKeySet.iterator();
    Integer key = null;
    while (iter.hasNext()) {
      key = iter.next();
      System.out.print(key + " : ");
      Iterator<Integer> values = networkTreeSet.get(key).iterator();
      while (values.hasNext()) {
        System.out.print(values.next() + " ");
      }
      System.out.println();
    }
  }

  /**
   * merges two sites after the checking algorithm assured this operation does
   * not violate the clique distribution.
   * 
   * @param toMerge a candidate to merge
   * @param toBeMerged a candidate to merge
   */
  public void merge(Integer toMerge, Integer toBeMerged) {
    TreeSet<Integer> neighboursOfB = networkTreeSet.get(toMerge);
    Iterator<Integer> iterateNeighboursOfB = neighboursOfB.iterator();

    while (iterateNeighboursOfB.hasNext()) {
      Integer i = iterateNeighboursOfB.next();
      if (networkTreeSet.containsKey(i)) {
        networkTreeSet.get(toBeMerged).add(i);
        networkTreeSet.get(i).add(toBeMerged);
        networkTreeSet.get(i).remove(toMerge);
      }
    }
    disconnect(toMerge, toBeMerged);
    networkTreeSet.remove(toMerge);
    if (debug)
      printNetwork();
  }

  /**
   * Check if a and b are neighbours/directly connected a---b
   * 
   * @param a a candidate to merge
   * @param b a candidate to merge
   * @return true if a and b are neighbours/directly connected, false otherwise
   */
  public boolean areNeighbours(Integer a, Integer b) {
    return networkTreeSet.get(a).contains(b);
  }

  /**
   * Check if a and b have one common neighbour/indirectly connected over one
   * site a-x-b. One condition must be fulfilled for this test: Sites must not
   * neighbours
   * 
   * @param a a candidate to merge
   * @param b a candidate to merge
   * @return true if a and b have one common neighbour/indirectly connected over
   * one site, false otherwise
   */
  public boolean haveOneCommonNeighbour(Integer a, Integer b) {
    TreeSet<Integer> neighboursOfA = networkTreeSet.get(a);
    TreeSet<Integer> neighboursOfB = networkTreeSet.get(b);
    Iterator<Integer> iterateNeighboursA = neighboursOfA.iterator();

    while (iterateNeighboursA.hasNext()) {
      if (neighboursOfB.contains(iterateNeighboursA.next())) {
        return true;
      }
    }
    return false;
  }

  /**
   * Gets common neighbours between two vertices.
   * 
   * @param a - a set of vertices
   * @param b - a vertice
   * @return TreeSet of vertices which are common neighbours to vertices a and b
   */
  public TreeSet<Integer> getCommonNeighbours(Integer a, Integer b) {
    TreeSet<Integer> neighboursOfA = networkTreeSet.get(a);
    TreeSet<Integer> neighboursOfB = networkTreeSet.get(b);
    Iterator<Integer> iterateNeighboursA = neighboursOfA.iterator();
    TreeSet<Integer> neighbours_in_common = new TreeSet<Integer>();
    while (iterateNeighboursA.hasNext()) {
      Integer tmp = iterateNeighboursA.next();
      if (neighboursOfB.contains(tmp)) {
        neighbours_in_common.add(tmp);
      }
    }
    return neighbours_in_common;
  }

  /**
   * Checks if a and b have one common neighbour of neighbours/indirectly
   * connected over two sites a-x-x-b. Two conditions must be fulfilled for this
   * test. Sites are not neighbours and have not one common neighbour. a-x-x-b
   * 
   * @param a a candidate to merge
   * @param b a candidate to merge
   * @return true if a and b have one common neighbour of neighbours/indirectly
   * connected over two sites, false otherwise
   */
  @SuppressWarnings("unused")
  private boolean haveOneCommonNeighbourOfNeighbours(Integer a, Integer b) {
    TreeSet<Integer> neighboursOfA = networkTreeSet.get(a);
    TreeSet<Integer> neighboursOfB = networkTreeSet.get(b);

    Iterator<Integer> iterateNeighboursB = neighboursOfB.iterator();

    while (iterateNeighboursB.hasNext()) {
      Integer i = iterateNeighboursB.next();
      Iterator<Integer> iterateNeighboursA = neighboursOfA.iterator();
      while (iterateNeighboursA.hasNext()) {
        Integer j = iterateNeighboursA.next();
        if (areNeighbours(i, j)) {
          return true;
        }
      }
    }
    return false;
  }

  /**
   * Checks wheather a vertice is fully connecte to a set of vertices
   * 
   * @param clique A set of vertices which define a clique
   * @param candidate A vertice which is to check wheather it is fully connected
   * to the clique
   */
  public boolean isConnectedTo(TreeSet<Integer> clique, Integer candidate) {
    if (clique.contains(candidate)) {
      return true;
    }
    Iterator<Integer> iter = clique.iterator();
    Integer key = null;

    while (iter.hasNext()) {
      key = iter.next();
      if (!networkTreeSet.get(key).contains(candidate)) {
        return false;
      }
    }
    return true;

  }

  /**
   * Gets the coordination number. The coordination number is the numer of links
   * a vertice has to another vertices.
   * 
   * @param vertice - a vertice to gets its coordination number.
   * @return count of links to vertice.
   */
  public Integer getCoordinationNumber(Integer vertice) {
    return networkTreeSet.get(vertice).size();
  }

  /**
   * Get the network as a TreeSet
   * 
   * @return network
   */
  public HashMap<Integer, TreeSet<Integer>> getNetworkTreeSet() {
    return networkTreeSet;
  }

  /**
   * sets a network
   * 
   * @param _networkTreeSet
   */
  public void setNetworkTreeSet(HashMap<Integer, TreeSet<Integer>> _networkTreeSet) {
    networkTreeSet = _networkTreeSet;
  }

  /**
   * Search for the smallest Clique (which has the minimal number of nodes)
   * which has node as one of its nodes
   * 
   * @param node
   * @return smallest clique as a TreeSet
   */
  private TreeSet<Integer> smallestCliqueOfNode(Integer node, TreeSet<Integer> subgraph) {
    cliques.remove(subgraph);
    Iterator<TreeSet<Integer>> cliquesIterator = cliques.iterator();
    TreeSet<Integer> smallestClique = null;
    while (cliquesIterator.hasNext()) {
      TreeSet<Integer> currentClique = cliquesIterator.next();
      /*******************************************************************
       * if the currentClique contains the node then check wheather it is
       * smaller then the other before
       ******************************************************************/
      if (currentClique.contains(node)) {
        if (smallestClique == null) {
          smallestClique = currentClique;
        } else if (currentClique.size() < smallestClique.size()) {
          smallestClique = currentClique;
        }
      }
    }
    cliques.add(subgraph);
    return smallestClique;
  }

  /**
   * Search for the smallest Clique (which has the minimal number of nodes)
   * which has node as one of its nodes
   * 
   * @param node
   * @return smallest clique as a TreeSet
   */
  @SuppressWarnings("unused")
  private TreeSet<Integer> biggestCliqueOfSubgraph(TreeSet<Integer> subgraph) {
    cliques.remove(subgraph);
    Iterator<TreeSet<Integer>> cliquesIterator = cliques.iterator();
    TreeSet<Integer> biggestClique = null;
    while (cliquesIterator.hasNext()) {
      TreeSet<Integer> currentClique = cliquesIterator.next();
      /*******************************************************************
       * if the currentClique contains the node then check wheather it is
       * smaller then the other before
       ******************************************************************/
      if (currentClique.containsAll(subgraph)) {
        if (biggestClique == null) {
          biggestClique = currentClique;
        } else if (currentClique.size() > biggestClique.size()) {
          biggestClique = currentClique;
        }
      }
    }
    cliques.add(subgraph);
    return biggestClique;
  }

  /**
   * Computes candidates to choose from to connect to node. They must not
   * destroy existing cliques when connected to node.
   * 
   * @param node - node which has been chosen
   * @param currentClique - the clique which is in construction process
   * @param cliquesize - the finite size of the clique in construction process
   * @return candidates to choose from to connect to parameter node. Candidates
   * must not destroy existing cliques when connected to node.
   */
  @SuppressWarnings("unused")
  private TreeSet<Integer> computeCandidates(Integer node, TreeSet<Integer> currentClique, int cliquesize) {
    Iterator<TreeSet<Integer>> cliquesIterator = cliques.iterator();
    /***********************************************************************
     * W�hle die kleinste Clique an der sich der Knoten beteiligt
     **********************************************************************/
    TreeSet<Integer> smallestClique = smallestCliqueOfNode(node, currentClique);
    TreeSet<Integer> biggestClique = new TreeSet<Integer>();
    TreeSet<Integer> candidates = new TreeSet<Integer>(networkTreeSet.keySet());

    while (cliquesIterator.hasNext()) {
      TreeSet<Integer> nextClique = cliquesIterator.next();
      nextClique.containsAll(currentClique);
      if (cliquesize - 1 != smallestClique.size()) {
        if (computeOverlapp(nextClique, smallestClique) < smallestClique.size()) {
          // FIXME
          // candidates.addAll(nextClique);
          candidates.removeAll(nextClique);
          // candidates.addAll(smallestClique);
          // System.out.println(candidates);
        }
      } else
      /*******************************************************************
       * Finde die gr��te Clique, deren Subgraph smallestClique ist
       ******************************************************************/
      if (nextClique.containsAll(smallestClique) && biggestClique.size() <= nextClique.size()) {
        // candidates.removeAll(nextClique);
        biggestClique = nextClique;
      }
    }
    candidates.removeAll(biggestClique);
    candidates.remove(node);
    return candidates;
  }

  /**
   * Computes the overlap between two cliques. An overlap is defined by the
   * number of common nodes the cliques share.
   * 
   * @param cliqueOne
   * @param cliqueTwo
   */
  private int computeOverlapp(TreeSet<Integer> cliqueOne, TreeSet<Integer> cliqueTwo) {
    /***********************************************************************
     * It is faster to iterate a smaller clique and do binary search in a bigger
     * clique
     **********************************************************************/
    TreeSet<Integer> smaller = null;
    TreeSet<Integer> bigger = null;
    if (cliqueOne.size() < cliqueTwo.size()) {
      smaller = cliqueOne;
      bigger = cliqueTwo;
    } else {
      bigger = cliqueOne;
      smaller = cliqueTwo;
    }
    Iterator<Integer> smallerIterator = smaller.iterator();
    int counter = 0;
    while (smallerIterator.hasNext()) {
      Integer i = smallerIterator.next();
      if (bigger.contains(i)) {
        counter++;
      }
    }
    return counter;
  }

  /**
   * Creates a local part of a this graph. The local part is cut from two sites,
   * their neighbours and their common neighbours.
   * 
   * @param siteA
   * @param siteB
   */
  @SuppressWarnings("unchecked")
  public BidirectionalGraph createLocalNetworkPart(Integer siteA, Integer siteB) {
    BidirectionalGraph localpart = new BidirectionalGraph();

    /***********************************************************************
     * Copy sites to the local part of the network and connect them if they are
     * connected
     **********************************************************************/
    localpart.addNode(siteA);
    localpart.addNode(siteB);

    /***********************************************************************
     * Copy all nodes to the local parts which are neighbours of siteA and siteB
     * and their common neighbours
     **********************************************************************/
    localpart.addNodes((TreeSet<Integer>) networkTreeSet.get(siteA).clone());
    localpart.addNodes((TreeSet<Integer>) networkTreeSet.get(siteB).clone());
    localpart.addNodes((TreeSet<Integer>) getCommonNeighbours(siteA, siteB).clone());

    if (areConnected(siteA, siteB)) {
      localpart.connect(siteA, siteB);
    }

    /***********************************************************************
     * Now copy all the connections. To do so get the key set and check
     * connections
     **********************************************************************/
    Iterator<Integer> nodesIterator = localpart.getSetOfNodes().iterator();
    Iterator<Integer> nodesIterator2 = null;
    Integer nodeA = null;
    Integer nodeB = null;
    while (nodesIterator.hasNext()) {
      nodeA = (Integer) nodesIterator.next();
      nodesIterator2 = localpart.getSetOfNodes().iterator();
      while (nodesIterator2.hasNext()) {
        nodeB = (Integer) nodesIterator2.next();
        if (areConnected(nodeA, nodeB)) {
          localpart.connect(nodeA, nodeB);
        }
      }
    }
    return localpart;
  }

  /**
   * Checks wheather two vertices are connected.
   * 
   * @param verticeA
   * @param verticeB
   * @return true if verticeAand verticeB are connected, false otherwise.
   */
  public boolean areConnected(Integer verticeA, Integer verticeB) {
    return networkTreeSet.get(verticeA).contains(verticeB);
  }

  /**
   * Returns the HashMap which holds the clique distribution.
   * 
   * @return the cliqueDistribution
   */
  public HashMap<Integer, Integer> getCliqueDistribution() {
    return cliqueDistribution;
  }

  /**
   * Returns the clique distribution used to construct the graph
   * 
   * @return clique distribution
   */
  public int[] getDistribution() {
    return distribution;
  }

  /**
   * Get all nodes of this graph as a TreeSet
   */
  public TreeSet<Integer> getSetOfNodes() {
    return new TreeSet<Integer>(networkTreeSet.keySet());
  }

  /**
   * Get all connections of a vertice to neighbours as a TreeSet
   * 
   * @param vertice
   */
  public TreeSet<Integer> getLinksOfNode(Integer vertice) {
    return networkTreeSet.get(vertice);
  }

  /**
   * Check wheather that node exists in the graph
   * 
   * @param integer
   * @return true if node does exist, false otherwise
   */
  public boolean getNodeExists(Integer integer) {
    return networkTreeSet.containsKey(integer);
  }

  /**
   * Returns the number of nodes in this graph.
   */
  public int getNodesCount() {
    return networkTreeSet.size();
  }

  public TreeSet<String> getConnections() {
    return connections;
  }

  public void setConnections(TreeSet<String> connections) {
    this.connections = connections;
  }

  @Override
  public void save(String path) {
    File fNetConfig = new File(path);
    FileWriter fNetConfigW = null;

    try {
      fNetConfigW = new FileWriter(fNetConfig);
      fNetConfigW.write("#graph file\n");
      fNetConfigW.write("#node,node\n");
      Iterator<Integer> nodeA_Iterator = networkTreeSet.keySet().iterator();
      while (nodeA_Iterator.hasNext()) {
        Integer nodeA = nodeA_Iterator.next();
        TreeSet<Integer> ts = networkTreeSet.get(nodeA);
        Iterator<Integer> nodeB_Iterator = ts.iterator();
        while (nodeB_Iterator.hasNext()) {
          Integer nodeB = nodeB_Iterator.next();
          fNetConfigW.write(nodeA + "," + nodeB + "\n");
        }
      }
      fNetConfigW.close();
    } catch (IOException e) {
      e.printStackTrace();
    }

  }

  public BidirectionalGraph loadPajek(String path) {
    File fNetConfig = new File(path);
    FileReader fNetConfigR = null;
    BufferedReader bfNetConfigR = null;
    BidirectionalGraph bg = new BidirectionalGraph();
    try {
      // *Edges
      fNetConfigR = new FileReader(fNetConfig);
      bfNetConfigR = new BufferedReader(fNetConfigR);
      String line = "";
      boolean readnodes = false;
      while ((line = bfNetConfigR.readLine()) != null) {

        if (readnodes) {
          String[] nodes = line.replaceAll("  ", " ").trim().split(" ");
          Integer nodeA = Integer.parseInt(nodes[0]);
          Integer nodeB = Integer.parseInt(nodes[1]);
          if (!bg.getNodeExists(nodeA))
            bg.addNode(nodeA);
          if (!bg.getNodeExists(nodeB))
            bg.addNode(nodeB);
          if (nodeA.compareTo(nodeB) != 0 && !bg.areConnected(nodeA, nodeB)) {
            bg.connect(nodeA, nodeB);
          }
          // connect(nodeA, nodeB);
        }
        if (line.startsWith("*Edges")) {
          readnodes = true;
        }
      }
      bfNetConfigR.close();
      // System.out.println(bg.getNodesCount());
    } catch (IOException e) {
      e.printStackTrace();
    }

    return bg;
  }

  public void savePajek(String path) {
    TreeSet<Integer> nodescheck = new TreeSet<Integer>();
    File fNetConfig = new File(path);
    FileWriter fNetConfigW = null;
    HashMap<Integer, Integer> nodeMap = createNodeMap();
    try {
      fNetConfigW = new FileWriter(fNetConfig);
      fNetConfigW.write("*Vertices    " + (networkTreeSet.keySet().size()) + "\r\n");
      Iterator<Integer> nodeA_Iterator = new TreeSet<Integer>(networkTreeSet.keySet()).iterator();
      while (nodeA_Iterator.hasNext()) {
        Integer nodeA = (Integer) nodeA_Iterator.next();
        fNetConfigW.write("    " + (nodeMap.get(nodeA)) + " \"v" + (nodeMap.get(nodeA)) + "\"                                     0.0 0.0 0.0\r\n");
      }

      fNetConfigW.write("*Edges\r\n");
      nodeA_Iterator = networkTreeSet.keySet().iterator();

      while (nodeA_Iterator.hasNext()) {
        Integer nodeA = (Integer) nodeA_Iterator.next();
        TreeSet<Integer> ts = networkTreeSet.get(nodeA);
        Iterator<Integer> nodeB_Iterator = ts.iterator();
        while (nodeB_Iterator.hasNext()) {
          Integer nodeB = (Integer) nodeB_Iterator.next();
          fNetConfigW.write("     " + (nodeMap.get(nodeA)) + " " + (nodeMap.get(nodeB)) + " 1\r\n");
          nodescheck.add(nodeA);
          nodescheck.add(nodeB);
        }
      }
      fNetConfigW.close();
    } catch (IOException e) {
      e.printStackTrace();
    }
    TreeSet<Integer> all = new TreeSet<Integer>(networkTreeSet.keySet());
    all.removeAll(nodescheck);
    System.out.println(all);
  }

  private HashMap<Integer, Integer> createNodeMap() {
    Iterator<Integer> nodeA_Iterator = networkTreeSet.keySet().iterator();
    Integer nodeCounter = 1;
    HashMap<Integer, Integer> nodeMap = new HashMap<Integer, Integer>();
    while (nodeA_Iterator.hasNext()) {
      Integer nodeA = nodeA_Iterator.next();
      nodeMap.put(nodeA, nodeCounter++);
    }
    return nodeMap;
  }

  /***************************************************************************
   * loads a graph of the format node white space node
   * 
   * @param path
   * @param edgecountlimit
   * @param nodecountlimit
   * 
   */
  public void load(String path, int edgecountlimit, int nodecountlimit) {
    int isZeroNode = isZeroNode(path) ? 1 : 0;
    networkTreeSet = new HashMap<Integer, TreeSet<Integer>>();
    File f = new File(path);
    FileReader fr = null;
    try {
      fr = new FileReader(f);
    } catch (FileNotFoundException e) {
      // TODO Auto-generated catch block
      e.printStackTrace();
    }
    BufferedReader br = new BufferedReader(fr);
    String line = null;

    try {
      while ((line = br.readLine()) != null) {
        Integer nodeA = isZeroNode + Integer.valueOf(line.substring(0, line.indexOf(" ")).trim());
        Integer nodeB = isZeroNode + Integer.valueOf(line.substring(line.indexOf(" "), line.length()).trim());
        if (networkTreeSet.get(nodeA) == null) {
          addNode(nodeA);
        }
        if (networkTreeSet.get(nodeB) == null) {
          addNode(nodeB);
        }
        if (nodeA.compareTo(nodeB) != 0 && !areConnected(nodeA, nodeB)) {
          connect(nodeA, nodeB);
        }
        if (getConnections().size() > edgecountlimit || networkTreeSet.size() > nodecountlimit) {
          break;
        }

      }
    } catch (IOException e) {
      // TODO Auto-generated catch block
      e.printStackTrace();
    }
  }

  public boolean isZeroNode(String path) {
    File f = new File(path);
    FileReader fr = null;
    try {
      fr = new FileReader(f);
    } catch (FileNotFoundException e) {
      // TODO Auto-generated catch block
      e.printStackTrace();
    }
    BufferedReader br = new BufferedReader(fr);
    String line = null;
    try {
      while ((line = br.readLine()) != null) {
        Integer nodeA = Integer.valueOf(line.substring(0, line.indexOf(" ")).trim());
        Integer nodeB = Integer.valueOf(line.substring(line.indexOf(" "), line.length()).trim());
        if (nodeA == 0 || nodeB == 0) {
          return true;
        }
      }
    } catch (IOException e) {
      // TODO Auto-generated catch block
      e.printStackTrace();
    }
    return false;
  }

  public Integer getNodeWithLowestDegree() {
    Iterator<Integer> nodeiterator = this.networkTreeSet.keySet().iterator();
    //BidirectionalGraph result = new BidirectionalGraph();
    Integer node = null;
    int currentdegree = Integer.MAX_VALUE;
    Integer currentnode = -1;
    while (nodeiterator.hasNext()) {
      node = (Integer) nodeiterator.next();
      if (networkTreeSet.get(node).size() < currentdegree) {
        currentdegree = networkTreeSet.get(node).size();
        currentnode = node;
      }
    }
    return currentnode;
  }

  public void loadBipartite(String path, int edgecountlimit, int nodecountlimit, boolean integrate) {
    networkTreeSet = new HashMap<Integer, TreeSet<Integer>>();
    File f = new File(path);
    FileReader fr = null;
    try {
      fr = new FileReader(f);
    } catch (FileNotFoundException e) {
      // TODO Auto-generated catch block
      e.printStackTrace();
    }
    BufferedReader br = new BufferedReader(fr);
    System.out.print("loading file:");
    String line = null;
    Integer nodeA = 0;
    Integer counter = 0;
    int totallines = 0;
    int tmpcounter = 0;
    try {
      if (!integrate) {
        while ((line = br.readLine()) != null) {
          nodeA++;
          line = line.trim();
          // String s[] = line.split(" ");
          addNode(nodeA);
        }
        Integer startFrom = nodeA;
        totallines = nodeA;
        br.close();
        try {
          fr = new FileReader(f);
        } catch (FileNotFoundException e) {
          // TODO Auto-generated catch block
          e.printStackTrace();
        }
        br = new BufferedReader(fr);

        while ((line = br.readLine()) != null) {
          if (counter == tmpcounter && totallines > 100) {
            System.out.print(counter * 100 / (totallines - (totallines % 10)) + "% ");
            tmpcounter += (totallines - (totallines % 10)) / 10;
          }
          counter++;
          line = line.trim();
          String s[] = line.split(" ");

          // addNode(nodeA);
          // System.out.println("line:"+line);
          for (int i = 0; i < s.length; i++) {
            Integer nodeB = Integer.valueOf(s[i]);
            if (networkTreeSet.get(startFrom + nodeB) == null) {
              addNode(startFrom + nodeB);
            }
            if (counter.compareTo(startFrom + nodeB) != 0 && !areConnected(counter, startFrom + nodeB)) {
              connect(counter, startFrom + nodeB);
              // System.out.println(startFrom+nodeB+", "+counter);
            }

            // for (int j = 0; i + j < s.length; j++) {
            // Integer nodeA = Integer.valueOf(s[j]);
            // if (networkTreeSet.get(nodeA) == null) {
            // addNode(nodeA);
            // }
            //
            // if (nodeA.compareTo(nodeB) != 0
            // && !areConnected(nodeA, nodeB)) {
            // connect(nodeA, nodeB);
            //
            // }
            //
            // }
            if (getConnections().size() > edgecountlimit || networkTreeSet.size() > nodecountlimit) {
              break;
            }
          }

          if (getConnections().size() > edgecountlimit || networkTreeSet.size() > nodecountlimit) {
            break;
          }
          // Integer nodeA = Integer.valueOf(line.substring(0, line
          // .indexOf(" ")).trim());
          // Integer nodeB = Integer.valueOf(line.substring(line
          // .indexOf(" "), line.length()).trim());
        }

      } else {

        while ((line = br.readLine()) != null) {
          totallines++;
        }
        //Integer startFrom = nodeA;
        br.close();
        try {
          fr = new FileReader(f);
        } catch (FileNotFoundException e) {
          // TODO Auto-generated catch block
          e.printStackTrace();
        }
        br = new BufferedReader(fr);

        try {
          fr = new FileReader(f);
        } catch (FileNotFoundException e) {
          // TODO Auto-generated catch block
          e.printStackTrace();
        }
        br = new BufferedReader(fr);

        while ((line = br.readLine()) != null) {

          if (counter == tmpcounter && totallines > 100) {
            System.out.print(counter * 100 / (totallines - (totallines % 10)) + "% ");
            tmpcounter += (totallines - (totallines % 10)) / 10;
          }
          counter++;
          line = line.trim();
          String s[] = line.split(" ");
          // addNode(nodeA);
          // System.out.println("line:"+line);
          for (int j = 0; j < s.length; j++) {
            nodeA = Integer.valueOf(s[j]);
            if (networkTreeSet.get(nodeA) == null) {
              addNode(nodeA);
            }
            for (int i = j + 1; i < s.length; i++) {
              Integer nodeB = Integer.valueOf(s[i]);

              if (networkTreeSet.get(nodeB) == null) {
                addNode(nodeB);
              }
              if (nodeA.compareTo(nodeB) != 0 && !areConnected(nodeA, nodeB)) {
                connect(nodeA, nodeB);
                // System.out.println(startFrom+nodeB+",
                // "+counter);
              }
            }
          }
          if (getConnections().size() > edgecountlimit || networkTreeSet.size() > nodecountlimit) {
            break;
          }
        }
      }
      System.out.println();
    } catch (IOException e) {
      // TODO Auto-generated catch block
      e.printStackTrace();
    }
  }

  public boolean existsNode(int j) {
    if (this.networkTreeSet.get(j) != null)
      return true;
    return false;
  }

  public BidirectionalGraph copy(TreeSet<Integer> name) {
    BidirectionalGraph bg = new BidirectionalGraph();
    // copy nodes first
    bg.addNodes(name);
    /***********************************************************************
     * Now copy all the connections. To do so get the key set and check
     * connections
     **********************************************************************/
    Iterator<Integer> nodesIterator = bg.getSetOfNodes().iterator();
    Iterator<Integer> nodesIterator2 = null;
    Integer nodeA = null;
    Integer nodeB = null;
    while (nodesIterator.hasNext()) {
      nodeA = nodesIterator.next();
      nodesIterator2 = bg.getSetOfNodes().iterator();
      while (nodesIterator2.hasNext()) {
        nodeB = nodesIterator2.next();
        if (areConnected(nodeA, nodeB)) {
          bg.connect(nodeA, nodeB);
        }
      }
    }

    // Iterator iter=name.iterator();
    // while(iter.hasNext()) {
    // bg.getC
    // }
    // copy connections between existing nodes

    return bg;
  }

  @SuppressWarnings("unchecked")
  public Vector<TreeSet<Integer>> detectIndependentSets() {
    visited = new TreeSet<Integer>();
    TreeSet<Integer> toVisit = new TreeSet<Integer>(getNetworkTreeSet().keySet());
    Vector<TreeSet<Integer>> isets = new Vector<TreeSet<Integer>>();
    int visitedNodesCount = 0;
    do {
      TreeSet<Integer> clone = (TreeSet<Integer>) detectIndependentSets(toVisit.first()).clone();
      visitedNodesCount += clone.size();
      isets.add(clone);
      toVisit.removeAll(clone);
      System.out.println("IDS:" + clone.size());
    } while (visitedNodesCount < getNetworkTreeSet().keySet().size());
    if (isets.size() == 1) {
      isets.clear();
      this.hasIdependentSets = false;
    } else {
      this.hasIdependentSets = true;
    }
    return isets;
  }

  private TreeSet<Integer> detectIndependentSets(Integer firstNode) {
    visited = new TreeSet<Integer>();
    traverse(firstNode);
    return visited;
  }

  private void traverse(Integer node) {
    // System.out.println("visiting:" + node);
    visited.add(node);
    Iterator<Integer> iterator = this.networkTreeSet.get(node).iterator();
    while (iterator.hasNext()) {
      Integer element = (Integer) iterator.next();
      if (!visited.contains(element))
        traverse(element);
    }
  }

  public boolean isHasIdependentSets() {
    return hasIdependentSets;
  }

}
