package brain.gui;

import java.awt.Component;
import java.awt.Dimension;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.lang.reflect.AccessibleObject;
import java.lang.reflect.Constructor;
import java.lang.reflect.Method;
import java.lang.reflect.Type;
import java.util.Set;
import java.util.logging.Level;
import java.util.logging.Logger;

import javax.swing.BoxLayout;
import javax.swing.DefaultListCellRenderer;
import javax.swing.JComboBox;
import javax.swing.JLabel;
import javax.swing.JList;
import javax.swing.JPanel;
import javax.swing.JTextField;
import javax.swing.SpringLayout;
import javax.swing.event.DocumentEvent;
import javax.swing.event.DocumentListener;

import brain.gui.table.ObjectTableModel;

/**
 * @author Mathias Linkerhand
 * @version 2011-04-05
 */
public class AccessibleObjectPreparationPanel extends JPanel implements ActionListener {
  private static final long   serialVersionUID = 1L;
  private JComboBox           comboBox;
  private Object[]            parameters;
  private JPanel              parametersPanel;

  private static final Logger logger           = Logger.getLogger(AccessibleObjectPreparationPanel.class.getName());

  public <T> AccessibleObjectPreparationPanel(final Set<T> accessibleObjects) {
    this.setSize(new Dimension(300, 300));
    this.setLayout(new BoxLayout(this, BoxLayout.PAGE_AXIS));
    final JPanel parametersPanel = new JPanel(new SpringLayout());
    this.parametersPanel = parametersPanel;

    final JComboBox comboBox = new JComboBox(accessibleObjects.toArray());
    this.comboBox = comboBox;
    comboBox.addActionListener(this);
    // Make sure the panel looks correct after initialization with all parameter text fields.
    comboBox.setSelectedIndex(0);
    comboBox.setRenderer(new DefaultListCellRenderer() {
      private static final long serialVersionUID = 1L;
      private JLabel            label            = new JLabel();

      @Override
      public Component getListCellRendererComponent(JList list, Object value, int index, boolean isSelected, boolean cellHasFocus) {
        label = (JLabel) super.getListCellRendererComponent(list, value, index, isSelected, cellHasFocus);
        if (value instanceof AccessibleObject) {
          final AccessibleObject accessibleObject = (AccessibleObject) value;
          if (accessibleObject instanceof Constructor<?>) {
            final Constructor<?> constructor = (Constructor<?>) accessibleObject;
            String text = constructor.getName() + "(";
            boolean isFirst = true;
            for (final Type type : constructor.getGenericParameterTypes()) {
              if (isFirst) {
                isFirst = false;
              } else {
                text += ", ";
              }
              if (type instanceof Class<?>) {
                text += ((Class<?>) type).getSimpleName();
              } else {
                text += type.toString();
              }
            }
            text += ")";
            label.setText(text);
          } else if (accessibleObject instanceof Method) {
            final Method method = (Method) accessibleObject;

            String text = method.getName() + "(";
            boolean isFirst = true;
            for (final Type type : method.getGenericParameterTypes()) {
              if (isFirst) {
                isFirst = false;
              } else {
                text += ", ";
              }
              if (type instanceof Class<?>) {
                text += ((Class<?>) type).getSimpleName();
              } else {
                text += type.toString();
              }
            }
            text += ")";
            label.setText(text);
          } else {
            label.setText(value.toString());
          }
        } else {
          label.setText(value.toString());
        }

        return label;
      }
    });

    this.add(comboBox);
    this.add(parametersPanel);
  }

  public AccessibleObject getAccessibleObject() {
    return (AccessibleObject) this.comboBox.getSelectedItem();
  }

  public Object[] getParameters() {
    return this.parameters;
  }

  private void updateParametersPanel() {
    parametersPanel.removeAll();
    final AccessibleObject accessibleObject = (AccessibleObject) comboBox.getSelectedItem();

    final Type[] types;
    if (accessibleObject instanceof Constructor<?>) {
      final Constructor<?> constructor = (Constructor<?>) accessibleObject;
      types = constructor.getGenericParameterTypes();
    } else if (accessibleObject instanceof Method) {
      final Method method = (Method) accessibleObject;
      types = method.getGenericParameterTypes();
    } else {
      types = new Type[0];
    }
    this.parameters = new Object[types.length];

    int i = -1;
    for (final Type type : types) {
      i++;
      final JTextField textField = new JTextField();
      final int j = i;

      textField.getDocument().addDocumentListener(new DocumentListener() {
        @Override
        public void removeUpdate(DocumentEvent e) {
          this.update();
        }

        @Override
        public void insertUpdate(DocumentEvent e) {
          this.update();
        }

        @Override
        public void changedUpdate(DocumentEvent e) {
        }

        private void update() {
          try {
            parameters[j] = ObjectTableModel.convertStringToType(textField.getText(), type);
            logger.log(Level.FINEST, "value changed to " + parameters[j]);
          } catch (Exception exception) {
          }
        }
      });

      final JLabel label = new JLabel(type instanceof Class<?> ? ((Class<?>) type).getSimpleName() : type.toString());
      label.setAlignmentX(Component.RIGHT_ALIGNMENT);
      label.setLabelFor(textField);

      parametersPanel.add(label);
      parametersPanel.add(textField);
    }

    // layout the components nicely
    SpringUtilities.makeCompactGrid(parametersPanel, types.length, 2, 0, 0, 5, 5);

    this.updateUI();
  }

  @Override
  public void actionPerformed(ActionEvent event) {
    this.updateParametersPanel();
  }
}