package brain.gui;

import java.awt.Color;
import java.awt.Graphics;
import java.awt.Graphics2D;
import java.awt.image.BufferedImage;
import java.awt.print.PageFormat;
import java.awt.print.Printable;
import java.awt.print.PrinterException;
import java.io.FileOutputStream;
import java.io.IOException;
import java.text.DecimalFormat;
import java.util.HashMap;
import java.util.Iterator;
import java.util.TreeSet;

import javax.print.Doc;
import javax.print.DocFlavor;
import javax.print.DocPrintJob;
import javax.print.PrintException;
import javax.print.SimpleDoc;
import javax.print.StreamPrintService;
import javax.print.StreamPrintServiceFactory;
import javax.print.attribute.HashPrintRequestAttributeSet;
import javax.print.attribute.PrintRequestAttributeSet;

/**
 * Represents a pattern i form of a key value formalism. A key is a site, node,
 * neuron or a pixel. Its value is a representation of weight/impact of a
 * stimulus
 * 
 * @version 2012-09-11
 * @since 2008
 */
public class PrintablePattern implements Printable {
  /**
   * In case the pattern should be represented as a matrix, its measures can be
   * defined by setting its width and height
   */
  protected int                    width          = 0;
  /**
   * In case the pattern should be represented as a matrix, its measures can be
   * defined by setting its width and height
   */
  protected int                    height         = 0;
  /** The actual pattern data */
  private HashMap<Integer, Double> pattern        = new HashMap<Integer, Double>();
  /** The graphical dimensions of a key value pair */
  protected int                    gwidth         = 40;
  /** The graphical dimensions of a key value pair */
  protected int                    gheight        = 40;
  /** file name of the pattern if exported to post script */
  protected String                 outputFileName = null;
  /** export pattern as a lattice */
  protected boolean                printAsLattice = false;

  /**
   * Constructor for instantiation of a pattern object
   * 
   * @param patt
   *          String array, of which its index represents the key and the data
   *          found at the index the value of the key - value pair representing a
   *          pattern.
   */

  public PrintablePattern(String patt[]) {
    height = width = patt.length;
    int patternlength = Integer.parseInt(patt[0]);
    if (patternlength < patt.length) {
      height = width = (int) Math.sqrt(patternlength);
      for (int i = 1; i <= patternlength; i++) {
        pattern.put(Integer.valueOf(Integer.parseInt(patt[i])), Double.valueOf(Double.parseDouble(patt[i + patternlength])));
      }
    } else {
      height = width = patt.length;
      for (int i = 0; i < patt.length; i++) {
        pattern.put(Integer.valueOf(Integer.parseInt(patt[i])), Double.valueOf(1d));
      }
    }

  }

  /**
   * Constructor for instantiation of a pattern object. The data representing the
   * pattern is taken from an image file.
   * 
   * @param imageFile -
   *          the image the data is taken from
   * @param treshold -
   *          simple treshold for data values
   */
  public PrintablePattern(BufferedImage imageFile, double treshold) {
    int counter = 0;

    for (int y = 0; y < imageFile.getHeight(); y++) {
      double color = 0;
      for (int x = 0; x < imageFile.getWidth(); x++) {
        color = (1 - (imageFile.getRaster().getSampleFloat(x, y, 0) / 255f));
        color = color > treshold ? color : 0d;
        pattern.put(Integer.valueOf(counter), Double.valueOf(color));
        counter++;
      }
    }

  }

  /**
   * Standard constructor
   */
  public PrintablePattern() {

  }

  /**
   * @see java.lang.Object#toString()
   */
  @Override
  public String toString() {
    StringBuffer sb = new StringBuffer(pattern.size() == 0 ? 1000 : pattern.size() * 3);
    DecimalFormat form = new DecimalFormat("#.##");
    TreeSet<Integer> keys = new TreeSet<Integer>();
    keys.addAll(pattern.keySet());
    sb = sb.append(pattern.size());
    // output used sites first
    for (Integer iterable : keys) {
      sb = sb.append(" ");
      sb = sb.append(iterable);
    }
    // now output all the values
    for (Iterator<Integer> iter = keys.iterator(); iter.hasNext();) {
      Integer element = iter.next();
      sb = sb.append(" ");
      sb = sb.append(form.format(pattern.get(element)));
    }
    return sb.toString();

  }

  /**
   * Sets a value for a key
   * 
   * @param key -
   *          identification for a site
   * @param value -
   *          sites value
   */
  public void set(Integer key, Double value) {
    pattern.put(key, value);
  }

  /**
   * Converts Integer in Double. In case there is no field for key, the method
   * returns 0
   * 
   * @param key -
   *          identification for a site
   * @return converted site identificator
   */
  public Double get(Integer key) {
    Double d;
    return (d = pattern.get(key)) == null ? Double.valueOf(0d) : d;
  }

  /**
   * 
   * 
   * @see java.awt.print.Printable#print(java.awt.Graphics,
   *      java.awt.print.PageFormat, int)
   */
  @Override
  public int print(Graphics graphics, PageFormat pageFormat, int pageIndex) throws PrinterException {
    if (pageIndex < 1) {
      Graphics2D g2d = (Graphics2D) graphics;
      g2d.translate(pageFormat.getImageableX(), pageFormat.getImageableY());

      g2d.setColor(Color.black);
      if (printAsLattice) {
        TreeSet<Integer> keys = new TreeSet<Integer>();
        keys.addAll(pattern.keySet());
        // now output all the values
        int f = 0;
        int counter = 0;
        int counter2 = 0;
        int average_over_pixel = keys.size() / 400;
        int widthtreshold = 400;
        for (Iterator<Integer> iter = keys.iterator(); iter.hasNext();) {
          counter++;
          Integer element = iter.next();

          if (pattern.size() < widthtreshold) {
            f = (int) (255d - ((255d) * pattern.get(element).doubleValue()));
            g2d.setColor(new Color(f, f, f, 255));
            g2d.fillRect(element.intValue() * gwidth, gheight, gwidth, gheight);
          } else {
            f += (int) (255d - ((255d) * pattern.get(element).doubleValue()));
            if (counter % average_over_pixel == 0) {
              f /= average_over_pixel;
              g2d.setColor(new Color(f, f, f, 255));
              g2d.fillRect(counter2 * gwidth, gheight, gwidth, gheight);
              f = 0;
              counter2++;
            }
          }

        }
        g2d.setColor(Color.black);
        if (pattern.size() < widthtreshold) {

          g2d.drawRect(gwidth, gheight, keys.size() * gwidth, gheight);
        } else {
          g2d.drawRect(gwidth, gheight, counter2 * gwidth, gheight);
        }
      } else {

        for (int y = 0; y < height; y++) {
          for (int x = 0; x < width; x++) {
            // System.out.println(get(x, y));
            int f = (int) (255d - ((255d) * get(x, y).doubleValue()));
            // System.out.print(f +", "+get(x, y)+ "\t");

            // if (get(x, y) == 0) {
            g2d.setColor(new Color(f, f, f, 255));
            g2d.fillRect(x * gwidth, y * gheight, gwidth, gheight);
            g2d.setColor(new Color(0, 0, 0, 255));
            g2d.drawRect(x * gwidth, y * gheight, gwidth, gheight);
            // } else {

            // }
            //					
            // if (get(x, y) == 0) {
            // g2d.setColor(Color.white);
            // } else {
            // g2d.setColor(Color.blue);
            // }
            // g2d.fillRect(x * gwidth, y * gheight, gwidth,
            // gheight);
          }
        }
        // g2d.setColor(Color.blue);
        // g2d.drawString("Hello the world : "+pageIndex, 72, 72);
        // g2d.setColor(Color.blue);
        // g2d.fillRect(0, 0, 72, 72);
      }

      return Printable.PAGE_EXISTS;
    }
    return Printable.NO_SUCH_PAGE;

  }

  /**
   * Recalculates data position from a matrix format to a lattice format
   * 
   * @return data position in lattice data format
   */
  protected Double get(int x, int y) {
    return get(Integer.valueOf(y * this.height + x));
  }

  /**
   * Exports the pattern to a postscript file.
   */
  public void print() {
    DocFlavor flavor = DocFlavor.SERVICE_FORMATTED.PRINTABLE;
    String psMimeType = DocFlavor.BYTE_ARRAY.POSTSCRIPT.getMimeType();
    StreamPrintServiceFactory[] factories = StreamPrintServiceFactory.lookupStreamPrintServiceFactories(flavor, psMimeType);

    // System.out.println("Combien de services disponible : "
    // + factories.length);

    if (factories.length == 0) {
      System.err.println("No suitable factories for printing !");
      // System.exit(0);
      return;
    }

    try {
      FileOutputStream fos = new FileOutputStream(outputFileName == null ? "pattern.ps" : outputFileName);

      // System.out
      // .println("Utilisation du premier service : " + factories[0]
      // + " // " + factories[0].getClass().getName());
      StreamPrintService sps = factories[0].getPrintService(fos);

      DocPrintJob pj = sps.createPrintJob();
      PrintRequestAttributeSet aset = new HashPrintRequestAttributeSet();

      // définie le nombre de copie du Job
      // modifie la commande /NumCopies dans le postscript de sortie
      aset.add(new javax.print.attribute.standard.Copies(1));

      // devrait permettre de choisir le paper tray mais ne fait rien :(
      // !!! THIS CODE DOESN'T CHANGE ANYTHING INTO THE POSTSCRIPT OUTPUT
      // :(
      // aset.add(javax.print.attribute.standard.MediaTray.MANUAL);

      // Définie la taille du papier
      // modifie la commande /PageSize dans le code postscript
      aset.add(javax.print.attribute.standard.MediaSizeName.ISO_A4);

      Doc doc = new SimpleDoc(this, flavor, null);

      pj.print(doc, aset);
      fos.close();
    } catch (PrintException pe) {
      System.err.println("PrintException : " + pe);
    } catch (IOException ioe) {
      System.err.println("IOexception : " + ioe);
    }
  }

  /**
   * Returns file name of the exported file
   * 
   * @return export file name
   */
  public String getOutputFileName() {
    return outputFileName;
  }

  /**
   * Sets file name of the exported file
   * 
   * @param _outputFileName
   *          the output to set
   */
  public void setOutputFileName(String _outputFileName) {
    this.outputFileName = _outputFileName;
  }

  /**
   * In case PrintablePattern is represented as a matrix the method returns the matrix
   * height
   * 
   * @return height of the matrix which represents this pattern
   */
  public int getHeight() {
    return height;
  }

  /**
   * In case PrintablePattern is represented as a matrix the method sets the matrix height
   * 
   * @param _height
   *          of the matrix which represents this pattern
   */
  public void setHeight(int _height) {
    this.height = _height;
  }

  /**
   * In case PrintablePattern is represented as a matrix the method returns the matrix
   * width
   * 
   * @return width of the matrix which represents this pattern
   */
  public int getWidth() {
    return width;
  }

  /**
   * In case PrintablePattern is represented as a matrix the method sets the matrix width
   * 
   * @param _width
   *          of the matrix which represents this pattern
   */
  public void setWidth(int _width) {
    this.width = _width;
  }

  /**
   * Returns the graphical height for one entity (key - value pair) of PrintablePattern
   * 
   * @return graphical height for one entity (key - value pair) of PrintablePattern
   */
  public int getGheight() {
    return gheight;
  }

  /**
   * Sets the graphical height for one entity (key - value pair) of PrintablePattern
   * 
   * @param gheight1
   *          the gheight to set
   */
  public void setGheight(int gheight1) {
    this.gheight = gheight1;
  }

  /**
   * Returns the graphical width for one entity (key - value pair) of PrintablePattern
   * 
   * @return graphical width for one entity (key - value pair) of PrintablePattern
   */
  public int getGwidth() {
    return gwidth;
  }

  /**
   * Sets the graphical width for one entity (key - value pair) of PrintablePattern
   * 
   * @param _gwidth -
   * 
   * 
   * the gwidth to set
   */
  public void setGwidth(int _gwidth) {
    this.gwidth = _gwidth;
  }

  /**
   * Exports the pattern to a postscript file as a Lattice
   * 
   * @param patternBarValue
   */
  public void saveAsLattice(String patternBarValue) {
    outputFileName = patternBarValue;
    printAsLattice = true;
    print();
    printAsLattice = false;
  }

  /**
   * Returns whether the pattern should be exported as a lattice or a matrix.
   * 
   * @return true if the pattern should be exported as a lattice, false if the
   *         pattern should be exported as a matrix
   */
  public boolean getPrintAsLattice() {
    return printAsLattice;
  }

  /**
   * Sets whether the pattern should be exported as a lattice or a matrix.
   * 
   * @param _printAsLattice
   *          if true the pattern will be exported as a lattice, if false the
   *          pattern will be exported as a matrix
   * 
   */
  public void setPrintAsLattice(boolean _printAsLattice) {
    this.printAsLattice = _printAsLattice;
  }

}
