package brain.gui;

import java.awt.Color;
import java.awt.Dimension;
import java.awt.Graphics;
import java.awt.Graphics2D;
import java.awt.Point;
import java.awt.print.PageFormat;
import java.awt.print.Printable;
import java.awt.print.PrinterException;
import java.util.ArrayList;
import java.util.Collections;

/**
 * Represents a receptive field of a clique.
 * 
 * @author Gregor Kaczor
 * @version September 2008
 */
public class ReceptiveField extends PrintablePattern implements Printable {

  /** The field itself */
  private ArrayList<Double> field          = null;

  /** x,y position on the screen */
  private Point             globalPosition = null;
  /** Minimum value of field */
  protected double          min            = 0;
  /** Maximum value of field */
  protected double          max            = 0;

  /** Standard constructor */
  public ReceptiveField() {
    super();
    field = new ArrayList<Double>();
    globalPosition = new Point(0, 0);
  }

  /** initiate a receptive field from a double array **/
  public ReceptiveField(double d[]) {
    super();
    field = new ArrayList<Double>();
    for (double value : d) {
      field.add(value);
    }
    max = Collections.max(field);
    min = Collections.min(field);
    globalPosition = new Point(0, 0);
  }

  /**
   * Constructs a receptive field from given data
   * @param _field - data values 
   * @param position - position in a receptive field collection
   * @param dimension - width and height of the receptive field
   */
  public ReceptiveField(ArrayList<Double> _field, Point position, Dimension dimension) {
    field = _field;
    globalPosition = position;
    max = Collections.max(field);
    min = Collections.min(field);
    this.height = dimension.height;
    this.width = dimension.width;
  }

  /**
   * Returns maximum value of receptive field
   * 
   * @return the max
   */
  public double getMax() {
    return max;
  }

  /**
   * Sets maximum value of receptive field
   * 
   * @param max1
   *          the max to set
   */
  public void setMax(double max1) {
    this.max = max1;
  }

  /**
   * Returns minimum value of receptive field
   * 
   * @return the min
   */
  public double getMin() {
    return min;
  }

  /**
   * Sets minimum value of receptive field
   * 
   * @param min1
   *          the min to set
   */
  public void setMin(double min1) {
    this.min = min1;
  }

  /**
   * 
   * @see brain.gui.PrintablePattern#print(java.awt.Graphics,
   *      java.awt.print.PageFormat, int)
   */
  @Override
  public int print(final Graphics graphics, final PageFormat pageFormat, final int pageIndex) throws PrinterException {
    if (pageIndex < 1) {
      Graphics2D g2d = (Graphics2D) graphics;
      g2d.translate(pageFormat.getImageableX(), pageFormat.getImageableY());

      g2d.setColor(Color.black);

      for (int y = 0; y < height; y++) {
        for (int x = 0; x < width; x++) {
          // if (get(x, y) == 0) {
          // g2d.drawRect(x * gwidth, y * gheight, gwidth, gheight);
          // } else {
          // float c = field.get(y * this.height + x).floatValue();
          int f = (int) ((255d) * ((field.get(y * this.height + x).doubleValue() - min) / (max - min)));
          // System.out.print(f + "\t");
          g2d.setColor(new Color(f, f, f, 255));
          // System.out.println("b:"+java.awt.Color.gray.getBlue());
          // System.out.println("r:"+java.awt.Color.gray.getRed());
          // System.out.println("g:"+java.awt.Color.gray.getRGB());

          g2d.fillRect(globalPosition.x * gwidth * width + x * gwidth, globalPosition.y * gheight * height + y * gheight, gwidth, gheight);
          // }
          // if (get(x, y) == 0) {
          // g2d.setColor(Color.white);
          // } else {
          // g2d.setColor(Color.blue);
          // }
          // g2d.fillRect(x * gwidth, y * gheight, gwidth, gheight);
        }
      }
      // g2d.setColor(Color.blue);
      // g2d.drawString("Hello the world : "+pageIndex, 72, 72);
      // g2d.setColor(Color.blue);
      // g2d.fillRect(0, 0, 72, 72);
      return Printable.PAGE_EXISTS;
    }
    return Printable.NO_SUCH_PAGE;

  }

  /**
   * 
   * @see brain.gui.PrintablePattern#get(java.lang.Integer)
   */
  @Override
  public Double get(Integer key) {
    Double d;
    return (d = field.get(key.intValue())) == null ? Double.valueOf(0d) : d;
  }

  /**
   * 
   * @see brain.gui.PrintablePattern#getOutputFileName()
   */
  @Override
  public String getOutputFileName() {
    return outputFileName;
  }

  /**
   * 
   * @see brain.gui.PrintablePattern#setOutputFileName(java.lang.String)
   */
  @Override
  public void setOutputFileName(String _outputFileName) {
    this.outputFileName = _outputFileName;
  }

  /**
   * Returns the receptive field.
   * 
   * @return the field
   */
  public ArrayList<Double> getField() {
    return field;
  }

  /**
   * Sets the receptive field.
   * 
   * @param field1
   *          the field to set
   */
  public void setField(ArrayList<Double> field1) {
    this.field = field1;
  }

  /**
   * Returns this receptive fields coordinate within a graphical representation.
   * 
   * @return the globalPositionInPsFile
   */
  public Point getGlobalPositionInPsFile() {
    return globalPosition;
  }

  /**
   * Sets Returns a coordinate within a graphical representation.
   * 
   * @param globalPositionInPsFile1
   *          the globalPositionInPsFile to set
   */
  public void setGlobalPositionInPsFile(Point globalPositionInPsFile1) {
    this.globalPosition = globalPositionInPsFile1;
  }

  /**
   * Method which provides a graphical representation of a receptive field.
   * 
   * @param g -
   *          a Graphics2D object to draw on.
   */
  public void mypaint(final Graphics2D g) {
    g.setColor(Color.black);

    for (int y = 0; y < height; y++) {
      for (int x = 0; x < width; x++) {
        int f = (int) ((255d) * ((field.get(y * this.height + x).doubleValue() - min) / (max - min)));
        // System.out.print(f + "\t");
        g.setColor(new Color(f, f, f, 255));
        g.fillRect(globalPosition.x * gwidth * width + x * gwidth, globalPosition.y * gheight * height + y * gheight, gwidth, gheight);
      }
    }
  }
}
