package brain.gui;

import java.awt.Color;
import java.awt.Graphics;
import java.awt.Graphics2D;
import java.awt.print.PageFormat;
import java.awt.print.Printable;
import java.awt.print.PrinterException;
import java.io.FileOutputStream;
import java.io.IOException;
import java.text.DecimalFormat;
import java.util.ArrayList;

import javax.print.Doc;
import javax.print.DocFlavor;
import javax.print.DocPrintJob;
import javax.print.PrintException;
import javax.print.SimpleDoc;
import javax.print.StreamPrintService;
import javax.print.StreamPrintServiceFactory;
import javax.print.attribute.HashPrintRequestAttributeSet;
import javax.print.attribute.PrintRequestAttributeSet;

/**
 * Every clique has a receptive field which is hold by this class. Additionally
 * this class contains information to paint or print all receptive fields.
 * 
 * @author Gregor Kaczor
 * @author Mathias Linkerhand
 * @version 2012-09-11
 * @since 2008
 */
public class ReceptiveFieldCollection implements Printable {

  /** an array which contains all receptive fields */
  ArrayList<ReceptiveField> receptiveFieldList = null;

  /**
   * Default horizontal distance between receptive field in a graphical
   * representation
   */
  private int               horizontalDistance = 5;
  /**
   * Default vertical distance between receptive field in a graphical
   * representation
   */
  private int               verticalDistance   = 5;
  /**
   * Should depend on the number of neurons from an input source. Example:
   * Math.sqrt(numberofneurons)
   */
  private int               heigth             = 20;
  /**
   * Should depend on the number of neurons from an input source. Example:
   * Math.sqrt(numberofneurons)
   */
  private int               width              = 20;

  /** absolute x position */
  private int               abs_x              = 0;

  /** absolute y position */
  private int               abs_y              = 0;
  /**
   * influences the color filter and can be used to adjust brightness in a
   * graphical representation of a feceptive field
   */
  private double            brightness;

  private Color[]           gradients;

  private String            title;

  /**
   * Standard constructor
   */
  public ReceptiveFieldCollection() {
    receptiveFieldList = new ArrayList<ReceptiveField>();
  }

  /**
   * Proprietary paint method, as this class does not extend or implement a AWT or
   * Swing component.
   * 
   * @param g2d
   */
  public void paint(Graphics2D g2d) {
    g2d.setColor(g2d.getBackground());
    g2d.clearRect(0, 0, this.getWidth(), this.getHeigth());
    if (gradients == null) {
      Color[] sourceColors = new Color[4];
      sourceColors[0] = Color.black;
      sourceColors[1] = Color.blue;
      sourceColors[2] = Color.red;
      sourceColors[3] = Color.yellow;
      gradients = createMultiGradient(sourceColors, 256);
    }
    g2d.setColor(Color.black);
    ReceptiveField rf = null;
    int maximum_x_GlobalPositionInPsFile = 0;
    int maximum_y_GlobalPositionInPsFile = 0;
    for (int ry = 0; ry < heigth; ry++) {
      for (int rx = 0; rx < width; rx++) {
        int index = ry * heigth + rx;
        if (!(index < receptiveFieldList.size())) {
          break;
        }
        rf = receptiveFieldList.get(ry * heigth + rx);
        if (rf == null)
          return;
        for (int y = 0; y < rf.getHeight(); y++) {
          for (int x = 0; x < rf.getWidth(); x++) {

            Double raw = Double.valueOf((((rf.get(Integer.valueOf(y * rf.getHeight() + x)).doubleValue() - rf.getMin()) / (rf.getMax() - rf.getMin()))));
            int f = (int) (filter(raw.doubleValue()) * (255d));

            g2d.setColor(gradients[f]);

            g2d.fillRect(abs_x + rf.getGlobalPositionInPsFile().x * rf.getGwidth() * rf.getWidth() + x * rf.getGwidth()
                + (rf.getGlobalPositionInPsFile().x > 0 ? horizontalDistance * rf.getGlobalPositionInPsFile().x : 0), abs_y + rf.getGlobalPositionInPsFile().y
                * rf.getGheight() * rf.getHeight() + (rf.getGlobalPositionInPsFile().y > 0 ? verticalDistance * rf.getGlobalPositionInPsFile().y : 0) + y
                * rf.getGheight(), rf.getGwidth(), rf.getGheight());
            maximum_x_GlobalPositionInPsFile = Math.max(maximum_x_GlobalPositionInPsFile, rf.getGlobalPositionInPsFile().x);
            maximum_y_GlobalPositionInPsFile = Math.max(maximum_y_GlobalPositionInPsFile, rf.getGlobalPositionInPsFile().y);
          }
        }
      }
    }
    /**
     * 0 is a valid position. so we have to add one
     */
    maximum_x_GlobalPositionInPsFile++;
    maximum_y_GlobalPositionInPsFile++;
    if (rf == null)
      return;
    double totalSteps = (rf.getHeight() * rf.getGheight() + (verticalDistance - 1)) * maximum_y_GlobalPositionInPsFile;
    totalSteps = totalSteps - g2d.getFontMetrics().getMaxAscent() * 2;

    int heigth1 = 1;
    final DecimalFormat format = new DecimalFormat("#.###");
    String maxValue = format.format(rf.max).toString();
    String minValue = format.format(rf.min).toString();
    char[] displayMax = maxValue.toCharArray();
    char[] displayMin = minValue.toCharArray();
    int width1 = Math.max(g2d.getFontMetrics().charsWidth(displayMax, 0, displayMax.length), g2d.getFontMetrics().charsWidth(displayMin, 0, displayMin.length));
    int y = g2d.getFontMetrics().getHeight() - 8;
    int x = (rf.getWidth() * rf.getGwidth() + horizontalDistance) * maximum_x_GlobalPositionInPsFile;

    g2d.setColor(Color.black);
    g2d.drawString(maxValue, x + abs_x, y + abs_y);
    y += 4;
    for (double i = 0; i < totalSteps; i++) {
      g2d.setColor(gradients[255 - (int) ((i / totalSteps) * 255)]);

      g2d.fillRect(x + abs_x, y + abs_y, width1, heigth1);
      y += heigth1;

    }
    y += g2d.getFontMetrics().getHeight() - 8;
    g2d.setColor(Color.black);
    g2d.drawString(minValue, x + 1 + abs_x, y + 2 + abs_y);
  }

  /**
   * 
   * @see java.awt.print.Printable#print(java.awt.Graphics,
   *      java.awt.print.PageFormat, int)
   */
  @Override
  public int print(Graphics graphics, PageFormat pageFormat, int pageIndex) throws PrinterException {
    if (pageIndex < 1) {
      int x_delta = 50;
      int y_delta = 90;
      this.abs_x += x_delta;
      this.abs_y += y_delta;
      if (gradients == null) {
        Color[] sourceColors = new Color[4];
        sourceColors[0] = Color.black;
        sourceColors[1] = new Color(0, 0, 255, 255);
        sourceColors[2] = Color.red;
        sourceColors[3] = Color.yellow;
        gradients = createMultiGradient(sourceColors, 256);
      }

      ReceptiveField rf = null;
      int maximum_x_GlobalPositionInPsFile = 0;
      int maximum_y_GlobalPositionInPsFile = 0;
      for (int ry = 0; ry < heigth; ry++) {
        for (int rx = 0; rx < width; rx++) {
          int index = ry * heigth + rx;
          if (!(index < receptiveFieldList.size())) {
            break;
          }
          rf = receptiveFieldList.get(ry * heigth + rx);
          if (rf == null)
            return Printable.NO_SUCH_PAGE;
          for (int y = 0; y < rf.getHeight(); y++) {
            for (int x = 0; x < rf.getWidth(); x++) {

              Double raw = Double.valueOf((((rf.get(Integer.valueOf(y * rf.getHeight() + x)).doubleValue() - rf.getMin()) / (rf.getMax() - rf.getMin()))));
              int f = (int) (filter(raw.doubleValue()) * (255d));

              graphics.setColor(gradients[f]);

              graphics.fillRect(abs_x + rf.getGlobalPositionInPsFile().x * rf.getGwidth() * rf.getWidth() + x * rf.getGwidth()
                  + (rf.getGlobalPositionInPsFile().x > 0 ? horizontalDistance * rf.getGlobalPositionInPsFile().x : 0), abs_y
                  + rf.getGlobalPositionInPsFile().y * rf.getGheight() * rf.getHeight()
                  + (rf.getGlobalPositionInPsFile().y > 0 ? verticalDistance * rf.getGlobalPositionInPsFile().y : 0) + y * rf.getGheight(), rf.getGwidth(), rf
                  .getGheight());
              maximum_x_GlobalPositionInPsFile = Math.max(maximum_x_GlobalPositionInPsFile, rf.getGlobalPositionInPsFile().x);
              maximum_y_GlobalPositionInPsFile = Math.max(maximum_y_GlobalPositionInPsFile, rf.getGlobalPositionInPsFile().y);
            }
          }
        }
      }
      /**
       * 0 is a valid position. so we have to add one
       */
      maximum_x_GlobalPositionInPsFile++;
      maximum_y_GlobalPositionInPsFile++;
      if (rf == null)
        return Printable.NO_SUCH_PAGE;
      double totalSteps = (rf.getHeight() * rf.getGheight() + (verticalDistance - 1)) * maximum_y_GlobalPositionInPsFile;
      totalSteps = totalSteps - graphics.getFontMetrics().getMaxAscent() * 2;

      int heigth1 = 1;
      String maxValue = Double.toString(rf.max) + "   ";
      maxValue = maxValue.substring(0, maxValue.indexOf(".") + 4).trim();
      String minValue = Double.toString(rf.min) + "   ";
      minValue = minValue.substring(0, minValue.indexOf(".") + 4).trim();
      char[] displayMax = maxValue.toCharArray();// new
      // Double(rf.max).toString().toCharArray();
      char[] displayMin = minValue.toCharArray();// new
      // Double(rf.min).toString().toCharArray();
      int width1 = Math.max(graphics.getFontMetrics().charsWidth(displayMax, 0, displayMax.length), graphics.getFontMetrics().charsWidth(displayMin, 0,
          displayMin.length));
      // int width1=20;
      int y = graphics.getFontMetrics().getHeight() - 8;
      int x = (rf.getWidth() * rf.getGwidth() + horizontalDistance) * maximum_x_GlobalPositionInPsFile;

      graphics.setColor(Color.black);
      graphics.drawString(maxValue, x + abs_x, y + abs_y);
      graphics.drawString(title, x + 1 + abs_x - 400, y + 2 + abs_y - 50);
      y += 4;
      for (double i = 0; i < totalSteps; i++) {
        graphics.setColor(gradients[255 - (int) ((i / totalSteps) * 255)]);

        graphics.fillRect(x + abs_x, y + abs_y, width1, heigth1);
        y += heigth1;

      }
      y += graphics.getFontMetrics().getHeight() - 8;
      graphics.setColor(Color.black);
      graphics.drawString(minValue.toString(), x + 1 + abs_x, y + 2 + abs_y);

      this.abs_x -= x_delta;
      this.abs_y -= y_delta;
      return Printable.PAGE_EXISTS;
    }
    return Printable.NO_SUCH_PAGE;

  }

  /**
   * A nonlinear color filter
   * 
   * @param y -
   *          a parameter
   * @return a non linear transformation of parameter y
   */
  public double filter(double y) {
    double bias = 0.05;
    brightness = 0.52;
    double dy = 0.12;

    return bias + (1 - bias) * (Math.tanh((y - brightness) / dy) - Math.tanh((0 - brightness) / dy))
        / (Math.tanh((1 - brightness) / dy) - Math.tanh((0 - brightness) / dy));

  }

  /**
   * Exports a all receptive fields to a postscript file
   * 
   * @param outputFileNamePath -
   *          path to the file to which {@link ReceptiveFieldCollection} is
   *          exported
   */
  public void print(String outputFileNamePath) {
    DocFlavor flavor = DocFlavor.SERVICE_FORMATTED.PRINTABLE;
    String psMimeType = DocFlavor.BYTE_ARRAY.POSTSCRIPT.getMimeType();
    StreamPrintServiceFactory[] factories = StreamPrintServiceFactory.lookupStreamPrintServiceFactories(flavor, psMimeType);

    if (factories.length == 0) {
      System.err.println("No suitable factories for printing !");
      return;
    }

    try {
      FileOutputStream fos = new FileOutputStream(outputFileNamePath == null ? "pattern.ps" : outputFileNamePath);

      StreamPrintService sps = factories[0].getPrintService(fos);

      DocPrintJob pj = sps.createPrintJob();
      PrintRequestAttributeSet aset = new HashPrintRequestAttributeSet();

      aset.add(new javax.print.attribute.standard.Copies(1));

      aset.add(javax.print.attribute.standard.MediaSizeName.ISO_A4);

      Doc doc = new SimpleDoc(this, flavor, null);

      pj.print(doc, aset);
      fos.close();
    } catch (PrintException pe) {
      System.err.println("PrintException : " + pe);
    } catch (IOException ioe) {
      System.err.println("IOexception : " + ioe);
    }
  }

  /**
   * Creates an array of Color objects for use as a gradient, using a linear
   * interpolation between the two specified colors.
   * 
   * @param one -
   *          Color used for the bottom of the gradient
   * @param two -
   *          Color used for the top of the gradient
   * @param numSteps -
   *          The number of steps in the gradient. 250 is a good number.
   * @return Array of colors with smooth transition between Color one and Color
   *         two
   */
  public static Color[] createGradient(Color one, Color two, int numSteps) {
    int r1 = one.getRed();
    int g1 = one.getGreen();
    int b1 = one.getBlue();

    int r2 = two.getRed();
    int g2 = two.getGreen();
    int b2 = two.getBlue();

    int newR = 0;
    int newG = 0;
    int newB = 0;

    Color[] gradient = new Color[numSteps];
    double iNorm;
    for (int i = 0; i < numSteps; i++) {
      iNorm = i / (double) numSteps; // a normalized [0:1] variable
      newR = (int) (r1 + iNorm * (r2 - r1));
      newG = (int) (g1 + iNorm * (g2 - g1));
      newB = (int) (b1 + iNorm * (b2 - b1));
      gradient[i] = new Color(newR, newG, newB);
    }

    return gradient;
  }

  /**
   * Creates an array of Color objects for use as a gradient, using an array of
   * Color objects. It uses a linear interpolation between each pair of points.
   * 
   * @param colors -
   *          An array of Color objects used for the gradient. The Color at index
   *          0 will be the lowest color.
   * @param numSteps -
   *          The number of steps in the gradient. 250 is a good number.
   * @return Array of colors with smooth transition between Color in the array
   *         colors
   */
  public static Color[] createMultiGradient(Color[] colors, int numSteps) {
    // we assume a linear gradient, with equal spacing between colors
    // The final gradient will be made up of n 'sections', where n =
    // colors.length - 1
    int numSections = colors.length - 1;
    int gradientIndex = 0; // points to the next open spot in the final
    // gradient
    Color[] gradient = new Color[numSteps];
    Color[] temp;

    if (numSections <= 0) {
      throw new IllegalArgumentException("You must pass in at least 2 colors in the array!");
    }

    for (int section = 0; section < numSections; section++) {
      // we divide the gradient into (n - 1) sections, and do a regular
      // gradient for each
      temp = createGradient(colors[section], colors[section + 1], numSteps / numSections);
      for (int i = 0; i < temp.length; i++) {
        // copy the sub-gradient into the overall gradient
        gradient[gradientIndex++] = temp[i];
      }
    }

    if (gradientIndex < numSteps) {
      // The rounding didn't work out in our favor, and there is at least
      // one unfilled slot in the gradient[] array.
      // We can just copy the final color there
      for (/* nothing to initialize */; gradientIndex < numSteps; gradientIndex++) {
        gradient[gradientIndex] = colors[colors.length - 1];
      }
    }

    return gradient;
  }

  /**
   * Returns a List with receptive fields.
   * 
   * @return the receptiveFieldList
   */
  public ArrayList<ReceptiveField> getReceptiveFieldList() {
    return receptiveFieldList;
  }

  /**
   * Sets a List with receptive fields.
   * 
   * @param _receptiveFieldList -
   *          the receptive field list to set
   */
  public void setReceptiveFieldList(ArrayList<ReceptiveField> _receptiveFieldList) {
    this.receptiveFieldList = _receptiveFieldList;
  }

  /**
   * Returns the vertical distance between receptive fields in pixels.
   * 
   * @return the vertical distance between receptive fields in pixels.
   */
  public int getVerticalDistance() {
    return verticalDistance;
  }

  /**
   * Sets the vertical distance between receptive fields in pixels.
   * 
   * @param distance -
   *          the vertical distance between receptive fields in pixels.
   */
  public void setVerticalDistance(int distance) {
    verticalDistance = distance;
  }

  /**
   * Returns horizontal distance between receptive fields in pixels.
   * 
   * @return horizontal distance between receptive fields in pixels.
   */
  public int getHorizontalDistance() {
    return horizontalDistance;
  }

  /**
   * Sets horizontal distance between receptive fields in pixels.
   * 
   * @param distance -
   *          horizontal distance between receptive fields in pixels.
   */
  public void setHorizontalDistance(int distance) {
    horizontalDistance = distance;
  }

  /**
   * Returns number of receptive fields in this ReceptiveFieldCollection.
   * 
   * @return number of receptive fields in this ReceptiveFieldCollection
   */
  public int size() {
    return receptiveFieldList.size();
  }

  /**
   * Sets absolute x position for a graphical representation of receptive field.
   * 
   * @return abs_x - absolute y position for a graphical representation of
   *         receptive field.
   */
  public int getAbs_x() {
    return abs_x;
  }

  /**
   * Sets absolute x position for a graphical representation of receptive field.
   * 
   * @param abs_x -
   *          absolute y position for a graphical representation of receptive
   *          field.
   */
  public void setAbs_x(int abs_x) {
    this.abs_x = abs_x;
  }

  /**
   * Returns absolute y position for a graphical representation of receptive
   * field.
   * 
   * @return absolute y position for a graphical representation of receptive
   *         field.
   */
  public int getAbs_y() {
    return abs_y;
  }

  /**
   * Sets absolute y position for a graphical representation of receptive field.
   * 
   * @param abs_y -
   *          absolute y position for a graphical representation of receptive
   *          field.
   */
  public void setAbs_y(int abs_y) {
    this.abs_y = abs_y;
  }

  /**
   * Returns number of sites in a receptive fields in a column.
   * 
   * @return number of sites in a receptive fields in a column.
   */
  public int getHeigth() {
    return heigth;
  }

  /**
   * Sets number of sites in a receptive fields in a column.
   * 
   * @param heigth -
   *          number of sites in a receptive fields in a column.
   */
  public void setHeigth(int heigth) {
    this.heigth = heigth;
  }

  /**
   * Returns number of sites in a receptive fields in a row.
   * 
   * @return number of sites in a receptive fields in a row.
   */
  public int getWidth() {
    return width;
  }

  /**
   * Sets number of sites in a receptive fields in a row.
   * 
   * @param width -
   *          number of sites in a receptive fields in a row.
   */
  public void setWidth(int width) {
    this.width = width;
  }

  /**
   * Returns the title of the receptive field collection
   * @return title as String
   */
  public String getTitle() {
    return title;
  }

  /**
   * Sets the title of the receptive field collection
   * @param title - a String as title
   */
  public void setTitle(String title) {
    this.title = title;
  }

}
