package brain.gui;

import java.awt.BorderLayout;
import java.awt.GridBagConstraints;
import java.awt.GridBagLayout;
import java.awt.Insets;
import java.awt.Point;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.io.File;
import java.io.IOException;
import java.util.Iterator;
import java.util.Set;
import java.util.TreeSet;
import java.util.Vector;

import javax.swing.JButton;
import javax.swing.JComboBox;
import javax.swing.JFileChooser;
import javax.swing.JFrame;
import javax.swing.JLabel;
import javax.swing.JPanel;
import javax.swing.JSpinner;
import javax.swing.SpinnerModel;
import javax.swing.SpinnerNumberModel;
import javax.swing.SwingConstants;
import javax.swing.event.ChangeEvent;
import javax.swing.event.ChangeListener;
import javax.swing.filechooser.FileNameExtensionFilter;

import brain.Module;
import brain.cliqueCalculation.BidirectionalGraph;
import brain.cliqueCalculation.GraphStatistics;

/**
 * Shows the receptive fields. A receptive field is the sum over all backward
 * links of a clique divided by its size.
 * 
 * @author Gregor Kaczor
 * @author Mathias Linkerhand
 * @version 2012-09-11
 * @since 2008
 */
public class ReceptiveFieldViewPanel extends ViewPanel {
  /**
   * 
   */
  private static final long        serialVersionUID         = 1L;
  /** Component for user action. Defines size of the smallest matrix element */
  private JSpinner                 zoomJSpinner;
  /**
   * Component for user action. Defines number of columns, when displaying
   * receptive fields.
   */
  private JSpinner                 spinnerColumns;

  /** Container for a receptive field of a clique */
  private ReceptiveFieldCollection receptiveFieldCollecion;
  /** Graphical panel where receptive fields are drawn on */
  private ReceptiveFieldPaintPanel receptiveFieldPaintPanel = new ReceptiveFieldPaintPanel();

  /**
   * Standard constructor
   * 
   * @param _title
   *            - title of the panel
   * @param _instance
   *            - instance of a module we paint from. in case we need
   *            inter-modular data then one can access written simulation
   *            files and read them for further analysis.
   * @param _methodName
   *            - method we need to call via Reflection API to get data
   */
  public ReceptiveFieldViewPanel(String _title, Module _instance, String _methodName) {
    super(_title, _instance, _methodName);
    init();
  }

  /**
   * Draws components on a {@link JFrame}
   */
  private void init() {

    this.setLayout(new BorderLayout());
    GridBagConstraints c = new GridBagConstraints();
    c.fill = GridBagConstraints.HORIZONTAL;
    c.gridx = 0;
    c.gridy = 0;
    c.insets = new Insets(3, 3, 3, 3);
    JPanel buttonPanel = new JPanel();
    buttonPanel.setLayout(new GridBagLayout());

    final Set<Module> backModules = module.getAfferentModules();
    final JComboBox backwardModulesJComboBox = new JComboBox(backModules.toArray());

    buttonPanel.add(backwardModulesJComboBox, c);

    /** Sets the grid constraints and add components to buttonPanel */
    final JButton update = new JButton("Update");
    update.setToolTipText("Recalculate receptive fields.");
    update.addActionListener(new ActionListener() {
      @Override
      public void actionPerformed(ActionEvent e) {

        //get internal links to calculate cliques
        // calculate cliques
        BidirectionalGraph bg = new BidirectionalGraph(module);
        GraphStatistics gs = new GraphStatistics(bg);

        Vector<TreeSet<Integer>> listOfCliques = gs.calculateCliques();
        //System.out.println("listOfCliques.size()" + listOfCliques.size());

        // to which backmodule do we calculate the receptive field?
        Module backModule = (Module) backwardModulesJComboBox.getSelectedItem();

        // get number of neurons in the back module
        int N = backModule.getSize();

        // reinitiate receptive field collection
        receptiveFieldCollecion = new ReceptiveFieldCollection();
        receptiveFieldCollecion.setVerticalDistance(5);
        int row = 0;
        int col = 0;

        double[][] receptiveField = new double[listOfCliques.size()][N];
        for (int j = 0; j < N; j++) {
          int cnumber = 0;
          for (TreeSet<Integer> clique : listOfCliques) {
            Iterator<Integer> siteIterator = clique.iterator();

            while (siteIterator.hasNext()) {
              Integer cliquesite = siteIterator.next();
              {
                receptiveField[cnumber][j] += module.getAfferentLink(backModule, cliquesite, j);
              }
            }
            receptiveField[cnumber][j] /= clique.size();
            cnumber++;
          }
        }

        for (int j = 0; j < listOfCliques.size(); j++) {
          double[] field = new double[N];
          for (int i = 0; i < field.length; i++) {
            field[i] = receptiveField[j][i];
          }
          ReceptiveField rc = new ReceptiveField(field);

          rc.setGlobalPositionInPsFile(new Point(col = j % Integer.parseInt(spinnerColumns.getValue().toString()) == 0 ? 0 : col + 1, j
              % Integer.parseInt(spinnerColumns.getValue().toString()) == 0
              && j > 0 ? ++row : row));
          rc.setGheight(Integer.parseInt(zoomJSpinner.getValue().toString()));
          rc.setGwidth(Integer.parseInt(zoomJSpinner.getValue().toString()));
          rc.setHeight((int) Math.sqrt(N));
          rc.setWidth((int) Math.sqrt(N));

          receptiveFieldCollecion.getReceptiveFieldList().add(rc);
        }

        receptiveFieldPaintPanel.setReceptiveFieldCollecion(receptiveFieldCollecion);
        receptiveFieldPaintPanel.repaint();

      }
    });

    c.gridx++;
    buttonPanel.add(update, c);

    JLabel zoomJLabel = new JLabel("Zoom", SwingConstants.CENTER);
    SpinnerModel zoomNumberModel = new SpinnerNumberModel(8, 1, 32, 1);

    zoomJSpinner = new JSpinner(zoomNumberModel);
    c.gridx++;
    buttonPanel.add(zoomJLabel, c);
    c.gridx++;
    buttonPanel.add(zoomJSpinner, c);
    zoomJSpinner.setToolTipText("Define the pixel size of the smallest element.");
    zoomJSpinner.addChangeListener(new ChangeListener() {
      @Override
      public void stateChanged(ChangeEvent e) {
        //calculateForPainting();
        receptiveFieldPaintPanel.repaint();
      }
    });

    JLabel strokeJLabel = new JLabel("Columns", SwingConstants.CENTER);
    SpinnerModel spinnerNumberModel = new SpinnerNumberModel(5, 1, 32, 1);

    spinnerColumns = new JSpinner(spinnerNumberModel);
    c.gridx++;
    buttonPanel.add(strokeJLabel, c);
    c.gridx++;
    buttonPanel.add(spinnerColumns, c);
    spinnerColumns.setToolTipText("Choose the number of columns.");
    spinnerColumns.addChangeListener(new ChangeListener() {

      @Override
      public void stateChanged(ChangeEvent e) {
        // calculateForPainting();
        receptiveFieldPaintPanel.repaint();
        test();
      }
    });

    JButton exportJButton = new JButton("Export");
    exportJButton.addActionListener(new ActionListener() {

      //@Override
      @Override
      public void actionPerformed(ActionEvent e) {
        JFileChooser fileChooser = null;
        try {
          File f = new File(new File(".").getCanonicalPath());
          fileChooser = new JFileChooser();
          fileChooser.setCurrentDirectory(f);
          fileChooser.setAcceptAllFileFilterUsed(true);
          fileChooser.setDialogTitle("Export receptive fields to file");
        } catch (IOException e2) {

          e2.printStackTrace();
          return;
        }

        String formats[] = new String[2];
        formats[1] = "ps";
        formats[0] = "eps";
        for (int i = 0; i < formats.length; i++) {
          FileNameExtensionFilter fileChooserMultiFilter = new FileNameExtensionFilter(formats[i] + " image", formats[i]);
          fileChooser.addChoosableFileFilter(fileChooserMultiFilter);
        }

        // Open file dialog.
        int result = fileChooser.showSaveDialog(null);

        // Determine which button was clicked to close the dialog
        switch (result) {
          case JFileChooser.APPROVE_OPTION:

            // Approve (Open or Save) was clicked
            receptiveFieldCollecion.print(fileChooser.getSelectedFile().getAbsolutePath());

            break;
          case JFileChooser.CANCEL_OPTION:
            // Cancel or the close-dialog icon was clicked
            return;

          case JFileChooser.ERROR_OPTION:
            // The selection process did not complete successfully
            return;
        }
      }
    });
    c.gridx++;
    buttonPanel.add(exportJButton, c);

    this.add(buttonPanel, BorderLayout.NORTH);
    this.add(receptiveFieldPaintPanel, BorderLayout.CENTER);
  }

  private void test() {
    this.updateUI();
  }
}
