package brain.gui;

import java.awt.BorderLayout;
import java.awt.GridBagConstraints;
import java.awt.GridBagLayout;
import java.awt.Insets;
import java.awt.Point;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.io.BufferedReader;
import java.io.File;
import java.io.FileNotFoundException;
import java.io.FileReader;
import java.io.IOException;
import java.util.ArrayList;

import javax.swing.JButton;
import javax.swing.JFileChooser;
import javax.swing.JLabel;
import javax.swing.JPanel;
import javax.swing.JSpinner;
import javax.swing.SpinnerModel;
import javax.swing.SpinnerNumberModel;
import javax.swing.SwingConstants;
import javax.swing.event.ChangeEvent;
import javax.swing.event.ChangeListener;
import javax.swing.filechooser.FileNameExtensionFilter;

import brain.Module;

/**
 * Shows the receptive fields. A receptive field is the sum over all backlinks
 * of a clique divided by its size.
 * 
 * @author Gregor Kaczor
 * @author Mathias Linkerhand
 * @version 2012-09-11
 * @since 2008
 */
public class ReferencePatternViewPanel extends ViewPanel {
  /**
   * 
   */
  private static final long        serialVersionUID         = 1L;
  /** Component for user action. Defines size of the smallest matrix element */
  private JSpinner                 zoomJSpinner;
  /**
   * Component for user action. Defines number of columns, when displaying
   * receptive fields.
   */
  private JSpinner                 spinnerColumns;
  /** Holds backlinks */
  private ArrayList<String[]>      backlinks;
  /** Container for a receptive field of a clique */
  private ReceptiveFieldCollection receptiveFieldCollecion;
  /** Graphical panel where receptive fields are drawn on */
  private ReceptiveFieldPaintPanel receptiveFieldPaintPanel = new ReceptiveFieldPaintPanel();
  /** File path where backlinks are read from */
  private String                   inputValue               = "Brain_input_structured.patt";

  /**
   * Standard constructor
   * 
   * @param _title -
   *          title of the panel
   * @param _instance -
   *          instance of a module we paint from. in case we need inter-modular
   *          data then one can access written simulation files and read them for
   *          further analysis.
   * @param _methodName -
   *          method we need to call via Reflection API to get data
   */
  ReferencePatternViewPanel(String _title, Module _instance, String _methodName) {
    super(_title, _instance, _methodName);
    init();
  }

  /**
   * Draws components on a {@link javax.swing.JFrame}
   */
  private void init() {

    this.setLayout(new BorderLayout());

    JPanel buttonPanel = new JPanel();
    buttonPanel.setLayout(new GridBagLayout());

    /** Sets the grid constraints and add components to buttonPanel */
    final JButton update = new JButton("Update");
    update.setToolTipText("Recalculate receptive fields.");
    update.addActionListener(new ActionListener() {

      private double max;
      private double min;

      @Override
      public void actionPerformed(ActionEvent e) {

        prepare();

      }

      /**
       * 
       */
      private void prepare() {
        File f = new File(inputValue);
        FileReader fr;
        BufferedReader bfr = null;

        try {
          fr = new FileReader(f);
          bfr = new BufferedReader(fr);
        } catch (FileNotFoundException e) {
          e.printStackTrace();
          System.err.println("failed to open " + f.getAbsolutePath());
          return;
        }

        String line = "";

        backlinks = new ArrayList<String[]>();
        try {
          while ((line = bfr.readLine()) != null) {
            line = line.trim();
            if (line.startsWith("#") || line.length() == 0) {
              continue;
            }
            do {
              line = line.replaceAll("  ", " ");
            } while (line.indexOf("  ") > -1);
            int startIndex = line.indexOf(" ");
            if (startIndex > -1) {
              line = line.substring(startIndex).trim();
            }
            backlinks.add(line.split(" "));
          }
        } catch (IOException e) {
          e.printStackTrace();
        }

        // Get Minimum and Maximum to calculate color ranges

        max = Double.MIN_VALUE;
        min = Double.MAX_VALUE;

        for (int i = 0; i < backlinks.size(); i++) {
          for (int j = 0; j < backlinks.get(i).length; j++) {
            max = Math.max(max, Double.parseDouble(backlinks.get(i)[j]));
            min = Math.min(min, Double.parseDouble(backlinks.get(i)[j]));
          }
        }

        receptiveFieldCollecion = new ReceptiveFieldCollection();
        receptiveFieldCollecion.setVerticalDistance(5);

        int row = 0;
        int col = 0;
        for (int j = 0; j < backlinks.get(0).length; j++) {
          ArrayList<Double> valueList = new ArrayList<Double>();
          ReceptiveField rc = new ReceptiveField();

          rc.setGlobalPositionInPsFile(new Point(col = j % Integer.parseInt(spinnerColumns.getValue().toString()) == 0 ? 0 : col + 1, j
              % Integer.parseInt(spinnerColumns.getValue().toString()) == 0
              && j > 0 ? ++row : row));
          rc.setMax(max);
          rc.setMin(min);
          rc.setGheight(Integer.parseInt(zoomJSpinner.getValue().toString()));
          rc.setGwidth(Integer.parseInt(zoomJSpinner.getValue().toString()));
          rc.setHeight((int) Math.sqrt(backlinks.size()));
          rc.setWidth((int) Math.sqrt(backlinks.size()));
          for (int i = 0; i < backlinks.size(); i++) {
            valueList.add(Double.valueOf(backlinks.get(i)[j]));
          }
          rc.setField(valueList);
          receptiveFieldCollecion.getReceptiveFieldList().add(rc);
          receptiveFieldCollecion.setHeigth((int) Math.sqrt(backlinks.size()));
          receptiveFieldCollecion.setWidth((int) Math.sqrt(backlinks.size()));
        }
        receptiveFieldPaintPanel.setReceptiveFieldCollecion(receptiveFieldCollecion);
        receptiveFieldPaintPanel.repaint();

      }
    });

    GridBagConstraints c = new GridBagConstraints();
    c.fill = GridBagConstraints.HORIZONTAL;
    c.gridx = 0;
    c.gridy = 0;
    c.insets = new Insets(3, 3, 3, 3);
    buttonPanel.add(update, c);

    JLabel zoomJLabel = new JLabel("Zoom", SwingConstants.CENTER);
    SpinnerModel zoomNumberModel = new SpinnerNumberModel(8, 1, 32, 1);

    zoomJSpinner = new JSpinner(zoomNumberModel);
    c.gridx++;
    buttonPanel.add(zoomJLabel, c);
    c.gridx++;
    buttonPanel.add(zoomJSpinner, c);
    zoomJSpinner.setToolTipText("Define the pixel size of the smallest element.");
    zoomJSpinner.addChangeListener(new ChangeListener() {
      @Override
      public void stateChanged(ChangeEvent e) {
        calculateForPainting();
        receptiveFieldPaintPanel.repaint();
      }
    });

    JLabel strokeJLabel = new JLabel("Columns", SwingConstants.CENTER);
    SpinnerModel spinnerNumberModel = new SpinnerNumberModel(5, 1, 32, 1);

    spinnerColumns = new JSpinner(spinnerNumberModel);
    c.gridx++;
    buttonPanel.add(strokeJLabel, c);
    c.gridx++;
    buttonPanel.add(spinnerColumns, c);
    spinnerColumns.setToolTipText("Choose the number of columns.");
    spinnerColumns.addChangeListener(new ChangeListener() {

      @Override
      public void stateChanged(ChangeEvent e) {
        calculateForPainting();
        receptiveFieldPaintPanel.repaint();
      }
    });

    JButton exportJButton = new JButton("Export");
    exportJButton.addActionListener(new ActionListener() {

      //@Override
      @Override
      public void actionPerformed(ActionEvent e) {
        JFileChooser fileChooser = null;
        try {
          File f = new File(new File(".").getCanonicalPath());
          fileChooser = new JFileChooser();
          fileChooser.setCurrentDirectory(f);
          fileChooser.setAcceptAllFileFilterUsed(true);
          fileChooser.setDialogTitle("Export receptive fields to file");
        } catch (IOException e2) {

          e2.printStackTrace();
          return;
        }

        String formats[] = new String[2];
        formats[1] = "ps";
        formats[0] = "eps";
        for (int i = 0; i < formats.length; i++) {
          FileNameExtensionFilter fileChooserMultiFilter = new FileNameExtensionFilter(formats[i] + " image", formats[i]);
          fileChooser.addChoosableFileFilter(fileChooserMultiFilter);
        }

        // Open file dialog.
        int result = fileChooser.showSaveDialog(null);

        // Determine which button was clicked to close the dialog
        switch (result) {
          case JFileChooser.APPROVE_OPTION:

            // Approve (Open or Save) was clicked
            receptiveFieldCollecion.print(fileChooser.getSelectedFile().getAbsolutePath());

            break;
          case JFileChooser.CANCEL_OPTION:
            // Cancel or the close-dialog icon was clicked
            return;

          case JFileChooser.ERROR_OPTION:
            // The selection process did not complete successfully
            return;
        }

      }

    });
    c.gridx++;
    buttonPanel.add(exportJButton, c);

    this.add("North", buttonPanel);
    this.add("OneCenter", receptiveFieldPaintPanel);
  }

  /**
   * Recalculates display parameters for zooming or columns. Its faster than
   * reinitiating the whole statistics process and redrawing receptive fields.
   */
  private void calculateForPainting() {
    if (receptiveFieldCollecion == null)
      return;
    int row = 0;
    int col = 0;
    for (int j = 0; j < backlinks.get(0).length; j++) {
      ReceptiveField rc = receptiveFieldCollecion.getReceptiveFieldList().get(j);
      rc.setGlobalPositionInPsFile(new Point(col = j % Integer.parseInt(spinnerColumns.getValue().toString()) == 0 ? 0 : col + 1, j
          % Integer.parseInt(spinnerColumns.getValue().toString()) == 0
          && j > 0 ? ++row : row));
      rc.setGheight(Integer.parseInt(zoomJSpinner.getValue().toString()));
      rc.setGwidth(Integer.parseInt(zoomJSpinner.getValue().toString()));
      rc.setHeight((int) Math.sqrt(backlinks.size()));
      rc.setWidth((int) Math.sqrt(backlinks.size()));
      receptiveFieldCollecion.getReceptiveFieldList().add(rc);
      receptiveFieldCollecion.setHeigth((int) Math.sqrt(backlinks.size()));
      receptiveFieldCollecion.setWidth((int) Math.sqrt(backlinks.size()));
    }
    receptiveFieldPaintPanel.setReceptiveFieldCollecion(receptiveFieldCollecion);
    receptiveFieldPaintPanel.repaint();
  }
}
