package brain.gui;

import java.awt.Color;
import java.awt.Graphics;
import java.awt.Graphics2D;
import java.awt.RenderingHints;

import brain.Module;

/**
 * Displays a module/object which implements Module. This class does access
 * data via Reflection API. Sites are represented by blue dots ordered on a
 * circle. Every site has a label which in this version is a integer starting
 * with 0. The relation between sites is represented by black lines. They
 * indicate an excitatory connection. The complete drawing scales with the size
 * of the panel which it is drawn on.
 * 
 * @author Gregor Kaczor
 * @author Mathias Linkerhand
 * @version 2010-09-10
 */
public class SystemViewPanel extends ViewPanel {
  private static final long serialVersionUID   = 1L;

  /**
   * critical activities indicates when a site is considered active or inactive.
   * Below that value it is defined inactive, above that value is is considered
   * active
   */
  // FIXME criticalActivation has to be read dynamically from the observed module
  private double            criticalActivation = 0.8;
  /**
   * The label sites are drawn on their own circle. And thats the difference
   * between the site circle and the label circle
   */
  private double            labelRadiusDelta   = 20;
  /** Switch for showing internal excitatory links or not */
  private boolean           showExcitatory     = true;

  /**
   * Instantiates <code>SystemViewPanel</code>
   * 
   * @param _title -
   *          title of the <code>JFrame</code> on which
   *          <code>SystemViewPanel</code> appears
   * @param _moduleInstance -
   *          instance of a module(object which implements <code>Module</code>
   * @param _methodName -
   *          method name of that method which is to be called via Reflection API
   *          in oder to access the data that is intended to be displayed.
   */
  public SystemViewPanel(String _title, Module _moduleInstance, String _methodName) {

    super(_title, _moduleInstance, _methodName);
    //this.setDoubleBuffered(true);
  }

  /**
   * Paints this component
   * 
   * @see javax.swing.JComponent#paint(java.awt.Graphics)
   */
  @Override
  public void paint(Graphics g1) {

    Graphics2D g = (Graphics2D) g1;
    double radius = 0;
    int drawAtX = 0;
    int drawAtY = 0;
    /** sites with their activities levels */
    drawAtX = this.getWidth() / 2;
    drawAtY = this.getHeight() / 2;
    radius = Math.min(drawAtX, drawAtY) - labelRadiusDelta * 2;
    final int size = module.getSize();
    g.setRenderingHint(RenderingHints.KEY_ANTIALIASING, RenderingHints.VALUE_ANTIALIAS_ON);
    g.clearRect(0, 0, this.getWidth(), this.getHeight());

    int[][] points = new int[size][2];
    int x1, x2, y1, y2;

    for (int i = 0; i < size; i++) {
      points[i][0] = (int) (drawAtX + radius * Math.cos(2.0 * i * Math.PI / size + 0.5 * Math.PI));
      points[i][1] = (int) (drawAtY + radius * Math.sin(2.0 * i * Math.PI / size + 0.5 * Math.PI));

      for (int j = 0; j < i; j++) {
        x1 = points[i][0];
        y1 = points[i][1];
        x2 = points[j][0];
        y2 = points[j][1];

        if (showExcitatory && module.getSize() > 0) {
          try {
            if (module.getLink(i, j) > 0) {
              g.setColor(Color.black);
              if (module.getActivity(i) > criticalActivation && module.getActivity(j) > criticalActivation) {
                g.setColor(Color.red);
              }
              g.drawLine(x1, y1, x2, y2);

            }
          } catch (UnsupportedOperationException exception) {
          }
        }
      }

      g.setColor(Color.blue);
      if (module.getActivity(i) > criticalActivation) {
        g.setColor(Color.red);
        g.fillOval(points[i][0] - 5, points[i][1] - 5, 10, 10);

      } else {
        g.fillOval(points[i][0] - 5, points[i][1] - 5, 10, 10);
      }

      int x = (int) (drawAtX + (radius + labelRadiusDelta) * Math.cos(2.0 * i * Math.PI / size + 0.5 * Math.PI));
      int y = (int) (drawAtY + (radius + labelRadiusDelta) * Math.sin(2.0 * i * Math.PI / size + 0.5 * Math.PI));
      String nodeNumber = Integer.toString(i);
      char data[] = new char[nodeNumber.length()];
      for (int letter = 0; letter < data.length; letter++) {
        data[letter] = nodeNumber.charAt(letter);
      }
      int width = g.getFontMetrics().charsWidth(data, 0, data.length);
      int height = g.getFontMetrics().getHeight();
      g.setColor(Color.black);
      g.drawString((Integer.valueOf(i)).toString(), x - width / 2, y + height / 2);
    }

  }

}
