package brain.gui.chart;

import java.awt.BorderLayout;
import java.awt.Component;
import java.awt.Dimension;
import java.awt.Graphics2D;
import java.awt.image.BufferedImage;
import java.io.File;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.OutputStream;
import java.lang.reflect.Field;
import java.util.Arrays;
import java.util.List;
import java.util.Vector;

import javax.imageio.ImageIO;
import javax.swing.JPanel;
import javax.swing.event.ChangeEvent;
import javax.swing.event.ChangeListener;

import org.apache.commons.lang3.ArrayUtils;

/**
 * @author Mathias Linkerhand
 * @version 2011-09-21
 * @version 2011
 */
abstract public class Chart extends JPanel implements ChangeListener {
  private static final long serialVersionUID = 1L;

  /**
   * Module instance of Module, which happens to be accessed via Reflection API
   * by mehtodName every time observable notifies this object
   */
  protected Object          object;

  /** The field which gets accessed if usePublicField is true */
  protected Field           fieldX;
  protected Field           fieldY;

  protected Component       chart;

  protected double          lastRepaintTime  = 0;
  protected double          repaintInterval  = 1;

  public Chart(final Field fieldX, final Field fieldY, final Object object) {
    this.object = object;
    this.fieldX = fieldX;
    this.fieldY = fieldY;

    this.setLayout(new BorderLayout());
  }

  @Override
  public void stateChanged(ChangeEvent e) {
    try {
      fieldX.setAccessible(true);
      double currentTime = (Double) fieldX.get(this.object);
      if (currentTime >= this.lastRepaintTime + this.repaintInterval) {
        final double[] values = this.getValues(this.fieldY);
        this.addPoints(currentTime, values);
        this.lastRepaintTime = currentTime;
      }
    } catch (IllegalArgumentException e1) {
      e1.printStackTrace();
    } catch (IllegalAccessException e1) {
      e1.printStackTrace();
    }
  }

  abstract protected void addPoints(double time, double[] values);

  abstract public void setTraceLength(int traceLength);

  abstract public boolean isZoomable();

  abstract public void zoomFull();

  abstract public void setStroke(int stroke);

  public void export(final File file, final String formatName) throws IOException {
    ChartFrame.exportComponent(this.chart, file, formatName);
  }

  public static void exportComponent(final Component component, final File file, final String formatName) throws IOException {
    Dimension size = component.getSize();
    BufferedImage myImage = new BufferedImage(size.width, size.height, BufferedImage.TYPE_INT_RGB);
    Graphics2D graphics2D = myImage.createGraphics();
    component.paint(graphics2D);
    OutputStream out = new FileOutputStream(file);
    ImageIO.write(myImage, formatName, out);
    out.close();
  }

  /**
   * Returns values which return type is an array
   * 
   * @return array of doubles which represent y values on the chart
   */
  protected double[] getValues(final Field field) {
    try {
      field.setAccessible(true);
      final Object values = field.get(object);
      final List<Double> result = getValueAsList(values);
      return ArrayUtils.toPrimitive(result.toArray(new Double[0]));
    } catch (IllegalArgumentException e) {
      e.printStackTrace();
    } catch (IllegalAccessException e) {
      e.printStackTrace();
    }
    return null;
  }

  public static List<Double> getValueAsList(final Object value) {
    // result list of doubles
    final List<Double> result = new Vector<Double>();

    if (value instanceof Number) {
      // Number?
      result.add(((Number) value).doubleValue());
    } else if (value instanceof double[]) {
      // double[]?
      result.addAll(Arrays.asList(ArrayUtils.toObject((double[]) value)));
    } else if (value instanceof Iterable<?>) {
      for (final Object innerValue : (Iterable<?>) value) {
        result.addAll(getValueAsList(innerValue));
      }
    } else if (value instanceof Object[]) {
      // Object[]? This also includes double[][], … since double[][] is instance of Object[]
      for (final Object object : (Object[]) value) {
        result.addAll(getValueAsList(object));
      }
    } else {
      System.err.println("Cannot convert value " + value);
    }
    return result;
  }

  public Component getChart() {
    return this.chart;
  }

  public void setChart(final Component chart) {
    this.add(chart, BorderLayout.CENTER);
    this.chart = chart;
  }
}