package brain.gui.table;

import java.awt.Color;
import java.awt.Component;
import java.awt.Font;
import java.lang.reflect.AccessibleObject;
import java.lang.reflect.Field;
import java.lang.reflect.Method;
import java.text.NumberFormat;
import java.util.Arrays;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Vector;

import javax.swing.JLabel;
import javax.swing.JTable;
import javax.swing.table.DefaultTableCellRenderer;

/**
 * @author Mathias Linkerhand
 * @version 2011-11-29
 */
public class AccessibleObjectTableCellRenderer extends DefaultTableCellRenderer {
  private static final long  serialVersionUID = 1L;

  private Map<Object, Color> colors           = new HashMap<Object, Color>();

  public static String getStringValue(final Object value) {
    final String string;
    if (value instanceof Number) {
      // Number?
      string = NumberFormat.getNumberInstance().format(value);
    } else if (value instanceof double[]) {
      // double[]?
      string = Arrays.toString((double[]) value);
    } else if (value instanceof Object[]) {
      // Object[]? This also includes double[][], … since double[][] is instance of Object[]
      final List<String> strings = new Vector<String>();
      for (final Object object : (Object[]) value) {
        strings.add(getStringValue(object));
      }
      string = strings.toString(); // […, …, …, …] or [[…, …], […, …], […, …], […, …], …]
    } else {
      string = value.toString();
    }
    return string;
  }

  @Override
  public Component getTableCellRendererComponent(final JTable table, final Object value, final boolean isSelected, final boolean hasFocus, final int row,
      final int column) {
    final Component component = super.getTableCellRendererComponent(table, value, isSelected, hasFocus, row, column);
    if (value instanceof AccessibleObject) {
      final JLabel label = (JLabel) component;

      switch (column) {
        case ObjectTableModel.ACCESSIBLE_OBJECT_COLUMN:
          label.setEnabled(true);
          label.setToolTipText(value.toString());
          label.setFont(label.getFont().deriveFont(Font.BOLD));

          String name = "no name";

          if (value instanceof Field) {
            final Field field = (Field) value;
            name = field.getName();
            label.setFont(label.getFont().deriveFont(-Font.ITALIC));
            label.setText(field.getName());
          } else if (value instanceof Method) {
            final Method method = (Method) value;
            name = method.getName();
            label.setFont(label.getFont().deriveFont(Font.ITALIC));
            label.setText(method.getName());
          }

          // set color
          final int prefixLength = 3;
          final String key = name.length() >= prefixLength ? name.substring(0, prefixLength) : name;
          if (!this.colors.containsKey(key)) {
            this.colors.put(key, Color.getHSBColor(0.265f * this.colors.size(), 1f, 0.7f));
          }
          label.setForeground(this.colors.get(key));
          break;
        case ObjectTableModel.VALUE_COLUMN:
          label.setEnabled(table.getModel().isCellEditable(row, column));
          if (value instanceof Field) {
            final Field field = (Field) value;
            try {
              field.setAccessible(true);
              final Object fieldValue = field.get(((ObjectTableModel) table.getModel()).getObject());

              if (fieldValue == null) {
                // null?
                label.setText("null");
                label.setFont(label.getFont().deriveFont(Font.ITALIC + Font.BOLD));
              } else {
                // not null?
                final String string = getStringValue(fieldValue);
                label.setText(string);
                label.setFont(label.getFont().deriveFont(-Font.ITALIC - Font.BOLD));
              }

            } catch (IllegalArgumentException e) {
              e.printStackTrace();
            } catch (IllegalAccessException e) {
              e.printStackTrace();
            }
            label.setForeground(this.colors.get(table.getForeground()));
          }
          break;
        default:
      }
    } else {
      System.out.println(this.getClass().getName() + " should not be used for type " + value.getClass().getName());
    }
    return component;
  }
}
