package brain.gui.table;

import java.lang.reflect.AccessibleObject;
import java.lang.reflect.Field;
import java.lang.reflect.InvocationTargetException;
import java.lang.reflect.Method;
import java.lang.reflect.Modifier;
import java.lang.reflect.ParameterizedType;
import java.lang.reflect.Type;
import java.text.NumberFormat;
import java.text.ParseException;
import java.util.Arrays;
import java.util.Comparator;
import java.util.HashSet;
import java.util.List;
import java.util.Set;
import java.util.SortedSet;
import java.util.StringTokenizer;
import java.util.TreeSet;
import java.util.Vector;
import java.util.logging.Level;
import java.util.logging.Logger;

import javax.swing.table.DefaultTableModel;

import sun.reflect.generics.reflectiveObjects.ParameterizedTypeImpl;
import utils.Utils;

/**
 * ObjectTableModel handles access to columns, cells and rows. This Model is
 * designed to support the alter values of a Object implementing Module
 * interface.
 * 
 * @author Mathias Linkerhand
 * @version 2012-09-11
 * @since 2010
 */
public class ObjectTableModel extends DefaultTableModel {
  private static final long         serialVersionUID         = 1L;
  private Object                    object;
  public static final Set<Class<?>> EDITABLE_TYPES           = new HashSet<Class<?>>(Arrays.asList(new Class<?>[] { boolean.class, boolean[].class,
      boolean[][].class, boolean[][][].class, short.class, short[].class, short[][].class, short[][][].class, int.class, int[].class, int[][].class,
      int[][][].class, long.class, long[].class, long[][].class, long[][][].class, float.class, float[].class, float[][].class, float[][][].class,
      double.class, double[].class, double[][].class, double[][][].class, Number.class, String.class, String[].class, String[][].class, String[][][].class }));
  public static final Set<Class<?>> OBSERVABLE_TYPES         = new HashSet<Class<?>>(Arrays.asList(new Class<?>[] { boolean.class, short.class, int.class,
      long.class, float.class, double.class, Number.class, double[].class, int[].class, Iterable.class }));

  public static final int           ACCESSIBLE_OBJECT_COLUMN = 0;
  public static final int           TYPE_COLUMN              = 1;
  public static final int           VALUE_COLUMN             = 2;
  public static final int           OBSERVE_COLUMN           = 3;

  private static final Logger       logger                   = Logger.getLogger(ObjectTableModel.class.getName());

  public ObjectTableModel(final Object object) {
    final String[] columnIdentifiers = { "Name", "Type", "Value", "Observe" };
    this.setColumnIdentifiers(columnIdentifiers);
    this.object = object;

    this.initData();
  }

  protected void initData() {
    final Comparator<AccessibleObject> compareByNameComparator = new Comparator<AccessibleObject>() {
      @Override
      public int compare(AccessibleObject o1, AccessibleObject o2) {
        return this.getName(o1).compareTo(this.getName(o2));
      }

      public String getName(final AccessibleObject accessibleObject) {
        return accessibleObject instanceof Method ? ((Method) accessibleObject).getName() : (accessibleObject instanceof Field ? ((Field) accessibleObject)
            .getName() : "no name");
      }
    };
    final SortedSet<AccessibleObject> accessibleObjects = new TreeSet<AccessibleObject>(compareByNameComparator);
    accessibleObjects.addAll(ObjectTableModel.filterAccessibleObjects(this.object));

    final Vector<Vector<Object>> dataVector = new Vector<Vector<Object>>();
    for (final AccessibleObject accessibleObject : accessibleObjects) {
      final Vector<Object> dataRow = new Vector<Object>();
      dataRow.setSize(4); // 4 Columns table data: Name, Type, Value, Observable

      dataRow.setElementAt(accessibleObject, ObjectTableModel.ACCESSIBLE_OBJECT_COLUMN);
      dataRow.setElementAt(accessibleObject, ObjectTableModel.VALUE_COLUMN);
      if (accessibleObject instanceof Field) {
        dataRow.setElementAt(((Field) accessibleObject).getGenericType(), ObjectTableModel.TYPE_COLUMN);
      }
      dataRow.setElementAt(Boolean.FALSE, ObjectTableModel.OBSERVE_COLUMN);

      dataVector.add(dataRow);
    }
    this.setDataVector(dataVector, this.columnIdentifiers);
  }

  /**
   * This method filters out all member variables which can be accessed either
   * directly or by get and set method and can be edited in the table
   * 
   * @param object Object to filter fields from.
   * @return An Set of fields which match the criteria.
   */
  public static Set<AccessibleObject> filterAccessibleObjects(final Object object) {
    final Set<AccessibleObject> filteredAccessibleObjects = new HashSet<AccessibleObject>();

    final Set<Class<?>> typesWhiteList = new HashSet<Class<?>>();
    typesWhiteList.addAll(ObjectTableModel.EDITABLE_TYPES);
    typesWhiteList.addAll(ObjectTableModel.OBSERVABLE_TYPES);

    // add fields
    final Set<Field> fields = Utils.getDeclaredFields(object.getClass());
    for (final Field field : fields) {
      if (/*!field.getName().startsWith("$") &&*/ObjectTableModel.isTypeAccepted(field.getType(), typesWhiteList) && !Modifier.isFinal(field.getModifiers())) {
        filteredAccessibleObjects.add(field);
      }
    }

    /*
    // add methods
    final Method methods[] = object.getClass().getDeclaredMethods();
    for (final Method method : methods) {
      if (method.getName().startsWith("set")) {
        filteredAccessibleObjects.add(method);
      }
    }
    */

    return filteredAccessibleObjects;
  }

  public static Object convertStringToType(final String string, final Type type) throws ParseException {
    Object result = null;

    if (string.equals("null")) {
      result = null;
    }
    if (type instanceof Class<?>) {
      final Class<?> clazz = (Class<?>) type;
      if (clazz.equals(String.class)) {
        result = string;
      } else if (clazz.equals(short.class)) {
        result = NumberFormat.getNumberInstance().parse(string).shortValue();
      } else if (clazz.equals(int.class)) {
        result = NumberFormat.getNumberInstance().parse(string).intValue();
      } else if (clazz.equals(long.class)) {
        result = NumberFormat.getNumberInstance().parse(string).longValue();
      } else if (clazz.equals(float.class)) {
        result = NumberFormat.getNumberInstance().parse(string).floatValue();
      } else if (clazz.equals(double.class)) {
        result = NumberFormat.getNumberInstance().parse(string).doubleValue();
      } else if (clazz.equals(boolean.class)) {
        result = Boolean.valueOf(string);
      } else if (clazz.equals(double[].class)) {
        final String trimedString = string.trim();
        final String[] tokens = trimedString.substring(1, trimedString.length() - 1).split(",");
        final double[] array = new double[tokens.length];
        for (int i = 0; i < tokens.length; i++) {
          array[i] = Double.parseDouble(tokens[i]);
        }
        result = array;
      } else if (clazz.equals(double[][].class)) {
        List<?> list = ObjectTableModel.parseStringToMultiDimensionalList(string.toString());
        if (list.isEmpty()) {
          result = new double[0][0];
        } else {
          result = new double[list.size()][((List<?>) list.get(0)).size()];
          for (int i = 0; i < list.size(); i++) {
            final List<?> subList = (List<?>) list.get(i);
            for (int j = 0; j < subList.size(); j++) {
              ((double[][]) result)[i][j] = Double.valueOf((String) subList.get(j));
            }
          }
        }
      } else if (clazz.equals(double[][][].class)) {
        List<?> list = ObjectTableModel.parseStringToMultiDimensionalList(string.toString());
        if (list.isEmpty()) {
          result = new double[0][0][0];
        } else {
          result = new double[list.size()][((List<?>) list.get(0)).size()][((List<?>) ((List<?>) list.get(0)).get(0)).size()];
          for (int i = 0; i < list.size(); i++) {
            final List<?> subList = (List<?>) list.get(i);
            for (int j = 0; j < subList.size(); j++) {
              final List<?> subSubList = (List<?>) subList.get(j);
              for (int k = 0; k < subSubList.size(); k++) {
                ((double[][][]) result)[i][j][k] = Double.valueOf((String) subSubList.get(k));
              }
            }
          }
        }
      }
    } else if (type instanceof ParameterizedTypeImpl) {
      if (type.toString().equals("java.util.List<java.lang.Double>")) {
        List<Double> list = new Vector<Double>();
        for (final Object v : ObjectTableModel.parseStringToMultiDimensionalList(string.toString())) {
          list.add(Double.valueOf((String) v));
        }
        result = list;
      } else if (type.toString().equals("java.util.Set<java.lang.Integer>")) {
        Set<Integer> set = new HashSet<Integer>();
        for (final Object v : ObjectTableModel.parseStringToMultiDimensionalList(string.toString())) {
          set.add(Integer.valueOf((String) v));
        }
        result = set;
      } else if (type.toString().equals("java.util.List<java.util.List<java.lang.Double>>")) {
        List<List<Double>> list = new Vector<List<Double>>();
        for (final Object row : ObjectTableModel.parseStringToMultiDimensionalList(string.toString())) {
          List<Double> subList = new Vector<Double>();
          list.add(subList);
          for (final Object v : (List<?>) row) {
            subList.add(Double.valueOf((String) v));
          }
        }
        result = list;
      } else if (type.toString().equals("java.util.Set<java.util.Set<java.lang.Integer>>")) {
        Set<Set<Integer>> set = new HashSet<Set<Integer>>();
        for (final Object row : ObjectTableModel.parseStringToMultiDimensionalList(string.toString())) {
          Set<Integer> subSet = new HashSet<Integer>();
          set.add(subSet);
          for (final Object v : (List<?>) row) {
            subSet.add(Integer.valueOf((String) v));
          }
        }
        result = set;
      }
    } else {
      result = string;
      throw new IllegalStateException("Converter for Type " + type + " not implemented.");
    }
    return result;
  }

  @Override
  public void setValueAt(final Object value, final int modelRow, final int modelColumn) {
    switch (modelColumn) {
      case ObjectTableModel.VALUE_COLUMN:
        final Field field = (Field) this.getValueAt(modelRow, ObjectTableModel.ACCESSIBLE_OBJECT_COLUMN);
        try {
          final Object correctedValue = ObjectTableModel.convertStringToType((String) value, field.getGenericType());
          try {
            this.setObjectFieldValue(field, correctedValue);
          } catch (IllegalAccessException e) {
            e.printStackTrace();
            return;
          }
        } catch (ParseException e1) {
          // TODO Auto-generated catch block
          e1.printStackTrace();
        }
        break;
      default:
        super.setValueAt(value, modelRow, modelColumn);
        break;
    }
  }

  /**
   * Assigns the given value to the specified field of this table’s object.
   * 
   * First a properly named setter method is called. If this fails or there is no such method the value will be set directly.
   * 
   * @param field
   * @param value
   * @throws IllegalAccessException
   */
  protected void setObjectFieldValue(final Field field, final Object value) throws IllegalAccessException {
    try {
      // Try setter method first.
      final String methodName = "set" + field.getName().substring(0, 1).toUpperCase() + field.getName().substring(1); // convert “valueA” to “setValueA”
      final Method methodToInvoke = this.getObject().getClass().getDeclaredMethod(methodName, field.getType()); // TODO: also handle superclass methods
      final Object result = methodToInvoke.invoke(this.getObject(), value);

      // Using the setter method the complete object could change due to the commands inside of the setter method. Therefore the whole table has to be updated.
      this.fireTableDataChanged();

      logger.log(Level.FINE, "method “" + methodToInvoke + "” invoked, result: " + result);
    } catch (SecurityException e) {
      e.printStackTrace();
    } catch (NoSuchMethodException e) {
      // The setter method failed.
      // Set field directly.
      logger.log(Level.FINE, "set field directly");
      field.setAccessible(true);
      field.set(this.getObject(), value);
    } catch (IllegalArgumentException e) {
      e.printStackTrace();
    } catch (InvocationTargetException e) {
      e.printStackTrace();
    }
  }

  @SuppressWarnings({ "unchecked", "rawtypes" })
  public static List parseStringToMultiDimensionalList(final String string) {
    // example: [[[1, 2], [3, 4]], [[5, 6], [7, 8]]]

    StringTokenizer tokenizer = new StringTokenizer(string, ", []", true);
    List<List> lists = new Vector<List>();
    int listIndex = -1;
    while (tokenizer.hasMoreTokens()) {
      final String nextToken = tokenizer.nextToken();
      if (nextToken.equals("[")) {
        listIndex++;
        final List subList = new Vector();
        lists.add(listIndex, subList);
      } else if (nextToken.equals("]")) {
        listIndex--;
        if (listIndex >= 0) {
          lists.get(listIndex).add(lists.get(listIndex + 1));
        }
      } else if (nextToken.equals(",")) {
        continue;
      } else if (nextToken.equals(" ")) {
        continue;
      } else {
        lists.get(listIndex).add(nextToken);
      }
    }

    return lists.get(0);
  }

  /**
   * Figures out whether the type is acceptable (editable, observable, …) for the GUI.
   * 
   * @param type
   * @param types
   * @return true if the type is accepted, false otherwise
   */
  public static boolean isTypeAccepted(final Type type, final Set<Class<?>> types) {
    for (final Class<?> clas : types) {
      if (type instanceof Class<?> && clas.isAssignableFrom((Class<?>) type)) {
        return true;
      } else if (type instanceof ParameterizedType) {
        final ParameterizedType parameterizedType = (ParameterizedType) type;
        if (type instanceof Class<?> && clas.isAssignableFrom((Class<?>) parameterizedType.getRawType())) {
          return true;
        } else if (List.class.isAssignableFrom((Class<?>) parameterizedType.getRawType())
            || Set.class.isAssignableFrom((Class<?>) parameterizedType.getRawType())) {
          // allow editing of lists and sets
          for (final Type typeArgument : parameterizedType.getActualTypeArguments()) {
            if (!isTypeAccepted(typeArgument, types)) {
              return false;
            }
          }
          return true;
        }
      }
    }
    return false;
  }

  @Override
  public boolean isCellEditable(int row, int column) {
    if (this.getValueAt(row, column) == null) {
      return false;
    }
    final Type type = (Type) this.getValueAt(row, ObjectTableModel.TYPE_COLUMN);

    switch (column) {
      case ObjectTableModel.ACCESSIBLE_OBJECT_COLUMN:
        // name column not editable
        return false;
      case ObjectTableModel.TYPE_COLUMN:
        // type column not editable
        return false;
      case ObjectTableModel.VALUE_COLUMN:
        // value column
        return ObjectTableModel.isTypeAccepted(type, ObjectTableModel.EDITABLE_TYPES);
      case ObjectTableModel.OBSERVE_COLUMN:
        // observable column
        return ObjectTableModel.isTypeAccepted(type, ObjectTableModel.OBSERVABLE_TYPES);
      default:
        throw new IllegalStateException("The program should not come to this point.");
    }
  }

  @Override
  public Class<?> getColumnClass(final int modelColumn) {
    switch (modelColumn) {
      case ObjectTableModel.ACCESSIBLE_OBJECT_COLUMN:
        return AccessibleObject.class;
      case ObjectTableModel.TYPE_COLUMN:
        return Type.class;
      case ObjectTableModel.VALUE_COLUMN:
        return AccessibleObject.class;
      case ObjectTableModel.OBSERVE_COLUMN:
        return Boolean.class;
      default:
        throw new IllegalStateException("The program should not come to this point.");
    }
  }

  public Object getObject() {
    return this.object;
  }
}