package calculation;

import java.io.File;
import java.io.FileNotFoundException;
import java.util.HashSet;
import java.util.Locale;
import java.util.Random;
import java.util.Set;

import brain.Brain;
import brain.ConstantSignalModule;
import brain.LeakyIntegratorWithIntrinsicPlasticityModule;
import brain.Module;
import brain.NoiseModule;

/**
 * Simulates a single neuron and its intrinsic plasticity.
 * 
 * @author Mathias Linkerhand
 * @version 2012-01-04
 */
public class IntrinsicPlasticity {

  public static void createSingleAdaptiveRelisiWithInput() {
    final Set<double[]> targetDistributions = new HashSet<double[]>();
    targetDistributions.add(new double[] { 0.0, 0.0 }); // flat
    targetDistributions.add(new double[] { -10.0, 0.0 }); // left convex
    targetDistributions.add(new double[] { 10.0, 0.0 }); // right convex
    targetDistributions.add(new double[] { -10.0, 10.0 }); // left & right convex symmetrical
    targetDistributions.add(new double[] { 20.0, -20.0 }); // hill
    targetDistributions.add(new double[] { -20.0, 20.0 }); // left & right convex symmetrical
    targetDistributions.add(new double[] { -20.0, 19.0 }); // left & right convex left skewed, stochastic escape
    targetDistributions.add(new double[] { -20.0, 18.5 }); // left & right convex left skewed

    final double[] intrinsicPlasticityAdaptationParameters = new double[] { 1e-5, 1e-4, 5e-4, 1e-3, 5e-3, 1e-2, 2e-2, 3e-2, 4e-2, 5e-2, 1e-1 };

    for (final double[] targetDistribution : targetDistributions) {
      for (final double intrinsicAdaptationParameter : intrinsicPlasticityAdaptationParameters) {
        final Brain brain = new Brain();
        final double timeDelta = 1e0;
        final double timeMax = 1e8;
        brain.setTimeDeltaAndDuration(timeDelta, timeMax);
        brain.setLogInterval(timeMax / 1e6); // 1e6 data points

        final Random random = new Random(0);
        final int size = 1;
        final double linkProbability = 0.0;
        final LeakyIntegratorWithIntrinsicPlasticityModule relisi = new LeakyIntegratorWithIntrinsicPlasticityModule(size, linkProbability, random);
        brain.addModule(relisi);
        brain.addLogger(relisi, "getActivities");
        brain.addLogger(relisi, "getMembranePotentials");
        brain.addLogger(relisi, "getTransferFunctionParameters");

        // intrinsic learning rate
        relisi.setIntrinsicAdaptationLearningRates(intrinsicAdaptationParameter);

        // transfer function
        relisi.setTransferFunctionParameters(1.0, 1.0); // gain, threshold
        //settransferfunctiontype

        // target distribution
        relisi.setTargetDistributionParameters(targetDistribution[0], targetDistribution[1]);
        final double leak = 1e0;
        relisi.setLeaks(leak);
        relisi.setName("Relisi");
        relisi.normalizeLinks();

        relisi.setPositiveLinkAverage(1.0);
        for (int i = 0; i < relisi.getSize(); i++) {
          for (int j = 0; j < relisi.getSize(); j++) {
            final double link = relisi.getLink(i, j);
            if (link < 0) {
              relisi.setLink(i, j, -1.0);
            }
          }
        }

        // input interval range, standard: [0, 1]
        final double inputBias = 0.0; // first shift by bias
        final double inputGain = 10.0; // then multiply by gain

        final boolean withInput = true;
        if (withInput) {
          int inputType = 1; // 1 = white noise
          final Module input;
          switch (inputType) {
            case 0:
              input = new NoiseModule(NoiseModule.GAUSSIAN_NOISE, 10, random);
              input.setBias(1);
              ((NoiseModule) input).setChangeInterval(1.0);
              break;
            case 1:
              input = new NoiseModule(NoiseModule.WHITE_NOISE, 10, random);
              input.setBias(inputBias);
              ((NoiseModule) input).setChangeInterval(1.0); // duration of constant tableaus
              break;
            case 2:
              input = new ConstantSignalModule(0.9);
              input.setBias(0);
              break;
            default:
              throw new IllegalArgumentException("wrong input type");
          }
          input.setName("Input");
          brain.addModule(input);
          brain.addLogger(input, "getActivities");

          relisi.addAfferentModule(input);
          for (int i = 0; i < relisi.getSize(); i++) {
            for (int j = 0; j < input.getSize(); j++) {
              // gain
              relisi.setAfferentLink(input, i, j, inputGain); // gain
            }
          }
        }

        final String filename = String.format(Locale.ENGLISH,
            "IntrinsicPlasticity Γ=%.1e Δt=%.1e t_max=%.1e λ=(%.2e,%.2e) g=1 ε=%.1e Ξ=[%.1e,%.1e] a=%.1e b=%.1e.brain.xml", relisi.getLeak(0), brain
                .getTimeDelta(), brain.getTimeStepNumber() * brain.getTimeDelta(), relisi.getTargetDistributionParameters(0).get(0), relisi
                .getTargetDistributionParameters(0).get(1), relisi.getIntrinsicAdaptationLearningRates(0).get(0), 0.0 + inputBias, (1.0 + inputBias) * inputGain, relisi
                .getTransferFunctionParameters(0).get(0), relisi.getTransferFunctionParameters(0).get(1));
        final File file = new File(Brain.DEFAULT_DATA_OUTPUT_DIRECTORY, filename);
        brain.save(file);
        System.out.println(file + " saved.");
      }
    }
  }

  public static void main(String[] args) throws FileNotFoundException {
    createSingleAdaptiveRelisiWithInput();
  }
}