/*
 * simulate: java -cp .:bin:lib/* calculation.LeakyIntegratorIntrinsicPlasticity
 * evaluate: date --utc; time gnuplot -e
 * "dir='data/output/LeakyIntegratorIntrinsicPlasticity'"
 * bin/network-leaky-integrator-time-series.gnuplot
 */
package calculation;

import java.io.File;
import java.io.FileNotFoundException;
import java.util.List;
import java.util.Random;
import java.util.Vector;

import brain.LeakyIntegratorWithIntrinsicPlasticityModule;
import brain.Brain;
import brain.ConstantSignalModule;
import brain.Module;
import brain.NoiseModule;

/**
 * This class implements a simulation of the leaky integrator with several
 * options: constant input, random input, autapse, intra-module connections.
 * Each option can be set using the corresponding intra- or inter-module
 * synaptic connections strengths.
 * 
 * @author Mathias Linkerhand
 * @version 2012-05-09
 */
public class LeakyIntegratorIntrinsicPlasticity {

  public static void simulateLeakyIntegratorIntrinsicPlasticity() {
    /*
     * random number generator setup
     */
    final long randomSeed = 8765094544301429395L; // or: 8765094544301429395L, new Random().nextLong()
    /*
     * seed for good autapse trial: 8765094544301429395L
     */
    System.out.println("random seed: " + randomSeed);
    final Random random = new Random(randomSeed);

    /*
     * brain setup
     */
    final Brain brain = new Brain();
    final double timeDelta = 1e-4;
    final double timeMax = 1e3;
    brain.setTimeDeltaAndDuration(timeDelta, timeMax);
    brain.setLogInterval(Math.max(timeDelta, timeMax / 1e5)); // max 1e5 data
    // points
    final File dataOutputDirectory = new File(Brain.DEFAULT_DATA_OUTPUT_DIRECTORY, LeakyIntegratorIntrinsicPlasticity.class.getSimpleName());
    dataOutputDirectory.mkdirs();
    brain.setDataOutputDirectory(dataOutputDirectory);

    /*
     * constant input module setup
     */
    final Module constantInput;
    {
      final double activity = 0.9;
      constantInput = new ConstantSignalModule(activity);
      constantInput.setBias(0);
      constantInput.setName("constant input");
    }

    /*
     * random noise input module setup
     */
    final Module randomInput;
    {
      final double inputBias = 0.0; // first shift by bias
      final double changeInterval = 1.0; // duration of constant plateaus

      final int noiseType = 1;
      switch (noiseType) {
        case 0:
          randomInput = new NoiseModule(NoiseModule.GAUSSIAN_NOISE, changeInterval, random);
          randomInput.setBias(inputBias);
          break;
        case 1:
          randomInput = new NoiseModule(NoiseModule.WHITE_NOISE, changeInterval, random);
          randomInput.setBias(inputBias);
          break;
        default:
          throw new IllegalArgumentException("wrong input type");
      }
      randomInput.setName("random input");
    }

    /*
     * leaky integrator module setup
     */
    final int size = 8;
    final double[][] adjacencyMatrix = new double[size][size];
    final LeakyIntegratorWithIntrinsicPlasticityModule leakyIntegrator = new LeakyIntegratorWithIntrinsicPlasticityModule(adjacencyMatrix, random);
    leakyIntegrator.setName("leaky integrator");

    // some target distributions
    final List<double[]> targetDistributions = new Vector<double[]>();
    targetDistributions.add(new double[] { 0.0, 0.0 }); // flat
    targetDistributions.add(new double[] { -10.0, 0.0 }); // left convex
    targetDistributions.add(new double[] { 10.0, 0.0 }); // right convex
    targetDistributions.add(new double[] { -10.0, 10.0 }); // left & right convex symmetrical
    targetDistributions.add(new double[] { 20.0, -20.0 }); // hill (concave)
    targetDistributions.add(new double[] { -20.0, 20.0 }); // left & right convex symmetrical
    targetDistributions.add(new double[] { -20.0, 19.0 }); // left & right convex left skewed
    targetDistributions.add(new double[] { -20.0, 18.5 }); // left & right convex left skewed

    for (int neuron = 0; neuron < leakyIntegrator.getSize(); neuron++) {
      // intrinsic adaptation learning rate $\epsilon_a$, $\epsilon_b$
      final double intrinsicAdaptationLearningRate = 1e-1;
      final double currentIntrinsicAdaptationLearningRate = intrinsicAdaptationLearningRate;
      leakyIntegrator.setIntrinsicAdaptationLearningRates(neuron, currentIntrinsicAdaptationLearningRate);
      leakyIntegrator.setIntrinsicAdaptationLearningRates(neuron, 0.0, intrinsicAdaptationLearningRate / Math.pow(2, neuron)); // TODO: remove to correct!x

      // membrane potential $x$
      final double defaultMembranePotential = 0.1;
      leakyIntegrator.setMembranePotential(neuron, defaultMembranePotential);

      // transfer function parameters gain $a$ and threshold $b$
      final double defaultTransferFunctionGain = 4.5; // $a$
      final double defaultTransferFunctionThreshold = 1.0; // $b$
      leakyIntegrator.setTransferFunctionParameters(neuron, defaultTransferFunctionGain, defaultTransferFunctionThreshold); // gain, threshold

      // target distribution parameters $\lambda_1$, $\lambda_2$
      leakyIntegrator.setTargetDistributionParameters(neuron, targetDistributions.get(neuron % leakyIntegrator.getSize())[0], targetDistributions.get(neuron
          % leakyIntegrator.getSize())[1]);
      leakyIntegrator.setTargetDistributionParameters(neuron, targetDistributions.get(0)[0], targetDistributions.get(0)[1]); // TODO: remove to correct!

      // leak $\Gamma$
      final double defaultLeak = 1e0;
      leakyIntegrator.setLeak(neuron, defaultLeak);

      // intra-module synaptic connections
      for (int afferentNeuron = 0; afferentNeuron < leakyIntegrator.getSize(); afferentNeuron++) {
        if (neuron == afferentNeuron) {
          // autapse
          final double autapseStrength = 1.0; // w = 0.0, 1.0, random.nextDouble()
          leakyIntegrator.setLink(neuron, neuron, autapseStrength);
        } else {
          final double synapticStrength = 0.0; // w = 0.0, 1.0, random.nextDouble()
          leakyIntegrator.setLink(neuron, afferentNeuron, synapticStrength);
        }
      }
    }

    // normalize links
    final boolean normalizeLinks = false;
    if (normalizeLinks) {
      leakyIntegrator.normalizeLinks();
      leakyIntegrator.setPositiveLinkAverage(1.0);
      for (int i = 0; i < leakyIntegrator.getSize(); i++) {
        for (int j = 0; j < leakyIntegrator.getSize(); j++) {
          final double link = leakyIntegrator.getLink(i, j);
          if (link < 0) {
            leakyIntegrator.setLink(i, j, -1.0);
          }
        }
      }
    }

    /*
     * connect modules
     */
    leakyIntegrator.addAfferentModule(constantInput);
    final double constantInputToLeakyIntegratorSynapticStrength = 0.0; // or: 0.0, 1.0, …
    leakyIntegrator.setAfferentLinks(constantInput, constantInputToLeakyIntegratorSynapticStrength);

    leakyIntegrator.addAfferentModule(randomInput);
    final double randomInputToLeakyIntegratorSynapticStrength = 0.0; // or: 0.0, 1.0, 10.0, …
    leakyIntegrator.setAfferentLinks(randomInput, randomInputToLeakyIntegratorSynapticStrength);

    /*
     * register modules and their loggers
     */
    brain.addModule(leakyIntegrator);
    brain.addLogger(leakyIntegrator, "getActivities");
    brain.addLogger(leakyIntegrator, "getMembranePotentials");
    brain.addLogger(leakyIntegrator, "getTransferFunctionParameters");
    brain.addLogger(leakyIntegrator, "getTransferFunctionParameterChanges");

    brain.addModule(constantInput);
    brain.addLogger(constantInput, "getActivities");

    brain.addModule(randomInput);
    brain.addLogger(randomInput, "getActivities");

    // show setup
    System.out.println("firing rates: " + leakyIntegrator.getActivities());
    System.out.println(leakyIntegrator.getCliques().size() + " clique(s): " + leakyIntegrator.getCliques());
    System.out.println("intra-module connections of " + leakyIntegrator.getName() + ": " + leakyIntegrator.getLinks());
    for (final Module afferentModule : leakyIntegrator.getAfferentModules()) {
      System.out.println("inter-module connections from " + afferentModule.getName() + " to " + leakyIntegrator.getName() + ": "
          + leakyIntegrator.getAfferentLinks(afferentModule));
    }

    // simulation
    brain.run();
  }

  public static void main(String[] args) throws FileNotFoundException {
    simulateLeakyIntegratorIntrinsicPlasticity();
  }
}