package calculation;

import java.io.File;
import java.io.FileNotFoundException;
import java.util.List;
import java.util.Random;
import java.util.Vector;

import utils.Graphs;

import brain.LeakyIntegratorWithIntrinsicPlasticityModule;
import brain.Brain;
import brain.ConstantSignalModule;
import brain.LeakyIntegratorModule;
import brain.Module;
import brain.NoiseModule;
import brain.gui.BrainGUI;

/**
 * This class implements a simulation of the leaky integrator with several
 * options: constant input, random input, autapse, intra-module connections.
 * Each option can be set using the corresponding intra- or inter-module
 * synaptic connections strengths.
 * 
 * running the simulation:
 * 
 * <pre>
 * java -cp .:bin:lib/*
 * </pre>
 * 
 * creating simulation plots:
 * 
 * <pre>
 * gnuplot -e "dir='data/output/LeakyIntegratorIntrinsicPlasticityCliqueGraph'" bin/network-leaky-integrator-time-series.gnuplot
 * </pre>
 * 
 * creating graph plots:
 * 
 * <pre>
 * bin/graphviz2pdf.sh data/output/LeakyIntegratorIntrinsicPlasticityCliqueGraph
 * </pre>
 * 
 * @author Mathias Linkerhand
 * @version 2012-05-30
 */
public final class LeakyIntegratorIntrinsicPlasticityCliqueGraph {

  public static void simulateLeakyIntegratorIntrinsicPlasticity() {
    /*
     * random number generator setup
     */
    final long randomSeed = new Random().nextLong(); // or: 8765094544301429395L, new Random().nextLong()
    System.out.println("random seed: " + randomSeed);
    final Random random = new Random(randomSeed);

    /*
     * brain setup
     */
    final Brain brain = new Brain();
    final double timeDelta = 1e-3;
    final double timeMax = 1e3;
    brain.setTimeDeltaAndDuration(timeDelta, timeMax);
    brain.setLogInterval(Math.max(timeDelta, timeMax / 1e5)); // max 1e5 data points
    final File dataOutputDirectory = new File(Brain.DEFAULT_DATA_OUTPUT_DIRECTORY, LeakyIntegratorIntrinsicPlasticityCliqueGraph.class.getSimpleName());
    dataOutputDirectory.mkdirs();
    brain.setDataOutputDirectory(dataOutputDirectory);

    /*
     * constant input module setup
     */
    final Module constantInput;
    {
      final double activity = 0.9;
      constantInput = new ConstantSignalModule(activity);
      constantInput.setBias(0);
      constantInput.setName("constant input");
    }

    /*
     * random noise input module setup
     */
    final Module randomInput;
    {
      final double inputBias = 0.0; // first shift by bias
      final double changeInterval = 1.0; // duration of constant plateaus
      randomInput = new NoiseModule(NoiseModule.WHITE_NOISE, changeInterval, random); // alternatives: {NoiseModule.GAUSSIAN_NOISE, NoiseModule.WHITE_NOISE, …}
      randomInput.setBias(inputBias);
      randomInput.setName("random input");
    }

    /*
     * leaky integrator module setup
     */
    final int size = 3;
    //final double[][] adjacencyMatrix = new double[size][size];
    //final double[][] adjacencyMatrix = Graphs.addInhibitoryLinks(Graphs.getCircleGraphAdjacencyMatrix(size)); // or: Graphs.getCircleGraphAdjacencyMatrix(size), Graphs.getTypicalDHAN4CliqueCircleGraphAdjacencyMatrix(size)
    //final double[][] adjacencyMatrix = Graphs.addInhibitoryLinks(Graphs.getTypicalDHAN4CliqueCircleGraphAdjacencyMatrix(size));
    final double[][] adjacencyMatrix = Graphs.addInhibitoryLinks(Graphs.getChainGraphAdjacencyMatrix(size));
    final LeakyIntegratorWithIntrinsicPlasticityModule leakyIntegrator = new LeakyIntegratorWithIntrinsicPlasticityModule(adjacencyMatrix, random);
    leakyIntegrator.setName("leaky integrator");

    // some target distributions
    final List<double[]> targetDistributions = new Vector<double[]>();
    targetDistributions.add(new double[] { 0.0, 0.0 }); // flat
    targetDistributions.add(new double[] { -10.0, 0.0 }); // left convex
    targetDistributions.add(new double[] { 10.0, 0.0 }); // right convex
    targetDistributions.add(new double[] { -10.0, 10.0 }); // left & right convex symmetrical
    targetDistributions.add(new double[] { 20.0, -20.0 }); // hill (concave)
    targetDistributions.add(new double[] { -20.0, 20.0 }); // left & right convex symmetrical
    targetDistributions.add(new double[] { -20.0, 19.0 }); // left & right convex left skewed
    targetDistributions.add(new double[] { -20.0, 18.5 }); // left & right convex left skewed

    for (int neuron = 0; neuron < leakyIntegrator.getSize(); neuron++) {
      // intrinsic adaptation learning rate $\epsilon_a$, $\epsilon_b$
      final double intrinsicAdaptationLearningRate = 1e-1;
      final double currentIntrinsicAdaptationLearningRate = intrinsicAdaptationLearningRate;
      leakyIntegrator.setIntrinsicAdaptationLearningRates(neuron, currentIntrinsicAdaptationLearningRate);

      // membrane potential $x$
      final double defaultMembranePotential = random.nextDouble(); // or: 0.0, 0.1, 0.5, random.nextDouble()
      leakyIntegrator.setMembranePotential(neuron, defaultMembranePotential);

      // transfer function parameters gain $a$ and threshold $b$
      final double defaultTransferFunctionGain = 1.0; // $a$
      final double defaultTransferFunctionThreshold = 0.0; // $b$
      leakyIntegrator.setTransferFunctionParameters(neuron, defaultTransferFunctionGain, defaultTransferFunctionThreshold); // gain,
      // threshold

      // target distribution parameters $\lambda_1$, $\lambda_2$
      final int targetDistribution = 3; // neuron % leakyIntegrator.getSize()
      final double[] targetDistributionParameters = targetDistributions.get(targetDistribution); // 
      leakyIntegrator.setTargetDistributionParameters(neuron, targetDistributionParameters[0], targetDistributionParameters[1]);

      // leak $\Gamma$
      final double defaultLeak = 1e0;
      leakyIntegrator.setLeak(neuron, defaultLeak);
      /*
       * // intra-module synaptic connections for (int afferentNeuron = 0;
       * afferentNeuron < leakyIntegrator.getSize(); afferentNeuron++) { if
       * (neuron == afferentNeuron) { // autapse final double autapseStrength =
       * 0.0; // w = 0.0, 1.0, random.nextDouble()
       * leakyIntegrator.setLink(neuron, neuron, autapseStrength); } else {
       * final double synapticStrength = 0.0; // w = 0.0, 1.0,
       * random.nextDouble() leakyIntegrator.setLink(neuron, afferentNeuron,
       * synapticStrength); } }
       */
    }
    /*
     * leakyIntegrator.setLink(0, 1, 1.0, true); leakyIntegrator.setLink(0, 2,
     * 1.0, true); leakyIntegrator.setLink(0, 3, 1.0, true);
     * leakyIntegrator.setLink(0, 7, 1.0, true); leakyIntegrator.setLink(1, 2,
     * 1.0, true); leakyIntegrator.setLink(2, 3, 1.0, true);
     * leakyIntegrator.setLink(2, 7, 1.0, true); leakyIntegrator.setLink(3, 4,
     * 1.0, true); leakyIntegrator.setLink(3, 5, 1.0, true);
     * leakyIntegrator.setLink(3, 7, 1.0, true); leakyIntegrator.setLink(4, 5,
     * 1.0, true); leakyIntegrator.setLink(5, 6, 1.0, true);
     * leakyIntegrator.setLink(5, 7, 1.0, true); leakyIntegrator.setLink(6, 7,
     * 1.0, true);
     */
    /*
     * // add inhibitory links for (int i = 0; i < leakyIntegrator.getSize();
     * i++) { for (int j = 0; j < leakyIntegrator.getSize(); j++) { if (i == j)
     * { // autapse } else if (leakyIntegrator.getLink(i, j) == 0) {
     * leakyIntegrator.setLink(i, j, -1.0); } } }
     */
    // normalize links
    final boolean normalizeLinks = true;
    if (normalizeLinks) {
      leakyIntegrator.normalizeLinks();
      leakyIntegrator.setPositiveLinkAverage(8.0);
      for (int i = 0; i < leakyIntegrator.getSize(); i++) {
        for (int j = 0; j < leakyIntegrator.getSize(); j++) {
          final double link = leakyIntegrator.getLink(i, j);
          if (link < 0) {
            leakyIntegrator.setLink(i, j, -5.0);
          }
        }
      }
    }

    /*
     * connect modules
     */
    leakyIntegrator.addAfferentModule(constantInput);
    final double constantInputToLeakyIntegratorSynapticStrength = 0.0; // or: 0.0, 1.0, …
    leakyIntegrator.setAfferentLinks(constantInput, constantInputToLeakyIntegratorSynapticStrength);

    leakyIntegrator.addAfferentModule(randomInput);
    final double randomInputToLeakyIntegratorSynapticStrength = 0.0; // or: 0.0, 1.0, 10.0, …
    leakyIntegrator.setAfferentLinks(randomInput, randomInputToLeakyIntegratorSynapticStrength);

    /*
     * register modules and their loggers
     */
    brain.addModule(leakyIntegrator);
    brain.addLogger(leakyIntegrator, "getActivities");
    brain.addLogger(leakyIntegrator, "getMembranePotentials");
    brain.addLogger(leakyIntegrator, "getTransferFunctionParameters");
    brain.addLogger(leakyIntegrator, "getTransferFunctionParameterChanges");
    //brain.addLogger(leakyIntegrator, "getActiveClique");

    brain.addModule(constantInput);
    brain.addLogger(constantInput, "getActivities");

    brain.addModule(randomInput);
    brain.addLogger(randomInput, "getActivities");

    /*
     * specific setup (to be removed? check it!)
     */
    leakyIntegrator.setActivities(leakyIntegrator.getCliques().iterator().next(), 0.9, 0.1);
    leakyIntegrator.setIntrinsicPlasticity(false);
    leakyIntegrator.setPositiveLinks(0.3);
    leakyIntegrator.setNegativeLinks(-1.0);
    //leakyIntegrator.setSilentInhibition(LeakyIntegratorModule.SilentInhibition.SILENT_INHIBITION_WITH_RESPECT_TO_MEMBRANE_POTENTIAL);
    leakyIntegrator.setSilentInhibition(LeakyIntegratorModule.SilentInhibition.INHIBITION_NOT_SILENT);

    /*
     * show setup
     */
    System.out.println("firing rates: " + leakyIntegrator.getActivities());
    System.out.println(leakyIntegrator.getCliques().size() + " clique(s): " + leakyIntegrator.getCliques());
    System.out.println("intra-module connections of " + leakyIntegrator.getName() + ": " + leakyIntegrator.getLinks());
    for (final Module afferentModule : leakyIntegrator.getAfferentModules()) {
      System.out.println("inter-module connections from " + afferentModule.getName() + " to " + leakyIntegrator.getName() + ": "
          + leakyIntegrator.getAfferentLinks(afferentModule));
    }

    /*
     * simulation
     */
    //brain.run();
    new BrainGUI(brain).setVisible(true);
  }

  public static void main(String[] args) throws FileNotFoundException {
    simulateLeakyIntegratorIntrinsicPlasticity();
  }
}