package calculation;

import java.io.BufferedOutputStream;
import java.io.File;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.PrintStream;
import java.util.Collections;
import java.util.List;
import java.util.Locale;
import java.util.Random;
import java.util.Set;
import java.util.Vector;
import java.util.concurrent.Callable;
import java.util.concurrent.ExecutionException;
import java.util.concurrent.ExecutorService;
import java.util.concurrent.Executors;
import java.util.concurrent.Future;

import utils.Graphs;
import utils.Utils;
import brain.LeakyIntegratorWithIntrinsicPlasticityModule;
import brain.Brain;
import brain.LeakyIntegratorModule;
import brain.Module;

public class PatternStability implements Callable<Double> {
  private final Module       module;
  private final Set<Integer> clique;

  public PatternStability(final byte[] serializedModule, final Set<Integer> clique, final double exhibitoryFactor, final double inhibitoryFactor) {
    final Module module = (Module) Utils.bytesToObject(serializedModule);
    module.multiplyExhibitoryAndInhibitoryLinksSeparately(exhibitoryFactor, inhibitoryFactor);

    final double highActivity = 0.9;
    final double lowActivity = 1.0 - highActivity;
    module.setActivities(clique, highActivity, lowActivity); // activate the clique, deactivate the rest

    this.module = module;
    this.clique = clique;
  }

  @Override
  public Double call() throws Exception {
    final double Δt = 1e-2;
    final double t_min = 0.0;
    final double t_max = 1e3;

    final int testForActiveCliqueInterval = 1;

    for (long timeStep = 0; timeStep < (t_max - t_min) / Δt; timeStep++) {
      final double t = t_min + (double) timeStep * (t_max - t_min) / Δt;
      this.module.performOneTimeStep(Δt);

      // Test for active clique after a certain time.
      if (timeStep / Δt % testForActiveCliqueInterval == 0) {
        final Set<Integer> activeClique = this.module.getActiveClique();
        if (activeClique == null || !this.clique.equals(activeClique)) {
          // Clique decayed at time t.
          return t;
        }
      }
    }

    // Clique remained active till time t.
    throw new Exception("Clique did not decay till t = " + t_max + ".");
  }

  public Set<Integer> getClique() {
    return this.clique;
  }

  public static void cliqueStability(final byte[] serializedModule, final Set<Set<Integer>> cliques, final double w, final double z,
      final ExecutorService executor, final PrintStream out) throws InterruptedException {
    // init calculation
    final List<Callable<Double>> tasks = new Vector<Callable<Double>>();
    for (final Set<Integer> clique : cliques) {
      final Callable<Double> task = new PatternStability(serializedModule, clique, w, -z);
      tasks.add(task);
      //break; // TODO: Remove this line!
    }

    // start calculation
    final List<Future<Double>> decayTimes = executor.invokeAll(tasks);

    // collect results
    double decayTimeMax = Double.NaN;
    int stableCliquesNumber = 0;
    final List<Double> minActivities = new Vector<Double>();
    final List<Double> maxActivities = new Vector<Double>();
    final List<Double> minMembranePotentials = new Vector<Double>();
    final List<Double> maxMembranePotentials = new Vector<Double>();
    for (int i = 0; i < tasks.size(); i++) {
      final Callable<Double> task = tasks.get(i);
      final Future<Double> decayTime = decayTimes.get(i);
      Double time = null;
      try {
        time = decayTime.get();
      } catch (ExecutionException e) {
      }

      if (time == null) {
        // stable clique
        stableCliquesNumber++;
      } else {
        // instable clique
        decayTimeMax = Math.max(decayTimeMax, time);
      }

      maxActivities.add(Collections.max(((PatternStability) task).module.getActivities()));
      minActivities.add(Collections.min(((PatternStability) task).module.getActivities()));
      maxMembranePotentials.add(Collections.max(((LeakyIntegratorModule) ((PatternStability) task).module).getMembranePotentials()));
      minMembranePotentials.add(Collections.min(((LeakyIntegratorModule) ((PatternStability) task).module).getMembranePotentials()));
    }
    final int precision = 6;
    out.printf(Locale.ENGLISH, "%." + precision + "f\t%." + precision + "f\t%3d\t%." + precision + "f\t%." + precision + "f\t%." + precision + "f\t%."
        + precision + "f", w, z, stableCliquesNumber, Collections.min(minActivities), Collections.max(maxActivities), Collections.min(minMembranePotentials),
        Collections.max(maxMembranePotentials));
    out.println();
  }

  public static void main(String[] args) throws InterruptedException, FileNotFoundException {
    final double[] leaks = new double[] { 0.0, 0.1, 1.0 };
    final int[] sizes = new int[] { 3, 4, 5, 6, 7, 8, 9, 10, 20, 30, 40, 50, 60, 70, 80, 90, 100 };
    final double[] thresholds = new double[] { 0.0, 1.0, 5.0, 10.0 };

    final int availableProcessors = Runtime.getRuntime().availableProcessors();
    final int usedProcessors = Math.max(1, availableProcessors);
    System.out.printf(Locale.ENGLISH, "# processors used/available: %d/%d\n", usedProcessors, availableProcessors);
    final ExecutorService executor = Executors.newFixedThreadPool(usedProcessors);

    for (final int size : sizes) {
      for (final double leak : leaks) {
        for (final double threshold : thresholds) {
          /*
           * module set-up
           */
          final Random random = new Random(0);
          final double[][] adjacencyMatrix = Graphs.addInhibitoryLinks(Graphs.getChainGraphAdjacencyMatrix(size)); // getCircleGraphAdjacencyMatrix
          final LeakyIntegratorWithIntrinsicPlasticityModule module = new LeakyIntegratorWithIntrinsicPlasticityModule(adjacencyMatrix, random);
          module.setTransferFunctionParameters(1.0, threshold);
          module.setSilentInhibition(LeakyIntegratorModule.SilentInhibition.INHIBITION_NOT_SILENT); // was: SILENT_INHIBITION_WITH_RESPECT_TO_FIRING_RATE
          module.setIntrinsicPlasticity(false);
          module.setName("leaky integrator");
          module.setLeaks(leak);
          if (module.getConnectedComponents().size() > 1) {
            System.out.println("WARNING! Graph has " + module.getConnectedComponents().size() + " connected components.");
          }
          System.out.println("# " + module.getCliques().size() + " cliques: " + module.getCliques());
          final Set<Set<Integer>> cliques = module.getCliques();
          if (cliques.size() == 1) {
            System.out.println("WARNING! Graph has only 1 clique.");
          }

          // Save module for later concurrent use of multiple copies.
          final byte[] serializedModule = Utils.objectToBytes(module);

          final double wMin = 0;
          final double wMax = leak == 0.0 ? 10 : 100.0 * leak; // 12
          final int wSteps = 25; // 40
          final double zMin = leak == 0.0 ? -10 : -100.0 * leak; // -12
          final double zMax = 0;
          final int zSteps = 25; // 40

          //final PrintStream out = System.out;
          final File file = new File(Brain.DEFAULT_DATA_OUTPUT_DIRECTORY, "clique-stability N=" + size + " leak=" + leak + " threshold=" + threshold + ".data");
          System.out.println(file);
          final PrintStream out = new PrintStream(new BufferedOutputStream(new FileOutputStream(file)));
          out.println("# size = " + module.getSize());
          out.println("# links = " + module.getLinks());
          out.println("# leaks = " + module.getLeaks());
          out.println("# transfer function parameters = " + module.getTransferFunctionParameters());
          out.println("# w, z, stable cliques, minActivity, maxActivity, minMembranePotential, maxMembranePotential");
          for (int wStep = 0; wStep <= wSteps; wStep++) {
            final double w = wMin + wStep * (wMax - wMin) / (double) wSteps;
            for (int zStep = 0; zStep <= zSteps; zStep++) {
              final double z = zMax - zStep * (zMax - zMin) / (double) zSteps;
              PatternStability.cliqueStability(serializedModule, cliques, w, z, executor, out);
            }
            out.println();
          }
          out.close();
        }
      }
    }
    executor.shutdown();
  }
}