package calculation;

import java.io.File;
import java.io.FileNotFoundException;
import java.io.PrintWriter;
import java.util.Arrays;
import java.util.Collections;
import java.util.Comparator;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Random;
import java.util.Set;
import java.util.TreeMap;
import java.util.Vector;
import java.util.concurrent.ExecutorService;
import java.util.concurrent.Executors;

import org.apache.commons.collections.keyvalue.MultiKey;

import utils.Graphs;
import utils.Utils;
import brain.Brain;
import brain.Module;
import brain.LeakyIntegratorModule;

/**
 * This class is for calculating the phase diagram of a leaky integrator.
 * 
 * @author Mathias Linkerhand
 */
public class PatternStabilityPhaseDiagram implements Runnable {
  private LeakyIntegratorModule        module;

  public PatternStabilityPhaseDiagram(final LeakyIntegratorModule module) {
    this.module = module;
  }

  public static void main(final String[] args) throws Exception {
    /*
    logger.setLevel(Level.ALL);
    Logger.getLogger("").getHandlers()[0].setLevel(Level.ALL);
    Logger.getLogger("").setLevel(Level.ALL);
    */

    final int availableProcessors = Runtime.getRuntime().availableProcessors();
    final int usedProcessors = Math.min(1, availableProcessors);
    final ExecutorService executor = Executors.newFixedThreadPool(usedProcessors);

    for (int size = 10; size <= 100; size += 10) {
      Random random = new Random(0);
      final LeakyIntegratorModule module;
      //module = new LeakyIntegratorModule(Graphs.addInhibitoryLinks(Graphs.grosGraphAdjacencyMatrix), random);
      //module = new LeakyIntegratorModule(Graphs.addInhibitoryLinks(Graphs.trivialGraphAdjacencyMatrix));
      //module = new LeakyIntegratorModule(Graphs.addInhibitoryLinks(Graphs.linkerhandGraphAdjacencyMatrix), new Random(0));
      final double linkProbability = 0.4;
      module = new LeakyIntegratorModule(size, linkProbability, random);
      System.out.println("size: " + module.getSize());
      System.out.println("cliques: " + module.getCliques().size());
      System.out.println("connected components: " + module.getConnectedComponents().size());
      if(module.getConnectedComponents().size() > 1){
        System.out.println("WARNING! Graph has " + module.getConnectedComponents().size() + " connected components.");
      }
      //module = new LeakyIntegratorModule(Graphs.addInhibitoryLinks(Graphs.getTypicalDHAN4CliqueCircleGraphAdjacencyMatrix(20)), random);

      //module = new LeakyIntegratorModule(weightedAdjacencyMatrix2);
      module.setLeaks(0.1); // default: 0.1 (0.03 for big graphs of size about 100)
      if (!Graphs.isSymmetric(module.getLinks())) {
        throw new Exception("graph not symmetric");
      }
      module.normalizeLinks();
      module.getCliques(); // precalculate cliques
      module.exportGraphvizGraph(new File(Brain.DEFAULT_DATA_OUTPUT_DIRECTORY, module.toString() + "-graph.gv"));
      module.setName("DELETE");

      Runnable task = new PatternStabilityPhaseDiagram(module);
      executor.execute(task);
    }

    executor.shutdown();
  }

  /**
   * Calculates the w-z phase diagram using iteration of the differential equations.
   * 
   * @throws FileNotFoundException
   */
  @Override
  public void run() {
    try {
      module.exportGraphvizGraph(new File(Brain.DEFAULT_DATA_OUTPUT_DIRECTORY, module.toString() + "-graph.gv"));
    } catch (FileNotFoundException e) {
      e.printStackTrace();
    }

    final double deltaTime = 1e-2; // 1e-2;
    final double duration = 1e+2; // 1e+2;
    final double gamma = module.getLeak(0); // assume all leaks are equal

    final double wMin = 0;
    final double wMax = 100 * gamma; // 12
    final int wSteps = 15; // 40

    final double zMin = -100 * gamma; // -12
    final double zMax = 0;
    final int zSteps = 15; // 40

    module.setName(module.getName() + " Δt=" + deltaTime + " t_max=" + duration);

    final File cliquesFile = new File(Brain.DEFAULT_DATA_OUTPUT_DIRECTORY, module.toString() + ", Γ=" + gamma + ", normalized_links-phase-diagram-cliques.data");
    final File activityFile = new File(Brain.DEFAULT_DATA_OUTPUT_DIRECTORY, module.toString() + ", Γ=" + gamma + ", normalized_links-phase-diagram-activities.data");

    // save module for later restore after manipulations/changes
    final byte[] serializedModule = Utils.objectToBytes(module);

    final Map<Set<Integer>, Integer> cliqueId = new HashMap<Set<Integer>, Integer>();
    for (final Set<Integer> clique : module.getCliques()) {
      cliqueId.put(clique, cliqueId.size());
    }
    //System.out.println("cliques: " + cliqueId);

    final Map<MultiKey, List<Integer>> cliqueData = new HashMap<MultiKey, List<Integer>>();
    final Map<MultiKey, List<Double>> activityData = new HashMap<MultiKey, List<Double>>();

    final int trialsNumber = 10; // module.getCliques().size(); // 50
    final Random random = new Random(0);
    for (int trial = 0; trial < trialsNumber; trial++) {
      System.out.println("Thread: " + Thread.currentThread() + "; " + module.toString() + ", Trial: " + trial);
      //for (final Set<Integer> clique : module.getCliques()) {
      //System.out.println("trial: " + trial);

      final List<Double> activities;
      //      if (trial < module.getCliques().size()) {
      //        activities = new Vector<Double>(module.getSize());
      //        //        final Set<Integer> clique = (Set<Integer>) module.getCliques().toArray(new Set[0])[trial];
      //        System.out.println("active clique: " + cliqueId.get(clique) + " (" + clique + ")");
      //        for (int vertex = 0; vertex < module.getSize(); vertex++) {
      //          activities.add(clique.contains(vertex) ? 1 - 1e-2 : 1e-2);
      //        }
      //      } else {
      activities = Collections.unmodifiableList(Utils.randomList(module.getSize(), random, 0, 1));
      //      }

      //System.out.println("activities: " + activities);

      for (double w = wMax; w >= wMin; w -= (wMax - wMin) / (double) wSteps) {
        for (double z = zMin; z <= zMax; z += (zMax - zMin) / (double) zSteps) {
          // restore module
          module = (LeakyIntegratorModule) Utils.bytesToObject(serializedModule);

          // set activities
          for (int i = 0; i < module.getSize(); i++) {
            module.setActivity(i, activities.get(i));
          }
          //module.setActivities(activities);

          // set new parameters w and z
          for (int i = 0; i < module.getSize(); i++) {
            for (int j = 0; j < module.getSize(); j++) {
              final double oldWeight = module.getLink(i, j);
              final double newWeight;
              if (oldWeight < 0) {
                // inhibitory link
                newWeight = -z * oldWeight;
              } else if (oldWeight > 0) {
                // exhibitory link
                newWeight = w * oldWeight;
              } else {
                // no link
                newWeight = 0;
              }
              module.setLink(i, j, newWeight);
            }
          }

          // simulation
          for (double time = 0; time <= duration; time += deltaTime) {
              module.performOneTimeStep(deltaTime);
          }

          // = 1 clique active?
          final Set<Integer> activeClique = module.getActiveClique();
          /*
          int cliqueNumber = 0;
          if (activeClique != null) {
            for (final Integer i : activeClique) {
              cliqueNumber += 1 << i;
            }
          }
          */

          if (activeClique != null) {
            System.out.print(Integer.toString(cliqueId.get(activeClique), Character.MAX_RADIX));
          } else {
            System.out.print(Collections.max(module.getActivities()) < 1e-8 ? "_" : ".");
          }

          //System.out.println(module.getMembranePotentials());

          final MultiKey coordinates = new MultiKey(z, w);

          final Integer value = activeClique == null ? 0 : 1;
          // increment number
          final List<Integer> newCliqueData;
          if (cliqueData.containsKey(coordinates)) {
            newCliqueData = cliqueData.get(coordinates);
          } else {
            newCliqueData = new Vector<Integer>(1 + cliqueId.size());
            for (int i = 0; i < 1 + cliqueId.size(); i++) {
              newCliqueData.add(0);
            }
          }
          newCliqueData.set(0, newCliqueData.get(0) + value);
          if (activeClique != null) {
            final int index = 1 + cliqueId.get(activeClique);
            newCliqueData.set(index, newCliqueData.get(index) + 1);
          }
          cliqueData.put(coordinates, newCliqueData);

          final double maxActivity = Collections.max(module.getActivities());
          final double minActivity = Collections.min(module.getActivities());
          final double meanActivity = (double) Utils.sum(module.getActivities()) / module.getSize();
          final Double[] sortedActivities = module.getActivities().toArray(new Double[0]);
          Arrays.sort(sortedActivities);
          final double medianActivity = sortedActivities[module.getSize() / 2];
          final List<Double> newActivityData;
          if (activityData.containsKey(coordinates)) {
            newActivityData = activityData.get(coordinates);
          } else {
            newActivityData = new Vector<Double>(6 + module.getSize());
            for (int i = 0; i < 6 + module.getSize(); i++) {
              newActivityData.add(0.0);
            }
            newActivityData.set(4, Double.POSITIVE_INFINITY);
          }
          newActivityData.set(0, newActivityData.get(0) + maxActivity);
          newActivityData.set(1, newActivityData.get(1) + minActivity);
          newActivityData.set(2, newActivityData.get(2) + meanActivity);
          newActivityData.set(3, Math.max(newActivityData.get(3), maxActivity));
          newActivityData.set(4, Math.min(newActivityData.get(4), minActivity));
          newActivityData.set(5, newActivityData.get(5) + medianActivity);
          for (int neuron = 0; neuron < module.getSize(); neuron++) {
            final int index = 6 + neuron;
            newActivityData.set(index, newActivityData.get(index) + module.getActivity(neuron));
          }
          activityData.put(coordinates, newActivityData);
        }
        System.out.println();
      }

      // output data
      module = (LeakyIntegratorModule) Utils.bytesToObject(serializedModule);
      try {
        savePhaseDiagram(cliquesFile, module, cliqueData);
        savePhaseDiagram(activityFile, module, activityData);
      } catch (FileNotFoundException e) {
        e.printStackTrace();
      }
    }
  }

  public static void savePhaseDiagram(final File activityFile, final Module module, final Map<MultiKey, ? extends List<?>> data) throws FileNotFoundException {
    final Map<MultiKey, List<?>> sortedData = new TreeMap<MultiKey, List<?>>(new Comparator<MultiKey>() {
      @SuppressWarnings({ "rawtypes", "unchecked" })
      @Override
      public int compare(MultiKey o1, MultiKey o2) {
        for (int i = 0; i < o1.size(); i++) {
          Comparable c1 = (Comparable) o1.getKey(i);
          Comparable c2 = (Comparable) o2.getKey(i);
          if (c1.compareTo(c2) != 0) {
            return c1.compareTo(c2);
          }
        }
        return 0;
      }
    });
    sortedData.putAll(data);

    final PrintWriter activityOut = new PrintWriter(activityFile);
    activityOut.println("# size: " + module.getSize());
    activityOut.println("# synaptic strength matrix: " + module.getLinks());
    activityOut.println("# cliques: " + module.getCliques());
    activityOut.println("# in degrees: " + module.getNodeInDegrees(0));
    activityOut.println("# in degrees (+): " + module.getNodeInDegrees(1));
    activityOut.println("# in degrees (-): " + module.getNodeInDegrees(-1));
    activityOut.println("# out degrees: " + module.getNodeOutDegrees(0));
    activityOut.println("# out degrees (+): " + module.getNodeOutDegrees(1));
    activityOut.println("# out degrees (-): " + module.getNodeOutDegrees(-1));
    activityOut.println("#");
    activityOut.println("# -z/Gamma   w/Gamma sum   cliques …");

    Object lastRow = null;
    for (final Map.Entry<MultiKey, List<?>> entry : sortedData.entrySet()) {
      final MultiKey coordinates = entry.getKey();
      final List<?> value = entry.getValue();

      // new x value?
      if (!coordinates.getKeys()[0].equals(lastRow)) {
        lastRow = coordinates.getKeys()[0];

        // new line
        activityOut.println(); // necessary for Gnuplot to avoid “Warning: Single isoline (scan) is not enough for a pm3d plot. Hint: Missing blank lines in the data file? See 'help pm3d' and FAQ.”
      }

      // x, y coordinate
      activityOut.print(Double.toString((Double) (coordinates.getKeys()[0])) + "\t" + Double.toString((Double) (coordinates.getKeys()[1])));

      // value
      for (final Object x : value) {
        activityOut.print("\t" + x);
      }

      // new line
      activityOut.println();
    }
    activityOut.close();
  }
}