package calculation;

import java.io.File;
import java.io.FileNotFoundException;
import java.util.Random;

import brain.LeakyIntegratorWithIntrinsicPlasticityModule;
import brain.Brain;

/**
 * Simulates a single neuron and its intrinsic plasticity.
 * 
 * @author Mathias Linkerhand
 * @version 2012-05-09
 */
public class SingleLeakyIntegratorIntrinsicPlasticityWithAutapse {

  public static void main(String[] args) throws FileNotFoundException {
    final Brain brain = new Brain();
    final double timeDelta = 1e-3;
    final double timeMax = 1e5;
    brain.setTimeDeltaAndDuration(timeDelta, timeMax);
    brain.setLogInterval(Math.max(timeDelta, timeMax / 1e5)); // 1e5 data points
    final File dataOutputDirectory = new File(Brain.DEFAULT_DATA_OUTPUT_DIRECTORY, SingleLeakyIntegratorIntrinsicPlasticityWithAutapse.class
        .getSimpleName());
    dataOutputDirectory.mkdirs();
    brain.setDataOutputDirectory(dataOutputDirectory);


    final Random random = new Random(0);
    final int size = 8;
    final double autapseStrength = 1.0;
    final double[][] adjacencyMatrix = new double[size][size];
    for (int i = 0; i < adjacencyMatrix.length; i++) {
      for (int j = 0; j < adjacencyMatrix[i].length; j++) {
        adjacencyMatrix[i][j] = i == j ? autapseStrength : 0.0; // autapse
      }
    }
    final LeakyIntegratorWithIntrinsicPlasticityModule leakyIntegrator = new LeakyIntegratorWithIntrinsicPlasticityModule(adjacencyMatrix, random);
    System.out.println(leakyIntegrator.getLinks());
    leakyIntegrator.setName("LeakyIntegrator");

    // intrinsic learning rate
    final double intrinsicAdaptationParameter = 1e-1;
    for (int neuron = 0; neuron < leakyIntegrator.getSize(); neuron++) {
      final double currentIntrinsicAdaptationParameter = intrinsicAdaptationParameter;
      leakyIntegrator.setIntrinsicAdaptationLearningRates(neuron, currentIntrinsicAdaptationParameter);
    }

    // leaks
    final double leak = 1e0;
    for (int neuron = 0; neuron < leakyIntegrator.getSize(); neuron++) {
      final double currentLeak = leak;
      leakyIntegrator.setLeak(neuron, currentLeak);
    }

    // default transfer function parameters
    leakyIntegrator.setTransferFunctionParameters(1.0, 1.0); // gain, threshold

    // default membrane potential
    for (int neuron = 0; neuron < leakyIntegrator.getSize(); neuron++) {
      leakyIntegrator.setMembranePotential(neuron, 0.1);
    }

    {
      // target distributions
      int neuron = 0;
      leakyIntegrator.setTargetDistributionParameters(neuron++, 0.0, 0.0);
      leakyIntegrator.setTargetDistributionParameters(neuron++, -10.0, 0.0);
      leakyIntegrator.setTargetDistributionParameters(neuron++, +10.0, 0.0);
      leakyIntegrator.setTargetDistributionParameters(neuron++, -10.0, +10.0);
      leakyIntegrator.setTargetDistributionParameters(neuron++, +20.0, -20.0);
      leakyIntegrator.setTargetDistributionParameters(neuron++, -20.0, +18.5);
      leakyIntegrator.setTargetDistributionParameters(neuron++, -20.0, +19.0);
      leakyIntegrator.setTargetDistributionParameters(neuron++, -20.0, +20.0);
    }

    // logger properties
    brain.addModule(leakyIntegrator);
    brain.addLogger(leakyIntegrator, "getActivities");
    brain.addLogger(leakyIntegrator, "getMembranePotentials");
    brain.addLogger(leakyIntegrator, "getTransferFunctionParameters");
    brain.addLogger(leakyIntegrator, "getTransferFunctionParameterChanges");

    // simulation
    brain.run();
  }
}