package calculation;

import java.io.File;
import java.io.FileNotFoundException;
import java.util.Random;

import brain.LeakyIntegratorWithIntrinsicPlasticityModule;
import brain.Brain;
import brain.ConstantSignalModule;

/**
 * Simulates several single neurons and its intrinsic plasticities due to a constant input.
 * 
 * @author Mathias Linkerhand
 * @version 2012-05-09
 */
public class SingleLeakyIntegratorIntrinsicPlasticityWithConstantInput {

  public static void main(String[] args) throws FileNotFoundException {
    final Brain brain = new Brain();
    final double timeDelta = 1e-2;
    final double timeMax = 1e4;
    brain.setTimeDeltaAndDuration(timeDelta, timeMax);
    brain.setLogInterval(Math.max(timeDelta, timeMax / 1e5)); // maximum 1e5 data points
    final File dataOutputDirectory = new File(Brain.DEFAULT_DATA_OUTPUT_DIRECTORY, SingleLeakyIntegratorIntrinsicPlasticityWithConstantInput.class
        .getSimpleName());
    dataOutputDirectory.mkdirs();
    brain.setDataOutputDirectory(dataOutputDirectory);

    final Random random = new Random();
    final int size = 8;
    final double[][] adjacencyMatrix = new double[size][size];
    final LeakyIntegratorWithIntrinsicPlasticityModule leakyIntegrator = new LeakyIntegratorWithIntrinsicPlasticityModule(adjacencyMatrix, random);
    leakyIntegrator.setName("LeakyIntegrator");

    // intrinsic learning rate
    final double intrinsicAdaptationParameter = 1e-1;
    for (int neuron = 0; neuron < leakyIntegrator.getSize(); neuron++) {
      final double currentIntrinsicAdaptationParameter = intrinsicAdaptationParameter;
      leakyIntegrator.setIntrinsicAdaptationLearningRates(neuron, currentIntrinsicAdaptationParameter);
    }

    // leaks
    final double leak = 1;
    for (int neuron = 0; neuron < leakyIntegrator.getSize(); neuron++) {
      final double currentLeak = leak;
      leakyIntegrator.setLeak(neuron, currentLeak);
    }

    // default transfer function parameters
    leakyIntegrator.setTransferFunctionParameters(1.0, 1.0); // gain, threshold

    // default membrane potential
    for (int neuron = 0; neuron < leakyIntegrator.getSize(); neuron++) {
      leakyIntegrator.setMembranePotential(neuron, 0.1);
    }

    {
      // target distributions
      int neuron = 0;
      leakyIntegrator.setTargetDistributionParameters(neuron++, 0.0, 0.0);
      leakyIntegrator.setTargetDistributionParameters(neuron++, -10.0, 0.0);
      leakyIntegrator.setTargetDistributionParameters(neuron++, +10.0, 0.0);
      leakyIntegrator.setTargetDistributionParameters(neuron++, -10.0, +10.0);
      leakyIntegrator.setTargetDistributionParameters(neuron++, +20.0, -20.0);
      leakyIntegrator.setTargetDistributionParameters(neuron++, -20.0, +18.5);
      leakyIntegrator.setTargetDistributionParameters(neuron++, -20.0, +19.0);
      leakyIntegrator.setTargetDistributionParameters(neuron++, -20.0, +20.0);
    }

    // input interval range, standard: [0, 1]
    final double inputGain = 0.0; // then multiply by gain
    final double activity = 1.0;
    final ConstantSignalModule input = new ConstantSignalModule(activity);
    input.setName("Input");

    // input → leaky integrator connections
    leakyIntegrator.addAfferentModule(input);
    for (int i = 0; i < leakyIntegrator.getSize(); i++) {
      //final double currentLink = inputGain * i / (leakyIntegrator.getSize() - 1);
      final double currentLink = inputGain;
      for (int j = 0; j < input.getSize(); j++) {
        leakyIntegrator.setAfferentLink(input, i, j, currentLink); // gain
      }
    }

    // logger properties
    brain.addModule(input);
    brain.addLogger(input, "getActivities");

    brain.addModule(leakyIntegrator);
    brain.addLogger(leakyIntegrator, "getActivities");
    brain.addLogger(leakyIntegrator, "getMembranePotentials");
    brain.addLogger(leakyIntegrator, "getTransferFunctionParameters");
    brain.addLogger(leakyIntegrator, "getTransferFunctionParameterChanges");

    // show setup
    System.out.println(leakyIntegrator.getCliques());
    System.out.println(leakyIntegrator.getAfferentLinks(input));

    // simulation
    brain.run();
  }
}