package calculation;

import java.io.File;
import java.io.FileNotFoundException;
import java.util.Random;

import brain.LeakyIntegratorWithIntrinsicPlasticityModule;
import brain.Brain;
import brain.NoiseModule;

/**
 * Simulates several single neurons and its intrinsic plasticity due to a random noise input.
 * 
 * This class (its simulation data) is used in the paper “Self-organized stochastic tipping in slow-fast dynamical systems” written by Mathias Linkerhand and Claudius Gros.
 * 
 * @author Mathias Linkerhand
 * @version 2012-03-06
 */
public class SingleLeakyIntegratorIntrinsicPlasticityWithRandomNoiseInput {

  public static void main(String[] args) throws FileNotFoundException {
    final Brain brain = new Brain();
    final double timeDelta = 1e-1;
    final double timeMax = 2e5;
    brain.setTimeDeltaAndDuration(timeDelta, timeMax);
    brain.setLogInterval(Math.max(timeDelta, timeMax / 1e5)); // max 1e5 data points
    final File dataOutputDirectory = new File(Brain.DEFAULT_DATA_OUTPUT_DIRECTORY, SingleLeakyIntegratorIntrinsicPlasticityWithRandomNoiseInput.class.getSimpleName());
    dataOutputDirectory.mkdirs();
    brain.setDataOutputDirectory(dataOutputDirectory);

    final Random random = new Random(0);
    final int size = 8;
    final double[][] adjacencyMatrix = new double[size][size];
    final LeakyIntegratorWithIntrinsicPlasticityModule relisi = new LeakyIntegratorWithIntrinsicPlasticityModule(adjacencyMatrix, random);
    relisi.setName("Relisi");

    // intrinsic learning rate
    final double intrinsicAdaptationParameter = 1e-1;
    for (int neuron = 0; neuron < relisi.getSize(); neuron++) {
      final double currentIntrinsicAdaptationParameter = intrinsicAdaptationParameter;
      relisi.setIntrinsicAdaptationLearningRates(neuron, currentIntrinsicAdaptationParameter);
    }

    // leaks
    final double leak = 1;
    for (int neuron = 0; neuron < relisi.getSize(); neuron++) {
      final double currentLeak = leak;
      relisi.setLeak(neuron, currentLeak);
    }

    // default transfer function parameters
    relisi.setTransferFunctionParameters(1.0, 1.0); // gain, threshold

    // default membrane potential
    for (int neuron = 0; neuron < relisi.getSize(); neuron++) {
      relisi.setMembranePotential(neuron, 0.1);
    }

    // target distributions
    relisi.setTargetDistributionParameters(0, 0.0, 0.0);
    relisi.setTargetDistributionParameters(1, -10.0, 0.0);
    relisi.setTargetDistributionParameters(2, +10.0, 0.0);
    relisi.setTargetDistributionParameters(3, -10.0, +10.0);
    relisi.setTargetDistributionParameters(4, +20.0, -20.0);
    relisi.setTargetDistributionParameters(5, -20.0, +18.5);
    relisi.setTargetDistributionParameters(6, -20.0, +19.0);
    relisi.setTargetDistributionParameters(7, -20.0, +20.0);

    // input interval range, standard: [0, 1]
    final double inputBias = 0.0; // first shift by bias
    final double inputGain = 10.0; // then multiply by gain
    final double inputChangeInterval = 1.0;
    final NoiseModule input = new NoiseModule(NoiseModule.WHITE_NOISE, inputChangeInterval, random);
    input.setName("Input");
    input.setChangeInterval(inputChangeInterval);
    input.setBias(inputBias);

    // input → relisi connections
    relisi.addAfferentModule(input);
    for (int i = 0; i < relisi.getSize(); i++) {
      final double currentLink = inputGain;
      for (int j = 0; j < input.getSize(); j++) {
        relisi.setAfferentLink(input, i, j, currentLink); // gain
      }
    }

    // logger properties
    brain.addModule(input);
    brain.addLogger(input, "getActivities");

    brain.addModule(relisi);
    brain.addLogger(relisi, "getActivities");
    brain.addLogger(relisi, "getMembranePotentials");
    brain.addLogger(relisi, "getTransferFunctionParameters");

    // show setup
    System.out.println(relisi.getCliques());
    System.out.println(relisi.getAfferentLinks(input));

    // simulation
    brain.run();
  }
}