package stochasticTipping;

import java.io.BufferedOutputStream;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.PrintStream;
import java.util.Arrays;
import java.util.InvalidPropertiesFormatException;
import java.util.Locale;
import java.util.Properties;

import org.apache.commons.io.FilenameUtils;
import org.apache.commons.math3.ode.FirstOrderDifferentialEquations;
import org.apache.commons.math3.ode.FirstOrderIntegrator;
import org.apache.commons.math3.ode.nonstiff.ClassicalRungeKuttaIntegrator;
import org.apache.commons.math3.ode.sampling.StepHandler;
import org.apache.commons.math3.ode.sampling.StepInterpolator;

import utils.Histogram;

/**
 * Numerical calculations for the scientific article "Self-organized stochastic
 * tipping in slow-fast dynamical systems" by Mathias Linkerhand and Claudius
 * Gros.
 * 
 * This class shows the stochastic tipping of a single leaky integrator neuron
 * with random white noise which remains constant for a specific time. The
 * neuron tries to adapt its internal parameters gain and threshold in order to
 * reach a desired target firing rate distribution.
 * 
 * @author Mathias Linkerhand
 * @version 2013-01-30
 * @since 2012
 */
// usage: java -cp bin:lib/* stochasticTipping.SelfOrganizedStochasticTippingInSlowFastDynamicalSystems > simulation.data
public class SelfOrganizedStochasticTippingInSlowFastDynamicalSystems {
  /**
   * Running the simulation.
   * 
   * @throws IOException
   * @throws FileNotFoundException
   * @throws InvalidPropertiesFormatException
   */
  public static void run(final Properties simulationProperties, final File timeSeriesFile, final File histogramFile, final File targetHistogramFile)
      throws InvalidPropertiesFormatException, FileNotFoundException, IOException {
    // Simulation constants.
    final double tMin = Double.parseDouble(simulationProperties.getProperty("t_min")); // $t_{\mathrm{min}}$
    final double tMax = Double.parseDouble(simulationProperties.getProperty("t_max"));// $t_{\mathrm{max}}$
    final double tDelta = Double.parseDouble(simulationProperties.getProperty("t_Delta")); // $\delta t$
    final double gamma = Double.parseDouble(simulationProperties.getProperty("Gamma")); // $\Gamma$
    final double w = 1.0; // $w$
    final double epsilon_a = Double.parseDouble(simulationProperties.getProperty("epsilon_a")); // $\epsilon_a$
    final double epsilon_b = Double.parseDouble(simulationProperties.getProperty("epsilon_b")); // $\epsilon_b$
    final double lambda_1 = Double.parseDouble(simulationProperties.getProperty("lambda_1")); // $\lambda_1$
    final double lambda_2 = Double.parseDouble(simulationProperties.getProperty("lambda_2")); // $\lambda_2$
    final double tNoiseConstantInterval = 1.0;
    final double xi_min = Double.parseDouble(simulationProperties.getProperty("xi_min"));
    final double xi_max = Double.parseDouble(simulationProperties.getProperty("xi_max"));

    // Initial state.
    final double[] state = new double[] { Double.parseDouble(simulationProperties.getProperty("x(0)")),
        Double.parseDouble((String) simulationProperties.get("a(0)")), Double.parseDouble(simulationProperties.getProperty("b(0)")) }; // $(x(t_{\mathrm{min}}), a(t_{\mathrm{min}}), b(t_{\mathrm{min}}))$

    final int binCount = 20;

    System.out.println(tMin + ", " + tMax + ", " + tDelta + ", " + gamma + ", " + w + ", " + epsilon_a + ", " + epsilon_b + ", " + lambda_1 + ", " + lambda_2
        + ", " + tNoiseConstantInterval + ", " + xi_min + ", " + xi_max + ", " + Arrays.toString(state));

    // Set up the ODE.
    final FirstOrderDifferentialEquations ode = new SingleLeakyIntegratorWithIntrinsicPlasticityAndNoiseInputODE(gamma, w, epsilon_a, epsilon_b, lambda_1,
        lambda_2, tNoiseConstantInterval, xi_min, xi_max);

    // Set up the integrator.
    final FirstOrderIntegrator integrator = new ClassicalRungeKuttaIntegrator(tDelta);

    // Output the time steps.
    final PrintStream timeSeriesPrintStream = new PrintStream(new BufferedOutputStream(new FileOutputStream(timeSeriesFile)));
    integrator.addStepHandler(new StepHandler() {
      private int               printStep            = 0;
      private final int         PRINT_EVERY_NTH_STEP = 1000;
      private Histogram         histogram            = new Histogram(binCount, 0.0, 1.0);
      private final PrintStream out                  = timeSeriesPrintStream;

      @Override
      public void handleStep(final StepInterpolator interpolator, final boolean isLast) {
        final double[] state = interpolator.getInterpolatedState();
        final double time = interpolator.getCurrentTime();

        if (time > tMin + (tMax - tMin) * 0.2) {
          // only add the last 80 % of data
          histogram.add(SingleLeakyIntegratorWithIntrinsicPlasticityAndNoiseInputODE.g(state[0], state[1], state[2])); // store firing rate in histogram
        }

        if (isLast) {
          try {
            histogram.export(histogramFile);

            // Create target distribution histogram.
            final Histogram targetHistogram = Histogram.createExponentialDistributionHistogram(binCount, lambda_1, lambda_2);
            targetHistogram.export(targetHistogramFile);

            // Write relative entropies to time series file.
            this.out.println("## relative entropy D_KL(p,q) = " + histogram.getKullbackLeiblerDivergence(targetHistogram));
            this.out.println("## relative entropy D_KL(q,p) = " + targetHistogram.getKullbackLeiblerDivergence(histogram));
          } catch (IOException e) {
            throw new IllegalStateException("could not save histogram to file");
          }
        } else {
          if (printStep == 0) {
            this.printStep(time, state, this.out);
          }
          printStep = (printStep + 1) % PRINT_EVERY_NTH_STEP;
        }
      }

      @Override
      public void init(final double t0, final double[] state, final double t) {
        this.out.println("## time t\tmembrane potential x\tgain a\tthreshold b\tfiring rate y\tnoise input xi");
        this.printStep(t0, state, this.out);
      }

      public void printStep(final double t, final double[] state, final PrintStream out) {
        out.printf(Locale.ENGLISH, "%.8f", t); // time $t$
        for (final double value : state) { // membrane potential $x$, transfer function gain $a$, transfer function threshold $b$
          out.printf(Locale.ENGLISH, "\t%.8f", value);
        }
        out.printf(Locale.ENGLISH, "\t%.8f", SingleLeakyIntegratorWithIntrinsicPlasticityAndNoiseInputODE.g(state[0], state[1], state[2])); // firing rate $y$
        out.printf(Locale.ENGLISH, "\t%.8f", ((SingleLeakyIntegratorWithIntrinsicPlasticityAndNoiseInputODE) ode).randomWhiteNoiseWithConstantTableaus(t,
            tNoiseConstantInterval, xi_min, xi_max)); // random noise input $y$
        out.println();
      }
    });

    // Do the integration.
    integrator.integrate(ode, tMin, state, tMax, state);

    // Close open streams.
    timeSeriesPrintStream.close();
  }

  /**
   * Running the application.
   * 
   * @param args A string array with the first argument being the properties
   * file with parameters to use for the simulation.
   * @throws IOException
   * @throws FileNotFoundException
   * @throws InvalidPropertiesFormatException
   */
  public static void main(final String[] args) throws InvalidPropertiesFormatException, FileNotFoundException, IOException {
    // Input/output files.
    final File propertiesFile = new File(args[0]);
    final File timeSeriesFile = new File(FilenameUtils.removeExtension(propertiesFile.getName()) + "TimeSeries.data");
    final File histogramFile = new File(FilenameUtils.removeExtension(propertiesFile.getName()) + "Histogram.data");
    final File targetHistogramFile = new File(FilenameUtils.removeExtension(propertiesFile.getName()) + "TargetHistogram.data");

    // Simulation properties.
    final Properties simulationProperties = new Properties();
    simulationProperties.loadFromXML(new FileInputStream(propertiesFile));

    // Running the simulation.
    SelfOrganizedStochasticTippingInSlowFastDynamicalSystems.run(simulationProperties, timeSeriesFile, histogramFile, targetHistogramFile);
  }
}