package stochasticTipping;

import java.util.Random;

import org.apache.commons.math3.ode.FirstOrderDifferentialEquations;

/**
 * This class implements a single leaky integrator with intrinsic plasticity and
 * noise driving.
 * 
 * It is used in the class @see
 * SingleLeakyIntegratorWithIntrinsicPlasticityAndNoiseInputODE in order to
 * reproduce the results from the scientific article "Self-organized stochastic
 * tipping in slow-fast dynamical systems" by Mathias Linkerhand and Claudius
 * Gros.
 * 
 * @author Mathias Linkerhand
 * @version 2013-01-30
 * @since 2012
 */
public class SingleLeakyIntegratorWithIntrinsicPlasticityAndNoiseInputODE implements FirstOrderDifferentialEquations {
  private final double epsilon_a;                                    // transfer function gain learning rate $\epsilon_a$
  private final double epsilon_b;                                    // transfer function threshold learning rate $\epsilon_b$
  private final double gamma;                                        // leak $\Gamma$
  private final double lambda_1;                                     // target distribution parameter $\lambda_1$
  private final double lambda_2;                                     // target distribution parameter $\lambda_2$
  private final double w;                                            // weighted adjacency matrix $w$
  private final double tNoiseConstantInterval;                       // constant noise plateau length $t_\xi$
  private final double xi_min;                                       // minimum noise value $\Xi_1$
  private final double xi_max;                                       // maximum noise value $\Xi_2$

  private double       lastRandomValue    = Double.NaN;
  private double       lastRandomInterval = Double.NEGATIVE_INFINITY;
  private final Random random             = new Random(0);            // default seed

  /**
   * Creates a new leaky integrator with random noise input according to the
   * specified parameters.
   * 
   * @param gamma leak
   * @param w weighted adjacency matrix
   * @param epsilon_a transfer function gain learning rate
   * @param epsilon_b transfer function threshold learning rate
   * @param lambda_1 target firing rate distribution parameter 1
   * @param lambda_2 target firing rate distribution parameter 2
   */
  public SingleLeakyIntegratorWithIntrinsicPlasticityAndNoiseInputODE(final double gamma, final double w, final double epsilon_a, final double epsilon_b,
      final double lambda_1, final double lambda_2, final double tNoiseConstantInterval, final double xi_min, final double xi_max) {
    this.epsilon_a = epsilon_a;
    this.epsilon_b = epsilon_b;
    this.gamma = gamma;
    this.lambda_1 = lambda_1;
    this.lambda_2 = lambda_2;
    this.w = w;
    this.tNoiseConstantInterval = tNoiseConstantInterval;
    this.xi_min = xi_min;
    this.xi_max = xi_max;
  }

  /**
   * Returns the dimension of the ODE.
   * 
   * @return 3
   */
  @Override
  public int getDimension() {
    return 3; // $(x, a, b)$
  }

  /**
   * Calculates the first derivatives of the ODE.
   * 
   * @param time time $t$
   * @param state vector $(x(t), a(t), b(t))$
   * @param stateDot vector $(\dot{x}(t), \dot{a}(t), \dot{b}(t))$
   */
  @Override
  public void computeDerivatives(final double time, final double[] state, final double[] stateDot) {
    final double firingRate = g(state[0], state[1], state[2]);

    /*
     * auxiliary terms
     */
    // $\frac{q\prime}{q}$
    final double qPrime_over_q = lambda_1 + 2.0 * lambda_2 * firingRate;
    // a term in $\dot{a}$ and $\dot{b}$.
    final double delta = 1.0 - 2.0 * firingRate + qPrime_over_q * (1.0 - firingRate) * firingRate;

    /*
     * derivatives
     */
    // membrane potential $\dot{x} = -\Gamma x + w y(t)$
    stateDot[0] = -gamma * state[0] + w * randomWhiteNoiseWithConstantTableaus(time, tNoiseConstantInterval, xi_min, xi_max);

    // transfer function gain $\dot{a} = $
    stateDot[1] = epsilon_a * (1.0 / state[1] + (state[0] - state[2]) * delta);

    // transfer function threshold $\dot{b} = $
    stateDot[2] = epsilon_b * (-state[1] * delta);
  }

  /**
   * Transfer function $g(x, a, b) = \frac{1}{1 + e^{-a(x-b)}}$.
   * 
   * @param x membrane potential $x$
   * @param a transfer function gain $a$
   * @param b transfer function threshold $b$
   * @return firing rate $y = g(x, a, b)$
   */
  public static double g(final double x, final double a, final double b) {
    return 1.0 / (1.0 + Math.exp(-a * (x - b)));
  }

  /**
   * Random white noise with constant tableaus. The noise remains constant for
   * $\delta t = 1$.
   * 
   * This functions works only if called with increasing parameter t. Random
   * access would give wrong results.
   * 
   * @param t time
   * @return random value of the corresponding plateau
   */
  public double randomWhiteNoiseWithConstantTableaus(final double t, final double tNoiseConstantInterval, final double xi_min, final double xi_max) {
    final double randomInterval = Math.floor(t / tNoiseConstantInterval);

    if (randomInterval < lastRandomInterval) {
      throw new IllegalStateException("No random access allowed.");
    }
    if (randomInterval > lastRandomInterval) {
      lastRandomValue = random.nextDouble() * (xi_max - xi_min) + xi_min;
      lastRandomInterval = randomInterval;
    }
    return lastRandomValue;
  }
}