package transientDynamics;

import java.io.BufferedOutputStream;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.PrintStream;
import java.util.Arrays;
import java.util.InvalidPropertiesFormatException;
import java.util.Locale;
import java.util.Properties;

import org.apache.commons.io.FilenameUtils;
import org.apache.commons.math3.ode.FirstOrderDifferentialEquations;
import org.apache.commons.math3.ode.FirstOrderIntegrator;
import org.apache.commons.math3.ode.nonstiff.ClassicalRungeKuttaIntegrator;
import org.apache.commons.math3.ode.nonstiff.EulerIntegrator;
import org.apache.commons.math3.ode.sampling.StepHandler;
import org.apache.commons.math3.ode.sampling.StepInterpolator;

/**
 * Numerical calculations for the single leaky integrator neuron with intrinsic plasticity and autapse.
 * 
 * @author Mathias Linkerhand
 * @version 2013-01-29
 * @since 2012
 */
// usage:
// java -cp bin:lib/* transientDynamics.Autapse > simulation.data
// gnuplot -e "set terminal png size 1*1680,800; set style data lp; set pointsize 0.3; EVERY=100; FILE='simulation.data'; set multiplot layout 4,1; plot FILE u 1:2 every EVERY lc rgb 'blue' t 'x', '' u 1:4 every EVERY lc rgb 'dark-red' t 'b'; plot FILE u 1:3 every EVERY lc rgb 'dark-green' t 'a'; plot FILE u 1:5 every EVERY lc rgb 'black' t 'y'; plot FILE u 1:6 every EVERY lc rgb 'orange' t 'xi'; unset multiplot" > simulation.data.png
public class Autapse {
  public static enum IntegrationMethod {
    EULER, CLASSICAL_RUNGE_KUTTA
  };

  /**
   * Running the simulation.
   * 
   * @throws FileNotFoundException 
   */
  public static void run(final Properties simulationProperties, final File timeSeriesFile) throws FileNotFoundException {
    // Simulation constants.
    final double tMin = Double.parseDouble(simulationProperties.getProperty("t_min")); // $t_{\mathrm{min}}$
    final double tMax = Double.parseDouble(simulationProperties.getProperty("t_max"));// $t_{\mathrm{max}}$
    final double tDelta = Double.parseDouble(simulationProperties.getProperty("t_Delta")); // $\delta t$
    final double gamma = Double.parseDouble(simulationProperties.getProperty("Gamma")); // $\Gamma$
    final double w = 1.0; // $w$
    final double epsilon_a = Double.parseDouble(simulationProperties.getProperty("epsilon_a")); // $\epsilon_a$
    final double epsilon_b = Double.parseDouble(simulationProperties.getProperty("epsilon_b")); // $\epsilon_b$
    final double lambda_1 = Double.parseDouble(simulationProperties.getProperty("lambda_1")); // $\lambda_1$
    final double lambda_2 = Double.parseDouble(simulationProperties.getProperty("lambda_2")); // $\lambda_2$

    // Initial state.
    final double[] state = new double[] { Double.parseDouble(simulationProperties.getProperty("x(0)")),
        Double.parseDouble((String) simulationProperties.get("a(0)")), Double.parseDouble(simulationProperties.getProperty("b(0)")) }; // $(x(t_{\mathrm{min}}), a(t_{\mathrm{min}}), b(t_{\mathrm{min}}))$

    System.out.println(tMin + ", " + tMax + ", " + tDelta + ", " + gamma + ", " + w + ", " + epsilon_a + ", " + epsilon_b + ", " + lambda_1 + ", " + lambda_2
        + ", " + Arrays.toString(state));

    // Set up the ODE.
    final FirstOrderDifferentialEquations ode = new AutapseODE(gamma, w, epsilon_a, epsilon_b, lambda_1, lambda_2);

    // Set the integration method.
    final IntegrationMethod integrationMethod = IntegrationMethod.EULER;
    final FirstOrderIntegrator integrator;
    switch (integrationMethod) {
      case EULER:
        integrator = new EulerIntegrator(tDelta);
        break;
      case CLASSICAL_RUNGE_KUTTA:
        integrator = new ClassicalRungeKuttaIntegrator(tDelta);
        break;
      default:
        throw new IllegalArgumentException("Integration method unknown.");
    }

    // Output the time steps.
    final PrintStream timeSeriesPrintStream = new PrintStream(new BufferedOutputStream(new FileOutputStream(timeSeriesFile)));
    integrator.addStepHandler(new StepHandler() {
      private final PrintStream out = timeSeriesPrintStream;

      @Override
      public void handleStep(final StepInterpolator interpolator, final boolean isLast) {
        if (!isLast) {
          this.printStep(interpolator.getCurrentTime(), interpolator.getInterpolatedState(), this.out);
        }
      }

      @Override
      public void init(final double t0, final double[] state, final double t) {
        this.out.println("## time\tmembrane potential\tgain\tthreshold\tfiring rate");
        this.printStep(t0, state, this.out);
      }

      public void printStep(final double t, final double[] state, final PrintStream out) {
        this.out.printf(Locale.ENGLISH, "%.8f", t); // time $t$
        for (final double value : state) { // membrane potential $x$, transfer function gain $a$, transfer function threshold $b$
          this.out.printf(Locale.ENGLISH, "\t%.8f", value);
        }
        this.out.printf(Locale.ENGLISH, "\t%.8f", AutapseODE.g(state[0], state[1], state[2])); // firing rate $y$
        this.out.println();
      }
    });

    // Do the integration.
    integrator.integrate(ode, tMin, state, tMax, state);
    
    // Close open streams.
    timeSeriesPrintStream.close();
  }

  /**
   * Running the application.
   * 
   * @param args No argument is used.
   * @throws IOException 
   * @throws FileNotFoundException 
   * @throws InvalidPropertiesFormatException 
   */
  public static void main(final String[] args) throws InvalidPropertiesFormatException, FileNotFoundException, IOException {
    // Input/output files.
    final File propertiesFile = new File(args[0]);
    final File timeSeriesFile = new File(FilenameUtils.removeExtension(propertiesFile.getName()) + "TimeSeries.data");

    // Simulation properties.
    final Properties simulationProperties = new Properties();
    simulationProperties.loadFromXML(new FileInputStream(propertiesFile));

    // Running the simulation.
    Autapse.run(simulationProperties, timeSeriesFile);
  }
}