package transientDynamics;

import org.apache.commons.math3.ode.FirstOrderDifferentialEquations;

/**
 * This class implements the ODE of a network of leaky integrators with
 * intrinsic plasticity.
 * 
 * It is used in the class
 * <code>TransientDynamicsOfLeakyIntegratorsWithIntrinsicPlasticity</code>.
 * 
 * @author Mathias Linkerhand
 * @version 2013-01-31
 * @since 2012
 */
public class NetworkLeakyIntegratorWithIntrinsicPlasticityODE implements FirstOrderDifferentialEquations {
  private final double[]   epsilon_a;
  private final double[]   epsilon_b;
  private final double[]   gamma;
  private final double[]   lambda_1;   // $\lambda_1$
  private final double[]   lambda_2;   // $\lambda_2$
  private final double[][] w;
  private final int        networkSize;

  /** Properties of each neuron: membrane potential, gain, threshold. */
  public static enum Properties {
    MEMBRANE_POTENTIAL(0), GAIN(1), THRESHOLD(2);

    private final int index;

    Properties(int index) {
      this.index = index;
    }

    public int getIndex() {
      return this.index;
    }
  };

  /**
   * Creates a new leaky integrator network according to the specified
   * parameters.
   * 
   * @param gamma leaks $\Gamma$
   * @param w weighted adjacency matrix $w$
   * @param epsilon_a transfer function gain learning rates $\epsilon_a$
   * @param epsilon_b transfer function threshold learning rates $\epsilon_b$
   * @param lambda_1 target firing rate distribution parameter 1 $\lambda_1$
   * @param lambda_2 target firing rate distribution parameter 2 $\lambda_2$
   */
  public NetworkLeakyIntegratorWithIntrinsicPlasticityODE(final double[] gamma, final double[][] w, final double[] epsilon_a, final double[] epsilon_b,
      final double[] lambda_1, final double[] lambda_2) {
    this.networkSize = gamma.length;
    this.epsilon_a = epsilon_a;
    this.epsilon_b = epsilon_b;
    this.gamma = gamma;
    this.lambda_1 = lambda_1;
    this.lambda_2 = lambda_2;
    this.w = w;
  }

  /**
   * Returns the dimension of the ODE.
   * 
   * @return 3 * networkSize
   */
  @Override
  public int getDimension() {
    return Properties.values().length * networkSize;
  }

  /**
   * Calculates the first derivatives of the ODE.
   * 
   * @param time time $t$
   * @param state vector $(x_1(t), a_1(t), b_1(t), x_2(t), a_2(t), b_2(t),
   * \dots, x_n(t), a_n(t), b_n(t))$
   * @param stateDot vector $(\dot x_1(t), \dot a_1(t), \dot b_1(t), \dots)$
   */
  @Override
  public void computeDerivatives(final double time, final double[] state, final double[] stateDot) {
    final double[] firingRates = new double[networkSize];
    for (int neuron = 0; neuron < networkSize; neuron++) {
      firingRates[neuron] = firingRate(neuron, state);
    }
    for (int neuron = 0; neuron < networkSize; neuron++) {
      /*
       * auxiliary terms
       */
      // $\frac{q'}{q}$
      final double qPrime_over_q = lambda_1[neuron] + 2.0 * lambda_2[neuron] * firingRates[neuron];
      // a term in $\dot a$ and $\dot b$.
      final double delta = 1.0 - 2.0 * firingRates[neuron] + qPrime_over_q * (1.0 - firingRates[neuron]) * firingRates[neuron];

      /*
       * input
       */
      double input = 0.0;
      for (int afferentNeuron = 0; afferentNeuron < networkSize; afferentNeuron++) {
        input += w[neuron][afferentNeuron] * firingRates[afferentNeuron]; // total network input $I_i = \sum\limits_{j=1}^{N} w_{i j} y_j$
      }

      /*
       * derivatives
       */
      // membrane potential $\dot x = -\Gamma x + w y(t)$
      setValue(neuron, stateDot, Properties.MEMBRANE_POTENTIAL, -gamma[neuron] * getValue(neuron, state, Properties.MEMBRANE_POTENTIAL) + input);

      // transfer function gain $\dot a$
      setValue(neuron, stateDot, Properties.GAIN, epsilon_a[neuron] * (1.0 / getValue(neuron, state, Properties.GAIN) + (getValue(neuron, state, Properties.MEMBRANE_POTENTIAL) - getValue(neuron, state, Properties.THRESHOLD)) * delta));

      // transfer function threshold $\dot b$
      setValue(neuron, stateDot, Properties.THRESHOLD, epsilon_b[neuron] * (-getValue(neuron, state, Properties.GAIN) * delta));
    }
  }

  /**
   * Get the value of a property of a neuron in the given state.
   * 
   * @param neuron index of the neuron in the network
   * @param state overall network state
   * @param property property to get (firing rate, gain, threshold)
   * @return value
   */
  public static double getValue(final int neuron, final double[] state, final Properties property) {
    return state[Properties.values().length * neuron + property.getIndex()];
  }

  /**
   * Set the value of a property of a neuron to the given state. The
   * corresponding field in the given state will be changed according to the
   * property type and the neuron index.
   * 
   * @param neuron index of the neuron in the network
   * @param state overall network state to change
   * @param property property to set (firing rate, gain, threshold)
   * @param value value to set
   */
  public static void setValue(final int neuron, final double[] state, final Properties property, final double value) {
    state[Properties.values().length * neuron + property.getIndex()] = value;
  }

  /**
   * Calculates the firing rate of the given neuron in the given state.
   * 
   * @param neuron index of neuron
   * @param state the whole network state as used in
   * <code>computeDerivatives</code> including membrane potentials, gains and
   * thresholds
   * @return firing rate $y = g(x,a,b)$
   */
  public static double firingRate(final int neuron, final double[] state) {
    return g(state[Properties.values().length * neuron + Properties.MEMBRANE_POTENTIAL.getIndex()],
        state[Properties.values().length * neuron + Properties.GAIN.getIndex()], state[Properties.values().length * neuron + Properties.THRESHOLD.getIndex()]);
  }

  /**
   * Transfer function $g(x, a, b) = \frac{1}{1 + e^{-a(x-b)}}$.
   * 
   * @param x membrane potential $x$
   * @param a transfer function gain $a$
   * @param b transfer function threshold $b$
   * @return firing rate $y = g(x, a, b)$
   */
  public static double g(final double x, final double a, final double b) {
    final double max = TransientDynamicsOfLeakyIntegratorsWithIntrinsicPlasticity.PATTERN_ACTIVE;
    final double min = TransientDynamicsOfLeakyIntegratorsWithIntrinsicPlasticity.PATTERN_INACTIVE;
    return 1.0 / (1.0 + Math.exp(-a * (x - b))) * (max - min) + min;
  }

  /**
   * Inverse transfer function $g^{-1}(y,a,b) = -\frac{1}{a} \log \left( \frac{y_{\mathrm{max}}-y_{\mathrm{min}}}{y-y_{\mathrm{min}}} - 1 \right) + b$.
   * 
   * @param y firing rate $y$
   * @param a transfer function gain $a$
   * @param b transfer function threshold $b$
   * @return firing rate $y = g(x, a, b)$
   */
  public static double gInverse(final double y, final double a, final double b) {
    final double max = TransientDynamicsOfLeakyIntegratorsWithIntrinsicPlasticity.PATTERN_ACTIVE;
    final double min = TransientDynamicsOfLeakyIntegratorsWithIntrinsicPlasticity.PATTERN_INACTIVE;
    return Math.log((y - min) / (max - y)) / a + b;
  }
}