package transientDynamics;

import java.io.BufferedOutputStream;
import java.io.File;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.PrintStream;
import java.util.Collection;
import java.util.List;
import java.util.Locale;

import org.apache.commons.math3.ode.sampling.StepHandler;
import org.apache.commons.math3.ode.sampling.StepInterpolator;

import transientDynamics.NetworkLeakyIntegratorWithIntrinsicPlasticityODE.Properties;

public class NetworkLeakyIntegratorWithIntrinsicPlasticityStepHandler implements StepHandler {
  final PrintStream              neuronStateOut;
  final PrintStream              patternActivityOut;
  final PrintStream              patternOverlapOut;

  final Collection<List<Double>> patterns;

  public NetworkLeakyIntegratorWithIntrinsicPlasticityStepHandler(final Collection<List<Double>> patterns, final File dataDirectory)
      throws FileNotFoundException {
    this.patterns = patterns;
    neuronStateOut = new PrintStream(new BufferedOutputStream(new FileOutputStream(new File(dataDirectory, "neuron.data"))));
    patternActivityOut = new PrintStream(new BufferedOutputStream(new FileOutputStream(new File(dataDirectory, "patternActivity.data"))));
    patternOverlapOut = new PrintStream(new BufferedOutputStream(new FileOutputStream(new File(dataDirectory, "patternOverlap.data"))));
  }

  @Override
  public void handleStep(final StepInterpolator interpolator, final boolean isLast) {
    if (isLast) {
      neuronStateOut.close();
      patternActivityOut.close();
      patternOverlapOut.close();
    } else {
      final int everyNthStepToPrint = 1;
      final double t = interpolator.getCurrentTime();
      final double tDelta = t - interpolator.getPreviousTime();
      if ((int) (t / tDelta) % everyNthStepToPrint == 0) {
        this.everyStep(t, interpolator.getInterpolatedState());
      }
    }
  }

  @Override
  public void init(final double t0, final double[] state, final double t) {
    final int networkSize = state.length / Properties.values().length;
    // print header
    neuronStateOut.print("## t");
    for (int neuron = 0; neuron < networkSize; neuron++) {
      neuronStateOut.print("\t\tx_" + neuron + "(t)");
      neuronStateOut.print("\t\ta_" + neuron + "(t)");
      neuronStateOut.print("\t\tb_" + neuron + "(t)");
      neuronStateOut.print("\t\ty_" + neuron + "(t)");
    }
    neuronStateOut.print("\t\txi(t)");
    neuronStateOut.println();

    // print header
    patternActivityOut.print("## t");
    for (int pattern = 0; pattern < patterns.size(); pattern++) {
      patternActivityOut.print("\t\tA_" + pattern + "(t)");
    }
    patternActivityOut.println();

    // print header
    patternOverlapOut.print("## t");
    for (int pattern = 0; pattern < patterns.size(); pattern++) {
      patternOverlapOut.print("\t\tA_" + pattern + "(t)");
    }
    patternOverlapOut.println();

    this.everyStep(t0, state);
  }

  public void everyStep(final double t, final double[] state) {
    // print current state
    this.printNeuronState(t, state, neuronStateOut);
    this.printPatternActivity(t, state, patternActivityOut);
    this.printPatternOverlap(t, state, patternOverlapOut);
  }

  public void printNeuronState(final double t, final double[] state, final PrintStream out) {
    final int networkSize = state.length / Properties.values().length;

    // Check for NaN.
    for (final double value : state) {
      if (Double.isNaN(value)) {
        throw new IllegalStateException("NaN occurred.");
      }
    }

    // print current state
    out.printf(Locale.ENGLISH, "%.3f", t); // time $t$
    final int maximumNumberOfNeuronsToPrint = 100;
    for (int neuron = 0; neuron < networkSize && neuron < maximumNumberOfNeuronsToPrint; neuron++) {
      out.printf(Locale.ENGLISH, "\t%.3f\t%.3f\t%.3f", state[Properties.values().length * neuron + Properties.MEMBRANE_POTENTIAL.getIndex()],
          state[Properties.values().length * neuron + Properties.GAIN.getIndex()], state[Properties.values().length * neuron + Properties.THRESHOLD.getIndex()]); // membrane potential $x$, transfer function gain $a$, transfer function threshold $b$
      out.printf(Locale.ENGLISH, "\t%.3f", NetworkLeakyIntegratorWithIntrinsicPlasticityODE.firingRate(neuron, state)); // firing rate $y$
    }
    out.println();
  }

  public void printPatternActivity(final double t, final double[] state, final PrintStream out) {
    // print pattern activity
    out.printf(Locale.ENGLISH, "%.8f", t); // time $t$
    for (final List<Double> pattern : patterns) {
      double patternActivity = TransientDynamicsOfLeakyIntegratorsWithIntrinsicPlasticity.getPatternActivity(pattern, state);
      out.printf(Locale.ENGLISH, "\t%.4f", patternActivity);
    }
    out.println();
  }

  /**
   * Print the overlap of the given state with all the patterns successively.
   * The overlap is the cosine of the angle between the state vector and the
   * pattern vector.
   * 
   * @param t time $t$
   * @param state network state
   * @param out stream to print to
   */
  public void printPatternOverlap(final double t, final double[] state, final PrintStream out) {
    // print pattern overlap
    out.printf(Locale.ENGLISH, "%.3f", t); // time $t$
    for (final List<Double> pattern : patterns) {
      final double patternOverlap = TransientDynamicsOfLeakyIntegratorsWithIntrinsicPlasticity.getPatternOverlap(pattern, state);
      out.printf(Locale.ENGLISH, "\t%.3f", patternOverlap);
    }
    out.println();
  }
}