package transientDynamics;

import java.io.BufferedOutputStream;
import java.io.File;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.PrintStream;
import java.util.Collection;
import java.util.Collections;
import java.util.List;
import java.util.Locale;
import java.util.Random;
import java.util.Set;
import java.util.Vector;
import java.util.concurrent.Callable;
import java.util.concurrent.ExecutionException;
import java.util.concurrent.ExecutorService;
import java.util.concurrent.Executors;
import java.util.concurrent.Future;

import org.apache.commons.math3.ode.FirstOrderDifferentialEquations;
import org.apache.commons.math3.ode.FirstOrderIntegrator;
import org.apache.commons.math3.ode.nonstiff.ClassicalRungeKuttaIntegrator;

import transientDynamics.NetworkLeakyIntegratorWithIntrinsicPlasticityODE.Properties;
import brain.Brain;
import brain.LeakyIntegratorModule;
import brain.Module;

class PatternsStability implements Callable<String> {
  final double[][]               adjacencyMatrix;
  final Collection<List<Double>> patterns;
  final double                   w;
  final double                   z;
  final double                   leak;
  final double                   gain;
  final double                   threshold;

  public PatternsStability(final double[][] adjacencyMatrix, final Collection<List<Double>> patterns, final double w, final double z, final double leak,
      final double gain, final double threshold) {
    this.adjacencyMatrix = adjacencyMatrix;
    this.patterns = patterns;
    this.w = w;
    this.z = z;
    this.leak = leak;
    this.gain = gain;
    this.threshold = threshold;
  }

  @Override
  public String call() throws Exception {
    final ExecutorService executor = Executors.newSingleThreadExecutor();

    final double[][] weights = new double[adjacencyMatrix.length][adjacencyMatrix[0].length];
    for (int i = 0; i < adjacencyMatrix.length; i++) {
      for (int j = 0; j < adjacencyMatrix[i].length; j++) {
        weights[i][j] = adjacencyMatrix[i][j] > 0.0 ? w : (adjacencyMatrix[i][j] < 0.0 ? z : 0.0);
      }
    }

    // init calculation
    final List<Callable<Double>> tasks = new Vector<Callable<Double>>();
    for (final List<Double> pattern : patterns) {
      final Callable<Double> task = new PatternStability(weights, pattern, leak, gain, threshold);
      tasks.add(task);
    }

    // start calculation
    final List<Future<Double>> decayTimes = executor.invokeAll(tasks);
    // collect results
    double decayTimeMax = Double.NEGATIVE_INFINITY;
    int stablePatternNumber = 0;
    double minActivity = Double.POSITIVE_INFINITY;
    double maxActivity = Double.NEGATIVE_INFINITY;
    double minMembranePotential = Double.POSITIVE_INFINITY;
    double maxMembranePotential = Double.NEGATIVE_INFINITY;
    final List<Double> allActivities = new Vector<Double>();
    for (int i = 0; i < tasks.size(); i++) {
      final Callable<Double> task = tasks.get(i);
      final Future<Double> decayTime = decayTimes.get(i);
      Double time = null;
      try {
        time = decayTime.get();
      } catch (ExecutionException e) {
      }

      if (time == null) {
        // stable pattern
        stablePatternNumber++;
      } else {
        // unstable pattern
        decayTimeMax = Math.max(decayTimeMax, time);
      }
      allActivities.addAll(((PatternStability) task).activities);
      minActivity = Math.min(minActivity, Collections.min(((PatternStability) task).activities));
      maxActivity = Math.max(maxActivity, Collections.max(((PatternStability) task).activities));
      minMembranePotential = Math.min(minMembranePotential, Collections.min(((PatternStability) task).membranePotentials));
      maxMembranePotential = Math.max(maxMembranePotential, Collections.max(((PatternStability) task).membranePotentials));
    }
    if (maxActivity < 1e-2) {
      stablePatternNumber = -1;
    }
    executor.shutdown();

    // Calculating the order parameter $\sqrt{\sum_{i=1}^3 (y_i^{(1)}-y_i^{(2)})^2}$
    double orderParameter = 0.0;
    for (int i = 0; i < adjacencyMatrix.length; i++) {
      orderParameter += Math.pow(allActivities.get(i + 0 * 3) - allActivities.get(i + 1 * 3), 2);
    }
    orderParameter = Math.sqrt(orderParameter);

    return String.format(Locale.ENGLISH, "%.10f\t%.10f\t%.10f\t%.10f\t%d\t%.10f\t%.10f\t%.10f\t%.10f\t%.10f", w, z, gain, threshold, stablePatternNumber,
        minActivity, maxActivity, minMembranePotential, maxMembranePotential, orderParameter);
  }
}

public class PatternStability implements Callable<Double> {
  private final List<Double> pattern;
  private final double[][]   w;
  final double               leak;
  final double               gain;
  final double               threshold;

  public List<Double>        membranePotentials;
  public List<Double>        activities;

  public PatternStability(final double[][] w, final List<Double> pattern, final double leak, final double gain, final double threshold) {
    this.w = w;
    this.pattern = pattern;
    this.leak = leak;
    this.gain = gain;
    this.threshold = threshold;
  }

  @Override
  public Double call() throws Exception {
    final double tDelta = 1e-1;
    final double tMin = 0.0;
    final double tMax = 1e4;
    final double highActivity = 0.99;
    final double lowActivity = 1.0 - highActivity;

    final int networkSize = this.w.length;

    final double[] gamma = new double[networkSize]; // $\Gamma$
    final double[] epsilon_a = new double[networkSize]; // $\epsilon_a$
    final double[] epsilon_b = new double[networkSize]; // $\epsilon_b$
    final double[] lambda_1 = new double[networkSize]; // $\lambda_1$
    final double[] lambda_2 = new double[networkSize]; // $\lambda_2$
    final double[] state = new double[Properties.values().length * networkSize];

    for (int neuron = 0; neuron < networkSize; neuron++) {
      gamma[neuron] = this.leak;
      epsilon_a[neuron] = 0.0; // disable intrinsic plasticity
      epsilon_b[neuron] = 0.0; // disable intrinsic plasticity
      lambda_1[neuron] = 0.0;
      lambda_2[neuron] = 0.0;

      //final double a = neuron == 1 ? 6.0 : this.gain; // TODO: replace by this.gain
      //final double b = neuron == 1 ? 1.0 : this.threshold; // TODO: replace by this.threshold
      final double a = this.gain;
      final double b = this.threshold;
      final double y = this.pattern.get(neuron) > 0.5 ? highActivity : lowActivity;
      final double x = NetworkLeakyIntegratorWithIntrinsicPlasticityODE.gInverse(y, a, b);
      NetworkLeakyIntegratorWithIntrinsicPlasticityODE.setValue(neuron, state, Properties.MEMBRANE_POTENTIAL, x); // $x(t_{\mathrm{min}})$
      NetworkLeakyIntegratorWithIntrinsicPlasticityODE.setValue(neuron, state, Properties.GAIN, a); // $a(t_{\mathrm{min}})$
      NetworkLeakyIntegratorWithIntrinsicPlasticityODE.setValue(neuron, state, Properties.THRESHOLD, b); // $b(t_{\mathrm{min}})$
    }

    // Set up the ODE.
    final FirstOrderDifferentialEquations ode = new NetworkLeakyIntegratorWithIntrinsicPlasticityODE(gamma, w, epsilon_a, epsilon_b, lambda_1, lambda_2);

    // Set the integration method.
    final FirstOrderIntegrator integrator = new ClassicalRungeKuttaIntegrator(tDelta);

    integrator.integrate(ode, tMin, state, tMax, state);

    activities = new Vector<Double>(networkSize);
    membranePotentials = new Vector<Double>(networkSize);
    for (int neuron = 0; neuron < networkSize; neuron++) {
      final double a = NetworkLeakyIntegratorWithIntrinsicPlasticityODE.getValue(neuron, state, Properties.GAIN); // $a(t_{\mathrm{min}})$
      final double b = NetworkLeakyIntegratorWithIntrinsicPlasticityODE.getValue(neuron, state, Properties.THRESHOLD); // $b(t_{\mathrm{min}})$
      final double x = NetworkLeakyIntegratorWithIntrinsicPlasticityODE.getValue(neuron, state, Properties.MEMBRANE_POTENTIAL);
      final double y = NetworkLeakyIntegratorWithIntrinsicPlasticityODE.g(x, a, b);
      membranePotentials.add(x);
      activities.add(y);
    }

    // Clique decayed at time t.
    final double patternOverlap = TransientDynamicsOfLeakyIntegratorsWithIntrinsicPlasticity.getPatternOverlap(pattern, state);
    //final double patternActivity = TransientDynamicsOfLeakyIntegratorsWithIntrinsicPlasticity.getPatternActivity(pattern, state);
    if (patternOverlap < 0.90) {
      return tMax;
    }

    // Clique remained active till time t.
    throw new Exception("Clique did not decay till t = " + tMax + ".");
  }

  public static void main(String[] args) throws InterruptedException, FileNotFoundException {
    //final double[] leaks = new double[] { 0.0, 0.1, 0.5, 1.0 };
    final double[] leaks = new double[] { 1.0 };
    //final int[] sizes = new int[] { 3, 4, 5, 6, 7, 8, 9, 10, 20, 30, 40, 50, 60, 70, 80, 90, 100 };
    final double[] thresholds = new double[] { 0.0, 1.0, 2.0, 4.0, 8.0, 16.0 };

    final int availableProcessors = Runtime.getRuntime().availableProcessors();
    final int usedProcessors = Math.max(1, availableProcessors);
    System.out.printf(Locale.ENGLISH, "# processors used/available: %d/%d\n", usedProcessors, availableProcessors);
    final ExecutorService executor = Executors.newFixedThreadPool(usedProcessors);

    /*
     * create network
     */
    //final double[][] adjacencyMatrix = Graphs.addInhibitoryLinks(Graphs.grosGraphAdjacencyMatrix);
    //final double[][] adjacencyMatrix = Graphs.addInhibitoryLinks(Graphs.gros2GraphAdjacencyMatrix);
    final double[][] adjacencyMatrix = new double[][] { { 0, 1, -1 }, { 1, 0, 1 }, { -1, 1, 0 } };

    /*
     * create patterns
     */
    final Collection<List<Double>> patterns = new Vector<List<Double>>();
    Module module = new LeakyIntegratorModule(adjacencyMatrix, new Random(0));
    for (Set<Integer> clique : module.getCliques()) {
      List<Double> pattern = new Vector<Double>(module.getSize());
      for (int neuron = 0; neuron < module.getSize(); neuron++) {
        pattern.add(clique.contains(neuron) ? 1.0 : 0.0);
      }
      patterns.add(pattern);
    }

    // 0: varying w+ and w-
    // 1: varying a and b
    final int type = 1;
    switch (type) {
      case 0:
        //for (final int size : sizes) {
        for (final double leak : leaks) {
          for (final double threshold : thresholds) {
            final double wMin = 0;
            final double wMax = leak == 0.0 ? 5 : 25.0 * leak; // 12
            final int wSteps = 25 * 2; // 40
            final double zMin = leak == 0.0 ? -5 : -25.0 * leak; // -12
            final double zMax = 0;
            final int zSteps = 25 * 2; // 40

            //final PrintStream out = System.out;
            final File directory = new File(Brain.DEFAULT_DATA_OUTPUT_DIRECTORY, "test2");
            directory.mkdirs();
            final File file = new File(directory, "pattern-stability_leak=" + leak + "_threshold=" + threshold + ".data");
            System.out.println(file);
            final PrintStream out = new PrintStream(new BufferedOutputStream(new FileOutputStream(file)));
            out.println("# w, z, gain, threshold, stable cliques, minActivity, maxActivity, minMembranePotential, maxMembranePotential");

            /*
             * create phase diagram
             */
            final List<Callable<String>> tasks = new Vector<Callable<String>>();
            for (int wStep = 0; wStep <= wSteps; wStep++) {
              final double w = wMin + wStep * (wMax - wMin) / (double) wSteps;
              for (int zStep = 0; zStep <= zSteps; zStep++) {
                final double z = zMax - zStep * (zMax - zMin) / (double) zSteps;
                // init calculation
                final double gain = 100;
                final Callable<String> task = new PatternsStability(adjacencyMatrix, patterns, w, z, leak, gain, threshold);
                tasks.add(task);
              }
            }

            // start calculation
            final List<Future<String>> results = executor.invokeAll(tasks);

            // collect results
            double wLast = Double.NaN;
            for (int i = 0; i < tasks.size(); i++) {
              final PatternsStability task = (PatternsStability) tasks.get(i);
              final Future<String> result = results.get(i);
              if (wLast != task.w) {
                wLast = task.w;
                out.println(); // gnuplot single isocline scan new lines
              }
              try {
                out.println(result.get());
              } catch (ExecutionException e) {
                e.printStackTrace();
              }
            }
            out.close();
          }
        }
        break;
      case 1:
        for (final double leak : leaks) {
          /*
           * setup
           */
          final File directory = new File(Brain.DEFAULT_DATA_OUTPUT_DIRECTORY, "pattern_stability");
          directory.mkdirs();

          final double aMin = 1;
          final double aMax = aMin + 10;
          final int aStepNumber = 200;
          final List<Double> aSteps = new Vector<Double>();
          for (int aStep = 0; aStep < aStepNumber; aStep++) {
            final double a = aMin + (double) aStep * (aMax - aMin) / (double) (aStepNumber + 1);
            aSteps.add(a);
          }
          final double bMin = -1;
          final double bMax = +2;
          final int bStepNumber = 200;
          final List<Double> bSteps = new Vector<Double>();
          for (int bStep = 0; bStep < bStepNumber; bStep++) {
            final double b = bMin + (double) bStep * (bMax - bMin) / (double) (bStepNumber - 1);
            bSteps.add(b);
          }
          final File file = new File(directory, "pattern_stability.data");

          /*
           * start
           */
          System.out.println(file);
          //System.out.println("a=" + aSteps);
          //System.out.println("b=" + bSteps);
          final PrintStream out = new PrintStream(new BufferedOutputStream(new FileOutputStream(file)));
          out.println("# w, z, gain, threshold, stable cliques, minActivity, maxActivity, minMembranePotential, maxMembranePotential");

          /*
           * create phase diagram
           */
          final List<Callable<String>> tasks = new Vector<Callable<String>>();
          for (final double a : aSteps) {
            for (final double b : bSteps) {
              // init calculation
              final double wPlus = +1.0;
              final double wMinus = -1.0;
              final Callable<String> task = new PatternsStability(adjacencyMatrix, patterns, wPlus, wMinus, leak, a, b);
              tasks.add(task);
            }
          }

          // start calculation
          final List<Future<String>> results = executor.invokeAll(tasks);

          // collect results
          double aLast = Double.NaN;
          for (int i = 0; i < tasks.size(); i++) {
            final PatternsStability task = (PatternsStability) tasks.get(i);
            final Future<String> result = results.get(i);
            if (aLast != task.gain) {
              aLast = task.gain;
              out.println(); // gnuplot single isocline scan new lines
            }
            try {
              out.println(result.get());
            } catch (ExecutionException e) {
              e.printStackTrace();
            }
          }
          out.close();
        }
        break;
    }
    executor.shutdown();
  }
}