package transientDynamics;

import java.io.BufferedOutputStream;
import java.io.File;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.PrintStream;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.List;
import java.util.Locale;
import java.util.Random;
import java.util.Set;
import java.util.Vector;

import org.apache.commons.math3.ode.FirstOrderDifferentialEquations;
import org.apache.commons.math3.ode.FirstOrderIntegrator;
import org.apache.commons.math3.ode.nonstiff.ClassicalRungeKuttaIntegrator;
import org.apache.commons.math3.ode.sampling.StepHandler;

import transientDynamics.NetworkLeakyIntegratorWithIntrinsicPlasticityODE.Properties;
import brain.Brain;
import brain.LeakyIntegratorModule;
import brain.Module;

/**
 * This class extends the calculations in
 * <code>SelfOrganizedStochasticTippingInSlowFastDynamicalSystems</code> from a
 * single neuron to a network of neurons.
 * 
 * @author Mathias Linkerhand
 * @version 2013-01-31
 * @since 2012
 */
// running simulation: java -cp bin:lib/* transientDynamics.TransientDynamicsOfLeakyIntegratorsWithIntrinsicPlasticity > data
public class TransientDynamicsOfLeakyIntegratorsWithIntrinsicPlasticity {
  public static final int PATTERN_ACTIVE   = +1; // binary: +1, biploar: +1
  public static final int PATTERN_INACTIVE = 0; // binary: 0, bipolar: -1

  public static enum IntegrationMethod {
    EULER, CLASSICAL_RUNGE_KUTTA
  };

  /**
   * Running the simulation.
   * 
   * @throws FileNotFoundException
   */
  public static void run(final int simulationParameters, final String... parameters) throws FileNotFoundException {
    final Random random = new Random(0);
    // Simulation constants.
    final double tMin; // $t_{\mathrm{min}}$
    final double tMax; // $t_{\mathrm{max}}$
    final double tDelta; // $\delta t$
    final double[][] w; // $w$
    final double[] gamma; // $\Gamma$
    final double[] epsilon_a; // $\epsilon_a$
    final double[] epsilon_b; // $\epsilon_b$
    final double[] lambda_1; // $\lambda_1$
    final double[] lambda_2; // $\lambda_2$
    final double[] state; // initial state $x((t_{\mathrm{min}}), a(t_{\mathrm{min}}), b(t_{\mathrm{min}}))$
    final Collection<List<Double>> patterns;
    final int networkSize;
    final int numberOfPatterns;

    switch (simulationParameters) {
      case 1: // normal latching transitions
        final double sparseness = 0.3;
        networkSize = Integer.parseInt(parameters[1]);
        numberOfPatterns = Integer.parseInt(parameters[2]);
        patterns = createRandomBinarySparsePatterns(sparseness, networkSize, numberOfPatterns, random);
        tMin = 0;
        tMax = 1e3;
        tDelta = 1e-1;
        w = TransientDynamicsOfLeakyIntegratorsWithIntrinsicPlasticity.hopfieldWeightedAdjacencyMatrix(networkSize, patterns, sparseness);
        gamma = new double[networkSize];
        epsilon_a = new double[networkSize];
        epsilon_b = new double[networkSize];
        lambda_1 = new double[networkSize];
        lambda_2 = new double[networkSize];
        state = new double[Properties.values().length * networkSize];

        for (int neuron = 0; neuron < networkSize; neuron++) {
          gamma[neuron] = 1.0;
          epsilon_a[neuron] = 0.1; // enable intrinsic plasticity
          epsilon_b[neuron] = 0.01; // enable intrinsic plasticity
          lambda_1[neuron] = -2.67210; // $\lambda_1 \approx -2.67210$ for $\mu=0.3$, if $\lambda_2 = 0$
          lambda_2[neuron] = 0.0;
          // TODO: use g inverse to set these values
          NetworkLeakyIntegratorWithIntrinsicPlasticityODE.setValue(neuron, state, Properties.MEMBRANE_POTENTIAL,
              (PATTERN_ACTIVE - PATTERN_INACTIVE) * random.nextDouble() + PATTERN_INACTIVE); // $x(t_{\mathrm{min}})$
          NetworkLeakyIntegratorWithIntrinsicPlasticityODE.setValue(neuron, state, Properties.GAIN, 2.0); // $a(t_{\mathrm{min}})$
          NetworkLeakyIntegratorWithIntrinsicPlasticityODE.setValue(neuron, state, Properties.THRESHOLD, sparseness * (PATTERN_ACTIVE - PATTERN_INACTIVE)
              + PATTERN_INACTIVE); // $b(t_{\mathrm{min}})$
        }
        break;
      case 2: // three vertex graph
        /*
         * create network
         */
        final double[][] adjacencyMatrix = new double[][] { { 0, 1, -1 }, { 1, 0, 1 }, { -1, 1, 0 } };

        /*
         * create patterns
         */
        patterns = new Vector<List<Double>>();
        Module module = new LeakyIntegratorModule(adjacencyMatrix, new Random(0));
        for (Set<Integer> clique : module.getCliques()) {
          List<Double> pattern = new Vector<Double>(module.getSize());
          for (int neuron = 0; neuron < module.getSize(); neuron++) {
            pattern.add(clique.contains(neuron) ? 1.0 : 0.0);
          }
          patterns.add(pattern);
        }
        networkSize = module.getSize();
        w = new double[networkSize][networkSize];
        for(int i = 0; i < networkSize; i++){
          for(int j = 0; j < networkSize; j++){
            w[i][j] = module.getLink(i, j);
          }
        }

        gamma = new double[networkSize];
        epsilon_a = new double[networkSize];
        epsilon_b = new double[networkSize];
        lambda_1 = new double[networkSize];
        lambda_2 = new double[networkSize];
        state = new double[Properties.values().length * networkSize];

        final double highActivity = 0.9;
        final double lowActivity = 1.0 - highActivity;
        
        final List<Double> initialPattern = patterns.iterator().next();

        for (int neuron = 0; neuron < networkSize; neuron++) {
          gamma[neuron] = 1.0;
          epsilon_a[neuron] = 0.1; // enable intrinsic plasticity
          epsilon_b[neuron] = 0.01; // enable intrinsic plasticity
          lambda_1[neuron] = 0.0; // $\lambda_1 = 0.0$ for $\mu=0.5$, if $\lambda_2 = 0$
          lambda_2[neuron] = 0.0;

          final double a = 5.0; // in the paper: a=5.0
          final double b = -0.5; // in the paper: b=-0.5
          final double y = initialPattern.get(neuron) > 0.5 ? highActivity : lowActivity;
          final double x = NetworkLeakyIntegratorWithIntrinsicPlasticityODE.gInverse(y, a, b);
          NetworkLeakyIntegratorWithIntrinsicPlasticityODE.setValue(neuron, state, Properties.GAIN, a); // $a(t_{\mathrm{min}})$
          NetworkLeakyIntegratorWithIntrinsicPlasticityODE.setValue(neuron, state, Properties.THRESHOLD, b); // $b(t_{\mathrm{min}})$
          NetworkLeakyIntegratorWithIntrinsicPlasticityODE.setValue(neuron, state, Properties.MEMBRANE_POTENTIAL, x); // $x(t_{\mathrm{min}})$
        }

        tMin = 0;
        tMax = 2e3;
        tDelta = 1e-1;
        break;
      default:
        throw new IllegalArgumentException("no such parameter set");
    }

    // Set up the ODE.
    final FirstOrderDifferentialEquations ode = new NetworkLeakyIntegratorWithIntrinsicPlasticityODE(gamma, w, epsilon_a, epsilon_b, lambda_1, lambda_2);

    // Set the integration method.
    final FirstOrderIntegrator integrator = new ClassicalRungeKuttaIntegrator(tDelta);

    // Output the time steps.
    final File dataDirectory = new File(Brain.DEFAULT_DATA_OUTPUT_DIRECTORY, parameters[3]);
    dataDirectory.mkdirs();

    {
      // Write all stored patterns to a file.
      final PrintStream patternsOut = new PrintStream(new BufferedOutputStream(new FileOutputStream(new File(dataDirectory, "patterns.data"))));
      patternsOut.println("## stored patterns");
      patternsOut.println("## each line a pattern");
      for (final List<Double> pattern : patterns) {
        for (int neuron = 0; neuron < pattern.size(); neuron++) {
          if (neuron != 0) {
            patternsOut.print("\t");
          }
          patternsOut.print(pattern.get(neuron));
        }
        patternsOut.println();
      }
      patternsOut.close();
    }

    {
      // Write all mutual pattern overlaps to a file.
      final PrintStream patternOverlapsOut = new PrintStream(new BufferedOutputStream(new FileOutputStream(
          new File(dataDirectory, "mutualPatternOverlaps.data"))));
      patternOverlapsOut.println("## mutual pattern overlaps");
      for (List<Double> pattern1 : patterns) {
        for (List<Double> pattern2 : patterns) {
          patternOverlapsOut.printf(Locale.ENGLISH, "%.4f", getPatternOverlap(pattern1, pattern2));
          patternOverlapsOut.print("\t");
        }
        patternOverlapsOut.println();
      }
      patternOverlapsOut.close();
    }

    {
      // Write all mutual pattern activities to a file.
      final PrintStream patternActivitiesOut = new PrintStream(new BufferedOutputStream(new FileOutputStream(new File(dataDirectory,
          "mutualPatternActivities.data"))));
      patternActivitiesOut.println("## mutual pattern activities");
      for (List<Double> pattern1 : patterns) {
        for (List<Double> pattern2 : patterns) {
          patternActivitiesOut.printf(Locale.ENGLISH, "%.4f", getPatternActivity(pattern1, pattern2));
          patternActivitiesOut.print("\t");
        }
        patternActivitiesOut.println();
      }
      patternActivitiesOut.close();
    }

    // Create step handler for numerical integration to save each step to a file.
    final StepHandler stepHandler = new NetworkLeakyIntegratorWithIntrinsicPlasticityStepHandler(patterns, dataDirectory);
    integrator.addStepHandler(stepHandler);

    // Do the integration.
    integrator.integrate(ode, tMin, state, tMax, state);
  }

  /**
   * Should be used for binary patterns only.
   * 
   * @param pattern
   * @return activity of the state according to the given pattern
   */
  public static double getPatternActivity(final List<Double> pattern, final double[] state) {
    double patternActivity = 0.0;
    int activePatternSize = 0;
    for (int neuron = 0; neuron < pattern.size(); neuron++) {
      final double firingRate = NetworkLeakyIntegratorWithIntrinsicPlasticityODE.firingRate(neuron, state);
      patternActivity += pattern.get(neuron) * firingRate;
      if (pattern.get(neuron) != 0) {
        activePatternSize++;
      }
    }
    patternActivity /= activePatternSize;
    return patternActivity;
  }

  /**
   * Should be used for binary patterns only.
   * 
   * @return activity of the given patterns
   */
  public static double getPatternActivity(final List<Double> pattern1, final List<Double> pattern2) {
    double patternActivity = 0.0;
    int activePatternSize = 0;
    for (int neuron = 0; neuron < pattern1.size(); neuron++) {
      patternActivity += pattern1.get(neuron) * pattern2.get(neuron);
      if (pattern1.get(neuron) != 0) {
        activePatternSize++;
      }
    }
    patternActivity /= activePatternSize;
    return patternActivity;
  }

  public static double getPatternOverlap(final List<Double> pattern, final double[] state) {
    double scalarAB = 0.0;
    double scalarAA = 0.0;
    double scalarBB = 0.0;
    for (int neuron = 0; neuron < pattern.size(); neuron++) {
      final double firingRate = NetworkLeakyIntegratorWithIntrinsicPlasticityODE.firingRate(neuron, state);
      scalarAB += firingRate * pattern.get(neuron);
      scalarAA += Math.pow(firingRate, 2);
      scalarBB += Math.pow(pattern.get(neuron), 2);
    }
    final double cosTheta = scalarAB / Math.sqrt(scalarAA) / Math.sqrt(scalarBB);
    return cosTheta;
  }

  public static double getPatternOverlap(final List<Double> pattern1, final List<Double> pattern2) {
    double scalarAB = 0.0;
    double scalarAA = 0.0;
    double scalarBB = 0.0;
    for (int neuron = 0; neuron < pattern1.size(); neuron++) {
      scalarAB += pattern1.get(neuron) * pattern2.get(neuron);
      scalarAA += Math.pow(pattern1.get(neuron), 2);
      scalarBB += Math.pow(pattern2.get(neuron), 2);
    }
    final double cosTheta = scalarAB / Math.sqrt(scalarAA) / Math.sqrt(scalarBB);
    return cosTheta;
  }

  public static double[][] hopfieldWeightedAdjacencyMatrix(final int networkSize, final Collection<List<Double>> patterns, final double sparseness) {
    final double[] meanActivityOfUnitOverAllMemoryPatterns = new double[networkSize];
    for (int neuron = 0; neuron < networkSize; neuron++) {
      for (final List<Double> pattern : patterns) {
        meanActivityOfUnitOverAllMemoryPatterns[neuron] += pattern.get(neuron);
      }
      meanActivityOfUnitOverAllMemoryPatterns[neuron] /= patterns.size();
    }

    final double[][] weightedAdjacencyMatrix = new double[networkSize][networkSize];
    for (final List<Double> pattern : patterns) {
      for (int i = 0; i < networkSize; i++) {
        for (int j = 0; j < networkSize; j++) {
          /*
           * Symmetric covariant learning rule. This is important for the
           * working simulation with binary patterns.
           */
          weightedAdjacencyMatrix[i][j] += i == j ? 0 : (pattern.get(i) - meanActivityOfUnitOverAllMemoryPatterns[i])
              * (pattern.get(j) - meanActivityOfUnitOverAllMemoryPatterns[j]);
        }
      }
    }
    for (int i = 0; i < networkSize; i++) {
      for (int j = 0; j < networkSize; j++) {
        /*
         * This is important for the working simulation with binary patterns.
         */
        weightedAdjacencyMatrix[i][j] /= networkSize - 1;
        weightedAdjacencyMatrix[i][j] /= sparseness;
        weightedAdjacencyMatrix[i][j] /= 0.2;
      }
    }
    return weightedAdjacencyMatrix;
  }

  public static Collection<List<Double>> createRandomContinuousSparsePatterns(final double sparseness, final int networkSize, final int numberOfPatterns,
      final Random random) {
    final List<List<Double>> patterns = new ArrayList<List<Double>>(numberOfPatterns);

    for (int patternIndex = 0; patternIndex < numberOfPatterns; patternIndex++) {
      final List<Double> pattern = createRandomContinuousSparsePattern(sparseness, networkSize, random);
      patterns.add(pattern);
    }
    return patterns;
  }

  public static List<Double> createRandomContinuousSparsePattern(final double sparseness, final int networkSize, final Random random) {
    final List<Double> pattern = new ArrayList<Double>(networkSize);
    for (int neuron = 0; neuron < networkSize; neuron++) {
      final double x = random.nextDouble();
      //pattern.add((double) (x < sparseness ? PATTERN_ACTIVE : PATTERN_INACTIVE)); // discrete pattern (binary)
      pattern.add(x < sparseness ? Math.min(1.0, -1.0 / 2.0 * Math.log10(1.0 - x / sparseness)) : 0.0); // continuous pattern
      //pattern.add(x < sparseness ? (-1.0 / 2.0 * Math.log10(1.0 - x / sparseness)) : 0.0); // continuous pattern
    }
    return pattern;
  }

  /**
   * Create randomly distributed binary sparse patterns with a given sparseness.
   * _______#_______#_____#_______#_##_#_#_#_#____#___#_#___#____#_____
   * #__#__####_##____#_#__#____________###__##__#____##_______#_#____#
   * ______________#_#___#___#___#_#_#_____#__#_##_#_#_____#____#______
   * 
   * @param sparseness
   * @param networkSize
   * @param numberOfPatterns
   * @param random
   * @return list of random binary patterns
   */
  public static Collection<List<Double>> createRandomBinarySparsePatterns(final double sparseness, final int networkSize, final int numberOfPatterns,
      final Random random) {
    final List<List<Double>> patterns = new ArrayList<List<Double>>(numberOfPatterns);

    for (int patternIndex = 0; patternIndex < numberOfPatterns; patternIndex++) {
      final List<Double> pattern = createRandomBinarySparsePattern(sparseness, networkSize, random);
      patterns.add(pattern);
    }
    return patterns;
  }

  /**
   * Create a randomly distributed binary sparse pattern with a given
   * sparseness.
   * _______#_______#_____#_______#_##_#_#_#_#____#___#_#___#____#_____
   * 
   * @param sparseness
   * @param networkSize
   * @param random
   * @return binary pattern
   */
  public static List<Double> createRandomBinarySparsePattern(final double sparseness, final int networkSize, final Random random) {
    final List<Double> pattern = new ArrayList<Double>(networkSize);
    for (int neuron = 0; neuron < networkSize; neuron++) {
      final double x = (double) neuron / networkSize;
      pattern.add((double) (x < sparseness ? PATTERN_ACTIVE : PATTERN_INACTIVE)); // discrete pattern (binary)
    }
    Collections.shuffle(pattern, random);
    return pattern;
  }

  /**
   * Running the application.
   * 
   * @throws FileNotFoundException
   */
  public static void main(String[] args) throws FileNotFoundException {
    TransientDynamicsOfLeakyIntegratorsWithIntrinsicPlasticity.run(Integer.parseInt(args[0]), args);
  }
}