package utils;

import java.io.File;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.OutputStream;
import java.io.PrintStream;
import java.util.LinkedList;
import java.util.List;
import java.util.Locale;

import org.apache.commons.math3.analysis.UnivariateFunction;
import org.apache.commons.math3.analysis.integration.SimpsonIntegrator;
import org.apache.commons.math3.analysis.integration.UnivariateIntegrator;

/**
 * A histogram for counting values. Alternative: AdaptiveHistogram of the
 * library Hist4j. Calculates also the relative entropy (Kullback-Leibler
 * divergence) of the histograms’s distributions.
 * 
 * @author Mathias Linkerhand
 * @version 2013-01-31
 * @since 2011
 */
public class Histogram {
  private long[]           frequencies;
  private long             maximumFrequency;
  private double           minimumValue;
  private double           maximumValue;
  private int              maximumFrequencyBin;
  private long             totalNumberOfValues;
  private double           sumOfValues;
  private int              binCount;
  private List<Double>     valuesHistory;
  private int              maximumNumberOfValues;

  private static final int BORDER_FLOOR = 0;
  private static final int BORDER_CEIL  = 1;
  private int              border       = BORDER_FLOOR;

  public Histogram(final int binCount, final double minValue, final double maxValue) {
    this(binCount, minValue, maxValue, -1);
  }

  public Histogram(final int binCount, final double minValue, final double maxValue, final int maximumNumberOfValues) {
    this.setBinCount(binCount);
    this.minimumValue = minValue;
    this.maximumValue = maxValue;
    this.maximumNumberOfValues = maximumNumberOfValues;
    this.maximumFrequency = 0;
    this.totalNumberOfValues = 0;

    // If maximumNumberOfValues <= 0 no old valuesHistory are to be removed.
    if (maximumNumberOfValues <= 0) {
      this.valuesHistory = new LinkedList<Double>();
    }

    this.frequencies = new long[binCount];
  }

  public static Histogram createExponentialDistributionHistogram(final int binCount, final double λ1, final double λ2) {
    final UnivariateIntegrator integrator = new SimpsonIntegrator(); // The intergrator could be replaced to get more accurate results.
    final UnivariateFunction targetDistribution = new UnivariateFunction() {
      /**
       * Calculates the target distribution given λ1 and λ2 according to the
       * equation:
       * 
       * $p(y) = \exp(\lambda_1 y + \lambda_2 y^{2}=$
       */
      @Override
      public double value(final double firingRate) {
        return Math.exp(λ1 * firingRate + λ2 * firingRate * firingRate);
      }
    };
    final int maxEval = 1000; // if reached, exception is thrown
    final double y_min = 0.0;
    final double y_max = 1.0;
    final double norm = integrator.integrate(maxEval, targetDistribution, y_min, y_max);

    final double[] frequencies = new double[binCount];
    double maximumFrequency = 0.0;
    for (int bin = 0; bin < binCount; bin++) {
      final double y_min_i = y_min + (y_max - y_min) * bin / binCount;
      final double y_max_i = y_min + (y_max - y_min) * (bin + 1) / binCount;
      final double frequency = integrator.integrate(maxEval, targetDistribution, y_min_i, y_max_i) / norm;
      frequencies[bin] = frequency;
      maximumFrequency = Math.max(maximumFrequency, frequency);
    }

    // Create histogram.
    final Histogram histogram = new Histogram(binCount, y_min, y_max);
    histogram.maximumFrequency = 0;
    for (int bin = 0; bin < binCount; bin++) {
      final long frequency = (long) (frequencies[bin] / maximumFrequency * Long.MAX_VALUE / binCount);
      histogram.frequencies[bin] = frequency;
      histogram.totalNumberOfValues += frequency;
      histogram.maximumFrequency = Math.max(histogram.maximumFrequency, frequency);
    }
    return histogram;
  }

  public void export(final OutputStream os) {
    final PrintStream out = new PrintStream(os);
    for (int bin = 0; bin < this.binCount; bin++) {
      out.printf(Locale.ENGLISH, "%f\t%f%n", this.binToValue(bin), (double) this.frequencies[bin] / this.getTotalNumberOfValues());
    }
  }

  public void export(final File file) throws IOException {
    final OutputStream os = new FileOutputStream(file);
    this.export(os);
    os.close();
  }

  private int valueToBin(final double value) {
    final double scaledValue = (value - this.minimumValue) / (this.maximumValue - this.minimumValue);
    switch (this.border) {
      case BORDER_FLOOR:
        return (int) Math.floor(scaledValue * this.getBinCount());
      case BORDER_CEIL:
        return (int) Math.ceil(scaledValue * this.getBinCount()) - 1;
      default:
        throw new IllegalStateException("Wrong border specification.");
    }
  }

  /**
   * Returns centered value of bin.
   * 
   * @param bin
   * @return value of the given bin
   */
  private double binToValue(final int bin) {
    return (0.5 + bin) / this.getBinCount() * (this.maximumValue - this.minimumValue) + this.minimumValue;
  }

  // Add one occurrence of a value. 
  public synchronized void add(final double value) {
    if (value >= this.maximumValue) {
      throw new IllegalArgumentException("value too large (maximumFrequency = " + this.maximumValue + ")");
    }
    if (value < this.minimumValue) {
      throw new IllegalArgumentException("value too small (min = " + this.minimumValue + ")");
    }

    this.sumOfValues += value;

    final int bin = this.valueToBin(value);
    this.frequencies[bin]++;

    if (this.frequencies[bin] > this.maximumFrequency) {
      this.maximumFrequency = this.frequencies[bin];
      this.maximumFrequencyBin = bin;
    }
    this.totalNumberOfValues++;

    // If maximumNumberOfValues == -1 no old valuesHistory are to be removed.
    if (this.maximumNumberOfValues != -1) {
      this.valuesHistory.add(value);

      if (this.valuesHistory.size() > this.maximumNumberOfValues) {
        final int eldestIndex = 0;
        final double eldestValue = this.valuesHistory.remove(eldestIndex);
        this.remove(eldestValue);
      }
    }
  }

  private synchronized void remove(final double value) {
    final int bin = this.valueToBin(value);

    this.sumOfValues -= value;

    this.frequencies[bin]--;

    if (this.maximumFrequencyBin == bin) {
      // TODO: recalculate maximum
      this.maximumFrequency--;
    }
    this.totalNumberOfValues--;
  }

  public int getBinCount() {
    return this.binCount;
  }

  private void setBinCount(final int binCount) {
    this.binCount = binCount;
  }

  public long getMaximumFrequency() {
    return this.maximumFrequency;
  }

  public double getSumOfValues() {
    return this.sumOfValues;
  }

  public long getTotalNumberOfValues() {
    return this.totalNumberOfValues;
  }

  public String getTextGraph() {
    String textGraph = "";
    final int width = 150;
    final String newline = "\n";

    for (int bin = 0; bin < this.binCount; bin++) {
      if (bin > 0) {
        textGraph += newline;
      }
      textGraph += String.format(Locale.ENGLISH, "%f", this.binToValue(bin)) + "|";
      if (this.totalNumberOfValues > 0) {
        for (int i = 1; i < (double) width * (double) this.frequencies[bin] / this.maximumFrequency; i++) {
          textGraph += '#';
        }
      }
      textGraph += ">";
      textGraph += this.frequencies[bin];
    }

    return textGraph;
  }

  public double getKullbackLeiblerDivergence(final Histogram targetHistogram) {
    if (this.getBinCount() != targetHistogram.getBinCount()) {
      throw new IllegalArgumentException("Differnt number of bin counts: " + this.getBinCount() + ", " + targetHistogram.getBinCount() + ".");
    }

    double D_KL = 0;

    for (int bin = 0; bin < this.binCount; bin++) {
      final double p = (double) this.frequencies[bin] / this.getTotalNumberOfValues();
      final double q = (double) targetHistogram.frequencies[bin] / targetHistogram.getTotalNumberOfValues();
      if (p != 0.0 && q != 0.0) {
        D_KL += p * Math.log(p / q);
      } else {
        // 0 log 0 is interpreted as zero.
      }
    }

    return D_KL;
  }

  @Override
  public String toString() {
    return this.frequencies.toString();
  }
}