package utils;

import java.io.ByteArrayInputStream;
import java.io.ByteArrayOutputStream;
import java.io.IOException;
import java.io.ObjectInputStream;
import java.io.ObjectOutputStream;
import java.io.Serializable;
import java.lang.reflect.Field;
import java.lang.reflect.Method;
import java.util.Arrays;
import java.util.Collection;
import java.util.HashSet;
import java.util.List;
import java.util.Random;
import java.util.Set;
import java.util.Vector;

/**
 * class for utilities
 * 
 * @author Mathias Linkerhand
 * @version 2011-09-01
 */
public final class Utils {
  private Utils() {
  }

  @SuppressWarnings("unchecked")
  public static <T extends Number> T sum(final Iterable<T> iterable) {
    T sum = null;
    for (final T value : iterable) {
      if (sum == null) {
        sum = value;
      } else {
        if (value instanceof Double) {
          sum = (T) (Double)((Double) sum + (Double) value);
        }
      }
    }
    return sum;
  }

  /**
   * deep copy arrays and cut/pad to size
   * 
   * @see java.util.Arrays#copyOf(double[], int)
   */
  public static double[][] copyOf(final double[][] array, final int newLength1, final int newLength2) {
    double[][] result = new double[newLength1][newLength2];
    for (int i1 = 0; i1 < Math.min(array.length, newLength1); i1++) {
      for (int i2 = 0; i2 < Math.min(array[0].length, newLength2); i2++) {
        result[i1][i2] = array[i1][i2];
      }
    }
    return result;
  }

  /**
   * deep copy arrays and cut/pad to size
   * 
   * @see java.util.Arrays#copyOf(double[], int)
   */
  public static double[][][] copyOf(final double[][][] array, final int newLength1, final int newLength2, final int newLength3) {
    double[][][] result = new double[newLength1][newLength2][newLength3];
    for (int i1 = 0; i1 < Math.min(array.length, newLength1); i1++) {
      for (int i2 = 0; i2 < Math.min(array[0].length, newLength2); i2++) {
        for (int i3 = 0; i3 < Math.min(array[0][0].length, newLength3); i3++) {
          result[i1][i2][i3] = array[i1][i2][i3];
        }
      }
    }
    return result;
  }

  public static void copyInto(final double[][][] from, double[][][] to) {
    for (int i = 0; i < Math.min(from.length, to.length); i++) {
      for (int j = 0; j < Math.min(from[i].length, to[i].length); j++) {
        for (int k = 0; k < Math.min(from[i][j].length, to[i][j].length); k++) {
          to[i][j][k] = from[i][j][k];
        }
      }
    }
  }

  public static double[] fill(final int size, final double value) {
    final double[] result = new double[size];
    Arrays.fill(result, value);
    return result;
  }

  public static double[] randomArray(final int size, final Random random, final double min, final double max) {
    final double[] result = new double[size];
    for (int i = 0; i < size; i++) {
      result[i] = min + (max - min) * random.nextDouble();
    }
    return result;
  }

  public static List<Double> randomList(final int size, final Random random, final double min, final double max) {
    final List<Double> result = new Vector<Double>(size);
    for (int i = 0; i < size; i++) {
      result.add(min + (max - min) * random.nextDouble());
    }
    return result;
  }

  public static void randomFill(double[] array, final Random random, final double min, final double max) {
    for (int i = 0; i < array.length; i++) {
      array[i] = min + (max - min) * random.nextDouble();
    }
  }

  public static void randomFill(double[][] array, final Random random, final double min, final double max) {
    for (int i = 0; i < array.length; i++) {
      Utils.randomFill(array[i], random, min, max);
    }
  }

  public static void randomFill(double[][][] array, final Random random, final double min, final double max) {
    for (int i = 0; i < array.length; i++) {
      Utils.randomFill(array[i], random, min, max);
    }
  }

  /**
   * changes extension to new extension
   * 
   * example: x = changeExtension("data.txt", ".java") will assign "data.java"
   * to x.
   * 
   * @param originalName
   * @param newExtension
   * @return filename
   */
  public static String changeExtension(final String originalName, final String newExtension) {
    final int lastDot = originalName.lastIndexOf(".");
    if (lastDot != -1) {
      return originalName.substring(0, lastDot) + newExtension;
    } else {
      return originalName + newExtension;
    }
  }

  /**
   * Returns the field of a given class also searching for the field in superclasses.
   * 
   * @see Class#getField(String)
   * @see Class#getDeclaredField(String)
   * 
   * @param clazz
   * @param fieldName
   * @return field
   * @throws NoSuchFieldException
   */
  public static Field getField(final Class<?> clazz, final String fieldName) throws NoSuchFieldException {
    try {
      return clazz.getDeclaredField(fieldName);
    } catch (NoSuchFieldException exception) {
      final Class<?> superClass = clazz.getSuperclass();
      if (superClass == null) {
        throw exception;
      } else {
        return getField(superClass, fieldName);
      }
    }
  }

  /**
   * Returns the method of a given class also searching for the method in superclasses.
   * 
   * @see Class#getField(String)
   * @see Class#getDeclaredField(String)
   * 
   * @param clazz
   * @param methodName
   * @return field
   * @throws NoSuchFieldException
   */
  public static Method getMethod(final Class<?> clazz, final String methodName) throws NoSuchMethodException {
    try {
      return clazz.getDeclaredMethod(methodName);
    } catch (NoSuchMethodException exception) {
      final Class<?> superClass = clazz.getSuperclass();
      if (superClass == null) {
        throw exception;
      } else {
        return getMethod(superClass, methodName);
      }
    }
  }

  /**
   * Returns all the fields of a given class including the fields of all superclasses.
   * 
   * @see Class#getFields()
   * @see Class#getDeclaredFields()
   * 
   * @param clazz
   * @return fields
   */
  public static Set<Field> getDeclaredFields(final Class<?> clazz) {
    final Set<Field> fields = new HashSet<Field>(Arrays.asList(clazz.getDeclaredFields()));
    final Class<?> superClass = clazz.getSuperclass();
    if (superClass != null) {
      fields.addAll(getDeclaredFields(superClass));
    }
    return fields;
  }

  /**
   * Returns all the fields of a given class including the fields of all superclasses.
   * 
   * @see Class#getFields()
   * @see Class#getDeclaredFields()
   * 
   * @param clazz
   * @return methods
   */
  public static Set<Method> getDeclaredMethods(final Class<?> clazz) {
    final Set<Method> methods = new HashSet<Method>(Arrays.asList(clazz.getDeclaredMethods()));
    final Class<?> superClass = clazz.getSuperclass();
    if (superClass != null) {
      methods.addAll(getDeclaredMethods(superClass));
    }
    return methods;
  }

  /**
   * Serializes the given object and returns a byte representation.
   * 
   * @param object
   * @return byte array
   */
  public static byte[] objectToBytes(final Serializable object) {
    byte[] bytes = null;

    try {
      ByteArrayOutputStream byteOut = new ByteArrayOutputStream();
      ObjectOutputStream objectOut = new ObjectOutputStream(byteOut);
      objectOut.writeObject(object);
      objectOut.close();
      bytes = byteOut.toByteArray();
    } catch (IOException e) {
      e.printStackTrace();
    }

    return bytes;
  }

  /**
   * Deserializes the given byte representation of an object and returns the corresponding object.
   * 
   * @param serializedObject
   * @return object
   */
  public static Object bytesToObject(final byte[] serializedObject) {
    Object object = null;

    try {
      ByteArrayInputStream byteIn = new ByteArrayInputStream(serializedObject);
      ObjectInputStream objectIn = new ObjectInputStream(byteIn);
      object = objectIn.readObject();
      objectIn.close();
    } catch (ClassNotFoundException e) {
      e.printStackTrace();
    } catch (IOException e) {
      e.printStackTrace();
    }

    return object;
  }

  /**
   * Creates a deep copy of a given object. The object has to be serializable and therefore has to implement the interface Serializable. The copy is created using ObjectInputStream and ObjectOutputStream.
   * 
   * @param <T>
   * @param object
   * @return object
   */
  @SuppressWarnings("unchecked")
  public static <T extends Serializable> T deepCopy(final T object) {
    return (T) bytesToObject(objectToBytes(object));
  }

  /**
   * Creates a flat copy of a (deep) collection hierarchy.
   */
  @SuppressWarnings("unchecked")
  public static <T> List<T> flatten(final T type, final Collection<?> collection) {
    final Vector<T> flatCopy = new Vector<T>();
    for (final Object element : collection) {
      if (element instanceof Collection<?>) {
        flatCopy.addAll(flatten(type, (Collection<?>) element));
      } else {
        flatCopy.add((T) element);
      }
    }
    return flatCopy;
  }
}